/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * $Log:	boot.c,v $
 * Revision 2.5.1.2  93/02/01  12:55:23  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:51:07  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:39:12  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/boot.c,v 2.5.1.2 93/02/01 12:55:23 kaufman Rel $";
#endif

#include <sys/types.h>
#include <sys/file.h>
#include <sys/filio.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <strings.h>
#include <stdio.h>

#include "embed.h"
#include "life.h"
#include "emblib.h"
#include "genera.h"

int verbose;
static int ivory_fd;

boot_ivory (device, fd, flags)
     char *device;
     int flags;
{
  int i, j, pid;
  int done, reset_request_count;
  short child_no;
  char buf0[BUFSIZ], *cp;
  IvoryConfig config;
  EmbQueue *q;
  EmbColdLoadChannel *cold_load;
  char *getenv();
  void fep_down(), disable_life_support();

  if (flags & FL_IVORY_PRELOCKED) {
    ivory_fd = fd;
  } else {
    /* Open device, map in communication area */
    if ((ivory_fd = MapEmbCommArea(device)) == -1)
      return(-1);

    /* Try to lock the Ivory device.
     * If we can't, and we weren't planning to INIT Ivory, then it's
     * not an error.  This just means we won't be able to grab the
     * Cold Load channel, because someone else already has it.  If
     * we were planning to INIT Ivory, then this is an error. */
    if (EmbCommAreaPtr->fep_status == HaltedFEPStatus)
      flags |= FL_LOCK_REQUIRED;
    switch (lock_ivory(ivory_fd, device, flags)) {
    case 1:
      fprintf(stderr, "Unable to lock %s against simultaneous access\n", device);
      break;
      
    case 0:
      break;
      
    case -1:
      return(-1);
    }
  }

  /* Check life support versions */
  if (!(flags & FL_SILENT))
    check_kernel_version();

  /* If the comm area is clobbered, we're hosed */
  if (EmbCommAreaPtr->version != 1 ||
      bcmp((char *) &EmbCommAreaPtr->identifier, "DBME", 4)) {
    fprintf(stderr, "%s: Embedding communication area clobbered\n", device);
    return(-1);
  }
  
  /* If life support is disabled, enable it */
  if (ioctl(ivory_fd, IVORYGCONF, &config) == -1) {
    perror("IVORYGCONF");
    return(-1);
  }
  if (config.enable_Ivory == 0) {
    config.enable_Ivory = 1;
    on_exit(disable_life_support, ivory_fd);
    if (ioctl(ivory_fd, IVORYSCONF, &config) == -1) {
      perror("IVORYSCONF");
      return(-1);
    }
  }

  /* If the FEP is not running, INIT the Ivory.  This will cause the Boot ROM
   * to run, loading the Device PROM via reset requests.  The Device PROM will
   * want to talk to the Cold Load... */
  if (flags & FL_FORCE || EmbCommAreaPtr->fep_status == HaltedFEPStatus) {
    /* Poke cold-load channel select state */
    cold_load = (EmbColdLoadChannel *) EmbHostPtr(EmbCommAreaPtr->cold_load_channel);
    if (cold_load != NULL) {
      /* We lose a signal here - this is dangerous! */
      q = (EmbQueue *) EmbHostPtr(cold_load->display_output_queue);
      q->signal = -1;
      cold_load->selproc = NULL;
      cold_load->command_history_index = 0;
      cold_load->command_history_wrapped_p = 0;
    }

    /* INIT Ivory */
    NOTE("Starting Ivory");
    for (i = 0, done = 0; i < 5 && !done; i++) {
      if (i)
	NOTE("Retrying...");
      reset_request_count = EmbCommAreaPtr->device_prom_reset_request_count;
      if (ioctl(ivory_fd, IVORYINIT, 0) == -1) {
	perror("IVORYINIT");
	return(-1);
      }
      /* Check at 1/4-second intervals for up to 5 seconds */
      for (j = 0; j < 20; j++) {
	if (j)
	  usleep(250000);
	if (EmbCommAreaPtr->device_prom_reset_request_count != reset_request_count) {
	  done = 1;
	  break;
	}
      }
    }

    if (done) {
      NOTE("The Device PROM is loading");

      /* Make sure this is turned off */
      EmbCommAreaPtr->restart_applications = 0;
    } else {
      fprintf(stderr, "%s: Couldn't start Ivory\n", device);
      return(-1);
    }
  }
  
  /* If the FEP isn't up within a minute, start griping */
  if (!(flags & FL_DISABLE_FEP_TIMER)) {
    signal(SIGALRM, fep_down);
    alarm(60);
  }

  return(ivory_fd);
}

lock_ivory (ivory_fd, device, flags)
     int ivory_fd;
     char *device;
     int flags;
{
  int fd;
  char buf[BUFSIZ];
  struct stat st;
  struct flock lock;

  if (fstat(ivory_fd, &st) == -1) {
    perror("fstat");
    return(1);
  }

  (void) sprintf(buf, "/etc/ivory%d.pid", minor(st.st_rdev));
  if ((fd = open(buf, O_RDWR)) == -1) {
    perror(buf);
    return(1);
  }

  lock.l_type = F_WRLCK;
  lock.l_whence = 0;
  lock.l_start = 0;
  lock.l_len = 0;
  if (fcntl(fd, F_SETLK, &lock) == -1) {
    if (flags & FL_LOCK_REQUIRED) {
      if (errno != EAGAIN && errno != EACCES) {
	perror("F_SETLK");
      } else {
	/* Who is locking this device? */
	lock.l_type = F_WRLCK;
	lock.l_whence = 0;
	lock.l_start = 0;
	lock.l_len = 0;
	if (fcntl(fd, F_GETLK, &lock) == -1) {
	  perror("F_GETLK");
	  lock.l_type = -1;
	}

	fprintf(stderr, "%s is locked", device);
	if (lock.l_type == F_WRLCK)
	  fprintf(stderr, " by pid %d", lock.l_pid);
	if (flags & FL_FORCE)
	  fprintf(stderr, "can't force Ivory");
	else if (EmbCommAreaPtr->fep_status == HaltedFEPStatus)
	  fprintf(stderr, " and Ivory is halted");
	fprintf(stderr, ".\n");
      }
      close(fd);
      return(-1);
    }
  }

  return(0);
}

void disable_life_support (status, ivory_fd)
     int status, ivory_fd;
{
  IvoryConfig config;

  if (status) {
    /* If life support is enabled, disable it */
    if (ioctl(ivory_fd, IVORYGCONF, &config) == -1) {
      perror("IVORYGCONF");
      abort();
    }
    if (config.enable_Ivory == 1) {
      config.enable_Ivory = 0;
      on_exit(disable_life_support, ivory_fd);
      if (ioctl(ivory_fd, IVORYSCONF, &config) == -1) {
	perror("IVORYSCONF");
	abort();
      }
    }
  }
}

#ifdef notdef
start_delayed_cold_x (ivory_fd, display)
     int ivory_fd;
     char *display;
{
  fd_set fds;
  EmbColdLoadChannel *cold_load = (EmbColdLoadChannel *) HostPointer(EmbCommAreaPtr->cold_load_channel);
  EmbQueue *display_queue = (EmbQueue *) HostPointer(cold_load->display_output_queue);

  while(!EmbQueueFilled(display_queue)) {
    FD_ZERO(&fds);
    FD_SET(ivory_fd, &fds);
    select(ivory_fd+1, fds, NULL, NULL, NULL);
  }

  start_cold_x(ivory_fd, display, NULL, 0);
}
#endif

void fep_down (sig, code, scp, addr)
     /*ARGSUSED*/
     int sig, code;
     struct sigcontext *scp;
     char *addr;
{
  IvoryConfig config;
  char buf[BUFSIZ];

  if (EmbCommAreaPtr->fep_status == HaltedFEPStatus) {

    /* FEP is still coming up, interact with user */
    fprintf(stderr, "\n%cThe FEP hasn't come up yet!\n", '\007');

    while (1) {
      printf("Disable life support, wait another minute, or stop checking? ");
      fflush(stdout);

      if (fgets(buf, BUFSIZ, stdin) == NULL) {
	clearerr(stdin);
	continue;
      }

      switch (*buf) {
      case 'd':
      case 'D':
	stop_cold_x();
	if (ioctl(ivory_fd, IVORYGCONF, &config) == -1) {
	  perror("IVORYGCONF");
	  exit(1);
	}
	if (config.enable_Ivory == 1) {
	  config.enable_Ivory = 0;
	  if (ioctl(ivory_fd, IVORYSCONF, &config) == -1) {
	    perror("IVORYSCONF");
	    exit(1);
	  }
	  puts("Ivory life support disabled");
	} else
	  puts("Ivory life support already disabled");
	exit(0);

      case 's':
      case 'S':
	puts("Disabling FEP check...");
	return;

      case 'w':
      case 'W':
	puts("Waiting for another minute...");
	alarm(60);
	return;

      default:
	puts("Please answer \"Disable\", \"Wait\", or \"Stop\"");
	break;
      }
    }
  }
}



EmbNetChannel *find_emb_network_channel ()
{
  EmbPtr cp;
  EmbChannel *ch;

  for (cp = EmbCommAreaPtr->channel_table; cp != NullEmbPtr; cp = ch->next) {
    ch = (EmbChannel *) HostPointer(cp);
    if (ch->type == EmbNetworkChannelType)
      return((EmbNetChannel *) ch);
  }
  return(NULL);
}
