/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * $Log:	config.c,v $
 * Revision 2.5.1.2  93/02/01  12:55:28  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:51:13  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:39:26  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/config.c,v 2.5.1.2 93/02/01 12:55:28 kaufman Rel $";
#endif

#include <sys/types.h>
#include <sys/file.h>
#include <sys/filio.h>
#include <ctype.h>
#include <errno.h>
#include <pwd.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>

#include "embed.h"
#include "life.h"
#include "config.h"

#ifdef unused
/* --- Describe contents of EmbConfig structure */
describe_config (cp)
     IvoryConfig *cp;
{
  printf(">Ivory %sabled\n", cp->enable_Ivory ? "en" : "dis");
}
#endif

open_prom (conf, name)
     Config *conf;
     char *name;
{
  char buf[BUFSIZ];
  int fd;
  sprintf(buf, "%s/%s", conf->prom_directory, name);
  if ((fd = open(buf, O_RDONLY)) == -1)
    log_error(LOG_ERR, buf);
  return(fd);
}

open_device_prom (conf)
     Config *conf;
{
  if (conf->kern.device_prom_fd == -1)
    if ((conf->kern.device_prom_fd = open_prom(conf, "solstice-device.prom")) == -1)
      return(-1);

  return(0);
}

void init_config (conf, fd, communication_area_nbytes)
     Config *conf;
     int fd;
     int communication_area_nbytes;
{
  struct passwd *pw;

  /* Set up configuration structure */
  bzero((char *) conf, sizeof(Config));
  if (ioctl(fd, IVORYGCONF, &conf->kern) == -1) {
    log_error(LOG_ERR, "IVORYGCONF");
    exit(1);
  }
  conf->kern.enable_Ivory = 0;
  conf->kern.slave_trigger_offset = (EmbPtr) communication_area_nbytes>>2;
  conf->kern.device_prom_fd = -1;
  conf->emb.communication_memory_size = communication_area_nbytes;
  conf->emb.host_major_version = 0x01000000;
  conf->emb.host_minor_version = 0;
  conf->emb.host_buffer_size = 0;
  conf->emb.guest_buffer_size = 40000;
  conf->emb.n_transmit_buffers = 2;		/* Arbitrary! */
  conf->emb.n_receive_buffers = 10;		/* Arbitrary! */
  conf->emb.network_filter = host_ip;
  conf->emb.debug_level = 0;	                /* By default, don't pester users */
  FD_ZERO(&conf->emb.network_interfaces);
  FD_ZERO(&conf->emb.disk_partitions);
  if ((pw = getpwnam("guest")) != NULL)
    conf->emb.host_file_uid = pw->pw_uid;
  else
    conf->emb.host_file_uid = 6; /* news, usually */
}

/* Read a configuration file */

static int config_debug(), config_device();
static int config_device_prom(), config_filter();
static int config_host_file(), config_interface();
static int config_partition();
static int config_prom(), check_prom();
#ifdef notdef
static int cmd_required(), check_interface(), check_partition();
#else
#define cmd_required 0
#define check_interface 0
#define check_partition 0
#endif

struct config_cmd {
  char *name;
  int (*fn)();
  int minargs, maxargs;		/* arg count includes command name */
  int (*ckfn)();
  int seen;
} commands[] = {
  /* name        fn     minargs  maxargs   ckfn   seen */
  "debug", config_debug, 1, MAX_ARGC, 0, 0,
  "device", config_device, 2, 2, 0, 0,
  "device-prom", config_device_prom, 2, 2, 0, 0,
  "host-file-user", config_host_file, 2, 2, NULL, 0,
  "interface", config_interface, 2, MAX_ARGC, check_interface, 0,
  "network-filter-mode", config_filter, 2, 2, NULL, 0,
  "partition", config_partition, 2, MAX_ARGC, check_partition, 0,
  "prom-directory", config_prom, 2, 2, check_prom, 0,
  "read-only-partition", config_partition, 2, MAX_ARGC, check_partition, 0,
/*  "slave-buffer-extra-words", config_extra, 2, 2, 0, 0,*/
  NULL, NULL, 0, 0, 0, 0,
};

void read_config (file, conf)
     char *file;
     Config *conf;
{
  FILE *fp;
  char buf[BUFSIZ], obuf[BUFSIZ], *cp, *argv[MAX_ARGC];
  register int i;
  int command_err = 0, syntax_err = 0, argc;

  if ((fp = fopen(file, "r")) == NULL) {
    log_error(LOG_ERR, file);
    exit(1);
  }

  /* Reset all seen flags */
  for (i = 0; commands[i].name != NULL; i++)
    commands[i].seen = 0;

  /* Read the file */
  while (fgets(buf, BUFSIZ, fp) != NULL) {
    if (buf[0] == '#' || buf[0] == ';')
      /* Support both Unixy and Lispy comment characters */
      continue;

    /* Trim trailing whitespace */
    for (cp = &buf[strlen(buf)-1]; isspace(*cp) && cp >= buf; cp--);
    *++cp = '\0';

    /* Build argument vector */
    for (cp = buf, argc = 0; *cp && argc < MAX_ARGC; argc++) {
      argv[argc] = cp;
      for (; *cp; cp++) {
	if (isspace(*cp)) {
	  /* End of word: terminate it, and skip any intervening whitespace */
	  *cp++ = '\0';
	  for (; isspace(*cp); cp++);
	  break;
	}
      }
    }
    if (argc == 0)
      continue;
    if (argc == MAX_ARGC) {
      sprintf(obuf, "Too many arguments to %s, ignoring extras", argv[0]);
      log_message(LOG_ERR, obuf);
    }

    /* Lowercase command name */
    for (cp = argv[0]; *cp; cp++)
      if (isupper(*cp))
	*cp = tolower(*cp);

    /* Call appropriate command */
    for (i = 0; commands[i].name != NULL; i++) {
      if (!strcmp(argv[0], commands[i].name)) {
        if (argc < commands[i].minargs) {
          sprintf(obuf, "%s: Too few arguments (must be at least %d)",
                  commands[i].name, commands[i].minargs-1);
	  log_message(LOG_ERR, obuf);
        } else if (argc > commands[i].maxargs) {
          sprintf(obuf, "%s: Too many arguments (must be no more than %d)",
                  commands[i].name, commands[i].maxargs-1);
	  log_message(LOG_ERR, obuf);
        } else
	  syntax_err |= (*commands[i].fn) (argc, argv, conf);
	break;
      }
    }

    if (commands[i].name == NULL) {
      sprintf(obuf, "%s is not a configuration-file command", buf);
      log_message(LOG_ERR, obuf);
      command_err = 1;
    }
  }

  fclose(fp);

  if (command_err) {
    fputs("Valid configuration-file commands are:\n", stderr);
    for (i=0; commands[i].name != NULL; i++)
      fprintf(stderr, "%s ", commands[i].name);
    fputs("\n", stderr);
  }

  /* Run sanity checkers */
  for (i = 0; commands[i].name != NULL; i++)
    if (commands[i].ckfn != 0)
      syntax_err |= (*commands[i].ckfn)(&commands[i], commands[i].seen, conf);

  if (command_err || syntax_err) {
    sprintf(obuf, "Unable to configure life-support for %s", conf->device);
    log_message(LOG_ERR, obuf);
    exit(1);
  }
}



static config_debug (argc, argv, conf)
     int argc;
     char **argv;
     Config *conf;
{
  register int i;
  char obuf[BUFSIZ];

  conf->emb.debug_level = 0;
  if (argc == 2 && !strcmp(argv[1], "all"))
    conf->emb.debug_level = -1;
  else
    for (i = 1; i < argc; i++)
      if (!debug_level(argv[i], &conf->emb.debug_level)) {
	sprintf(obuf, "%s is not a debug flag, ignored", argv[i]);
	log_message(LOG_ERR, obuf);
      }

  return(0);
}

static config_device (argc, argv, conf)
     /*ARGSUSED*/
     int argc;
     char **argv;
     Config *conf;
{
  char obuf[BUFSIZ];

  if (strcmp(argv[1], conf->device)) {
    sprintf(obuf,
	    "Configuration file %s specifies device %s, but is being read for device %s",
	    conf->config_filename, argv[1], conf->device);
    log_message(LOG_ERR, obuf);
    return(1);
  }
  return(0);
}

/* Obsolete, retained for compatibility */
static config_device_prom (argc, argv, conf)
     /*ARGSUSED*/
     int argc;
     char **argv;
     Config *conf;
{
  int fd;

  if ((fd = open(argv[1], O_RDONLY)) == -1 || ioctl(fd, FIOCLEX, 0) < 0) {
    log_error(LOG_ERR, argv[1]);
    log_message(LOG_ERR, "   the FEP will be unable to boot");
    return(-1);
  }
  conf->kern.device_prom_fd = fd;

  return(0);
}

#ifdef notdef
/* This command allows users to allocate the low half of the slave
 * buffer as part of the communication area.  (Genera doesn't use it yet.) */
static config_extra (argc, argv, conf)
     /*ARGSUSED*/
     int argc;
     char **argv;
     Config *conf;
{
  char *cp;
  int i;

  if (!strcmp(argv[1], "all"))
    conf->emb.guest_extra_buffer_size = conf->kern.comm_area_base;
  else {
    for (cp = argv[1]; *cp; cp++)
      if (!isdigit(*cp)) {
	fprintf(stderr, "%s: bad value \"%s\"\n", argv[0], argv[1]);
	return(1);
      }
    i = atoi(argv[1]) * sizeof(long);
    if (i > conf->kern.comm_area_base) {
      fprintf(stderr, "%s: %d too big, using %d\n",
	      argv[0], i/sizeof(long), conf->kern.comm_area_base/sizeof(long));
      i = conf->kern.comm_area_base;
    }
    conf->emb.guest_extra_buffer_size = i;
  }
  return(0);
}
#endif

static config_filter (argc, argv, conf)
     /*ARGSUSED*/
     int argc;
     char **argv;
     Config *conf;
{
  char obuf[BUFSIZ];

  if (!strcmp("all-ip", argv[1]))
    conf->emb.network_filter = all_ip;
  else if (!strcmp("host-ip", argv[1]))
    conf->emb.network_filter = host_ip;
  else if (!strcmp("none", argv[1]))
    conf->emb.network_filter = none;
  else {
    sprintf(obuf, "Unknown network filter mode %s, ignored", argv[1]);
    log_message(LOG_ERR, obuf);
  }

  return(0);
}

static config_host_file (argc, argv, conf)
     /*ARGSUSED*/
     int argc;
     char **argv;
     Config *conf;
{
  struct passwd *pw;
  char obuf[BUFSIZ];

  if ((pw = getpwnam(argv[1])) == NULL) {
    sprintf(obuf, "%s: No such user", argv[1]);
    log_message(LOG_ERR, obuf);
  } else
    conf->emb.host_file_uid = pw->pw_uid;

  return(0);
}

static config_interface (argc, argv, conf)
     /*ARGSUSED*/
     int argc;
     char **argv;
     Config *conf;
{
  int fd;
  char obuf[BUFSIZ];

  if (!strncmp(argv[1], "lo", 2) && isdigit(argv[1][2])) {
    sprintf(obuf, "%s: Invalid interface argument", argv[1]);
    log_message(LOG_ERR, obuf);
  } else if ((fd = nit_init(argv[1], conf->emb.network_filter)) < 0) {
    sprintf(obuf, "   for interface %s", argv[1]);
    log_message(LOG_ERR, obuf);
  } else if (ioctl(fd, FIOCLEX, 0) < 0)
    log_error(LOG_ERR, argv[1]);
  else
    FD_SET(fd, &conf->emb.network_interfaces);

  return(0);
}

static config_partition (argc, argv, conf)
     int argc;
     char **argv;
     Config *conf;
{
  int i, fd, mode;
  int zero = 0;

  if (argv[0][0] == 'r')
    mode = O_RDONLY|O_EXCL;
  else
    mode = O_RDWR|O_EXCL;

  for (i = 1; i < argc; i++) {
    if ((fd = open(argv[i], mode)) < 0 ||
	ioctl(fd, FIOCLEX, 0) < 0 ||
	ioctl(fd, FIOASYNC, &zero) < 0)
      log_error(LOG_ERR, argv[i]);
    else
      FD_SET(fd, &conf->emb.disk_partitions);
  }

  return(0);
}

static config_prom (argc, argv, conf)
     int argc;
     char *argv[];
     Config *conf;
{
  int tmp;
  char buf[BUFSIZ];
  
  if (access(argv[1], X_OK)) {
    tmp = errno;
    sprintf(buf, "%s %s", argv[0], argv[1]);
    errno = tmp;
    log_error(LOG_ERR, buf);
    return(1);
  }

  conf->prom_directory = strdup(argv[1]);
  return(0);
}

static int check_prom (argc, argv, conf)
     int argc;
     char **argv;
     Config *conf;
{
  if (conf->prom_directory == NULL)
    conf->prom_directory = DEFAULT_PROM_DIRECTORY;
  return(0);
}
