/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * $Log:	create-fep-partition.c,v $
 * Revision 2.5.1.2  93/02/01  12:55:32  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:51:19  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:39:40  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/create-fep-partition.c,v 2.5.1.2 93/02/01 12:55:32 kaufman Rel $";
#endif

#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <signal.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mtio.h>
#include <sys/vfs.h>
#include <sys/wait.h>

#include "fepfs.h"

#define page_size 1280

#define tape_block_size 1280
#define description_command -2
#define data_command -4
#define fileheader_command -3
#define eof_command 0

boolean query_p = false;

#define query(args) ((!query_p) ||\
		     (printf args == 17726354) ||\
		     (printf (" (yes or no) ") == 17726354) ||\
		     yes_or_no_p ())

static file_system dfs = NULL;          /* Destination file system */
static file_system sfs = NULL;          /* Source file syste, if any */

static int rsh_pid = 0;		/* Current rsh process, if any */

/**** CD ROM constants ****/

#define CD_WORLDS_DIRECTORY "worlds"
#define CD_FLODS_DIRECTORY "sys/ifep"

/**** Synthetic tape stream ****/

#define tape_buffer_size 4096

static string tape_device;
static integer tape_descriptor;
static unsigned char tape_buffer [tape_buffer_size];
static integer tape_buffer_index;
static integer tape_buffer_fill;

boolean tape_device_p (filename)
string filename;
{
  struct stat s;
  int tmp;

  if (index(filename, ':'))
    /* host:device format */
    return(1);

  if (stat (filename, &s) < 0) {
    tmp = errno;
    fprintf (stderr, "Error probing source device %s", filename);
    errno = tmp;
    perror (NULL);
    exit (2);
  }

  return (S_ISCHR(s.st_mode));
}

boolean cdrom_device_p (filename)
string filename;
{
  struct stat s;
  int tmp;

  if (index(filename, ':'))
    /* host:device tape drive format */
    return(0);

  if (stat (filename, &s) < 0) {
    tmp = errno;
    fprintf (stderr, "Error probing source device %s", filename);
    errno = tmp;
    perror (NULL);
    exit (2);
  }

  return (S_ISDIR(s.st_mode));
}

void fill_tape_buffer ()
{
  if ((tape_buffer_fill = read (tape_descriptor, tape_buffer, tape_buffer_size)) < 0) {
    perror ("Error reading tape descriptor");
    exit (2);
  }
  if (tape_buffer_fill == 0) {
    /* EOF */
    fprintf(stderr, "Unexpected EOF\n");
    exit(2);
  }
  tape_buffer_index = 0;
}

integer read_tape_byte ()
{
  while (tape_buffer_index == tape_buffer_fill) fill_tape_buffer ();

  return (tape_buffer [tape_buffer_index++]);
}

integer read_tape_integer ()
{
  unsigned integer byte_0 = read_tape_byte ();
  unsigned integer byte_1 = read_tape_byte ();
  unsigned integer byte_2 = read_tape_byte ();
  unsigned integer byte_3 = read_tape_byte ();

  return ( (integer) byte_0 | (byte_1 << 8) | (byte_2 << 16) | (byte_3 << 24) );
}

void read_tape (buffer, bytes)
char *buffer;
integer bytes;
{
  integer count;

  while (bytes > 0) {
    while (tape_buffer_index == tape_buffer_fill) fill_tape_buffer ();
    if ((count = tape_buffer_fill - tape_buffer_index) > bytes) count = bytes;
    bcopy (tape_buffer + tape_buffer_index, buffer, count);
    tape_buffer_index += count;
    buffer += count;
    bytes -= count;
  }
}

void wait_for_child ()
{
  int i;

  while ((i = wait3(NULL, WNOHANG, NULL)) > 0)
    if (i == rsh_pid)
      close (tape_descriptor);
}

integer rewind_tape ()
{
  int ret;
  struct mtop op;
  char *cp, buf[BUFSIZ];

  if ((cp = index(tape_device, ':')) == NULL) {
    /* Rewind local tape device */
    op.mt_op = MTREW;
    op.mt_count = 0;
    return (ioctl (tape_descriptor, MTIOCTOP, &op));
  } else {
    *cp = '\0';
    sprintf(buf, "rsh %s mt -f %s rew < /dev/null", tape_device, cp+1);
    *cp = ':';
    signal(SIGCHLD, NULL);
    return(system(buf)>>8);
  }
}

boolean open_tape (filename)
string filename;
{
  char *cp, buf[BUFSIZ];
  int fds[2], i;
  tape_device = filename;

  if ((cp = index(filename, ':')) == NULL) {
    /* Open local tape device, then rewind */
    if ((tape_descriptor = open (filename, O_RDONLY)) < 0) {
      fprintf (stderr, "Error opening tape device %s: ", filename);
      perror (NULL);
      return (false);
    }
    if (rewind_tape () < 0) {
      fprintf (stderr, "Error rewinding tape device %s: ", filename);
      perror (NULL);
      close (tape_descriptor);
      return (false);
    }
  } else {
    /* Rewind remote tape device, then open */
    if (rewind_tape () < 0) {
      fprintf (stderr, "Error rewinding tape device %s: ", filename);
      perror (NULL);
      return (false);
    }
    /* Set up a pipe */
    if (pipe(fds) == -1) {
      perror("pipe");
      return(false);
    }
    /* Fork */
    if ((i = fork()) == -1) {
      /* Error */
      perror("fork");
      return(false);
    } else if (i == 0) {
      /* Child */
      if (dup2(fds[1], fileno(stdout)) == -1) {
	perror("dup2");
	exit(-1);
      }
      freopen("/dev/null", "r", stdin);
      *cp = '\0';
      sprintf(buf, "if=%s", cp+1);
      execlp("rsh", "rsh", filename, "dd", buf, NULL);
      perror("execlp(\"rsh\")");
      exit(-1);
    } else {
      /* Parent */
      tape_descriptor = fds[0];
      rsh_pid = i;
      signal(SIGCHLD, wait_for_child);
    }
  }
  tape_buffer_index = tape_buffer_fill = 0;
  return (true);
}

void next_tape ()
{
  close (tape_descriptor);
  if (rsh_pid != 0) {
    kill(rsh_pid, SIGTERM);
    rsh_pid = 0;
  }
  printf ("\n");
  do {
    do
      printf ("Is the next tape in place and are you ready to continue? (yes or no) ");
    while (!yes_or_no_p ());
  } while (!open_tape (tape_device));
}


/**** Restore from FEP-Tape ****/

property_list parse_header_property_list (header)
string header;
{
  unsigned char pathname [256];
  unsigned char name [32 + 1];
  unsigned char type [4 + 1];
  integer version;
  unsigned char author [64];
  unsigned char comment [256];
  string s;
  integer i;
  struct tm t;
  integer bytes, blocks = -1;
  integer month, day, year, hour, minute, second;
  property_list properties;

  pathname [0] = NULL;
  name [0] = NULL;
  type [0] = NULL;
  author [0] = NULL;
  comment [0] = NULL;
  s = header;
  while (*s != NULL) {
    sscanf (s, "Filename: %s", pathname);
    sscanf (s, "Length: %d bytes (%d blocks)", &bytes, &blocks);
    sscanf (s, "Created: %d/%d/%d %d:%d:%d", &month, &day, &year, &hour, &minute, &second);
    sscanf (s, "Author: %s", author);
    if (sscanf (s, "Comment: %s", comment)) {
      for (i = 0; s[i+10] != 012; i++) comment[i] = s[i+10];
      comment[i] = NULL;
    }
    while (*s >= ' ') s++;
    while (*s == 012) s++;
  }
  s = (char *) pathname;
  for (i = strlen (s) - 1; i > 0; i--)
    if (s [i] == '>' || s [i] == ':' || s [i] == ';') {
      s = s + i + 1;
      break;
    }
  sscanf (s, "%[^.].%[^.].%d", name, type, &version);
  t.tm_mon = month - 1;
  t.tm_mday = day;
  t.tm_year = year;
  t.tm_hour = hour;
  t.tm_min = minute;
  t.tm_sec = second;
  t.tm_wday = t.tm_yday = t.tm_isdst = t.tm_gmtoff = 0;
  properties = (property_list) malloc (sizeof (struct property_list_representation));
  properties->name = strdup (name);
  properties->type = strdup (type);
  properties->version = version;
  properties->creation_date = (integer) ((unsigned integer) timegm (&t) + 2208988800);
  properties->author = strdup (author);
  properties->comment = strdup (comment);
  properties->length_in_bytes = bytes;
  properties->length_in_blocks = blocks;
  properties->directory = false;
  properties->deleted = false;
  properties->dont_delete = false;
  return (properties);
}

property_list setup_file_restore ()
{
  unsigned char buffer [tape_block_size];
  property_list properties;
  integer i;

  read_tape (buffer, tape_block_size);
  for (i = 0; i < tape_block_size; i++)
    if (buffer [i] == 0215) buffer [i] = 012;
  return (parse_header_property_list (buffer));
}

void install_fep_kernel_if_appropriate (name, type, version)
string name;
string type;
integer version;
{
  char filename [64];

  /* If the file looks like a FEP kernel, offer to install it in the disk label */
  if (string_equal (type, "FEP") &&
      ((name[0] == 'i') || (name[0] == 'I')) &&
      string_equal (name + 4, "-KERNEL")) {
    sprintf (filename, "%s.%s.%d", name, type, version);
    /* Make sure the file is actually there before offering to install it */
    if (lookup_fep_file (dfs, filename))
      if (query (("Install %s as the FEP kernel to boot on this partition?", filename))) {
	printf ("Installing %s as the FEP kernel to boot on this partition...\n", filename);
	install_fep_kernel (dfs, filename);
      }
  }
}

boolean assure_free_space (filename, blocks)
string filename;
integer blocks;
{
  integer free = dfs->disk_size - bitmap_allocated (dfs->free_bitmap);

  /* Allow a few blocks slop for datamaps, etc */
  if (blocks > (free + 8)) {
    printf ("There isn't enough room in the file system to create %s.\n", filename);
    printf ("%s needs %d blocks, only %d free blocks remain.\n", filename, blocks, free);
    printf ("Continue, skipping this file? (yes or no) ");
    if (yes_or_no_p ())
      return (false);
    else exit (0);
  }
  return (true);
}

void update_file_properties (properties)
property_list properties;
{
  unsigned char buffer [tape_block_size];
  char filename [64];

  /* We've already extracted the properties from the initial description, skip this buffer */
  read_tape (buffer, tape_block_size);
  install_fep_kernel_if_appropriate (properties->name, properties->type, properties->version);
}

void get_continuation_tape (properties)
property_list properties;
{
  property_list new_properties;
  integer command_word, check_word;

 retry:
  next_tape ();
  command_word = read_tape_integer ();
  check_word = read_tape_integer ();
  if ((command_word != description_command) || (check_word != description_command)) {
    printf ("This tape was supposed to start with a description of %s.%s.%d\n",
	    properties->name, properties->type, properties->version);
    printf ("It doesn't start with a file description operation.\n");
    printf ("This is the wrong tape.\n");
    goto retry;
  }
  new_properties = setup_file_restore ();
  if (strcmp (properties->name, new_properties->name) ||
      strcmp (properties->type, new_properties->type) ||
      (properties->version != new_properties->version) ||
      (properties->creation_date != new_properties->creation_date)) {
    printf ("This tape was supposed to start with a description of %s.%s.%d\n",
	    properties->name, properties->type, properties->version);
    printf ("It actually starts with a description of %s.%s.%d\n",
	    new_properties->name, new_properties->type, new_properties->version);
    printf ("This is the wrong tape.\n");
    goto retry;
  }
  command_word = read_tape_integer ();
  check_word = read_tape_integer ();
  if ((command_word != data_command) || (check_word != data_command)) {
    printf ("This tape was supposed to start with a description of %s.%s.%d\n",
	    properties->name, properties->type, properties->version);
    printf ("It does, but the description isn't followed by a valid data segment.\n");
    printf ("This is the wrong tape.\n");
    goto retry;
  }
  free (new_properties);
}

void supervise_file_transfer (stream, properties)
fep_stream stream;
property_list properties;
{
  integer offset;
  integer count;
  integer total = 0;
  unsigned char *buffer;
  integer buffer_size = tape_block_size * 1024;
  integer transfer_size = 0;
  integer bytes;
  boolean first_time = true;

  buffer = (unsigned char *) malloc (buffer_size);
  while (true) {
    offset = read_tape_integer ();
    count = read_tape_integer ();
    if (first_time) {
      if (offset != 0) {
	printf ("\n");
	printf ("This tape file doesn't start at the beginning of %s.%s.%d.\n",
		properties->name, properties->type, properties->version);
	printf ("You may have the wrong tape.\n");
	printf ("Are you sure you want to proceed? (yes or no) ");
	if (!yes_or_no_p ()) {
	  printf ("The file has been created already, not deleting it.\n");
	  exit (0);
	}
	printf ("\n");
      }
      printf ("%s %s.%s.%d (%d blocks)...\n",
	      (stream == NULL) ? "Skipping" : "Restoring",
	      properties->name, properties->type, properties->version,
	      properties->length_in_blocks);
      first_time = false;
    }
    if (count == 0)
      if (total < properties->length_in_bytes) {
	printf ("End of tape reached, file %s.%s.%d continues on next tape.\n",
		properties->name, properties->type, properties->version);
	get_continuation_tape (properties);
	printf ("\n");
	printf ("Still %s %s.%s.%d (%d blocks remaining)...\n",
		(stream == NULL) ? "skipping" : "restoring",
		properties->name, properties->type, properties->version,
		(properties->length_in_bytes - total) / tape_block_size);
      } else break;
    else {
      if (stream != NULL)
	seek_fep_stream (stream, offset * tape_block_size);
      bytes = count * tape_block_size;
      while (bytes > 0) {
	if ((transfer_size = buffer_size) > bytes) transfer_size = bytes;
	read_tape (buffer, transfer_size);
	if (stream != NULL) {
	  bswap (buffer, buffer, transfer_size);
	  write_fep_stream (stream, buffer, transfer_size);
	}
	bytes -= transfer_size;
	total += transfer_size;
      }
      fsync (dfs->descriptor);
    }
  }
  free (buffer);
}

restore_from_tape ()
{
  integer command_word;
  integer check_word;
  fep_stream stream = NULL;
  property_list properties = NULL;
  char filename [64];

  while (true) {
    command_word = read_tape_integer (tape_descriptor);
    check_word = read_tape_integer (tape_descriptor);
    if (command_word != check_word) {
      printf ("Command word (%d) doesn't match check word (%d)\n", command_word, check_word);
      printf("This is the wrong tape.\n");
      close (tape_descriptor);
      return(-1);
    }
    switch (command_word) {
    case description_command:
      properties = setup_file_restore ();
      stream = NULL;
      if (query (("Restore file %s.%s.%d?",
		  properties->name, properties->type, properties->version))) {
	sprintf (filename, "%s.%s.%d",
		 properties->name, properties->type, properties->version);
	if (assure_free_space (filename, properties->length_in_blocks)) {
	  stream = open_fep_stream (dfs, filename, true);
	  set_fep_stream_properties (stream, properties);
	}
      }
      break;
    case fileheader_command:
      update_file_properties (properties);
      stream = NULL;
      properties = NULL;
      break;
    case data_command:
      supervise_file_transfer (stream, properties);
      break;
    case eof_command:
      printf ("End of tape encountered.\n");
      close (tape_descriptor);
      return(0);
    default:
      printf ("Unrecognized FEP-Tape command %d.\n", command_word);
      close (tape_descriptor);
      return(-1);
    }
  }
}


/**** Restore from disk ****/

/* FEP file system internals used to read through a directory */

disk_buffer allocate_disk_buffer ();
disk_address lookup_fep_directory ();
datamap read_fep_file_datamap ();
integer datamap_blocks ();
disk_address datamap_translate ();
void read_disk_block ();
void check_block_id ();
property_list get_fep_file_properties ();
integer fepd_nentries ();
string fepd_entry_name ();
string fepd_entry_type ();
integer fepd_entry_version ();
integer fepd_entry_address ();

void restore_from_disk ()
{
  disk_address directory_address;
  datamap directory_map;
  integer directory_blocks;
  integer block;
  disk_address fepd_address;
  disk_buffer fepd_buffer;
  integer entry;
  string name;
  string type;
  integer version;
  property_list properties;
  unsigned char *buffer;
  integer buffer_size = page_size * 16;
  integer transfer_size = 0;
  integer bytes;
  char filename [64];
  fep_stream from_stream, to_stream;

  buffer = (unsigned char *) malloc (buffer_size);
  fepd_buffer = allocate_disk_buffer ();
  directory_address = sfs->root_directory_address;
  directory_map = read_fep_file_datamap (sfs, directory_address);
  directory_blocks = datamap_blocks (directory_map);
  for (block = 0; block < directory_blocks; block++) {
    fepd_address = datamap_translate (directory_map, block);
    read_disk_block (sfs, fepd_buffer, fepd_address);
    check_block_id (fepd_buffer, "FEPD");
    for (entry = 0; entry < fepd_nentries (fepd_buffer); entry++) {
      name = fepd_entry_name (fepd_buffer, entry);
      type = fepd_entry_type (fepd_buffer, entry);
      version = fepd_entry_version (fepd_buffer, entry);
      if (string_equal (type, "FLOD") ||
	  string_equal (type, "ILOD") ||
	  (string_equal (type, "FEP") && string_equal (name + 4, "-KERNEL")))
	if (query (("Copy file %s.%s.%d?", name, type, version))) {
	  sprintf (filename, "%s.%s.%d", name, type, version);
	  from_stream = open_fep_stream (sfs, filename, false);
	  properties = get_fep_stream_properties (from_stream);
	  if (assure_free_space (filename, properties->length_in_blocks)) {
	    to_stream = open_fep_stream (dfs, filename, true);
	    set_fep_stream_properties (to_stream, properties);
	    printf ("Copying %s.%s.%d (%d blocks)...\n",
		    properties->name, properties->type, properties->version,
		    properties->length_in_blocks);
	    bytes = properties->length_in_blocks * page_size;
	    while (bytes > 0) {
	      if ((transfer_size = buffer_size) > bytes) transfer_size = bytes;
	      read_fep_stream (from_stream, buffer, transfer_size);
	      write_fep_stream (to_stream, buffer, transfer_size);
	      bytes -= transfer_size;
	    }
	    fsync (dfs->descriptor);
	    install_fep_kernel_if_appropriate
	      (properties->name, properties->type, properties->version);
	  }
	}
    }
  }
  deallocate_disk_buffer (fepd_buffer);
  free (buffer);
}

static property_list get_unix_file_properties (path, filename)
string path;
string filename;
{
  static struct property_list_representation plist_rep;
  static char buf[BUFSIZ];
  struct stat st;
  char *cp, *dp, *calloc();
  boolean capitalize_next = true;

  if (stat(path, &st) == -1) {
    perror("stat");
    return(NULL);
  }
  plist_rep.name = calloc(strlen(filename)+1, sizeof(char));
  for (cp = filename, dp = plist_rep.name; *cp != '\0' && *cp != '.'; cp++)
    /* Rudimentary name translation */
    if (*cp == '_') {
      *dp++ = '-';
      /* Special capitalization: f in from */
      if (*(cp+1) != 'f')
	capitalize_next = true;
    } else if (isalpha(*cp) && capitalize_next == true) {
      *dp++ = toupper(*cp);
      /* Special capitalization: X in UX */
      if (*cp == 'u')
	capitalize_next = true;
      else
	/* Default case */
	capitalize_next = false;
    } else {
      *dp++ = *cp;
      if (*cp == 'c' && *(cp+1) == 'i')
	/* Special capitalization: I in MacIvory */
	capitalize_next = true;
    }
  plist_rep.type = *cp == '.' ? cp+1 : "";
  plist_rep.version = 1;	/* *** */
  plist_rep.creation_date = st.st_ctime + 2208988800; /* If KHS can hardcode it, so can I */
  plist_rep.author = ""; /* *** */
  plist_rep.comment = "Restored from CD-ROM";
  plist_rep.length_in_bytes = st.st_size;
  plist_rep.length_in_blocks = ceiling (st.st_size, page_size);
  plist_rep.directory = false;
  plist_rep.deleted = false;
  plist_rep.dont_delete = false;

  return(&plist_rep);
}

restore_from_cdrom (cdrom, subdirectory, include_delivery_worlds)
string cdrom;
string subdirectory;
boolean include_delivery_worlds;
{
  property_list properties;
  unsigned char *buffer;
  integer buffer_size = page_size * 16;
  integer transfer_size = 0;
  integer bytes;
  char path[BUFSIZ], filename [64], *source, *type;
  fep_stream to_stream;
  DIR *dir;
  struct dirent *file;
  int source_fd, cc;

  sprintf(path, "%s/%s/", cdrom, subdirectory);
  source = &path[strlen(path)];
  if ((dir = opendir(path)) == NULL) {
    perror(path);
    return(-1);
  }
  buffer = (unsigned char *) malloc (buffer_size);
  for (file = readdir(dir); file != NULL; file = readdir(dir)) {
    strcpy(source, file->d_name);
    properties = get_unix_file_properties(path, file->d_name);
    /* Restore .FLOD and .FEP files.  Restore .ILODs if they
     * don't have "-MacIvory-" in their names (that signifies the
     * MacIvory Delivery world), and either they don`t have "-Del"
     * in their names (that signifies Delivery worlds in general),
     * or include_delivery_worlds is true. */
    if ((string_equal (properties->type, "flod") ||
	 (string_equal (properties->type, "ilod") &&
	  string_search (properties->name, "-MacIvory-") == false &&
	  (include_delivery_worlds == 
	   string_search (properties->name, "-Del"))) ||
	 (string_equal (properties->type, "fep") &&
	  string_equal (properties->name + 5, "Kernel")))) {
      sprintf (filename, "%s.%s.%d",
	       properties->name, properties->type, properties->version);
      if (query (("Copy file %s?", filename))) {
	if ((source_fd = open(path, O_RDONLY)) == -1) {
	  perror(path);
	} else {
	  if (assure_free_space (filename, properties->length_in_blocks)) {
	    to_stream = open_fep_stream (dfs, filename, true);
	    set_fep_stream_properties (to_stream, properties);
	    printf ("Copying %s.%s.%d (%d blocks)...\n",
		    properties->name, properties->type, properties->version,
		    properties->length_in_blocks);
	    bytes = properties->length_in_bytes;
	    while (bytes > 0) {
	      if ((transfer_size = buffer_size) > bytes)
		transfer_size = bytes;
	      if ((transfer_size = read(source_fd, buffer, transfer_size)) == -1) {
		perror("read");
		bytes = -1;
		break;
	      } else if (transfer_size == 0)
		break;
	      bswap (buffer, buffer, transfer_size);
	      write_fep_stream (to_stream, buffer, transfer_size);
	      bytes -= transfer_size;
	    }
	    fsync (dfs->descriptor);
	    install_fep_kernel_if_appropriate
	      (properties->name, properties->type, properties->version);
	  }
	  close(source_fd);
	}
      }
    }
  }
  closedir(dir);
  free (buffer);
  return(0);
}


/**** Boot file generation ****/

/* Support for writing out lines of the boot file, which have to be byte swapped */

static unsigned char *boot_stream_buffer;
static integer boot_stream_index;

fep_stream open_boot_stream (fs, name)
file_system fs;
string name;
{
  boot_stream_buffer = (unsigned char *) malloc (1280);
  bzero (boot_stream_buffer, 1280);
  boot_stream_index = 0;
  return (open_fep_stream (fs, name, true));
}

void write_boot_line (stream, s)
fep_stream stream;
string s;
{

  sprintf (boot_stream_buffer + boot_stream_index, "%s", s);
  boot_stream_index += strlen (s);
  boot_stream_buffer [boot_stream_index] = 0215;
  boot_stream_index += 1;
}

void close_boot_stream (stream)
fep_stream stream;
{
  unsigned char *buffer = boot_stream_buffer;
  property_list properties;

  bswap (buffer, buffer, 1280);
  write_fep_stream (stream, buffer, 1280);
  properties = get_fep_stream_properties (stream);
  properties->length_in_bytes = boot_stream_index;
  properties->author = "Create FEP Partition";
  set_fep_stream_properties (stream, properties);
  fsync (stream->fs->descriptor);
  free (boot_stream_buffer);
  boot_stream_buffer = NULL;
}

string get_internet_address ()
{
  char s [32];
  integer a, b, c, d;

  while (true) {
    a = b = c = d = -1;
    printf ("Enter the Internet network address for the UX: (kk.ll.mm.nn) ");
    scanf ("%d.%d.%d.%d", &a, &b, &c, &d);
    if ((a >= 0) && (b >= 0) && (c >= 0) && (d >= 0) &&
	(a < 256) && (b < 256) && (c < 256) && (d < 256)) {
      sprintf (s, "%d.%d.%d.%d", a, b, c, d);
      return (strdup (s));
    }
    printf ("An Internet address is a sequence of four decimal integers, each\n");
    printf ("less than 256, separated by periods.\n");
  }
}

string get_chaosnet_address ()
{
  char s [32];
  integer address;

  while (true) {
    address = -1;
    printf ("Enter the Chaosnet network address for the UX: (octal integer) ");
    scanf ("%o", &address);
    if ((address > 0) && (address < 65536)) {
      sprintf (s, "%o", address);
      return (strdup (s));
    }
    printf ("A Chaosnet address is an octal integer that is the concatenation\n");
    printf ("of an 8-bit subnet number and an 8-bit host number.\n");
  }
}

void create_hello_file (fep_version, chaosnet, internet)
integer fep_version;
string chaosnet;
string internet;
{
  char s [256];
  fep_stream stream;

  stream = open_boot_stream (dfs, "Hello.Boot.1");
  write_boot_line (stream, sprintf(s, "Hello I%d.Boot", fep_version));
  write_boot_line (stream, "Hello Local.Boot");
  close_boot_stream (stream);
  stream = open_boot_stream(dfs, sprintf(s, "I%d.Boot.1", fep_version));
  write_boot_line (stream, sprintf (s, "Scan I%d-Lisp", fep_version));
  write_boot_line (stream, sprintf (s, "Scan I%d-Loaders", fep_version));
  write_boot_line (stream, sprintf (s, "Scan I%d-Info", fep_version));
  write_boot_line (stream, sprintf (s, "Scan I%d-Debug", fep_version));
  write_boot_line (stream, "Initialize Hardware Tables");
  close_boot_stream (stream);
  stream = open_boot_stream(dfs, "Local.Boot.1");
  /* If there's a Paging-2.Page, assume there's a Paging-1.Page */
  if (lookup_fep_file (dfs, "Paging-2.Page.1"))
    write_boot_line (stream, "Declare Paging Files Paging-1,Paging-2");
  else if (lookup_fep_file (dfs, "Paging-1.Page.1"))
    write_boot_line (stream, "Declare Paging Files Paging-1");
  if (internet == NULL)
    internet = get_internet_address ();
  if (chaosnet == NULL)
    chaosnet = get_chaosnet_address ();
  write_boot_line (stream, sprintf (s, "Set Network Address CHAOS|%s,Internet|%s", chaosnet, internet));
  write_boot_line (stream, "Set Boot Options :IDS Enable");
  /*
  printf ("Do you want to use the default Standalone site? (yes or no) ");
  if (yes_or_no_p ()) {
    write_boot_line (stream, "Set Boot Options :Site Name Standalone");
    write_boot_line (stream, "Set Boot Options :Namespace Descriptor File UX400S|HOST:/usr/share/local/ivory/standalone-site.sct/standalone-namespace.text");
    printf ("\nThe Hello.Boot file now refers to namespace files in the\n");
    printf ("/usr/share/local/ivory/standalone-site.sct/* directory on the\n");
    printf ("host containing the UX400 processor.  Those files need to be\n");
    printf ("updated with the correct Internet address.  If the standalone site\n");
    printf ("directory was created elsewhere, the Hello.Boot file in the FEP\n");
    printf ("partition will need to be fixed.\n\n");
  }
  */
  close_boot_stream (stream);
}

void create_boot_file (automatic_p)
boolean automatic_p;
{
  disk_address directory_address;
  datamap directory_map;
  integer directory_blocks;
  integer block;
  disk_address fepd_address;
  disk_buffer fepd_buffer;
  integer entry;
  string name;
  string type;
  integer version;
  fep_stream stream;
  char s [64];

  if (automatic_p) {
    stream = open_boot_stream (dfs, "Boot.Boot.1");
    /*write_boot_line (stream, "Clear Machine");*/
    write_boot_line (stream, "Load World");
    write_boot_line (stream, "Start");
    close_boot_stream (stream);
  } else {
    fepd_buffer = allocate_disk_buffer ();
    directory_address = dfs->root_directory_address;
    directory_map = read_fep_file_datamap (dfs, directory_address);
    directory_blocks = datamap_blocks (directory_map);
    for (block = 0; block < directory_blocks; block++) {
      fepd_address = datamap_translate (directory_map, block);
      read_disk_block (dfs, fepd_buffer, fepd_address);
      check_block_id (fepd_buffer, "FEPD");
      for (entry = 0; entry < fepd_nentries (fepd_buffer); entry++) {
	name = fepd_entry_name (fepd_buffer, entry);
	type = fepd_entry_type (fepd_buffer, entry);
	version = fepd_entry_version (fepd_buffer, entry);
	if (string_equal (type, "ILOD")) {
	  printf ("Boot world %s.%s.%d by default? (yes or no) ", name, type, version);
	  if (yes_or_no_p ()) {
	    stream = open_boot_stream (dfs, "Boot.Boot.1");
	    /*write_boot_line (stream, "Clear Machine");*/
	    write_boot_line (stream, sprintf (s, "Load World %s", name));
	    write_boot_line (stream, "Start");
	    close_boot_stream (stream);
	    return;
	  }
	}
      }
    }
  }
}


/**** User interface ****/

void create_partition_file (name, bytes)
string name;
integer bytes;
{
  integer descriptor;
  struct stat stat;
  integer buffer_size = 1280 * 256;
  unsigned char *buffer;
  integer count;

  if ((descriptor = open (name, O_CREAT | O_RDWR | O_EXCL, 0777)) < 0) {
    perror ("Error creating file");
    exit (2);
  }
  buffer = (unsigned char *) malloc (buffer_size);
  bzero (buffer, buffer_size);
  /* A Sun 3/260 with SCSI disks writes local files at about 1/4 megabyte per second */
  printf ("Creating %d megabyte file %s...\n", bytes / (1024 * 1024), name);
  if (bytes > 10 * 1024 * 1024)
    printf ("This will take a few minutes (about %d, I'd guess).\n",
	    (bytes / (1024 * 1024 / 4)) / 60);
  while (bytes > 0) {
    if ((count = buffer_size) > bytes) count = bytes;
    if (write (descriptor, buffer, count) < 0) {
      perror ("Error writing file");
      exit (2);
    }
    bytes -= count;
  }
  fsync (descriptor);
  if (fstat (descriptor, &stat) < 0) {
    perror ("Error probing file");
    exit (2);
  }
  printf ("Created file %s, stored in %d Unix blocks (%d bytes).\n\n",
	  name,
	  stat.st_blocks/(stat.st_blksize/512),
	  stat.st_blksize);
  close (descriptor);
  free (buffer);
}

integer file_system_free_space (pathname, root_available_bytes)
string pathname;
integer *root_available_bytes;
{
  struct statfs s;
  integer i, length;
  char directory_name [256];
  integer descriptor;

  for (length = strlen (pathname); length > 0; length--)
    if (pathname[length] == '/') break;
  for (i = 0; i < length; i++)
    directory_name [i] = pathname [i];
  directory_name [i] = NULL;
  if ((descriptor = open (directory_name, O_RDONLY)) < 0) {
    fprintf (stderr, "Error opening directory %s: ", directory_name);
    perror (NULL);
    exit (2);
  }
  if (fstatfs (descriptor, &s) < 0) {
    perror ("Error determining file system parameters");
    exit (2);
  }
  close (descriptor);
  *root_available_bytes = s.f_bfree * s.f_bsize - 262144;
  return ( (getuid () == 0) ? *root_available_bytes : (s.f_bavail * s.f_bsize) - 262144 );
}

boolean bootstrap_partition_file (name)
string name;
{
  integer blocks = -1;
  integer available_blocks;
  integer available_bytes;
  integer root_available_bytes;
  integer created_p = false;

  while ((dfs = bootstrap_disk (name, false)) == NULL) {
    if (errno != ENOENT) {
      fprintf (stderr, "Couldn't open FEP partition file %s: ", name);
      perror (NULL);
      exit (2);
    }
    printf ("\n");
    printf ("The FEP partition file %s doesn't exist.\n", name);
    printf ("\n");
    available_bytes = file_system_free_space (name, &root_available_bytes);
    if (available_bytes < 0) available_bytes = 0;
    available_blocks = available_bytes / 1280;
    printf ("The file system containing %s has %d megabytes available.\n",
	    name, available_bytes / (1024 * 1024));
    printf ("This is enough space for %d Genera blocks (1280 bytes).\n", available_blocks);
    if ((getuid () != 0) && (root_available_bytes > available_bytes))
      /* Careful, the percentage calculation has to fit in 31 bits... */
      printf ("If you were logged in as `root', you could allocate %d%% more space.\n",
	      (root_available_bytes - available_bytes) / (root_available_bytes / 100));
    printf ("\n");
    while (blocks < 0) {
      printf ("How many blocks (1280 bytes) in the partition file? (integer) ");
      scanf ("%d", &blocks);
      if (blocks == 0) exit (0);
      else if (blocks < 1000) {
	printf ("A %d block partition file is too small to be useful.\n", blocks);
	blocks = -1;
      } else if (blocks * page_size > available_bytes) {
	/* The user may have deleted something in another process */
	available_bytes = file_system_free_space (name);
	available_blocks = available_bytes / 1280;
	if (blocks * page_size > available_bytes) {
	  printf ("Sorry, there's not enough space for a file that size.\n");
	  blocks = -1;
	}
      }
    }
    if (blocks < 50000)
      printf ("A %d block partition file is small by Genera standards.\n", blocks);
    else if (blocks > 200000)
      printf ("A %d block partition file is large by Genera standards.\n", blocks);
    printf ("Create partition file %s with %d blocks? (yes or no) ", name, blocks);
    if (!yes_or_no_p ()) exit (0);
    create_partition_file (name, (blocks * 1280));
    created_p = true;
  }
  return (created_p);
}

void create_paging_files (paging_space)
integer paging_space;
{
  integer free_space;
  integer paging_1;
  integer paging_2;

  if (!lookup_fep_file (dfs, "Paging-1.Page.1")) {
    /* --- should actually compute free space */
    free_space = dfs->disk_size - 1000;
    /* If paging space wasn't specified by the user, use approximately half the disk, rounded
       to the nearest 10000 blocks, unless the disk is unusually small or large. */
    if (paging_space <= 0) {
      if (free_space < 100000)
	paging_space = 30000;
      else
	paging_space = ((free_space / 2) / 10000) * 10000;
      if (paging_space > 120000) paging_space = 120000;
    }
    if (paging_space <= 30000) {
      paging_1 = paging_space;
      paging_2 = 0;
    } else {
      paging_2 = ((paging_space / 2) / 10000) * 10000;
      paging_1 = paging_space - paging_2;
    }
    if (paging_1 > 0) {
      if (assure_free_space ("Paging-1.Page.1", paging_1))
	if (query (("Create %d block paging file Paging-1.Page.1?", paging_1))) {
	  printf ("Creating paging file Paging-1.Page.1 (%d blocks)...\n", paging_1);
	  create_fep_file (dfs, "Paging-1.Page.1", paging_1);
	} else return;
    }
    if (paging_2 > 0) {
      if (assure_free_space ("Paging-2.Page.1", paging_2))
	if (query (("Create %d block paging file Paging-2.Page.1?", paging_2))) {
	  printf ("Creating paging file Paging-2.Page.1 (%d blocks)...\n", paging_2);
	  create_fep_file (dfs, "Paging-2.Page.1", paging_2);
	} else return;
    }
  }
}

void create_fep_partition (destination, source, paging_space, erase_p,
			   automatic_p, chaos_address, ip_address)
string destination;
string source;
integer paging_space;
boolean erase_p;
boolean automatic_p;
string chaos_address;
string ip_address;
{
  boolean created_p, first_time;
  integer fep_version;

  /* Check existence and validity of the specified partition file */
  created_p = bootstrap_partition_file (destination);

  if (erase_p && !created_p) {
    printf ("\n");
    printf ("You have specified that the FEP file system in %s be erased.\n", dfs->device);
    printf ("This will destroy the entire contents of the partition.\n");
    printf ("Are you sure you want to proceed? (yes or no) ");
    if (!yes_or_no_p ()) exit (0);
    printf ("Creating initial FEP file system in partition file %s...\n", dfs->device);
    create_initial_file_system (dfs);
  }

  /* Try to bootstrap the partition as a FEP file system, creating it if necessary */
  while (!bootstrap_file_system (dfs)) {
    if (!created_p && !automatic_p) {
      printf ("\n");
      printf ("Partition %s does not contain a viable FEP file system.\n", dfs->device);
      printf ("Creating a FEP partition in %s will destroy its present contents.\n", dfs->device);
      printf ("Are you sure you want to proceed? (yes or no) ");
      if (!yes_or_no_p ()) exit (0);
    }
    printf ("Creating initial FEP file system in partition file %s...\n", dfs->device);
    create_initial_file_system (dfs);
  }

  /* Create paging files */
  if (paging_space != 0)
    create_paging_files (paging_space);

  /* Restore FEP software and worlds */
  if (source != NULL) {
    if (tape_device_p (source)) {
      printf ("\n");
      first_time = true;
      do {
	if (automatic_p == false || first_time == false)
	  do {
	    printf ("Is the first tape in place and are you ready to proceed? (yes or no) ");
	  } while (!yes_or_no_p ());
	first_time = false;
      } while (!open_tape (source) ||
	       restore_from_tape() == -1);
    } else if (cdrom_device_p (source)) {
      if (restore_from_cdrom(source, CD_FLODS_DIRECTORY, false) == -1)
	exit(2);
      if (restore_from_cdrom(source, CD_WORLDS_DIRECTORY, false) == -1)
	exit(2);
      if (restore_from_cdrom(source, CD_WORLDS_DIRECTORY, true) == -1)
	exit(2);
    } else {
      if ((sfs = bootstrap_disk (source, true)) == NULL) {
	fprintf (stderr, "Couldn't open source partition file %s: ", source);
	perror (NULL);
	exit (2);
      }
      if (!bootstrap_file_system (sfs)) {
	printf ("Sorry, partition file %s does not contain a viable FEP file system.\n", source);
	exit (2);
      }
      restore_from_disk ();
    }
  }
  fep_version = lookup_fep_version (dfs);
  if ((automatic_p == false || chaos_address != NULL) &&
      (fep_version > 310) && (fep_version < 400)) {
    if (query (("Do you wish to create new Hello.Boot and Boot.Boot files?"))) {
      printf ("Creating Hello.Boot and Boot.Boot files...\n");
      create_hello_file (fep_version, chaos_address, ip_address);
      create_boot_file (automatic_p);
    }
  } else
    printf ("Warning: there is no FEP kernel installed in this partition.\n");
}

void main (argc, argv)
int argc;
char **argv;
{
  string destination = NULL;
  string source = NULL;
  string chaos_address = NULL, ip_address = NULL;
  integer paging = -1;
  integer size;
  boolean erase_p = false;
  boolean automatic_p = false;
  integer i;

  for (i = 1; i < argc && argv[i][0] == '-'; i++)
    switch (argv[i][1]) {
    case 'a':			/* Called from install-solstice */
      automatic_p = true;
      i++;
      paging = -1;
      erase_p = false;
      size = atoi(argv[i++]);
      switch(argv[i][0]) {
      case 'y':
	chaos_address = argv[++i];
	ip_address = argv[++i];
	break;
      case 'n':
	break;
      default:
	usage(argv[0]);
	break;
      }
      break;
    case 'q':
      query_p = true;
      break;
    case 'p':
      paging = atoi(argv[++i]);
      break;
    case 'e':
      erase_p = true;
      break;
    default:
      usage (argv[0]);
    }
  if (i == argc) usage (argv[0]);         /* Not enough args */
  destination = argv[i++];
  if (i < argc) source = argv[i++];
  if (i < argc) usage (argv[0]);          /* Too many args */

  if (automatic_p == true)
    create_partition_file(destination, size);
  create_fep_partition (destination, source, paging, erase_p,
			automatic_p, chaos_address, ip_address);
  exit(0);
}

usage (name)
string name;
{
  fprintf (stderr, "Usage: %s [-query] [-paging blocks] [-erase] partition-filename [source-filename]\n", name);
  exit (1);
}
