/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * $Log:	debug.c,v $
 * Revision 2.5.1.2  93/02/01  12:55:35  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:51:22  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:39:46  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/debug.c,v 2.5.1.2 93/02/01 12:55:35 kaufman Rel $";
#endif

#include <ctype.h>
#include <signal.h>
#include <strings.h>
#include <stdio.h>

static long ivory_debug_level = 0;

static char *dbg_levels[] = {
  "xmit", "drop", "recv", "init", "life", "buf",
  "if", "sig", "dbg1", "dbg2", "dbg3", "dbg4",
  "queue", "disk", "select", "process",
};
#define N_DBG_LEVELS (sizeof(dbg_levels)/sizeof(char *))


/* If this is a debug level, OR it in and return 1, otherwise return 0 */ 
debug_level (arg, debugp)
     char *arg;
     long *debugp;
{
  int j;

  if (debugp == NULL)
    debugp = &ivory_debug_level;

  for (j = 0; j < N_DBG_LEVELS; j++)
    if (!strcmp(arg, dbg_levels[j])) {
      *debugp |= 1<<j;
      return(1);
    }

  return(0);
}

#ifdef notdef
void set_debug_level (fd)
     int fd;
{
  IOCTL(MERLINDEBUG, &ivory_debug_level, "MERLINDEBUG")
}
#endif

void print_debug_levels (fp, which, terminator)
     FILE *fp;
     int which;
     char *terminator;
{
  int i;

  for (i = 0; i < N_DBG_LEVELS; i++)
    if (which == -1 || which & (1 << i))
      fprintf(fp, " %s", dbg_levels[i]);
  if (terminator != NULL)
    fputs(terminator, fp);
}

/* Set debug level interactively */
#ifdef SIGDEBUG
int cached_fd;

static void interactive_set_debug ()
{
  char buf[BUFSIZ], *cp, *dp;
  int current, new;
  int i, modifier, fd = cached_fd;
  EmbConfig config;
  
  IOCTL(IVORYGCONF, &config, "IVORYGCONF");
  current = config.debug_level;

  puts("\nInteractive debugging level frobber.  Type ^D to quit, Return to set levels.");
  puts("Warning: life support is disabled in this loop!");

  for (;;) {
    if (!current)
      puts("\nNo debug flags on");
    else {
      printf("\n%s debug flags are",
	     current == config.debug_level ? "Current" : "Proposed");
      print_debug_levels(stdout, current, "\n");
    }

    printf("Changes: ");
    if (fgets(buf, BUFSIZ, stdin) == NULL) {
      puts("\nNo changes made\n");
      clearerr(stdin);
      break;
    }
    for (cp = &buf[strlen(buf)-1]; isspace(*cp) && cp >= buf; cp--);
    *++cp = '\0';

    cp = buf;
    if (*cp == '\0') {
      config.debug_level = current;
      IOCTL(IVORYSCONF, &config, "IVORYSCONF");
      break;
    } else if (*cp == '+') {
      cp++;
      modifier = 1;
    } else if (*cp == '-') {
      cp++;
      modifier = -1;
    } else
      modifier = 0;

    new = 0;
    for (; *cp; cp = dp) {
      if ((dp = index(cp, ' ')) == NULL)
	dp = cp+strlen(cp);
      else
	*dp++ = '\0';
      
      for (i = 0; i < N_DBG_LEVELS; i++)
	if (!strcmp(cp, dbg_levels[i])) {
	  new |= 1<<i;
	  break;
	}
      if (i == N_DBG_LEVELS) {
	printf("%s is not a valid debugging level.  valid levels are:\n", cp);
	print_debug_levels(stdout, -1, "\n");
	modifier = -100;
	break;
      }
    }

    switch (modifier) {
    case 1:
      current |= new;
      break;

    case 0:
      current = new;
      break;

    case -1:
      current &= ~new;
      break;

    case -100:
      break;
    }
  }

  enable_interactive_set_debug(cached_fd);
}

void enable_interactive_set_debug (fd)
{
  cached_fd = fd;
  signal(SIGDEBUG, interactive_set_debug);
}
  
#endif
