/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * $Log:	describe.c,v $
 * Revision 2.5.1.2  93/02/01  12:55:36  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:51:24  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:39:50  kaufman
 * Baseline for Genera 8.3
 * 
 * 
 */

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/describe.c,v 2.5.1.2 93/02/01 12:55:36 kaufman Rel $";
#endif

#include <sys/types.h>
#include <sys/file.h>
#include <sys/mman.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>

#define INCLUDE_FILE_HEADERS
#include "embed.h"
#include "life.h"
#undef INCLUDE_FILE_HEADERS

main (argc, argv)
     int argc;
     char *argv[];
{
  char *ivory = NULL;
  char *getenv();
  int all_ivories_p = FALSE, channel_table_p = TRUE;
  int i, j;

  for (i = 1; i < argc; i++) {
    if (argv[i][0] != '-')
      if (ivory == NULL)
	ivory = argv[i];
      else
	usage();
    else
      for (j = 1; argv[i][j] != '\0'; j++)
	switch(argv[i][j]) {
	case 'a':
	  all_ivories_p = TRUE;
	  break;
	case 'h':
	  channel_table_p = FALSE;
	  break;
	default:
	  usage();
	  /*NOTREACHED*/
	}
  }

  if (ivory == NULL)
    if ((ivory = getenv("IVORY")) == NULL)
      ivory = DEFAULT_GUEST();

  if (all_ivories_p)
    describe_all_ivories(channel_table_p);
  else
    describe_ivory(ivory, channel_table_p, TRUE);

  exit(0);
}

describe_all_ivories (channel_table_p)
     int channel_table_p;
{
  DIR *dev;
  struct dirent *dir;
  char *ivory = "/dev/ivoryXXX";

  if ((dev = opendir("/dev")) == NULL) {
    perror("/dev");
    exit(1);
  }

  while ((dir = readdir(dev)) != NULL) {
    if (strncmp("ivory", dir->d_name, 5))
      continue;

    strcpy(ivory+5, dir->d_name);
    if (describe_ivory(ivory, channel_table_p, FALSE) != 1)
      putchar('\n');
  }

  closedir(dev);
}

/* returns -1 for error, 0 for success, 1 for silent error */
describe_ivory (ivory, channel_table_p, errors_fatal_p)
     char *ivory;
     int channel_table_p;
     int errors_fatal_p;
{
  IvoryConfig config;
  char buf[BUFSIZ];
  caddr_t ptr;
  int fd, locker;
  static int checked_kernel_version = FALSE;

  if ((fd = open(ivory, 0)) < 0) {
    if (errors_fatal_p || errno != ENXIO)
      perror(ivory);
    if (errors_fatal_p)
      exit(1);
    else
      return(errno == ENXIO ? 1 : -1);
  }
  if (!checked_kernel_version) {
    check_kernel_version();
    checked_kernel_version = TRUE;
  }

  if (ioctl(fd, IVORYGCONF, &config) < 0) {
    perror("IVORYGCONF");
    exit(1);
  }

  if ((int) (ptr = mmap(0, config.mmap_size, PROT_READ,
			MAP_SHARED, fd, 0)) == -1) {
    perror("mmap");
    exit(1);
  }
  ptr += config.comm_area_base;

  /* Check lock status */
  locker = ivory_locker(fd);
  if (locker == 0)
    printf("%s is not locked; ", ivory);
  else if (locker > 0)
    printf("%s is locked by pid %3d; ", ivory, locker);
  
  printf("life support is %sabled\n", config.enable_Ivory ? "en" : "dis");
  map_comm_area(ptr, config.host_ether_addr_offset, channel_table_p);
  return(0);
}

usage ()
{
  fputs("usage: describe-ivory [-a] [-h] [/dev/ivoryN]\n", stderr);
  exit(1);
}

ivory_locker(ivory_fd)
     int ivory_fd;
{
  struct flock lock;
  struct stat st;
  int fd;
  char buf[BUFSIZ];

  if (fstat(ivory_fd, &st) == -1) {
    perror("fstat");
    return(1);
  }

  (void) sprintf(buf, "/etc/ivory%d.pid", minor(st.st_rdev));
  if ((fd = open(buf, O_RDONLY)) == -1) {
    perror(buf);
    return(1);
  }

  lock.l_type = F_WRLCK;
  lock.l_whence = 0;
  lock.l_start = 0;
  lock.l_len = 0;
  if (fcntl(fd, F_GETLK, &lock) == -1) {
    perror("F_GETLK");
    close(fd);
    return(-1);
  }
  
  close(fd);
  if (lock.l_type != F_WRLCK)
    return(0);
  else
    return(lock.l_pid);
}

#ifdef notdef
/* Dump embedding status */
ivory_status(fd, snooze)
     int fd, snooze;
{
  caddr_t ptr;
  register EmbCommArea *ep;
  char *cp;
  int i;

  if ((ptr = mmap(0, IVORY_BUFSIZ, PROT_READ,
		  MAP_SHARED, fd, 0)) == (caddr_t) -1) {
    perror("mmap");
    exit(1);
  }

  ptr += IVORY_HALF_BUFSIZ;
  ep = (EmbCommArea *) ptr;

  printf("Embedding version %d\n", ep->version);
  printf("System type %s, host version %d.%d, guest %d.%d, FEP %d.%d\n",
	 assoc(ep->system_type, system_types),
	 ep->host_major_version, ep->host_minor_version,
	 ep->guest_major_version, ep->guest_minor_version,
	 ep->fep_major_version, ep->fep_minor_version);
  printf("Board version %d.%d #%d\n",
	 ep->board_major_version,
	 ep->board_minor_version,
	 ep->board_serial_number);
  printf("\n");
  
  for (i = 0; ; i++) {
    if (i % 20 == 0) {
      /* print header */
      puts("");
      puts("  Status           Disk physical:logical");
      puts("FEP   Lisp   Rst       read       write    rop wop mrg");
    }

    printf("%4s %6s  %2d   %5d:%5d %5d:%5d %3d %3d %3d\n",
	   assoc(ep->fep_status, fep_status),
	   assoc(ep->guest_status, guest_status),
	   ep->reset_request_count,
	   ep->NumberOfPhysicalReads,
	   ep->NumberOfLogicalReads,
	   ep->NumberOfPhysicalWrites,
	   ep->NumberOfLogicalWrites,
	   ep->NumberOfPhysicalReadOps,
	   ep->NumberOfPhysicalWriteOps,
	   ep->NumberOfPhysicalMerges);

    if (snooze == -1)
      break;
    else
      sleep(snooze);
  }

}
#endif
