/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * Basic subroutines for embedded host
 *
 * $Log:	embed.c,v $
 * Revision 2.5.1.2  93/02/01  12:55:43  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:51:32  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:40:08  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/embed.c,v 2.5.1.2 93/02/01 12:55:43 kaufman Rel $";
#endif

#include <sys/types.h>
#include <signal.h>

#include "embed.h"
#include "life.h"

#define INCLUDE_FILE_HEADERS
#include "symbolics_env.h"
#include "emblib.h"
#undef INCLUDE_FILE_HEADERS

static void ckq(), ckp();
static void panic();

EmbCommArea *EmbCommAreaPtr = NULL;

#ifdef sun

#include <sys/types.h>
#include <sys/uio.h>
#include <sys/file.h>
#include <sys/mman.h>
#include <sys/ioccom.h>
#include <errno.h>

static int fd = -1;

#endif /* sun */

/*** General ***/

MapEmbCommArea(device)
char *device;
{
  int i;
  
  if (EmbCommAreaPtr != NULL) panic("EmbCommArea already mapped.");
  
#ifdef sun
  {
    caddr_t ptr;
    IvoryConfig config;
    
    /* Open merlin device, map in slave buffer */
    if ((fd = open(device, O_RDWR)) < 0) {
      perror(device);
      return(-1);
    }
    if (ioctl(fd, IVORYGCONF, &config)<0) {
      perror("IVORYGCONF");
      return(-1);
    }
    if ((ptr = mmap(0, config.mmap_size, PROT_READ|PROT_WRITE,
		    MAP_SHARED, fd, 0)) == (caddr_t) -1) {
      perror("mmap");
      return(-1);
    }
    ptr += config.comm_area_base;
    EmbCommAreaPtr = (EmbCommArea *) ptr;
  }
#endif
  
  return(fd);
}

/* For thems as like to roll their own */
void RegisterEmbCommArea (fd_arg, EmbCommAreaPtr_arg)
     int fd_arg;
     EmbCommArea *EmbCommAreaPtr_arg;
{
  fd = fd_arg;
  EmbCommAreaPtr = EmbCommAreaPtr_arg;
}

void UnMapEmbCommArea ()
{
  EmbCommAreaPtr = NULL;
#ifdef sun
  if (fd != -1)
    close(fd);
  fd = -1;
#endif
}

EmbSendGuestInit ()
{
  return(ioctl(fd, IVORYINIT, 0));
}


EmbSendGuestReset ()
{
  return(ioctl(fd, IVORYRESET, 0));
}

void halt_lisp(notify_fn,force_p)
     void(*notify_fn)();
     int force_p;
{
#define NOTE(arg) if (notify_fn != NULL) (*notify_fn)(arg)
  int i;

  EmbCommAreaPtr->stop_request = -1;

  if (!force_p) {
    NOTE("Asking Lisp to stop...");

    /* Wait a bit before checking the first time */
    usleep(250000);
  
    /* Give Lisp 5 seconds to stop itself gracefully */
    for (i = 0; i < 10; i++) {
      if (EmbCommAreaPtr->fep_status == RunningFEPStatus) {
	NOTE("  Lisp stopped.");
	return;
      }
      usleep(500000);
    }

    NOTE("Lisp not stopped, asking FEP to take control...");
    if (EmbSendGuestReset() == -1) {
      perror("IVORYRESET");
      exit(1);
    }
  }

  for (i = 0; i < 10; i++) {
    if (EmbCommAreaPtr->fep_status == RunningFEPStatus) {
      if (!force_p)
	NOTE("  ");
      NOTE("Lisp forcibly stopped.");
      return;
    }
    usleep(500000);
  }

  NOTE("Failed to stop Lisp.");
  return;
#undef NOTE
}

#ifdef sun
/* Kludge kludge kludge */
static EmbPtr cached_select_chanptr = NullEmbPtr;

select_disable_child_hook ()
{
  static void handle_signals();

  cached_select_chanptr = NullEmbPtr;
  handle_signals(SIG_DFL);
}

select_disable_on_exit()
{
  if (cached_select_chanptr != NullEmbPtr)
    if (select_disable(cached_select_chanptr) == -1 && errno != ENODEV)
      perror("select_disable");
}

static void select_disable_on_signal (sig, code, scp, addr)
     /*ARGSUSED*/
     int sig, code;
     struct sigcontext *scp;
     char *addr;
{
  int mask;

  select_disable_on_exit();

  switch (sig) {
  case SIGQUIT:
  case SIGILL:
  case SIGTRAP:
  case SIGABRT:
  case SIGEMT:
  case SIGFPE:
  case SIGBUS:
  case SIGSEGV:
  case SIGSYS:
  case SIGLOST:
    /* Don't deprive user of a core dump */
    mask = sigblock(sigmask(SIGKILL));
    mask &= ~sig;
    sigsetmask(mask);
    signal(sig, SIG_DFL);
    kill(0, sig);
    /* Shouldn't get here */
    abort();

  default:
    /* Don't keep running */
    exit(1);
  }
}

static void handle_signals (handler)
     void (*handler)();
{
  signal(SIGHUP, handler);
  signal(SIGINT, handler);
  signal(SIGQUIT, handler);
  signal(SIGILL, handler);
  signal(SIGTRAP, handler);
  signal(SIGABRT, handler);
  signal(SIGEMT, handler);
  signal(SIGFPE, handler);
  signal(SIGKILL, handler);
  signal(SIGBUS, handler);
  signal(SIGSEGV, handler);
  signal(SIGSYS, handler);
  signal(SIGPIPE, handler);
/*  signal(SIGALRM, handler); */
  signal(SIGTERM, handler);
  signal(SIGXCPU, handler);
  signal(SIGXFSZ, handler);
  signal(SIGVTALRM, handler);
  signal(SIGPROF, handler);
  signal(SIGUSR1, handler);
  signal(SIGUSR2, handler);
}

select_enable (chanptr)
     EmbPtr chanptr;
{
  on_exit(select_disable_on_exit);
  handle_signals(select_disable_on_signal);
  cached_select_chanptr = chanptr;
  if (ioctl(fd, IVORYSELCHANNEL, &chanptr) == -1)
    return(-1);
  else
    return(fd);
}

select_disable (chanptr)
     EmbPtr chanptr;
{
  handle_signals(SIG_DFL);
  cached_select_chanptr = NullEmbPtr;
  return(ioctl(fd, IVORYUNSELCHANNEL, &chanptr));
}

EmbReset (ep)
     EmbCommArea *ep;
{
  static int reset_request_count = 0;

  if (EmbCommAreaPtr->reset_request_count != reset_request_count) {
    reset_request_count = EmbCommAreaPtr->reset_request_count;
    return(1);
  }
  return(0);
}

#endif /* sun */

/*** Pointers ***/

static void ckp(p)
EmbPtr p;
{
  if ((0 > (long)p) || ((long)p >= EmbCommAreaPtr->comm_memory_size))
    panic("Invalid embedded pointer.");
}  

void *EmbHostPtr(guest_ptr)
EmbPtr guest_ptr;
{
  ckp(guest_ptr);
  return(((long *)EmbCommAreaPtr) + guest_ptr);
}

EmbPtr EmbGuestPtr(host_ptr)
void *host_ptr;
{
  return(((long *)host_ptr) - ((long *)EmbCommAreaPtr));
}


/*** Signals ***/

void EmbSendSignal (num)
EmbSignalNumber num;
{
  if (num > -1)
    if (ioctl(fd, IVORYSIGNAL, &num) == -1)
      perror("IVORYSIGNAL");
}

void InstallEmbSignalHandler (software_signal, signal_ptr)
int software_signal;
EmbSignalNumber *signal_ptr;
{
}


/*** Queues ***/

static void ckq(qp)
EmbQueue *qp;
{
  ckp(EmbGuestPtr(qp));
}

long EmbQueueSpace (qp)	/* Number of free elements */
EmbQueue *qp;
{
  register long put = qp->put_index;
  register long take = qp->take_index;
  
  ckq(qp);
  
  if ( take > put )
    return ( take - put - 1 );
  else return ( take - put - 1 + qp->queue_size );
}
  
long EmbQueueFilled (qp) /* Number of non-free elements */
EmbQueue *qp;
{
  register long put = qp->put_index;
  register long take = qp->take_index;
  
  ckq(qp);
  
  if ( put >= take )
    return ( put - take );
  else return ( put - take + qp->queue_size );
}
  
void EmbQueuePut (qp_arg, ep)	/* Put element into queue */
EmbQueue *qp_arg;
void *ep;
{
  register EmbQueue *qp = qp_arg;
  register long put = qp->put_index;
  long original_put = put;
  register byte* element_array = (byte*) qp->first_element;
  
  ckq(qp);
  
  /* Fill in the element at the put index */
  memcpy(&element_array[put * qp->element_size], ep, qp->element_size);
  
  /* Compute the incremented put index */
  put++;
  if (put >= qp->queue_size)
    put = 0;
  
    /* Wait if queue is full.  Queue sizes will be chosen so that this is rare. */
  while (put == qp->take_index);
  
  /* Store the incremented put index into the queue header */
  qp->put_index = put;
  
  /* If the queue had been empty, send a signal to the taker. */
  if (original_put == qp->take_index)
    EmbSendSignal(qp->signal);
}
  
void EmbQueuePutWord (qp_arg, elt)	/* Put element into word queue */
EmbQueue *qp_arg;
long elt;
{ register EmbQueue *qp = qp_arg;
  register long put = qp->put_index;
  long original_put = put;
  register long* element_array = (long*) qp->first_element;
  
  ckq(qp);
  
  /* Fill in the element at the put index */
  element_array[put] = elt;
  
  /* Compute the incremented put index */
  put++;
  if ( put >= qp->queue_size )
    put = 0;
  
    /* Wait if queue is full.  Queue sizes will be chosen so that this is rare. */
  while ( put == qp->take_index );
  
  /* Store the incremented put index into the queue header */
  qp->put_index = put;
  
  /* If the queue had been empty, send a signal to the taker. */
  if ( original_put == qp->take_index )
    EmbSendSignal(qp->signal);
}
  
void EmbQueuePutByte (qp_arg, elt)	/* Put element into byte queue */
EmbQueue *qp_arg;
byte elt;
{ register EmbQueue *qp = qp_arg;
  register long put = qp->put_index;
  long original_put = put;
  register byte *element_array = (byte*) qp->first_element;
  
  ckq(qp);
  
  /* Fill in the element at the put index */
  element_array[put] = elt;
  
  /* Compute the incremented put index */
  put++;
  if ( put >= qp->queue_size )
    put = 0;
  
    /* Wait if queue is full.  Queue sizes will be chosen so that this is rare. */
  while ( put == qp->take_index );
  
  /* Store the incremented put index into the queue header */
  qp->put_index = put;
  
  /* If the queue had been empty, send a signal to the taker. */
  if ( original_put == qp->take_index )
    EmbSendSignal(qp->signal);
}
  
EmbBool EmbQueueTake (qp_arg, ep) /* Take element from queue */
EmbQueue *qp_arg;
void *ep;
{ register EmbQueue *qp = qp_arg;
  register long put = qp->put_index;
  register long take = qp->take_index;
  register byte *element_array = (byte*) qp->first_element;
  
  ckq(qp);
  
  /* Check for empty queue */
  if ( put == take )
    return ( EmbFalse );
  
    /* Copy the element at the take index */
  memcpy(ep, &element_array[take * qp->element_size], qp->element_size);
  
  /* Increment the take index and store it back */
  take++;
  if ( take >= qp->queue_size )
    take = 0;
  qp->take_index = take;
  return ( EmbTrue );
}
  
long EmbQueueTakeWord (qp_arg)	/* Take element from word queue */
EmbQueue *qp_arg;
{ register EmbQueue *qp = qp_arg;
  register long put = qp->put_index;
  register long take = qp->take_index;
  register long* element_array = (long*) qp->first_element;
  register long elt;
  
  ckq(qp);
  
  /* Check for empty queue--should not happen */
  if ( put == take )
    return ( EmbFalse );
  
    /* Copy the element at the take index */
  elt = element_array[take];
  
  /* Increment the take index and store it back */
  take++;
  if ( take >= qp->queue_size )
    take = 0;
  qp->take_index = take;
  return ( elt );
}
  
byte EmbQueueTakeByte (qp_arg)	/* Take element from byte queue */
EmbQueue *qp_arg;
{ register EmbQueue *qp = qp_arg;
  register long put = qp->put_index;
  register long take = qp->take_index;
  register byte *element_array = (byte*) qp->first_element;
  register byte elt;
  
  ckq(qp);
  
  /* Check for empty queue--should not happen */
  if ( put == take )
    return ( EmbFalse );
  
    /* Copy the element at the take index */
  elt = element_array[take];
  
  /* Increment the take index and store it back */
  take++;
  if ( take >= qp->queue_size )
    take = 0;
  qp->take_index = take;
  return ( elt );
}
  
  
  /*** Transferring multiple queue elements at a time ***/
  
  /* Each of these routines returns the number of elements actually transferred */
  /* None of these routines ever waits */
  
long EmbQueuePutBytes (qp_arg, buffer, length)	/* Put multiple bytes */
EmbQueue *qp_arg;
byte* buffer;
long length;
{ register EmbQueue *qp = qp_arg;
  register long put = qp->put_index;
  long original_put = put;
  register long take;
  register byte *element_array = (byte*) qp->first_element;
  long actual_length = 0;
  register long count;
  
  ckq(qp);
  
  /* Loop transferring contiguous blocks of elements */
  while ( length > 0 )
    { /* Fill in as many elements as we can */
      take = qp->take_index;
      if ( take > put )
	count = take - put - 1;
      else if ( take == 0 )
	  count = qp->queue_size - put - 1;
	else count = qp->queue_size - put;
      if ( count > length ) count = length;
      memcpy(&element_array[put], buffer, count);
      buffer += count;
      length -= count;
      actual_length += count;
      put += count;
      if ( put == qp->queue_size) put = 0;
      qp->put_index = put;
    };
  
    /* If the queue had been empty, send a signal to the taker. */
  if ( original_put == qp->take_index )
    EmbSendSignal(qp->signal);
  return ( actual_length );
}
  
long EmbQueuePutWords (qp_arg, buffer, length)	/* Put multiple words */
EmbQueue *qp_arg;
long* buffer;
long length;
{ register EmbQueue *qp = qp_arg;
  register long put = qp->put_index;
  long original_put = put;
  register long take;
  register long* element_array = (long*) qp->first_element;
  long actual_length = 0;
  register long count;
  
  ckq(qp);
  
  /* Loop transferring contiguous blocks of elements */
  while ( length > 0 )
    { /* Fill in as many elements as we can */
      take = qp->take_index;
      if ( take > put )
	count = take - put - 1;
      else if ( take == 0 )
	  count = qp->queue_size - put - 1;
	else count = qp->queue_size - put;
      if ( count > length ) count = length;
      memcpy(&element_array[put], buffer, count << 2);
      buffer += count;
      length -= count;
      actual_length += count;
      put += count;
      if ( put == qp->queue_size) put = 0;
      qp->put_index = put;
    };
  
    /* If the queue had been empty, send a signal to the taker. */
  if ( original_put == qp->take_index )
    EmbSendSignal(qp->signal);
  return ( actual_length );
}
  
long EmbQueueTakeBytes (qp_arg, buffer, length)	/* Take multiple bytes */
EmbQueue *qp_arg;
byte* buffer;
long length;
{ register EmbQueue *qp = qp_arg;
  register long take = qp->take_index;
  register long put;
  register byte *element_array = (byte*) qp->first_element;
  long actual_length = 0;
  register long count;
  
  ckq(qp);
  
  /* Loop transferring contiguous blocks of elements */
  while ( length > 0 )
    { /* Fill in as many elements as we can */
      put = qp->put_index;
      if ( put >= take )
	count = put - take;
      else count = qp->queue_size - take;
      if ( count > length ) count = length;
      memcpy(buffer, &element_array[take], count);
      buffer += count;
      length -= count;
      actual_length += count;
      take += count;
      if ( take == qp->queue_size) take = 0;
      qp->take_index = take;
    };
  return ( actual_length );
}

long EmbQueueTakeWords (qp_arg, buffer, length)	/* Take multiple words */
EmbQueue *qp_arg;
long* buffer;
long length;
{ register EmbQueue *qp = qp_arg;
  register long take = qp->take_index;
  register long put;
  register long* element_array = (long*) qp->first_element;
  long actual_length = 0;
  register long count;
  
  ckq(qp);
  
  /* Loop transferring contiguous blocks of elements */
  while ( length > 0 )
    { /* Fill in as many elements as we can */
      put = qp->put_index;
      if ( put >= take )
	count = put - take;
      else count = qp->queue_size - take;
      if ( count > length ) count = length;
      memcpy(buffer, &element_array[take], count << 2);
      buffer += count;
      length -= count;
      actual_length += count;
      take += count;
      if ( take == qp->queue_size) take = 0;
      qp->take_index = take;
    };
  return ( actual_length );
}


/*** Debugging ***/

describe_emb_queue (qp, name)
EmbQueue *qp;
char *name;
{
  int n = EmbQueueFilled(qp);
  
  /* Do it all in only two printf() calls, to save csw's */
  printf("Queue %s at emb %d\n  element-size  %d\n  queue-size    %d\n  put-index     %d\n  take-index    %d\n  signal-number %d\n",
	 name, EmbGuestPtr(qp),
	 qp->element_size,
	 qp->queue_size,
	 qp->put_index,
	 qp->take_index,
	 (long) qp->signal);
  if (n)
    printf("  contains %d elements\n", n);
  else
    printf("  is empty\n");
}

static void panic(message)
char *message;
{
  fprintf(stderr, "%s", message);
  exit(1);
}
