/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/* This file defines the various common data structures, including the
 * communication area and the channels.
 *
 * $Log:	embed.h,v $
 * Revision 2.5.1.2  93/02/01  12:55:45  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:51:34  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:40:15  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifndef _EMBED_INCLUDED
#define _EMBED_INCLUDED

#ifdef INCLUDE_FILE_HEADERS
#ifndef lint
static char *embed_rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/embed.h,v 2.5.1.2 93/02/01 12:55:45 kaufman Rel $";
#endif
char *embed_version = "@(#)$Revision: 2.5.1.2 $";
#endif

/*** Unpleasant kludges ***/

#define NINTERFACES_PER_IVORY 4	/* Max EMB_IFs per ivory device */


/*** Miscellaneous Types ***/

typedef long EmbPtr;		   /* "Pointer" to communication area = word offset */
typedef unsigned long SignalMask;  /* 32-bit bit mask of signals */
typedef long SignalNumber;	   /* Index into that bit mask */
typedef long bool;		   /* True if success, false if failure */
typedef unsigned char byte;	   /* byte = unsigned 8-bit byte */
typedef long Ptr;
typedef void* PtrV;		   /* PtrV is like Ptr but with better error checking */
typedef void (*ProcPtrV)();	   /* ProcPtrV is like ProcPtr but returns nothing */


/*** Communication Area ***/


/* Values for EmbCommArea.system_type */

enum system_type {
  SystemTypeUX400G = 01003,	/* SGI box with Merlin board in it */
  SystemTypeXL400 = 01004,	/* Standalone Merlin-I system */
  SystemTypeMacIvory = 01006,	/* Macintosh II running MacOS with Ivory board in it */
  SystemTypeUX400S = 01007,	/* Sun-3 or Sun-4 running SunOS with Merlin board in it */
  SystemTypeXL1200 = 01011,	/* Standalone Merlin-II system */
  SystemTypeUX1200S = 01012,	/* Sun-3 or Sun-4 running SunOS with Merlin-II board in it */
  SystemTypeUX1200G = 01013,	/* SGI box with Merlin-II board in it */
};

/* Values for EmbCommArea.reset_request */

enum ResetRequest {
  AreYouThereResetRequest = -4,	/* NOOP to see if host is alive */
  BootResetRequest = -3,	/* Boot ROM reset: initialize Device-PROM loading */
  DevicePROMResetRequest = -2,	/* Requests next FEP-buffer-size words */
  FEPResetRequest = -1,		/* Reset request issued by the FEP doesn't clear everything */
  NoResetRequest = 0,		/* Normal state */
  LispResetRequest = 1		/* Reset request issued by Lisp proper clears everything */
};

/* Values for EmbCommArea.guest_status */
enum GuestStatus {
  NonexistentGuestStatus = -2,	/* no guest hardware present */
  BrokenGuestStatus = -1,	/* host or guest initialization failed */
  UninitializedGuestStatus = 0,	/* communication area okay, hardware not initialized yet */
  InitializingGuestStatus,	/* an application is initializing the hardware */
  InitializedGuestStatus,	/* hardware initialized, IFEP loaded, not yet started */
  StartedGuestStatus,		/* Machine started and, no doubt, running in the IFEP */
  CrashedGuestStatus,		/* was fine, then the guest crashed and now in IFEP */
  RunningGuestStatus		/* everything is fine (i.e., Lisp is running) */
};

/* Values for EmbCommArea.fep_status */
#define HaltedFEPStatus 255	/* FEP has returned to the SPY (i.e., machine is halted) */
#define RunningFEPStatus  0	/* FEP is running */
#define IdleFEPStatus 1		/* FEP is idle (i.e., Lisp is running) */


typedef struct {
  caddr_t proc;			/* Process that handles signal */
  void (*fn)();			/* Function it should call */
  PtrV arg;			/* Argument to pass that function */
  short flag;			/* SIGF flags */
  short count;			/* Number of queues using this signal */
} SignalHandler;

/* Signal handler flag bits */
#define SIGF_INTERRUPT       001 /* Run handler at interrupt time */
#define SIGF_RESET           002 /* Run handler on resets */
#define SIGF_UNINITIALIZED   010 /* proc * is actually a PID (startup condition) */
#define SIGF_EMBPTRARG       020 /* arg is an EmbPtr that needs fixing at IVORYLIFE time */

#define InstallSignalHandler(ep,fn,arg) InstallSignalHandlerInternal(ep,fn,arg,0)

struct version {
  int trunk_major:8,
 trunk_minor:8,
 branch_major:8,
 branch_minor:8;
};


typedef struct {
  /* Overall Information */
  long identifier;		/* char[4] 'DBME' verifies this is the communication area */
				/* 'DBME' = 'EMBD' spelled backwards; the 68020 is big-ender */
  long version;			/* 1 identifies the communication architecture version */
  enum system_type system_type;	/* SystemTypeXxx identifies the host & guest systems jointly */

  long number_of_slots;		/* Number of guest-visible 32-bit words in this structure */
  long comm_memory_size;	/* Total number of 32-bit words in communication memory */

  /* Version Numbers of various software entities, in case anyone cares */
  long host_major_version;
  long host_minor_version;
  long guest_major_version;
  long guest_minor_version;
  long fep_major_version;
  long fep_minor_version;

  /* Memory Allocation */
  EmbPtr guest_buffer_start;	/* Portion of communication memory allocated by guest */
  long guest_buffer_size;	/* Number of 32-bit words in that */
  EmbPtr host_buffer_start;	/* Portion of communication memory allocated by host */
  long host_buffer_size;	/* Number of 32-bit words in that */
  EmbPtr fep_buffer_start;	/* Portion of communication memory used by IFEP/RDBG */
  long fep_buffer_size;		/* Number of 32-bit words in that */

  /* Signals */
  SignalMask guest_to_host_signals;		/* 1 if signal bit wants attention */
  SignalMask live_guest_to_host_signals;	/* 1 if signal bit is in use */
  SignalMask host_to_guest_signals;		/* 1 if signal bit wants attention */
  SignalMask live_host_to_guest_signals;	/* 1 if signal bit is in use */

  /* Channels */
  EmbPtr channel_table;		/* Head of threaded list of all channels */
  EmbPtr boot_disk_channel;	/* Disk channel from which the IFEP boots itself */
  EmbPtr cold_load_channel;	/* Cold load stream channel */
  EmbPtr command_channel;	/* Master remote-procedure-call channel */

  /* Specific to MacIvory */
  EmbPtr main_memory_map;	/* Pointer to map of Ivory locations that exist */
  long main_memory_map_size;	/* Number of 64-bit entries in that map */
  EmbPtr bad_memory_map;	/* Pointer to map of holes in Ivory memory */
  long bad_memory_map_size;	/* Number of 64-bit entries in that map */

  /* Stuff that was added later and isn't really in the right place */
  SignalNumber clock_signal;	/* Send this signal to simulate a clock interrupt */
  long clock_interval;		/* Approximate clock interrupt interval in microseconds */

  long run_lights;		/* Right-justified bit mask of run lights */
  enum ResetRequest reset_request; /* Set by guest to request reset, cleared by host */
  long board_serial_number;	/* Guest board serial number */
  long board_major_version;	/* Guest board major revision level */
  long board_minor_version;	/* Guest board minor revision level */
  long spy_command;		/* Debugging */
  long spy_status;		/* Debugging */
  long stop_request;		/* Set to force guest to halt */
  union {
    long fepstat_word;
    struct {
      int fepstat_pad:20,	/* FEP status bit fields: */
    fepstat_lisp_is_loaded:1,	/*   Set if FEP thinks Lisp will work */
    fepstat_error:1,		/*   Set when a FEP command errs */
    fepstat_busy:1,		/*   Set when FEP not at command prompt */
    fepstat_cursor:1,		/*   Toggled when FEP cursor should blink */
    fepstat_status:8;		/*   FEP's idea of guest's status */
    } bits;
  } fepstat;
#define fep_lisp_is_loaded fepstat.bits.fepstat_lisp_is_loaded
#define fep_error fepstat.bits.fepstat_error
#define fep_busy fepstat.bits.fepstat_busy
#define fep_cursor fepstat.bits.fepstat_cursor
#define fep_status fepstat.bits.fepstat_status
  long restart_applications;	/* Set by guest to request host restart its applications */
  union {
    long siv_word;		/* Tells guest more precisely how to interrupt the host */
    struct {
      int siv_pad:21,
    siv_level:3,		/*   (Solstice: VME interrupt priority) */
    siv_vector:8;		/*   (Solstice: VME interrupt vector) */
    } bits;
  } siv;
#define signal_interrupt_level siv.bits.siv_level
#define signal_interrupt_vector siv.bits.siv_vector
  long base_register;		/* More MacIvory garbage */
  struct version host_os_version; /* UNIX version numbers */
  struct version boot_version;	/* Version of program that booted Ivory */

  /* Remainder of structure is intra-Unix globals, not known to the Ivory */

  /* Padding to allow for convenient expansion of shared portion later */
  long pad0,			/* Special padding slot used by INIT code */
       pad1[10];

  /* To add slots to the shared portion of the communications area, add the slots before 
     the pad0 field, and decrease the size of the pad1 field by the number of slots added.
     As long as spare slots remain, you won't have to recompile all of the host software;
     just those files which you modify to use the new slots.

     However, if you do use up all the remaining pad slots, instead of leaving no padding,
     leave a minimum of 16 pad slots (one in pad0, 15 in pad1).

     Be sure that any changes you make to the shared portion of the communications area
     are reflected in the sources of all host systems (Solstice, MacIvory) as well as by 
     appropriate patches in Lisp.
  */

  /* Embedding hardware */
  enum GuestStatus guest_status; /* A GuestStatus code */
	/* --- Other hardware-oriented stuff goes here */

  /* Table of Signal Handlers */
  SignalHandler SignalHandler[32]; /* How to handle each signal */

  /* Clock interrupts */
  int poll_tick;		/* Counts ticks between polling all signal handlers */
  int signal_tick;		/* Counts ticks between clock interrupts sent to Ivory */
  SignalMask reawaken;		/* Signals to try again on the next clock tick */

#ifdef MACIVORY_DISK_HACKS
  /* Disk Cache */
    /* These variables are the parameters that we can twiddle */
    /* Don't make these ints, even though they would fit, LSC just loses totally... */
  long CacheSize;		/* Number of lines */
  long LineSize;		/* Number of physical sectors per line */
  long GapSize;			/* Minimum gap to avoid missing rotation between requests */

    /* Arrays indexed by cache line number */
  long* CacheTag;		/* First sector# in line, or -1 if invalid */
  byte** CacheData;		/* Array of buffers (must be allocated contiguously!) */
  long FifoPointer;		/* FIFO replacement algorithm: next line to reuse */

    /* Metering */
  long NumberOfHits;
  long NumberOfMisses;
  long NumberOfLogicalReads;
  long NumberOfLogicalWrites;
  long NumberOfPhysicalReads;
  long NumberOfPhysicalWrites;
  long NumberOfPhysicalMerges;
  long NumberOfPhysicalReadOps;
  long NumberOfPhysicalWriteOps;
#endif
  
  long reset_request_count;	/* # of times that guest's Lisp (not FEP) requested a reset */
  long restart_applications_count;	/* # of times we've been requested to restart */

  bool inhibit_disk;		/* If true, disk life support doesn't do anything */
  int debug_level;		/* If true, debugging breakpoints are turned on */

#ifdef MACIVORY_DISK_HACKS
  long maximum_disk_io_time;	/* Maximum time (in 1/60ths) for continuous disk I/O */
  long disk_io_throttle_interval;	/* Time (in 1/60ths) to "rest" during heavy disk I/O */
  long remaining_disk_io_throttle_time;	/* Amount of time left to throttle disk I/O */
  long number_of_disk_throttles;	/* Counter of how often we've throttled disk I/O */
#endif

  /* Network stuff */
  PtrV host_ether_addrs;	/* Pointer to table of host ethernet addresses */
  long n_host_ether_addrs;	/* Length of table */

  /* Pointer to location to frob to interrupt Genera */
  long *slave_trigger;

  long device_prom_reset_request_count; /* Count of Boot and DevicePROM resets */

  /*  Add new things at the end of this structure since several
      applications/code resources use this data.
   */
} EmbCommArea;


/*** Pointers ***/

/* Null pointer of EmbPtr type */
#define NullEmbPtr (EmbPtr)(-1)

/* Convert a guest pointer (EmbPtr) to a host pointer */
/* Assumes EmbCommAreaPtr is an accessible variable */
#define HostPointer(p) (PtrV)(&((long*)EmbCommAreaPtr)[p])

/* Convert a host pointer to a guest pointer (EmbPtr) */
/* Assumes EmbCommAreaPtr is an accessible variable */
#define GuestPointer(p) (EmbPtr)((long*)p - (long*)EmbCommAreaPtr)


/*** Signals ***/

/* Send a host-to-guest signal */
void EmbSendSignal();
SignalNumber InstallSignalHandlerInternal();


/*** Queues ***/

typedef struct {
  long element_size;		/* Number of bytes per queue element */
  long queue_size;		/* Number of elements in the queue */
  long put_index;		/* Index of next element to be written */
  long take_index;		/* Index of next element to be read */
  SignalNumber signal;		/* Send this signal on empty => non-empty transition
				   -1 if no signal used */
  long first_element[1];	/* Array of elements starts here */
} EmbQueue;

void EmbQueuePut();		/* Put element into queue */
bool EmbQueueTake();		/* Take element from queue */
long EmbQueueSpace();		/* Number of free elements */
long EmbQueueFilled();		/* Number of non-free elements */
void EmbQueuePutWord();		/* Put element into word queue */
long EmbQueueTakeWord();	/* Take element from word queue */
void EmbQueuePutByte();		/* Put element into byte queue */
byte EmbQueueTakeByte();	/* Take element from byte queue */
long EmbQueuePutWords();	/* Multi-word */
long EmbQueueTakeWords();
long EmbQueuePutBytes();	/* Multi-byte */
long EmbQueueTakeBytes();


/*** Channels ***/

enum EmbChannelType {
  EmbDiskChannelType = 1,	/* Disk drive/partition */
  EmbConsoleChannelType,	/* Display, keyboard, and mouse */
  EmbNetworkChannelType,	/* Packet Network */
  EmbCommandChannelType,	/* Remote Procedure Call */
  EmbSCSIChannelType,		/* IFEP SCSI I/O */
  EmbColdLoadChannelType, 	/* Cold load stream */
  EmbHostFileChannelType,	/* IFEP access to host file system */
  EmbMessageChannelType,	/* General purpose stream */
  EmbKludgeChannelType = 66000,	/* Force this to be a long word type */
};

typedef struct {
  enum EmbChannelType type;	/* Code number of type of channel */
  long unit;			/* Distinguishes multiple channels of same type */
  EmbPtr next;			/* Next channel in list of all channels */
  /* Remainder of structure depends on the channel type */
} EmbChannel;


/*** Disk Channels ***/

typedef struct {
  /* Standard channel fields */
  enum EmbChannelType type;	/* Code number = EmbDiskChannelType */
  long unit;			/* Logical drive number, may not equal physical drive numb */
  EmbPtr next;			/* Next channel in list of all channels */

  /* Disk-specific channel fields */
  long number_of_pages;		/* Size of disk in Ivory pages */
    /* -- more disk characteristics fields to be added later */
  EmbPtr command_queue;		/* Guest to Host */
  EmbPtr status_queue;		/* Host to Guest */

  /* Fields not visible to Ivory */
  EmbQueue *command_queue_ptr;	/* Redundant but more convenient pointer */
  EmbQueue *status_queue_ptr;	/* .. */
  int initialized_p;

  long error_pending;		/* Nonzero if channel is in error and needs to be reset */
  struct file *FileStructurePointer;	/* Unix file partition is in */
  /* These have to come last */
#if defined(_UIO_) || defined(_sys_uio_h)
  struct uio UserIOBlock;	/* Parameter Block for I/O operations */
  struct iovec IOVectorTable[16]; /* Scatter/gather vectors */
#endif
} EmbDiskChannel;

/* The queues contain Pointers to disk command blocks defined as follows: */

typedef struct {	/* Disk operation, assume bit fields packed left to right */
  unsigned int pad1 : 16;	/* Padding */
  unsigned int pad2 : 7;	/* Padding */
  unsigned int suppress_error_recovery : 1;  /*  (not yet implemented, but should be) */
  unsigned int pad3 : 3;	/* Padding */
  unsigned int buffered : 1;	/* 0 => to main memory (not implemented), 1 => to buffer */
  unsigned int tagged : 1;	/* 0 => 32-bit data, 1 => 40-bit data (not implemented) */
  unsigned int cmd : 3;		/* EmbDiskCmd code */
} EmbDiskOperation;

enum EmbDiskCmd { ReadCmd = 1, WriteCmd, ResetCmd, InitializeCmd };

enum EmbDiskStatus { WonStatus = 1, LostStatus, AbortStatus };

typedef struct {
  long id;		/* 32 bits meaningful only to guest */
  long sync;		/* Two commands with the same value here must be executed in order */
  EmbDiskOperation op;	/* Command */
  long page;		/* Starting disk address in pages (-- not words??) */  
  long count;		/* Length of transfer in pages */
  long n_addresses;	/* Number of memory address pairs */
  enum EmbDiskStatus status;
  long error_code;	/* If status == LostStatus, this is an OSErr */
  struct EmbAddressPair {
    EmbPtr address;	/* Address of buffer */
    long n_words;	/* Number of words at that address */
   } addresses[1];		/* Actually of indefinite size */
} EmbDiskQueueElement;

/* 
 * Network channel defs
 */

typedef struct {
  /* the canned channel preamble */
  enum EmbChannelType type;	/* channel type; EmbNetworkChannelType */
  long unit;			/* net int'fc num, probably only one */
  EmbPtr next;			/* link to next channel */
  /* the rest is specific to network channels.  For now I assume ethernet, this will
     probably want to get more structured later */
  long status;			/* status bits, defined below */
  EmbPtr xmit_guest_to_host_queue; /* guest to host xmitter que handle */
  EmbPtr xmit_host_to_guest_queue; /* host to guest xmitter que handle */
  EmbPtr recv_guest_to_host_queue; /* guest to host receiver que handle */
  EmbPtr recv_host_to_guest_queue; /* host to guest reciever que handle */
  /* Names and addresses */
  long name_0;			/* First part of interface name */
  long name_1;			/* Second part of interface name */
  long hardware_address_high;	/* high 32 bits; sys:net-address-1 */
  long hardware_address_low;	/* low 16 bits; sys:net-address-2 */
  long host_primary_protocol, host_primary_address;
  long guest_primary_protocol, guest_primary_address;
  /* Meters (first three from struct ifnet in <net/if.h>) */
  int if_ierrors;		/* input errors on interface */
  int if_oerrors;		/* output errors on interface */
  int if_collisions;		/* collisions on csma interfaces */
  int if_dropped;		/* packets dropped due to emb buf underflow */
  long other_meters[4];
  /* fields used only by host */
  EmbQueue *xmit_guest_to_host_queue_ptr; /* more useful ptr to above queue */
  EmbQueue *xmit_host_to_guest_queue_ptr;
  EmbQueue *recv_guest_to_host_queue_ptr;
  EmbQueue *recv_host_to_guest_queue_ptr;
  int initialized_p;
  /* Network interface stuff */
  PtrV interface;
  PtrV network_queue;
#ifdef IVORY_EMBIF
  long emb_if_unit;
#else
  long place_holder;
#endif
} EmbNetChannel;

/* Status bits that appear in the 'status' word of the above struct */
#define EmbNet_status_host_ready 1		/* means host is ready for business */
#define EmbNet_status_guest_ready 2

/* Maximum size of the network frames that pass through this channel */
#define EPACKETDATALEN 1514


/*** Cold Load Stream ***/

#define COLD_LOAD_STRING_TOTAL_SIZE 256
#define COLD_LOAD_HISTORY_SIZE 128

typedef struct {
  enum EmbChannelType type;	/* EmbColdLoadChannelType */
  long unit;			/* 0 (not used) */
  EmbPtr next;			/* link to next channel */
  EmbPtr keyboard_input_queue;	/* words containing keystrokes */
  EmbPtr display_output_queue;	/* words containing display commands */
  long display_width;		/* width in pixels of window */
  long display_height;		/* height in pixels of window */
  long character_width;		/* width in pixels of a character */
  long line_height;		/* height in pixels of a line */
  long progress_numerator;	/* as in real progress notes */
  long progress_denominator;	/* " */
  long string_total_size;	/* Max length of string in string-char's */
  long string_length;		/* When non-zero, displays a note */
  char string[COLD_LOAD_STRING_TOTAL_SIZE]; /* String */
  /* The remaining fields are not visible to Ivory */
  struct proc *selproc;		/* State for select() */
  long command_history_index, command_history_wrapped_p;
  long command_history[COLD_LOAD_HISTORY_SIZE];	/* History of CL commands */
} EmbColdLoadChannel;


/*** Remote Procedure Call  ***/

typedef struct {
  enum EmbChannelType type;			/* EmbCommandChannelType */
  long unit;					/* probably not used, always zero */
  EmbPtr next;					/* link to next channel */
  EmbPtr guest_to_host_queue;
  EmbPtr host_to_guest_free_queue;
  EmbPtr host_to_guest_queue;
  EmbPtr guest_to_host_free_queue;
  bool listening;				/* True if the host is listening */
  /* fields used only by host */
  EmbQueue* guest_to_host_queue_ptr;		/* more useful ptr to above queue */
  EmbQueue* host_to_guest_free_queue_ptr;
  EmbQueue* host_to_guest_queue_ptr;
  EmbQueue* guest_to_host_free_queue_ptr;
} EmbCommandChannel;

/*** SCSI I/O Channels ***/

typedef struct {
  /* Standard channel fields */
  enum EmbChannelType type;	/* EmbSCSIChannelType */
  long unit;			/* SCSI device unit number */
  EmbPtr next;			/* Next channel in list of all channels */

  /* SCSI channel specific fields */
  EmbPtr command_queue;		/* Guest to Host */
  EmbPtr reply_queue;		/* Host to Guest */

  /* Fields not visible to Ivory */
  EmbQueue* command_queue_ptr;	/* Redundant but more convenient pointer */
  EmbQueue* reply_queue_ptr;	/* .. */
  long program_id;		/* unique ID of "owner" of the SCSI channel */
} EmbSCSIChannel;

/* Queues contain pointers to the following command/reply blocks */

typedef struct {
  long opcode;			/* Type of operation (an EmbSCSIComandOpCode) */
  long uid;			/* Unique ID for consistency checks */
  long buffer_address;		/* Host address of buffer for I/O */
  long buffer_size;		/* Size of buffer in bytes */
  long error_code;		/* Host specific error code if non-zero */
  long error_data;		/* Host dependent additional info about error */
  long status;			/* SCSI status returned by operation */
  long message;			/* SCSI message returned by operation */
  long command_word_1;		/* First word of actual SCSI command ... */
  long command_word_2;		/* ... 2nd word ... */
  long command_word_3;		/* ... 3rd word */
} EmbSCSICommand;

enum EmbSCSICommandOpCode { 
  EmbSCSICommandNOOP,		/* Synchronize */
  EmbSCSICommandStartup,	/* Setup for SCSI operations */
  EmbSCSICommandShutdown,	/* Cleanup after SCSI operations */
  EmbSCSICommandMaximumBufferSize,	/* Return maximum size of host buffer */
  EmbSCSICommandAllocateBuffer,		/* Allocate a buffer in host memory */
  EmbSCSICommandDeallocateBuffer,	/* Release said buffer */
  EmbSCSICommandIssueControlCommand,	/* Execute a SCSI control class command */
  EmbSCSICommandIssueReadCommand,	/* Execute a SCSI read class command */
  EmbSCSICommandIssueWriteCommand,	/* Execute a SCSI write class command */
  EmbSCSICommandResetSCSIBus		/* Reset the SCSI bus */
 };

/*** Host File access from IFEP ***/

typedef struct {
  long type;			/* EmbHostFileChannelType */
  long unit;			/* 0 (not used) */
  EmbPtr next;			/* link to next channel */
  EmbPtr command_queue;		/* EmbPtr to queue of commands from guest */
  EmbPtr reply_queue;		/* EmbPtr to queue of replies from host */
  /* The remaining fields are not visible to Ivory */
  struct proc *selproc;		/* State for select() */
} EmbHostFileChannel;

/*  Common structure of both host file access commands and their replies. */
#define EmbHostFileCommandHeader \
  long opcode;				  /* Type of operation (see below) */       \
  long uid;				  /* Unique ID for consistency checks */    \
  long host_reference;			  /* Host's reference for target opening */ \
  long error_code;			  /* Host independent error code if non-zero */ \
  long host_error_code			  /* Host dependent error code */

typedef struct {
  EmbHostFileCommandHeader;		/* common header */
} EmbHostFileCommand;

enum EmbHostFileCommandOpcodes {
  EmbHostFileCommandNoop,		/* Synchronize */
  EmbHostFileCommandStartup,		/* Setup for host file operations */
  EmbHostFileCommandShutdown,		/* Cleanup after host file operations */
  EmbHostFileCommandOpen,		/* Open a file/directory */
  EmbHostFileCommandClose,		/* Close the file/directory */
  EmbHostFileCommandReadData,		/* Read data */
  EmbHostFileCommandWriteData,		/* Write data */
  EmbHostFileCommandReadDirectory,	/* Read directory */
};

typedef struct {
  EmbHostFileCommandHeader;		/* common header */
  long direction;			/* One of the EmbHostFileOpenDirections below */
  long pathname_length;			/* Length in bytes of following pathname */
  unsigned char pathname[1];		/* First word of the pathname */
} EmbHostFileOpenCommand;

enum EmbHostFileOpenDirections {
  EmbHostFileOpenInput,
  EmbHostFileOpenOutput,
  EmbHostFileOpenDirectory,		/* Directory listing (implicitly input). */
};

typedef struct {
  EmbHostFileCommandHeader;		/* common header */
  long close_mode;
  long open_direction;			/* One of the EmbHostFileOpenDirections above */
} EmbHostFileCloseCommand;

enum EmbHostFileCloseModes {
  EmbHostFileCloseNormal,
  EmbHostFileCloseAbort,
};

typedef struct {
  EmbHostFileCommandHeader;		/* common header */
  long transfer_position;		/* Byte offset in file or -1 for current offset */
  long transfer_length;			/* Length in bytes of requested transfer */
					/* For read, overwritten with actual count */
  char transfer_data[1];		/* First word of the buffer */
} EmbHostFileTransferCommand;

typedef struct {
  EmbHostFileCommandHeader;		/* common header */
  long index;				/* Which file in directory is desired */
  long length;				/* Length in bytes of requested transfer, */
							/* overwritten with actual count */
  long size_in_blocks;			/* Host dependent size */
  long length_in_bytes;			/* 8-bit bytes, that is */
  long flags; 					/* (see below) */
  long creation_date;			/* Universal time (host must convert) */
  long name_length;				/* Length of name,type,version string */
  long comment_length;			/* Length of comment string */
  long author_length;			/* Length of author string */
  char text[1];		       	/* First word of name, comment, author concatenated data */
} EmbHostFileDirectoryEntry;

enum EmbHostFileDirectoryEntryFlags {
  EmbHostFileDirectoryEntryFlagDirectory   = 1,
  EmbHostFileDirectoryEntryFlagDeleted     = 2,
  EmbHostFileDirectoryEntryFlagFixed_pages = 4,
  EmbHostFileDirectoryEntryFlagRead_only   = 8,
  EmbHostFileDirectoryEntryFlagLink        = 16,
};

enum EmbHostFileErrorCodes {
  EmbHostFileNoError = 0,
  EmbHostFileUnknownOpcode,
  EmbHostFileUnknownOpenMode,
  EmbHostFileUnknownCloseMode,
  EmbHostFileBadReference,	/* IFEP supplied an invalid file reference */
  EmbHostFileNotAFile,		/* Attempt to do a file operation on a directory */
  EmbHostFileNotADirectory,	/* Attempt to do a directory operation on a file */
  EmbHostFileWrongDirection,	/* Attempt to read from output or vice-versa */
  EmbHostFileOutOfBounds,	/* Attempt to read outside file or directory */
  EmbHostFileFileNotFound,
  EmbHostFileEOF,
  EmbHostFileIOError,		/* Catch all for other I/O errors */
};



/*** General-purpose stream ***/

enum EmbMessageChannelSubtype {
  EmbSerialMessageChannelSubtype = 1,
  EmbKludgeMessageChannelSubtype = 66000,	/* Force this to be a long word type */
};

typedef struct {
  enum EmbChannelType type;	/* EmbMessageChannelType */
  long unit;			/* 0 (not used) */
  EmbPtr next;			/* link to next channel */
  enum EmbMessageChannelSubtype subtype;
  /* Guest transmits messages, as EmbPtrs, through this queue */
  EmbPtr guest_to_host_queue;
  /* Host sends the above messages back when it's done, to recycle buffers */
  EmbPtr guest_to_host_return_queue;
  /* Out of band messages (opcode only) from guest to host */
  long guest_to_host_impulse;
  /* Host transmits messages, as EmbPtrs, through this queue */
  EmbPtr host_to_guest_queue;
  /* Guest provides buffers for the above queue through this one */
  EmbPtr host_to_guest_supply_queue;
  /* Out of band messages (opcode only) from host to guest */
  long host_to_guest_impulse;
  /* Following fields are host-private */
  struct proc *selproc;		/* host_state_0: State for select() */
  long host_state_1;		/* host_state_1: unused */
} EmbMessageChannel;

/* A message contains a one word header containing the opcode and byte length */
typedef struct {
  unsigned int opcode: 8;
  unsigned int length: 24;
  byte data[1];
} EmbMessage;

/* These message opcodes are common to all channels, others are application specific */
enum EmbMessageOpcode {
  EmbMessageOpcodeNOOP,
  EmbMessageOpcodeReset,
  EmbMessageOpcodeClose,
  EmbMessageOpcodeAllocate,
  EmbMessageOpcodeData,
  EmbMessageOpcodeError
  };


/*** Serial ***/

#define EMB_SERIAL_PROTOCOL_VERSION 2

enum SerialOpcode {
  SerialOpcodeNOOP,
  SerialOpcodeData,
  SerialOpcodeSetParameters,
  SerialOpcodeSetControlLines,
  SerialOpcodeReadControlLines,
  SerialOpcodeReceiveError,
  SerialOpcodeShutdown,
  SerialOpcodeSendBreak,
  SerialOpcodeKludge = 66000,	/* Force this to be a long word type */
};

enum SerialImpulse {
  SerialImpulseNone,
  SerialImpulseFlush,
  SerialImpulseWakeup,
  SerialImpulseKludge = 66000,	/* Force this to be a long word type */
};

#define SerialControlLineRTS  1
#define SerialControlLineCTS  2
#define SerialControlLineDSR  4
#define SerialControlLineDTR  8
#define SerialControlLineRI  32
#define SerialControlLineCD  64

#define SerialErrorBreak   1
#define SerialErrorOverrun 2
#define SerialErrorFraming 4
#define SerialErrorParity  8

#define EmbSerialBufferHeader unsigned int opcode: 8;  \
                              unsigned int length: 24; \
                              long return_code

typedef struct {
  EmbSerialBufferHeader;
} EmbSerialBuffer;

typedef struct {
  EmbSerialBufferHeader;
  long nbytes;
} EmbSerialDataBufferHeader;

#define EmbSerialBufferData(buf) (((char *) buf) + sizeof(EmbSerialDataBufferHeader))

enum EmbSerialMode {
  SerialModeAsync,
  SerialModeKludge = 66000,	/* Force this to be a long word type */
};

#define SerialErrorBreak   1
#define SerialErrorOverrun 2
#define SerialErrorFraming 4
#define SerialErrorParity  8

enum SerialParity {
  SerialParityNone,
  SerialParityOdd,
  SerialParityEven,
  SerialParitySpace,
  SerialParityMark,
  SerialParityKludge = 66000,	/* Force this to be a long word type */
};

typedef struct {
  EmbSerialBufferHeader;
  enum EmbSerialMode mode;
  long baud_rate;
  long data_bits;
  long stop_bits;
  enum SerialParity parity;
  long respect_xon_xoff_p;
  long input_xon_char;
  long input_xoff_char;
  long generate_xon_xoff_p;
  long output_xon_char;
  long output_xoff_char;
  long rts_cts_flow_control_p;
  long check_errors;
  long input_error_character;	/* --- Remove? */
} EmbSerialParameterBuffer;

typedef struct {
  EmbSerialBufferHeader;
  long dtr;
  long dsr;
  long cts;
  long rts;
  long cd;
  long ri;
} EmbSerialStatusBuffer;

typedef struct {
  long raw_character;
  long error_bits;
} EmbSerialReceiveErrorBuffer;

typedef struct {
  EmbSerialBufferHeader;
  long duration;
} EmbSerialBreakBuffer;

#endif
