/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * Life-support initialization
 *
 * $Log:	embinit.c,v $
 * Revision 2.5.1.2  93/02/01  12:55:48  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:51:37  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:40:21  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/embinit.c,v 2.5.1.2 93/02/01 12:55:48 kaufman Rel $";
#endif

#include <sys/types.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <net/nit_if.h>
#include <nlist.h>
#include <stdio.h>
#include <stropts.h>
#include <syslog.h>

#include "embed.h"
#include "life.h"
#include "config.h"

/* These could migrate into the configuration structure */
#define ColdLoadInputQueueSize 100
#define ColdLoadOutputQueueSize 50
#define RPCHostToGuestQueueSize 15
#define RPCGuestToHostQueueSize 30
#define SCSIQueueSize 10
#define DiskQueueSize 32
#define NET_RECV_QUEUE_SIZE 20		/* n elts in a net packet queue, incoming */
#define NET_XMIT_QUEUE_SIZE 20		/* n elts in a net packet queue, outgoing */
#define HOST_FILE_COMMAND_QUEUE_SIZE 5
#define HOST_FILE_REPLY_QUEUE_SIZE 5
  
#define LO_IFNAME "lo0"
#define MAX_INTERFACE_N 20

extern int verbose;

static struct nlist nl[] = {
#define NL_DiskLife 0
  {"_DiskLife"},
#define NL_NetReceiverWakeup 1
  {"_NetReceiverWakeup"},
#define NL_NetTransmitter 2
  {"_NetTransmitter"},
#define NL_HostFileLife 3
  {"_HostFileLife"},
  0,
};  


static EmbPtr EmbCommAreaAllocPtr;	/* For allocation during initialization */
static long EmbCommAreaAllocSize;

/* Allocate structure in communication area */
EmbPtr EmbCommAreaAlloc(EmbCommAreaPtr, n_bytes) 
     EmbCommArea *EmbCommAreaPtr; /*ARGSUSED*/
     long n_bytes;
{
  long n_words = (n_bytes+3)>>2;
  EmbPtr result = EmbCommAreaAllocPtr;
  char obuf[BUFSIZ];

  if ( n_words > EmbCommAreaAllocSize || n_bytes <= 0) {
    sprintf(obuf, "EmbCommAreaAlloc: Couldn't fulfill request for %d words", n_words);
    log_message(LOG_ERR, obuf);
    return(-1);
  }
  if (verbose)
    printf("EmbCommAreaAlloc: Allocated %d words\n", n_words);
  EmbCommAreaAllocSize -= n_words;
  EmbCommAreaAllocPtr += n_words;
  return (result);
}


/* Emb signal handling */

/* Signal handler table management */

void InitializeSignalHandlers (EmbCommAreaPtr)
     EmbCommArea *EmbCommAreaPtr;
{
  register int i;

  for (i=0; i<32; i++) {
    EmbCommAreaPtr->SignalHandler[i].proc = (caddr_t) 0;
    EmbCommAreaPtr->SignalHandler[i].fn   = (void (*)()) 0;
    EmbCommAreaPtr->SignalHandler[i].arg  = (PtrV) 0;
    EmbCommAreaPtr->SignalHandler[i].flag = 0;
  }
}

/* Allocate a guest-to-host signal, install a handler for it, and return the signal number */
SignalNumber InstallSignalHandlerInternal (EmbCommAreaPtr, handler, argument, flag)
     EmbCommArea *EmbCommAreaPtr;
     int handler;
     PtrV argument;
     short flag;
{
  register EmbCommArea* ep = EmbCommAreaPtr;
  register SignalNumber signal;
  char obuf[BUFSIZ];

  if (nl[handler].n_value == 0) {
    sprintf(obuf, "Kernel function %s not found!", nl[handler].n_name);
    log_message(LOG_ERR, obuf);
    return(-1);
  }

  for (signal=0; signal<32; signal++) {
    if ((ep->live_guest_to_host_signals & (1<<signal)) == 0) {
      ep->SignalHandler[signal].fn = (void (*)()) nl[handler].n_value;
      ep->SignalHandler[signal].arg = argument;
      ep->SignalHandler[signal].proc = (caddr_t) getpid();
      ep->SignalHandler[signal].flag = flag|SIGF_UNINITIALIZED;
      ep->SignalHandler[signal].count = 1;
      ep->live_guest_to_host_signals |= 1<<signal;
      return (signal);
    }
  }

  log_message(LOG_ERR,"Signal table overflow");
  exit(1);
  /*NOTREACHED*/
}


/* Create the cold load stream */

void InitializeColdLoadStream (EmbCommAreaPtr, config)
     EmbCommArea *EmbCommAreaPtr;
     EmbConfig* config;		/*ARGSUSED*/
{ register EmbCommArea* ep = EmbCommAreaPtr;
  EmbPtr cp = EmbCommAreaAlloc(EmbCommAreaPtr, (long)sizeof(EmbColdLoadChannel));
  register EmbColdLoadChannel *p = (EmbColdLoadChannel*)HostPointer(cp);
  register EmbQueue* q;
  
  p->type = EmbColdLoadChannelType;
  p->unit = 0;
  p->next = ep->channel_table;			/* Thread into list of all channels */
  ep->channel_table = cp;
  ep->cold_load_channel = cp;			/* Make it easy to find */

  cp = EmbCommAreaAlloc(EmbCommAreaPtr, (long)sizeof(EmbQueue) + ColdLoadInputQueueSize * sizeof(EmbPtr));
  q = (EmbQueue*)HostPointer(cp);
  p->keyboard_input_queue = cp;
  q->element_size = sizeof(EmbPtr);
  q->queue_size = ColdLoadInputQueueSize;
  q->put_index = 0;
  q->take_index = 0;
  q->signal = -1;

  cp = EmbCommAreaAlloc(EmbCommAreaPtr, (long)sizeof(EmbQueue) + ColdLoadOutputQueueSize * sizeof(EmbPtr));
  q = (EmbQueue*)HostPointer(cp);
  p->display_output_queue = cp;
  q->element_size = sizeof(EmbPtr);
  q->queue_size = ColdLoadOutputQueueSize;
  q->put_index = 0;
  q->take_index = 0;
  q->signal = -1;

  p->string_total_size = COLD_LOAD_STRING_TOTAL_SIZE;
  p->string_length = 0;
  p->selproc = NULL;
}


/* Create the disk channels */

void InitializeDiskLife (EmbCommAreaPtr, config, fd)
     EmbCommArea* EmbCommAreaPtr;
     EmbConfig* config;
     int fd;			/*ARGSUSED*/
{ 
  EmbChannel *ch;
  EmbDiskChannel *p;
  EmbQueue *q;
  EmbPtr cp;
  int i = -1, disk_fd;
  struct stat stat_buf;

#ifdef MACIVORY_DISK_HACKS
  /* Might not use these */
  EmbCommAreaPtr->maximum_disk_io_time = config->maximum_disk_io_time;
  EmbCommAreaPtr->disk_io_throttle_interval = config->disk_io_throttle_interval;
  EmbCommAreaPtr->remaining_disk_io_throttle_time = 0;
  EmbCommAreaPtr->number_of_disk_throttles = 0;
#endif

  /* Create disk channels for every mounted Ivory partition */
  for (disk_fd = 0; disk_fd < FD_SETSIZE; disk_fd++)
    if (FD_ISSET(disk_fd, &config->disk_partitions)) {
      i++;
      cp = EmbCommAreaAlloc(EmbCommAreaPtr, (long)sizeof(EmbDiskChannel));
      p = (EmbDiskChannel *) HostPointer(cp);
      p->type = EmbDiskChannelType;
      p->unit = i;
      p->initialized_p = 0;
      p->FileStructurePointer = (struct file *) disk_fd;
      if (fstat(disk_fd, &stat_buf) == -1) {
	log_error(LOG_ERR, "stat");
	exit(1);
      }
      p->number_of_pages = stat_buf.st_size/PageSize;

      /* This kludge is to make sure that you can't compile this code
       * such that it'll create the disk channels without allocating space
       * for the UIO structures.
       */
#ifndef lint
      p->UserIOBlock;
#endif

      if (i == 0)	/*--- Fix later, how do we know which partition is first? */
	EmbCommAreaPtr->boot_disk_channel = cp; /* First disk channel is boot chan */
      p->next = EmbCommAreaPtr->channel_table; /* Thread into list of all channels */
      EmbCommAreaPtr->channel_table = cp;

      cp = EmbCommAreaAlloc(EmbCommAreaPtr,
			    (long)sizeof(EmbQueue) + DiskQueueSize * sizeof(EmbPtr));
      q = (EmbQueue *) HostPointer(cp);
      p->command_queue = cp;
      p->command_queue_ptr = NULL;
      q->element_size = sizeof(EmbPtr);
      q->queue_size = DiskQueueSize;
      q->put_index = 0;
      q->take_index = 0;
      q->signal = -1;
    
      cp = EmbCommAreaAlloc(EmbCommAreaPtr,
			    (long)sizeof(EmbQueue) + DiskQueueSize * sizeof(EmbPtr));
      q = (EmbQueue*)HostPointer(cp);
      p->status_queue = cp;
      p->status_queue_ptr = NULL;
      q->element_size = sizeof(EmbPtr);
      q->queue_size = DiskQueueSize;
      q->put_index = 0;
      q->take_index = 0;
      q->signal = -1;				/* Guest can fill this in */
    }

  if (i == -1) {
    log_message(LOG_ERR, "No Ivory disk partitions");
    exit(1);
  }

  /* Fork, assign signals, and enter the kernel
   * Once we fork we can't allocate any Emb Comm area */
  if ((i = fork()) == -1) {
    log_error(LOG_ERR, "fork disk");
    exit(-1);
  } else if (i != 0) {
    /* Parent */
    return;
  } else {
    /* Child */
    for (cp = EmbCommAreaPtr->channel_table; cp != NullEmbPtr; cp = ch->next) {
      ch = (EmbChannel *) HostPointer(cp);
      if (ch->type == EmbDiskChannelType) {
	p = (EmbDiskChannel *) ch;
	q = (EmbQueue *) HostPointer(p->command_queue);
	q->signal = InstallSignalHandlerInternal(EmbCommAreaPtr, NL_DiskLife,
						 (PtrV) cp, SIGF_EMBPTRARG);
      }
    }

    IvoryLife(fd, PROC_DISK);
  }
  /*NOTREACHED*/
}


/* Create the Network channels */

static InitializeNetworkChannelNameLoc(loc, name)
     long *loc;
     char *name;
{
  int len, i, j;

  len = strlen(name);
  *loc = 0;

  for (i = 0, j = 0; i < 4 && i < len; j+=8, i++)
    *loc |= name[i] << j;

  return(i < len ? len-i : 0);
}

void InitializeNetworkChannelAddressesInternal (EmbCommAreaPtr, channel, ifname,
						ether_addr, ip_addr)
     EmbCommArea *EmbCommAreaPtr;
     EmbNetChannel *channel;
     char *ifname;
     u_char *ether_addr;
     long ip_addr;
{

  /* Store the ether address in the channel */
  channel->hardware_address_high = ether_addr[0] | (ether_addr[1] << 8) |
    (ether_addr[2] << 16) | (ether_addr[3] << 24);
  channel->hardware_address_low = ether_addr[4] | (ether_addr[5] << 8);

  /* Store the host IP address in the channel */
  channel->host_primary_protocol = ETHERTYPE_IP;
  channel->host_primary_address = ip_addr;

  /* Write interface name into network channel */
  if (strlen(ifname))
    if (InitializeNetworkChannelNameLoc(&channel->name_0, ifname))
      InitializeNetworkChannelNameLoc(&channel->name_1, ifname+4);
    else
      channel->name_1 = 0;
  else
    channel->name_0 = 0;
}

/* Set the address fields in the network channel */
void InitializeNetworkChannelAddresses (EmbCommAreaPtr, channel, fd)
     EmbCommArea *EmbCommAreaPtr;
     EmbNetChannel *channel;
     int fd;
{
  struct ifreq ifr;
  struct sockaddr_in *sin;
  char *ifname, *get_if_name();
  u_char addr[sizeof(struct ether_addr)];
  int s;

  /* Look up ethernet address */
  if (ioctl(fd, SIOCGIFADDR, &ifr) == -1) {
    log_error(LOG_ERR, "SIOCGIFADDR AF_UNSPEC");
    exit(1);
  }
  bcopy((u_char *) ifr.ifr_addr.sa_data, addr, 6);

  /* Look up the internet address.
   * Thanks to Unix lameness, we need the interface name, so we get it
   * from a kludge in network.c. */
  ifname = get_if_name(fd);
  if ((s = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    log_error(LOG_ERR, "socket");
    exit(1);
  }
  strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
  ifr.ifr_addr.sa_family = AF_INET;
  if (ioctl(s, SIOCGIFADDR, &ifr) == -1) {
    log_error(LOG_ERR, "SIOCGIFADDR AF_INET");
    exit(1);
  }
  sin = (struct sockaddr_in *) &ifr.ifr_addr;

  /* Store addresses in the network channel */
  InitializeNetworkChannelAddressesInternal(EmbCommAreaPtr, channel, ifname, addr,
					    sin->sin_addr.s_addr);
}

void InitializeNetworkChannelFakeAddresses (EmbCommAreaPtr, channel, fd)
     EmbCommArea *EmbCommAreaPtr;
     EmbNetChannel *channel;
     int fd;
{
  struct ifreq ifr;
  struct sockaddr_in *sin;
  int s;
  u_char addr[sizeof(struct ether_addr)];

  /* Default ethernet address is all zeros */
  bzero(addr, sizeof(addr));

  /* Look up the loopback internet address */
  if ((s = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    log_error(LOG_ERR, "socket");
    exit(1);
  }
  strncpy(ifr.ifr_name, LO_IFNAME, sizeof(ifr.ifr_name));
  ifr.ifr_addr.sa_family = AF_INET;
  if (ioctl(s, SIOCGIFADDR, &ifr) == -1) {
    log_error(LOG_ERR, "SIOCGIFADDR AF_INET");
    exit(1);
  }
  sin = (struct sockaddr_in *) &ifr.ifr_addr;

  /* Store addresses in the network channel */
  InitializeNetworkChannelAddressesInternal(EmbCommAreaPtr, channel, LO_IFNAME, addr,
					    sin->sin_addr.s_addr);
}

/* Set up an EmbNetChannel */
/* Later, observe settings in the config data.  For now, just try to start up */
EmbNetChannel *InitializeNetworkChannel(EmbCommAreaPtr, net_fd, channel_unit)
     EmbCommArea *EmbCommAreaPtr;
     int net_fd, channel_unit;
{
  register EmbCommArea *ep = EmbCommAreaPtr;
  EmbNetChannel *channel;
  EmbQueue *queue;
  EmbPtr ptr;

  ptr = EmbCommAreaAlloc(EmbCommAreaPtr, (long) sizeof(EmbNetChannel));	/* alloc the channel struct */
  channel = (EmbNetChannel *) HostPointer(ptr);		/* make a pointer we can use */
  channel->type = EmbNetworkChannelType;		/* say what kind it is */
  channel->unit = channel_unit;
  channel->next = ep->channel_table;		/* link it onto global list */
  ep->channel_table = ptr;
  channel->if_ierrors = channel->if_oerrors =
    channel->if_collisions = channel->if_dropped = 0;

  /* Make the queues and set them up */
  ptr = EmbCommAreaAlloc(EmbCommAreaPtr, (long )sizeof(EmbQueue) + (sizeof(EmbPtr) *
								    NET_XMIT_QUEUE_SIZE));
  queue = (EmbQueue *) HostPointer(ptr);
  channel->xmit_host_to_guest_queue = ptr;
  channel->xmit_host_to_guest_queue_ptr = NULL;
  queue->element_size = sizeof(EmbPtr);		/* these queues deal in pointers */
  queue->queue_size = NET_XMIT_QUEUE_SIZE;
  queue->take_index = 0;
  queue->put_index = 0;
  queue->signal = -1;				/* Ivory fills this in if it wants */

  ptr = EmbCommAreaAlloc(EmbCommAreaPtr, (long )sizeof(EmbQueue) + (sizeof(EmbPtr) *
								    NET_XMIT_QUEUE_SIZE));
  queue = (EmbQueue *) HostPointer(ptr);
  channel->xmit_guest_to_host_queue = ptr;
  channel->xmit_guest_to_host_queue_ptr = NULL;
  queue->element_size = sizeof(EmbPtr);		/* these queues deal in pointers */
  queue->queue_size = NET_XMIT_QUEUE_SIZE;
  queue->take_index = 0;
  queue->put_index = 0;
  queue->signal = -1;

  ptr = EmbCommAreaAlloc(EmbCommAreaPtr, (long )sizeof(EmbQueue) + (sizeof(EmbPtr)
								    * NET_RECV_QUEUE_SIZE));
  queue = (EmbQueue *) HostPointer(ptr);
  channel->recv_host_to_guest_queue = ptr;
  channel->recv_host_to_guest_queue_ptr = NULL;
  queue->element_size = sizeof(EmbPtr);		/* these queues deal in pointers */
  queue->queue_size = NET_RECV_QUEUE_SIZE;
  queue->take_index = 0;
  queue->put_index = 0;
  queue->signal = -1;				/* Ivory will fill this in */

  ptr = EmbCommAreaAlloc(EmbCommAreaPtr, (long )sizeof(EmbQueue) + (sizeof(EmbPtr) *
								    NET_RECV_QUEUE_SIZE));
  queue = (EmbQueue *) HostPointer(ptr);
  channel->recv_guest_to_host_queue = ptr;
  channel->recv_guest_to_host_queue_ptr = NULL;
  queue->element_size = sizeof(EmbPtr);		/* these queues deal in pointers */
  queue->queue_size = NET_RECV_QUEUE_SIZE;
  queue->take_index = 0;
  queue->put_index = 0;
  queue->signal = -1;

  /* Clear the status word, clear various other bits */
  channel->status = 0;
  channel->emb_if_unit = -1;
  channel->initialized_p = 0;
  channel->network_queue = (PtrV) net_fd;
  channel->guest_primary_protocol = -1;
  channel->guest_primary_address = -1;

  if (net_fd == -1)
    InitializeNetworkChannelFakeAddresses(EmbCommAreaPtr, channel, net_fd);
  else
    InitializeNetworkChannelAddresses(EmbCommAreaPtr, channel, net_fd);

  return(channel);
}

EmbPtr InitializeHostEthernetAddresses (EmbCommAreaPtr, require_fake_address_p)
     EmbCommArea *EmbCommAreaPtr;
     int require_fake_address_p;
{
  int j, fd, nif, ninterfaces = 0;
  struct ifconf ifc;
  struct ifreq ifr[MAX_INTERFACE_N];
  char obuf[BUFSIZ];
  struct strioctl si;
  struct ether_addr *ether_addrp;
  EmbPtr cp;

  /* Get interface list */
  if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    log_error(LOG_ERR, "socket for interface list");
    return(-1);
  }
  ifc.ifc_len = sizeof(ifr);
  ifc.ifc_req = ifr;
  if (ioctl(fd, SIOCGIFCONF, &ifc) == -1) {
    log_error(LOG_ERR, "SIOCGIFCONF");
    return(-1);
  }
  nif = ifc.ifc_len/sizeof(struct ifreq);
  (void) close(fd);

  /* Don't offer loopback or emb interfaces */
  for (j = 0; j < nif; j++) {
    if ((!strncmp(ifr[j].ifr_name, "lo", 2) &&
	 isdigit(ifr[j].ifr_name[2])) ||
	(!strncmp(ifr[j].ifr_name, "emb", 3) &&
	 isdigit(ifr[j].ifr_name[3])))
      ifr[j].ifr_name[0] = '\0';
    else
      ninterfaces++;
  }

  /* There always has to be at least one ethernet address */
  if (ninterfaces == 0) {
    ninterfaces = 1;
    nif = -1;
  } else if (require_fake_address_p)
    ninterfaces++;

  /* Allocate space for the table */
  cp = EmbCommAreaAlloc(EmbCommAreaPtr, ninterfaces * sizeof(struct ether_addr));
  EmbCommAreaPtr->n_host_ether_addrs = ninterfaces;
  ether_addrp = (struct ether_addr *) HostPointer(cp);

  if (nif == -1 || require_fake_address_p) {
    /* Table entry for ethernet interface with no external connection */
    bzero(ether_addrp++, sizeof(struct ether_addr));
  }

  if (nif > 0) {
    /* Open NIT so we can get at ethernet interface addresses */
    if ((fd = open("/dev/nit", O_RDONLY)) == -1) {
      log_error(LOG_ERR, "/dev/nit");
      exit(1);
    }

    /* Fill in the table */
    for (j = 0; j < nif; j++) {
      if (ifr[j].ifr_name[0] == '\0')
	continue;

      /* Bind the NIT device to this interface */
      si.ic_cmd = NIOCBIND;
      si.ic_len = sizeof(struct ifreq);
      si.ic_timout = 0;
      si.ic_dp = (char *) &ifr[j]; 
      if (ioctl(fd, I_STR, (char *) &si) == -1) {
	log_error(LOG_ERR, "NIOCBIND");
	continue;
      }

      /* Look up ethernet address */
      if (ioctl(fd, SIOCGIFADDR, &ifr[j]) == -1) {
	sprintf(obuf, "SIOCGIFADDR %s", ifr[j].ifr_name);
	log_error(LOG_ERR, obuf);
      } else {
	bcopy((u_char *) ifr[j].ifr_addr.sa_data, ether_addrp++, sizeof(struct ether_addr));
      }
    }
    (void) close(fd);
  }

  return(cp);
}

EmbPtr InitializeNetworkLife(EmbCommAreaPtr, config, ivory_fd)
     EmbCommArea *EmbCommAreaPtr;
     EmbConfig *config;
     int ivory_fd;
{
  EmbPtr cp;
  EmbQueue *q;
  EmbChannel *ch;
  EmbNetChannel *channel;
  int i, net_fd, ninterfaces = 0, nchannels;

  /* Count the interfaces */
  for (i = getdtablesize(); i >= 0; i--)
    if (FD_ISSET(i, &config->network_interfaces))
      ninterfaces++;
  nchannels = ninterfaces == 0 ? 1 : ninterfaces;

  /* Allocate and fill in array of host ethernet addresses */
  cp = InitializeHostEthernetAddresses(EmbCommAreaPtr, ninterfaces == 0);

  /* Create network channels (one per network fd) */
  if (ninterfaces == 0) {
    (void) InitializeNetworkChannel(EmbCommAreaPtr, -1, 0);
  } else
    for (net_fd = i = 0; net_fd < FD_SETSIZE; net_fd++)
      if (FD_ISSET(net_fd, &config->network_interfaces))
	(void) InitializeNetworkChannel(EmbCommAreaPtr, net_fd, i++);

  /* Fork, assign signals, and enter the kernel
   * Once we fork we can't allocate any Emb Comm area */
  if ((i = fork()) == -1) {
    log_error(LOG_ERR, "fork network");
    exit(-1);
  } else if (i != 0) {
    /* Parent */
    return(cp);
  } else {
    /* Child */
    for (cp = EmbCommAreaPtr->channel_table; cp != NullEmbPtr; cp = ch->next) {
      ch = (EmbChannel *) HostPointer(cp);
      if (ch->type == EmbNetworkChannelType) {
	channel = (EmbNetChannel *) ch;
	q = (EmbQueue *) HostPointer(channel->xmit_guest_to_host_queue);
	q->signal = InstallSignalHandlerInternal(EmbCommAreaPtr, NL_NetTransmitter,
						 (PtrV) cp, SIGF_EMBPTRARG);
	q = (EmbQueue *) HostPointer(channel->recv_guest_to_host_queue);
	q->signal = InstallSignalHandlerInternal(EmbCommAreaPtr, NL_NetReceiverWakeup,
						 (PtrV) cp, SIGF_EMBPTRARG);
      }
    }

    IvoryLife(ivory_fd, PROC_NETWORK);
  }
  /*NOTREACHED*/
}


/* Set up the Host File channel */

void InitializeHostFileAccess(EmbCommAreaPtr, config)
     EmbCommArea *EmbCommAreaPtr;
     EmbConfig *config;
{
  EmbHostFileChannel *channel;
  EmbQueue *queue;
  EmbPtr cp, ptr;

  cp = EmbCommAreaAlloc(EmbCommAreaPtr, (long) sizeof(EmbHostFileChannel));
  channel = (EmbHostFileChannel *) HostPointer(cp);
  channel->type = EmbHostFileChannelType;
  channel->unit = 0;
  channel->next = EmbCommAreaPtr->channel_table;
  EmbCommAreaPtr->channel_table = cp;

  ptr = EmbCommAreaAlloc(EmbCommAreaPtr, (long) sizeof(EmbQueue) +
			 (sizeof(EmbPtr) * HOST_FILE_COMMAND_QUEUE_SIZE));
  queue = (EmbQueue *) HostPointer(ptr);
  queue->element_size = sizeof(long);
  queue->queue_size = HOST_FILE_COMMAND_QUEUE_SIZE;
  queue->put_index = 0;
  queue->take_index = 0;
  queue->signal = -1;
  channel->command_queue = ptr;

  ptr = EmbCommAreaAlloc(EmbCommAreaPtr, (long) sizeof(EmbQueue) +
			 (sizeof(EmbPtr) * HOST_FILE_REPLY_QUEUE_SIZE));
  queue = (EmbQueue *) HostPointer(ptr);
  queue->element_size = sizeof(long);
  queue->queue_size = HOST_FILE_REPLY_QUEUE_SIZE;
  queue->put_index = 0;
  queue->take_index = 0;
  queue->signal = -1;
  channel->reply_queue = ptr;
}


void InitializeCommunicationArea (ptr, fd, conf)
     caddr_t ptr;
     int fd;
     Config *conf;
{
  register EmbCommArea* EmbCommAreaPtr = (EmbCommArea *) ptr;
  EmbConfig *config = &conf->emb;
  struct stat st;
  int interrupt_level, interrupt_vector;
  short os_major, os_minor, os_patch;
  extern char *life_version, *embed_version, *lookup_kernel_version();
  char obuf[BUFSIZ];

  /* Dig handler function addresses out of kernel */
  if (nlist("/vmunix", nl) == -1) {
    log_error(LOG_ERR, "nlist");
    exit(1);
  }

  /* Look up minor device number */
  if (fstat(fd, &st) == -1) {
    log_error(LOG_ERR, "fstat");
    exit(-1);
  }

  /* Initialize miscellaneous communication area fields */
  EmbCommAreaPtr->identifier = 'ENON';	/* Set to 'DBME' while Life Support is running */
  EmbCommAreaPtr->version = 1;
  EmbCommAreaPtr->number_of_slots = ((Ptr)&EmbCommAreaPtr->pad0 - (Ptr)EmbCommAreaPtr) >>2; /* Ivory-visible portion only */
  EmbCommAreaPtr->board_serial_number = minor(st.st_rdev);
  EmbCommAreaPtr->comm_memory_size = config->communication_memory_size>>2;

  EmbCommAreaPtr->system_type = conf->kern.system_type;
  set_version(&EmbCommAreaPtr->host_major_version, lookup_kernel_version());
  set_version(&EmbCommAreaPtr->host_minor_version, life_version);
  set_version(&EmbCommAreaPtr->boot_version, "-1.-1.-1.-1");
  if (get_os_version(&os_major, &os_minor, &os_patch) >= 0) {
    char buf[50];

    sprintf(buf, "%d.%d.%d.-1", os_major, os_minor, os_patch);
    set_version(&EmbCommAreaPtr->host_os_version, buf);
  }
#ifdef POLL_EMB_COMM_AREA
  EmbCommAreaPtr->siv.siv_word = -1;	/* Don't interrupt */
  if (verbose)
    printf("init-ivory: interrupts disabled\n");
#else /* POLL_EMB_COMM_AREA */
  if (lookup_interrupt_vector(fd, &interrupt_level, &interrupt_vector) == -1) {
    log_message(LOG_ERR, "ivory-life: couldn't find interrupt handler, interrupts won't work");
    EmbCommAreaPtr->siv.siv_word = -1;	/* Don't interrupt */
  } else {
    EmbCommAreaPtr->signal_interrupt_level = interrupt_level;
    EmbCommAreaPtr->signal_interrupt_vector = interrupt_vector;
    if (verbose)
      printf("ivory-life: interrupt level %d vector 0x%x\n",
	     interrupt_level, interrupt_vector);
  }
#endif

  EmbCommAreaPtr->channel_table = NullEmbPtr;
  EmbCommAreaPtr->boot_disk_channel = NullEmbPtr;
  EmbCommAreaPtr->command_channel = NullEmbPtr;
  EmbCommAreaPtr->clock_signal = -1;	/* No signal allocated until guest needs it */
  EmbCommAreaPtr->debug_level = config->debug_level;

  /* Set up for dynamic allocation within the communication area */
  EmbCommAreaAllocPtr = sizeof(*EmbCommAreaPtr)>>2;
  EmbCommAreaAllocSize = (config->communication_memory_size>>2) - EmbCommAreaAllocPtr;

  InitializeSignalHandlers(EmbCommAreaPtr);

  /* Initialize life support channels & processes */
  InitializeColdLoadStream(EmbCommAreaPtr, config);
  InitializeDiskLife(EmbCommAreaPtr, config, fd);
  conf->kern.host_ether_addr_offset = InitializeNetworkLife(EmbCommAreaPtr, config, fd);
  InitializeHostFileAccess(EmbCommAreaPtr, config);

  /* Allocate remaining portion of communication area to buffers */
  EmbCommAreaPtr->host_buffer_start = EmbCommAreaAllocPtr;
  EmbCommAreaPtr->host_buffer_size = config->host_buffer_size>>2;
  EmbCommAreaPtr->fep_buffer_start = EmbCommAreaAllocPtr + EmbCommAreaPtr->host_buffer_size;
  /* FEP gets enough for an Ethernet packet or a disk page, plus 100 words of overhead */ 
  EmbCommAreaPtr->fep_buffer_size = 512;
  EmbCommAreaPtr->guest_buffer_start = EmbCommAreaAllocPtr + EmbCommAreaPtr->host_buffer_size + EmbCommAreaPtr->fep_buffer_size;
  EmbCommAreaPtr->guest_buffer_size = EmbCommAreaAllocSize - EmbCommAreaPtr->host_buffer_size - EmbCommAreaPtr->fep_buffer_size;
  EmbCommAreaPtr->base_register = config->guest_extra_buffer_size;
  if (EmbCommAreaPtr->guest_buffer_size < config->guest_buffer_size>>2) {
    sprintf(obuf, "%s communication area %d words too small.  Requested size of host and guest buffers won't fit.",
	    conf->device,
	    (config->guest_buffer_size>>2) - EmbCommAreaPtr->guest_buffer_size);
    log_message(LOG_ERR, obuf);
    return;
  }

#ifdef notdef
  InitializeEmbMemory(EmbCommAreaPtr, unit);
#endif

}
