/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */
/*
 * $Log:	fepfs.c,v $
 * Revision 2.5.1.2  93/02/01  12:55:51  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:51:42  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:40:33  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/fepfs.c,v 2.5.1.2 93/02/01 12:55:51 kaufman Rel $";
#endif

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/timeb.h>

#include "fepfs.h"


/**** Utilities ****/

void report_error (message)
string message;
{
  fprintf (stderr, message);
  abort ();
}

/* Unix measures time since 1970, Genera measures from 1900 */
integer get_universal_time ()
{
  /* 1/1/00 and 1/1/70 are 2208988800 seconds apart, according to Genera */
  return ((integer) ((unsigned integer) time (NULL) + 2208988800));
}

/* Exactly like bcopy, except it swaps bytes within each word, and requires word alignment */
void bswap (source, destination, bytes)
string source;
string destination;
integer bytes;
{
  register unsigned integer *s = (unsigned integer *) source;
  register unsigned integer *d = (unsigned integer *) destination;
  register unsigned integer word;
  integer words = ceiling (bytes, 4);

  while (words--) {
    word = *s++;
    *d++ = (((word << 24)) |
	    ((word << 8) & 0x00FF0000) |
	    ((word >> 8) & 0x0000FF00) |
	    ((word >> 24)));
  }
}

/* without-aborts facility to protect critical regions */
/* Note well: these aren't nestable! */

static void (*previous_signal)() = NULL;

void disable_aborts ()
{
  if (previous_signal == NULL)
    previous_signal = signal (SIGINT, SIG_IGN);
}

void enable_aborts ()
{
  if (previous_signal != NULL)
    signal (SIGINT, previous_signal);
  previous_signal = NULL;
}

/* Case insensitive string compare */
boolean string_equal (a, b)
register string a;
register string b;
{
  register char ac;
  register char bc;

  while (true) {
    ac = *a++;
    bc = *b++;
    if (((ac ^ bc) & 0337) != 0) return (false);
    if ((ac == NULL) || (bc == NULL)) return (true);
  }
}

/* Case insensitive string search */
boolean string_search (a, pattern)
register string a;
string pattern;
{
  register char *pp, *cp;
  char pc;

#define FORCE_UPPER(ch) (islower(ch) ? toupper(ch) : ch)

  pc = FORCE_UPPER(*pattern);
  for (; *a != '\0'; a++)
    if (FORCE_UPPER(*a) == pc) {
      for (cp = a, pp = pattern; *pp != '\0' && *cp != '\0'; cp++, pp++)
	if (FORCE_UPPER(*cp) != FORCE_UPPER(*pp))
	  break;
      if (*pp == '\0')
	return(true);
    }
  return(false);
}

boolean yes_or_no_p ()
{
  char s [64];

  while (true) {
    bzero (s, 16);
    scanf (" %s", s);
    if (string_equal (s, "Yes")) return (true);
    if (string_equal (s, "No")) return (false);
    printf ("Please answer yes or no: ");
  }
}

void translate_genera_to_unix (p, count)
register unsigned char *p;
register integer count;
{
  register unsigned char c;

  while (count--) {
    switch (c = *p) {
    case 0210: c = 010; break;
    case 0211: c = 011; break;
    case 0215: c = 012; break;
    case 0212: c = 015; break;
    case 0207: c = 0177; break;
    case 0214: c = 014; break;
    default: break;
    }
    *p++ = c;
  }
}

void translate_unix_to_genera (p, count)
register unsigned char *p;
register integer count;
{
  register unsigned char c;

  while (count--) {
    switch (c = *p) {
    case 010: c = 0210; break;
    case 011: c = 0211; break;
    case 012: c = 0215; break;
    case 015: c = 0212; break;
    case 0177: c = 0207; break;
    case 014: c = 0214; break;
    default: break;
    }
    *p++ = c;
  }
}


/**** Disk buffers ****/

/* These should be parameterized per file system, so we can vary the block size */

static integer block_size = 320;

disk_buffer allocate_disk_buffer ()
{
  disk_buffer result = malloc (block_size*4);
  bzero (result, block_size*4);
  return (result);
}

void deallocate_disk_buffer (buffer)
disk_buffer buffer;
{
  free (buffer);
}

void initialize_disk_buffer (buffer)
disk_buffer buffer;
{
  bzero (buffer, block_size*4);
}


/**** Disk I/O ****/

void read_disk_block (fs, buffer, address)
file_system fs;
disk_buffer buffer;
disk_address address;
{
  if (lseek (fs->descriptor, (address & 0777777777) * block_size * 4, L_SET) < 0)
    perror ("Seek failed");
  if (read (fs->descriptor, buffer, block_size*4) < 0)
    perror ("Read failed");
}

void write_disk_block (fs, buffer, address)
file_system fs;
disk_buffer buffer;
disk_address address;
{
  if (lseek (fs->descriptor, (address & 0777777777) * block_size * 4, L_SET) < 0)
    perror ("Seek failed");
  if (write (fs->descriptor, buffer, block_size*4) < 0)
    perror ("Write failed");
}

void read_disk_block_portion (fs, buffer, address, offset, bytes)
file_system fs;
disk_buffer buffer;
disk_address address;
integer offset;
integer bytes;
{
  if (lseek (fs->descriptor, (address & 0777777777) * block_size * 4 + offset, L_SET) < 0)
    perror ("Seek failed");
  if (read (fs->descriptor, buffer, bytes) < 0)
    perror ("Read failed");
}

void write_disk_block_portion (fs, buffer, address, offset, bytes)
file_system fs;
disk_buffer buffer;
disk_address address;
integer offset;
integer bytes;
{
  bytes = 4 * ceiling(bytes, 4);
  if (lseek (fs->descriptor, (address & 0777777777) * block_size * 4 + offset, L_SET) < 0)
    perror ("Seek failed");
  if (write (fs->descriptor, buffer, bytes) < 0)
    perror ("Write failed");
}

file_system bootstrap_disk (filename, read_only_p)
string filename;
boolean read_only_p;
{
  struct stat buffer;
  integer descriptor;
  file_system fs;

  block_size = 320;
  if ((descriptor = open (filename, (read_only_p) ? O_RDONLY : O_RDWR|O_EXCL)) < 0)
    return (NULL);
  if (fstat (descriptor, &buffer) < 0)
    return (NULL);
  fs = (file_system) malloc (sizeof (struct file_system_representation));
  fs->disk_size = buffer.st_size / (block_size * 4);
  fs->device = strdup (filename);
  fs->descriptor = descriptor;
  return (fs);
}

void force_disk_output (fs)
file_system fs;
{
  fsync (fs->descriptor);
}


/**** Data maps ****/

datamap allocate_datamap (size)
integer size;
{
  datamap result = (datamap) malloc ((size * 2 + 2) * sizeof (integer));
  result->size = size;
  result->entries = 0;
  return (result);
}

#define deallocate_datamap(map) free (map)

datamap copy_datamap (map)
datamap map;
{
  datamap result = (datamap) malloc ((map->entries * 2 + 2) * sizeof (integer));
  bcopy (map->array, result->array, map->entries * 2 * sizeof (integer));
  result->size = result->entries = map->entries;
  return (result);
}

datamap truncate_datamap (map)
datamap map;
{
  datamap result = copy_datamap (map);
  deallocate_datamap (map);
  return (result);
}

#define datamap_size(map) (map->size)
#define datamap_entries(map) (map->entries)
#define datamap_entry_address(map, i) (map->array [(i) * 2])
#define datamap_entry_size(map, i) (map->array [(i) * 2 + 1])

integer datamap_blocks (map)
datamap map;
{
  integer i;
  integer sum = 0;

  for (i = 0; i < datamap_entries (map); i++)
    sum += datamap_entry_size (map, i);

  return (sum);
}

datamap extend_datamap (map, address, size)
datamap map;
integer address;
integer size;
{
  integer entries = map->entries;

  if ((entries != 0) &&
      (address == map->array [(entries - 1) * 2] + map->array [(entries - 1) * 2 + 1]))
    /* Just extend the last entry */
    map->array [(entries - 1) * 2 + 1] += size;
  else if (entries < map->size) {
    /* Add a new entry */
    map->array [entries * 2] = address;
    map->array [entries * 2 + 1] = size;
    map->entries = entries + 1;
  } else
    report_error("Datamap growing not yet implemented");
  return(map);
}

integer datamap_translate (map, offset)
datamap map;
integer offset;
{
  register integer *p = map->array;
  register integer entry_address, entry_size;
  register integer entry_offset = 0;
  integer entries = map->entries;

  while (entries--) {
    entry_address = *p++;
    entry_size = *p++;
    if ((entry_offset + entry_size) > offset)
      return (entry_address + offset - entry_offset);
    entry_offset += entry_size;
  }
  report_error ("Datamap address translation failed");
}

/* Same as the above, but `returns' two values: the address and size of the containing extent */
void datamap_translate_block (map, offset, extent_address, extent_size)
datamap map;
integer offset;
integer *extent_address;
integer *extent_size;
{
  register integer *p = map->array;
  register integer entry_address, entry_size;
  register integer entry_offset = 0;
  integer entries = map->entries;

  while (entries--) {
    entry_address = *p++;
    entry_size = *p++;
    if ((entry_offset + entry_size) > offset) {
      *extent_address = entry_address + offset - entry_offset;
      *extent_size = entry_offset + entry_size - offset;
      return;
    }
    entry_offset += entry_size;
  }
  report_error ("Datamap address translation failed");
}

void describe_datamap (map)
datamap map;
{
  integer i;

  printf("Datamap has %d entries out of %d possible\n", map->entries, map->size);
  for (i = 0; i < map->entries; i++)
    printf(" Entry %d: address %d, size %d\n", i,
	   datamap_entry_address (map, i),
	   datamap_entry_size (map, i));
}

#ifdef notdef
void test_datamap ()
{
  integer i;
  datamap map = allocate_datamap (64);
  extend_datamap (map, 0, 10);
  extend_datamap (map, 12, 10);
  extend_datamap (map, 22, 10);
  extend_datamap (map, 40, 5);
  describe_datamap (map);
  for (i = 0; i < 35; i++)
    printf ("Block %d at %d\n", i, datamap_translate (map, i));
}
#endif


/**** Bitmaps ****/

boolean test_bit (map, index)
bitmap map;
integer index;
{
  if (index > map->bits)
    report_error ("Bitmap reference out of bounds");
  else
    return ((map->array [index / 32] >> (index % 32)) & 1);
}

void set_bit (map, index)
bitmap map;
integer index;
{
  if (index > map->bits)
    report_error ("Bitmap reference out of bounds");
  else
    map->array [index / 32] |= (1 << index % 32);
}

void clear_bit (map, index)
bitmap map;
integer index;
{
  if (index > map->bits)
    report_error ("Bitmap reference out of bounds");
  else
    map->array [index / 32] &= ~(1 << index % 32);
}

bitmap allocate_bitmap (bits)
integer bits;
{
  integer words = ceiling (bits, 32);
  bitmap result = (bitmap) malloc (sizeof (integer) * 2 + words * 4);
  integer i;
  /* Clear all the bits in the bitmap except set the ones above the boundary */
  bzero (result->array, words * 4);
  result->bits = words * 32;
  for (i = bits; i < words * 32; i++)
    set_bit (result, i);
  result->bits = bits;
  return (result);
}

datamap find_free_bits (freemap, bits)
bitmap freemap;
integer bits;
{
  datamap map = allocate_datamap (64);
  register unsigned long *p = freemap->array;
  integer bit;

  /* Quickly scan for the first unfilled word in the bitmap */
  register integer words;
  for (words = ceiling (freemap->bits, 32); words > 0; words--)
    if (*p == ~0) p++; else break;
  bit = (p - freemap->array) * 32;

  /* Now loop allocating bits, slowly */
  while (bits > 0) {
    if (bit >= freemap->bits) {
      fprintf (stderr, "Not enough room to allocate %d FEP blocks.\n", bits);
      exit (2);
    }
    if (!test_bit (freemap, bit)) {
      set_bit (freemap, bit);
      extend_datamap (map, bit, 1);
      bits--;
    }
    bit++;
  }
  return (truncate_datamap (map));
}

/* Count the number of on bits in a bitmap */
integer bitmap_allocated (map)
bitmap map;
{
  register unsigned long *p = map->array;
  register integer words = ceiling (map->bits, 32);
  register unsigned integer word;
  register integer count = 0;

  /* Count all the on bits in the bitmap; read whole words, check quickly for the 0 and
     -1 cases, degenerate to counting bits */
  while ((words--) > 0) {
    if (word = *p++)
      if (word == -1)
	count += 32;
      else
	while (word) {
	  count += word & 1;
	  word >>= 1;
	}
  }

  return (count);
}

#ifdef notdef
void test_bitmap ()
{
  bitmap free = allocate_bitmap (1024);
  describe_datamap (find_free_bits (free, 64));
  describe_datamap (find_free_bits (free, 17));
  describe_datamap (find_free_bits (free, 93));
  describe_datamap (find_free_bits (free, 117));
  describe_datamap (find_free_bits (free, 3));
  describe_datamap (find_free_bits (free, 74));
}
#endif


/**** FEP filesystem structure definitions ****/

#define current_feph_header_version 1
#define current_labl_header_version 2
#define current_fepd_header_version 1

typedef struct {
  /* These fields are present for all FEP structure blocks */
  char id [4];
  integer header_version;
  integer npages;
  integer timestamp;
  integer sequence_number;
  /* These fields are only present for FEP file headers */
  integer owning_directory;
  integer version;
  char type [4];
  integer creation_date;
  char author [28];
  char name [32];
  char comment [96];
  integer byte_length;
  integer flags;
  /* This field is present for FEPF and FEPP blocks */
  integer nentries;
} feph;

#define define_integer_accessor(type,element,reader,writer) \
  integer reader (buffer)\
  disk_buffer buffer;\
  { return ( ((type *) buffer)->element ); }\
  void writer (buffer, value)\
  disk_buffer buffer;\
  integer value;\
  { ((type *) buffer)->element = value; }

#define define_character_accessor(type,element,reader,writer,bytes) \
  string reader (buffer)\
  disk_buffer buffer;\
  {\
    return (read_character_field (((type *) buffer)->element, bytes));\
  }\
  void writer (buffer, value)\
  disk_buffer buffer;\
  string value;\
  {\
    write_character_field (((type *) buffer)->element, bytes, value);\
  }

/* Return a C string containing a copy of the specified character field */
/* The characters in the field are stored little-endian, which is what the i^3 is about */
string read_character_field (field, bytes)
char *field;
integer bytes;
{
  string result;
  register integer length;
  register integer i;

  for (length = 0; (length < bytes); length++)
    if (field [length^3] == NULL) break;
  result = (string) malloc (length + 1);
  for (i = 0; i < length; i++)
    result [i] = field [i^3];
  result [length] = NULL;

  return (result);
}

/* Set the specified character field to a value specified as a C string */
/* The characters in the field are stored little-endian, which is what the i^3 is about */
void write_character_field (field, bytes, value)
char *field;
integer bytes;
string value;
{
  register integer i;

  for (i = 0; i < bytes; i++)
    if ((field [i^3] = value [i]) == NULL) break;
  for ( ; i < bytes; i++)
    field [i^3] = NULL;
}

/* Compare the specified character field against a specified C string */
/* The characters in the field are stored little-endian, which is what the i^3 is about */
/* Obviously we do a care-insensitive compare; chars are allowed to differ in bit 5 */
boolean compare_character_field (field, bytes, value)
char *field;
integer bytes;
string value;
{
  register integer i;

  for (i = 0; (i < bytes) && (value [i] != NULL); i++)
    if (((field [i^3] ^ value [i]) & 0337) != 0)
      return (false);

  return (true);
}

define_character_accessor(feph, id, feph_id, set_feph_id, 4);
define_integer_accessor(feph, header_version, feph_header_version, set_feph_header_version);
define_integer_accessor(feph, npages, feph_npages, set_feph_npages);
define_integer_accessor(feph, timestamp, feph_timestamp, set_feph_timestamp);
define_integer_accessor(feph, sequence_number, feph_sequence_number, set_feph_sequence_number);
define_integer_accessor(feph, owning_directory, fepf_owning_directory, set_fepf_owning_directory);
define_integer_accessor(feph, version, fepf_version, set_fepf_version);
define_character_accessor(feph, type, fepf_type, set_fepf_type, 4);
define_integer_accessor(feph, creation_date, fepf_creation_date, set_fepf_creation_date);
define_character_accessor(feph, author, fepf_author, set_fepf_author, 28);
define_character_accessor(feph, name, fepf_name, set_fepf_name, 32);
define_character_accessor(feph, comment, fepf_comment, set_fepf_comment, 96);
define_integer_accessor(feph, byte_length, fepf_byte_length, set_fepf_byte_length);
define_integer_accessor(feph, flags, fepf_flags, set_fepf_flags);
define_integer_accessor(feph, nentries, fepp_nentries, set_fepp_nentries);

#define define_flag_accessor(reader,writer,bit) \
  boolean reader (buffer)\
  disk_buffer buffer;\
  {\
    return(ldb(1,bit,fepf_flags(buffer)));\
  }\
  void writer (buffer, value)\
  disk_buffer buffer;\
  boolean value;\
  {\
    set_fepf_flags(buffer, dpb(value,1,bit,fepf_flags(buffer)) );\
  }

define_flag_accessor(fepf_directory, set_fepf_directory, 0);
define_flag_accessor(fepf_read_only, set_fepf_read_only, 1);
define_flag_accessor(fepf_fixed, set_fepf_fixed, 2);
define_flag_accessor(fepf_deleted, set_fepf_deleted, 3);

/* Starting index for the FEP page table */
#define fepp_entry_offset 51
#define fepp_entry_size 2

integer fepp_entry_npages (buffer, index)
disk_buffer buffer;
integer index;
{
  return ((*((integer *) buffer + fepp_entry_offset + index*fepp_entry_size)) & 077777777);
}

disk_address fepp_entry_dpn (buffer, index)
disk_buffer buffer;
integer index;
{
  return ((*((integer *) buffer + fepp_entry_offset + index*fepp_entry_size + 1)) & 0377777777);
}

boolean fepp_entry_indirect (buffer, index)
disk_buffer buffer;
integer index;
{
  return (ldb (1, 24, *((integer *) buffer + fepp_entry_offset + index*fepp_entry_size)));
}

void set_fepp_entry (buffer, index, address, blocks, indirect)
disk_buffer buffer;
integer index;
disk_address address;
integer blocks;
boolean indirect;
{
  register integer *p = (integer *) buffer + fepp_entry_offset + index*fepp_entry_size;

  *p++ = blocks | (indirect << 24);
  *p = address;
}

typedef struct {
  /* These fields are present for all FEP structure blocks */
  char id [4];
  integer header_version;
  integer npages;
  integer timestamp;
  integer sequence_number;
  /* These fields are specific to disk label blocks */
  integer ncylinders;
  integer nheads;
  integer npages_per_track;
  integer creation_date;
  char author [28];
  char pack_name [32];
  char comment [96];
  integer root_directory_dpn;
  integer pack_id;
  integer fast_mode;
  integer disk_type;
  integer diagnostic_dpn;
  integer fep_dpn;
} labl;

define_integer_accessor(labl, ncylinders, labl_ncylinders, set_labl_ncylinders);
define_integer_accessor(labl, nheads, labl_nheads, set_labl_nheads);
define_integer_accessor(labl, npages_per_track, labl_npages_per_track, set_labl_npages_per_track);
define_integer_accessor(labl, creation_date, labl_creation_date, set_labl_creation_date);
define_character_accessor(labl, author, labl_author, set_labl_author, 28);
define_character_accessor(labl, pack_name, labl_pack_name, set_labl_pack_name, 32);
define_character_accessor(labl, comment, labl_comment, set_labl_comment, 96);
define_integer_accessor(labl, root_directory_dpn, labl_root_directory_dpn, set_labl_root_directory_dpn);
define_integer_accessor(labl, diagnostic_dpn, labl_diagnostic_dpn, set_labl_diagnostic_dpn);
define_integer_accessor(labl, fep_dpn, labl_fep_dpn, set_labl_fep_dpn);

typedef struct {
  /* These fields are present for all FEP structure blocks */
  char id[4];
  integer header_version;
  integer npages;
  integer timestamp;
  integer sequence_number;
  /* These fields are specific to directory page blocks */
  integer page_number;
  integer undefined[7];
  integer nentries;
} fepd;

define_integer_accessor(fepd, page_number, fepd_page_number, set_fepd_page_number);
define_integer_accessor(fepd, nentries, fepd_nentries, set_fepd_nentries);

#define fepd_entry_offset 14*4
#define fepd_entry_size 11*4
#define fepd_entry_name_size 32
#define fepd_entry_type_size 4

typedef struct {
  char name [32];
  char type [4];
  integer version;
  integer address;
} fepd_entry;

string fepd_entry_name (buffer, index)
disk_buffer buffer;
integer index;
{
  fepd_entry *entry = (fepd_entry *) (buffer + fepd_entry_offset + index*fepd_entry_size);

  return (read_character_field (entry->name, fepd_entry_name_size));
}

void set_fepd_entry_name (buffer, index, value)
disk_buffer buffer;
integer index;
string value;
{
  fepd_entry *entry = (fepd_entry *) (buffer + fepd_entry_offset + index*fepd_entry_size);

  write_character_field (entry->name, fepd_entry_name_size, value);
}

string fepd_entry_type (buffer, index)
disk_buffer buffer;
integer index;
{
  fepd_entry *entry = (fepd_entry *) (buffer + fepd_entry_offset + index*fepd_entry_size);

  return (read_character_field (entry->type, fepd_entry_type_size));
}

void set_fepd_entry_type (buffer, index, value)
disk_buffer buffer;
integer index;
string value;
{
  fepd_entry *entry = (fepd_entry *) (buffer + fepd_entry_offset + index*fepd_entry_size);

  write_character_field (entry->type, fepd_entry_type_size, value);
}

integer fepd_entry_version (buffer, index)
disk_buffer buffer;
integer index;
{
  fepd_entry *entry = (fepd_entry *) (buffer + fepd_entry_offset + index*fepd_entry_size);

  return (entry->version);
}

void set_fepd_entry_version (buffer, index, value)
disk_buffer buffer;
integer index;
integer value;
{
  fepd_entry *entry = (fepd_entry *) (buffer + fepd_entry_offset + index*fepd_entry_size);

  entry->version = value;
}

integer fepd_entry_address (buffer, index)
disk_buffer buffer;
integer index;
{
  fepd_entry *entry = (fepd_entry *) (buffer + fepd_entry_offset + index*fepd_entry_size);

  return (entry->address);
}

void set_fepd_entry_address (buffer, index, value)
disk_buffer buffer;
integer index;
integer value;
{
  fepd_entry *entry = (fepd_entry *) (buffer + fepd_entry_offset + index*fepd_entry_size);

  entry->address = value;
}

void set_fepd_entry (buffer, index, name, type, version, address)
disk_buffer buffer;
integer index;
string name;
string type;
integer version;
disk_address address;
{
  set_fepd_entry_name (buffer, index, name);
  set_fepd_entry_type (buffer, index, type);
  set_fepd_entry_version (buffer, index, version);
  set_fepd_entry_address (buffer, index, address);
}

boolean compare_fepd_entry (buffer, index, name, type, version)
disk_buffer buffer;
integer index;
string name;
string type;
integer version;
{
  fepd_entry *entry = (fepd_entry *) (buffer + fepd_entry_offset + index*fepd_entry_size);

  return (compare_character_field (entry->name, fepd_entry_name_size, name) &&
	  compare_character_field (entry->type, fepd_entry_type_size, type) &&
	  (version == entry->version));
}

integer sequence_number (buffer)
disk_buffer buffer;
{
  return (* (integer *) buffer);
}

/* Set all 32-bit words in a buffer to the specified integer */
void set_sequence_number (buffer, value)
disk_buffer buffer;
integer value;
{
  integer *p = (integer *) buffer;
  integer i;

  for (i = 0; i < block_size; i++)
    *p++ = value;
}


/**** Tools for debugging file system structure ****/

void check_block_id (buffer, id)
disk_buffer buffer;
string id;
{
  if ((buffer [0] != id [3]) ||
      (buffer [1] != id [2]) ||
      (buffer [2] != id [1]) ||
      (buffer [3] != id [0]))
    report_error ("Incorrect FEP ID on block");
}

void print_feph (buffer)
disk_buffer buffer;
{
  printf ("ID: %s\n", feph_id (buffer));
  printf ("Header version: %d\n", feph_header_version (buffer));
  printf ("Number of pages: %d\n", feph_npages (buffer));
  printf ("Timestamp: %d\n", feph_timestamp (buffer));
  printf ("Sequence number: %d\n", feph_sequence_number (buffer));
}

void print_fepf (buffer)
disk_buffer buffer;
{
  printf ("Name: %s.%s.%d\n", fepf_name (buffer), fepf_type (buffer), fepf_version (buffer));
  printf ("Owning directory: %d\n", fepf_owning_directory (buffer));
  printf ("Creation date: %d\n", fepf_creation_date (buffer));
  printf ("Author: %s\n", fepf_author (buffer));
  printf ("Comment: %s\n", fepf_comment (buffer));
  printf ("Byte length: %d\n", fepf_byte_length (buffer));
}

void print_labl (buffer)
disk_buffer buffer;
{
  printf ("Cylinders: %d\n", labl_ncylinders (buffer));
  printf ("Heads: %d\n", labl_nheads (buffer));
  printf ("Sectors/track: %d\n", labl_npages_per_track (buffer));
  printf ("Creation Date: %d\n", labl_creation_date (buffer));
  printf ("Author: %s\n", labl_author (buffer));
  printf ("Pack name: %s\n", labl_pack_name (buffer));
  printf ("Comment: %s\n", labl_comment (buffer));
  printf ("Root Directory: %d\n", labl_root_directory_dpn (buffer));
  printf ("Fep kernel: %d\n", labl_fep_dpn (buffer));
}

void print_fepp (buffer)
disk_buffer buffer;
{
  integer i;
  printf ("Page table entries: %d\n", fepp_nentries (buffer));
  for (i = 0; i < fepp_nentries (buffer); i++) {
    printf ("Entry %d: %d blocks at %d", i,
	    fepp_entry_npages (buffer, i), fepp_entry_dpn (buffer, i));
    if (fepp_entry_indirect (buffer, i))
      printf(" (indirect)\n");
    else
      printf("\n");
  }
}

void print_fepd (buffer)
disk_buffer buffer;
{
  integer i;
  printf ("Directory page entries: %d\n", fepd_nentries (buffer));
  for (i = 0; i < fepd_nentries (buffer); i++)
    printf ("Entry %d: %s.%s.%d at %d\n", i,
	    fepd_entry_name (buffer, i), fepd_entry_type (buffer, i),
	    fepd_entry_version (buffer, i), fepd_entry_address (buffer, i));
}


/**** Primitive file system operations ****/

disk_address lookup_fep_file();
datamap read_fep_file_datamap();
void write_fep_file_datamap();
void create_initial_file_system();

/* Write the specified bitmap to the blocks specified by the datamap */
/* This is used to maintain the free bitmap (and create it during IFS) */
void write_bitmap (fs, bmap, dmap)
file_system fs;
bitmap bmap;
datamap dmap;
{
  unsigned integer *p = bmap->array;
  integer bytes = ceiling (bmap->bits, 8);
  integer blocks = ceiling (bytes, block_size * 4);
  disk_buffer buffer = allocate_disk_buffer ();
  integer b = 0;

  for (b = 0; b < blocks; b++) {
    if (b != blocks - 1)
      bcopy (p + b * block_size, buffer, block_size * 4);
    else {
      bzero (buffer, block_size * 4);
      bcopy (p + b * block_size, buffer, bytes % (block_size * 4));
    }
    write_disk_block (fs, buffer, datamap_translate (dmap, b));
  }
  deallocate_disk_buffer (buffer);
}

void read_bitmap (fs, bmap, dmap)
file_system fs;
bitmap bmap;
datamap dmap;
{
  unsigned integer *p = bmap->array;
  integer bytes = ceiling (bmap->bits, 8);
  integer blocks = ceiling (bytes, block_size * 4);
  disk_buffer buffer = allocate_disk_buffer ();
  integer b = 0;

  for (b=0; b<blocks; b++) {
    read_disk_block (fs, buffer, datamap_translate (dmap, b));
    if (b != blocks - 1)
      bcopy (buffer, p + b * block_size, block_size * 4);
    else
      bcopy (buffer, p + b * block_size, bytes % (block_size * 4));
  }
  deallocate_disk_buffer (buffer);
}

datamap allocate_blocks (fs, blocks)
file_system fs;
integer blocks;
{
  datamap map = find_free_bits (fs->free_bitmap, blocks);
  write_bitmap (fs, fs->free_bitmap, fs->free_datamap);
  return (map);
}

void deallocate_blocks (fs, map)
file_system fs;
datamap map;
{
  report_error ("Block deallocation not yet implemented");
}

integer allocate_sequence_number (fs)
file_system fs;
{
  integer sequence;
  disk_address address;
  disk_buffer buffer;

  address = lookup_fep_file (fs, "SEQUENCE-NUMBER.FEP.1");
  if (address == NULL)
    report_error ("Couldn't find SEQUENCE-NUMBER.FEP.1");
  buffer = allocate_disk_buffer ();
  read_disk_block (fs, buffer, address);
  check_block_id (buffer, "FEPF");
  if (fepp_nentries (buffer) != 1)
    report_error ("SEQUENCE-NUMBER.FEP has more than one page table entry");
  if (fepp_entry_npages (buffer, 0) != 1)
    report_error ("SEQUENCE-NUMBER.FEP has more than one data block");
  if (fepp_entry_indirect (buffer, 0) != false)
    report_error ("SEQUENCE-NUMBER.FEP has an indirect data block");
  address = fepp_entry_dpn (buffer, 0);
  read_disk_block (fs, buffer, address);
  sequence = sequence_number (buffer) + 1;
  set_sequence_number (buffer, sequence);
  write_disk_block (fs, buffer, address);
  deallocate_disk_buffer (buffer);

  return (sequence);
}

void install_fep_kernel (fs, pathname)
file_system fs;
string pathname;
{
  disk_address header;
  disk_buffer buffer;
  integer i;

  if ((header = lookup_fep_file (fs, pathname)) == NULL)
    report_error ("Couldn't find specified FEP kernel");
  buffer = allocate_disk_buffer ();
  disable_aborts ();
  for (i = 0; i < 64; i++) {
    read_disk_block (fs, buffer, i);
    check_block_id (buffer, "LABL");
    set_labl_fep_dpn (buffer, header);
    write_disk_block (fs, buffer, i);
  }
  force_disk_output (fs);
  enable_aborts ();
  deallocate_disk_buffer (buffer);
}

integer lookup_fep_version (fs)
file_system fs;
{
  disk_buffer buffer = allocate_disk_buffer ();
  disk_address address;
  string name;
  integer version = 0;

  read_disk_block (fs, buffer, 0);
  check_block_id (buffer, "LABL");
  address = labl_fep_dpn (buffer);
  if ((address < 64) || (address >= fs->disk_size))
    goto unknown;
  read_disk_block (fs, buffer, address);
  if ((buffer [0] != 'F') ||
      (buffer [1] != 'P') ||
      (buffer [2] != 'E') ||
      (buffer [3] != 'F'))
    goto unknown;
  name = fepf_name (buffer);
  if (((name [0] == 'i') || (name [0] == 'I')) && string_equal (name + 4, "-KERNEL"))
    version = atoi (name + 1);
  deallocate_disk_buffer (buffer);
  return (version);
 unknown:
  deallocate_disk_buffer (buffer);
  return (0);
}

boolean bootstrap_file_system (fs)
file_system fs;
{
  disk_buffer buffer;
  disk_address free_address;

  buffer = allocate_disk_buffer ();
  read_disk_block (fs, buffer, 0);
  if ((buffer [0] != 'L') ||
      (buffer [1] != 'B') ||
      (buffer [2] != 'A') ||
      (buffer [3] != 'L') ||
      (feph_header_version (buffer) != current_labl_header_version))
    goto failure;
  fs->root_directory_address = labl_root_directory_dpn (buffer);
  if (fs->root_directory_address < 64) goto failure;
  read_disk_block (fs, buffer, fs->root_directory_address);
  if ((buffer [0] != 'F') ||
      (buffer [1] != 'P') ||
      (buffer [2] != 'E') ||
      (buffer [3] != 'F') ||
      (feph_header_version (buffer) != current_feph_header_version))
    goto failure;
  fs->free_bitmap = allocate_bitmap (fs->disk_size);
  free_address = lookup_fep_file (fs, "FREE-PAGES.FEP.1");
  if (free_address == NULL) goto failure;
  fs->free_datamap = read_fep_file_datamap (fs, free_address);
  read_bitmap (fs, fs->free_bitmap, fs->free_datamap);
 success:
  deallocate_disk_buffer (buffer);
  return (true);
 failure:
  deallocate_disk_buffer (buffer);
  return (false);
}


/**** File datamap hacking ****/

/* Determine the number of entries needed in the datamap for the specified file header */
integer count_fep_file_datamap (fs, buffer)
file_system fs;
disk_buffer buffer;
{
  integer count = 0;
  integer entries = fepp_nentries (buffer);
  integer i = 0;

  for (i=0; i<entries; i++) {
    if (!fepp_entry_indirect (buffer, i))
      count++;
    else {
      disk_buffer page_buffer = allocate_disk_buffer();
      read_disk_block (fs, page_buffer, fepp_entry_dpn (buffer, i));
      check_block_id (page_buffer, "FEPP");
      count += count_fep_file_datamap (fs, page_buffer);
      deallocate_disk_buffer (page_buffer);
    }
  }
  return (count);
}

datamap collect_fep_file_datamap (fs, buffer, map)
file_system fs;
disk_buffer buffer;
datamap map;
{
  integer entries = fepp_nentries (buffer);
  integer i = 0;

  for (i=0; i<entries; i++) {
    if (!fepp_entry_indirect (buffer, i))
      map = extend_datamap (map, fepp_entry_dpn (buffer, i), fepp_entry_npages (buffer, i));
    else {
      disk_buffer page_buffer = allocate_disk_buffer ();
      read_disk_block (fs, page_buffer, fepp_entry_dpn (buffer, i));
      check_block_id (page_buffer, "FEPP");
      map = collect_fep_file_datamap (fs, page_buffer, map);
      deallocate_disk_buffer (page_buffer);
    }
  }
  return (map);
}

datamap read_fep_file_datamap (fs, address)
file_system fs;
disk_address address;
{
  disk_buffer buffer;
  datamap map;

  buffer = allocate_disk_buffer ();
  read_disk_block (fs, buffer, address);
  check_block_id (buffer, "FEPF");
  map = allocate_datamap (count_fep_file_datamap (fs, buffer) + 4);
  map = collect_fep_file_datamap (fs, buffer, map);
  deallocate_disk_buffer (buffer);
  return (map);
}

void write_fep_file_datamap (fs, address, map)
file_system fs;
disk_address address;
datamap map;
{
  disk_buffer buffer;
  integer entries;
  integer i;

  buffer = allocate_disk_buffer ();
  read_disk_block (fs, buffer, address);
  entries = datamap_entries (map);
  if (entries <= (block_size - fepp_entry_offset) / fepp_entry_size) {
    set_fepp_nentries (buffer, entries);
    for (i = 0; i < entries; i++)
      set_fepp_entry
	(buffer, i, datamap_entry_address (map, i), datamap_entry_size (map, i), false);
  } else
    report_error ("Writing file datamaps with indirect blocks isn't supported yet");
  write_disk_block (fs, buffer, address);
  deallocate_disk_buffer (buffer);
}

datamap grow_fep_file_datamap (fs, address, map, blocks)
file_system fs;
disk_address address;
datamap map;
integer blocks;
{
  integer i;
  datamap new;
  integer entries;

  disable_aborts ();
  new = allocate_blocks (fs, blocks);
  entries = datamap_entries (new);
  for (i = 0; i < entries; i++)
    map = extend_datamap (map, datamap_entry_address (new, i), datamap_entry_size (new, i));
  deallocate_datamap (new);
  write_fep_file_datamap (fs, address, map);
  force_disk_output (fs);
  enable_aborts ();

  return (map);
}

datamap shrink_fep_file_datamap (fs, address, map, blocks)
file_system fs;
disk_buffer address;
datamap map;
integer blocks;
{
  report_error ("Datamap shrinking not supported yet");
}


/**** File operations ****/

disk_address lookup_fep_file_internal (fs, directory_address, name, type, version)
file_system fs;
disk_address directory_address;
string name;
string type;
integer version;
{
  datamap directory_map;
  integer directory_blocks;
  disk_address fepd_address;
  disk_buffer fepd_buffer;
  integer block, entry;
  disk_address result;

  fepd_buffer = allocate_disk_buffer ();
  directory_map = read_fep_file_datamap (fs, directory_address);
  directory_blocks = datamap_blocks (directory_map);
  for (block = 0; block < directory_blocks; block++) {
    fepd_address = datamap_translate (directory_map, block);
    read_disk_block (fs, fepd_buffer, fepd_address);
    check_block_id (fepd_buffer, "FEPD");
    for (entry = 0; entry < fepd_nentries (fepd_buffer); entry++)
      if (compare_fepd_entry (fepd_buffer, entry, name, type, version)) {
	result = fepd_entry_address (fepd_buffer, entry);
	deallocate_disk_buffer (fepd_buffer);
	return (result);
      }
  }
  deallocate_disk_buffer (fepd_buffer);
  return (false);
}

disk_address lookup_fep_directory (fs, pathname, name, type, version)
file_system fs;
string pathname;
char *name;
char *type;
integer *version;
{
  char *position;
  disk_address directory_address = fs->root_directory_address;

  pathname = strdup (pathname);                /* destructive below */
  if (pathname [0] == '>') pathname++;
  while (position = strchr (pathname, '>')) {
    *position = '\0';
    directory_address = lookup_fep_file_internal (fs, directory_address, pathname, "DIR", 1);
    pathname = &position [1];
  }
  bzero (name, 32);
  bzero (type, 4);
  *version = 1;
  sscanf (pathname, "%[^.].%[^.].%d", name, type, version);

  return (directory_address);
}

disk_address lookup_fep_file (fs, pathname)
file_system fs;
string pathname;
{
  disk_address directory_address;
  char name [32+1];
  char type [4+1];
  integer version;

  directory_address = lookup_fep_directory (fs, pathname, name, type, &version);

  return (lookup_fep_file_internal (fs, directory_address, name, type, version));
}

disk_address create_fep_file (fs, pathname, blocks)
file_system fs;
string pathname;
integer blocks;
{
  disk_address fepd_address;
  disk_buffer fepd_buffer;
  integer fepd_entries;
  disk_address directory_address;
  datamap directory_map;
  integer directory_blocks;
  integer block;
  disk_address file_address;
  disk_buffer file_buffer;
  char name [32+1];
  char type [4+1];
  integer version;

  directory_address = lookup_fep_directory (fs, pathname, name, type, &version);
  directory_map = read_fep_file_datamap (fs, directory_address);
  directory_blocks = datamap_blocks (directory_map);
  /* Search all the directory pages for one with a free entry we can use */
  for (block = 0; block < directory_blocks; block++) {
    fepd_address = datamap_translate (directory_map, block);
    fepd_buffer = allocate_disk_buffer ();
    read_disk_block (fs, fepd_buffer, fepd_address);
    check_block_id (fepd_buffer, "FEPD");
    fepd_entries = fepd_nentries (fepd_buffer);
    if (fepd_entries < (block_size - fepd_entry_offset) / fepd_entry_size) {
      disable_aborts ();
      /* Allocate a block for the new file header */
      file_address = datamap_entry_address (allocate_blocks (fs, 1), 0);
      file_buffer = allocate_disk_buffer ();
      set_feph_id (file_buffer, "FEPF");
      /* Initialize the new file header, and write it out to the disk */
      set_feph_header_version (file_buffer, current_feph_header_version);
      set_feph_npages (file_buffer, 1);
      set_feph_timestamp (file_buffer, get_universal_time ());
      set_feph_sequence_number (file_buffer, allocate_sequence_number (fs));
      set_fepf_owning_directory (file_buffer, feph_sequence_number (fepd_buffer));
      set_fepf_name (file_buffer, name);
      set_fepf_type (file_buffer, type);
      set_fepf_version (file_buffer, version);
      set_fepf_author (file_buffer, "Benson");
      set_fepf_comment (file_buffer, "");
      set_fepf_byte_length (file_buffer, 0);
      set_fepf_directory (file_buffer, string_equal (type, "DIR"));
      set_fepf_read_only (file_buffer, 0);
      set_fepf_fixed (file_buffer, 0);
      set_fepf_deleted (file_buffer, 0);
      set_fepf_creation_date (file_buffer, get_universal_time ());
      set_fepp_nentries (file_buffer, 0);
      write_disk_block (fs, file_buffer, file_address);
      deallocate_disk_buffer (file_buffer);
      /* Initialize the directory entry pointing to the new file header, and write it out */
      set_fepd_entry (fepd_buffer, fepd_entries, name, type, version, file_address);
      set_fepd_nentries (fepd_buffer, fepd_entries + 1);
      write_disk_block (fs, fepd_buffer, fepd_address);
      force_disk_output (fs);
      enable_aborts ();
      deallocate_disk_buffer (fepd_buffer);
      if (blocks > 0)
	grow_fep_file_datamap (fs, file_address, allocate_datamap (64), blocks);

      return (file_address);
    }
  }
  report_error ("Growing directories isn't supported yet");
}

/* Determine the number of blocks specified by the datamap for a given file */
integer fep_file_block_length (fs, buffer)
file_system fs;
disk_buffer buffer;
{
  integer blocks = 0;
  integer i;

  for (i = 0; i < fepp_nentries (buffer); i++)
    if (!fepp_entry_indirect (buffer, i))
      blocks += fepp_entry_npages (buffer, i);
    else {
      disk_buffer page_buffer = allocate_disk_buffer ();
      read_disk_block (fs, page_buffer, fepp_entry_dpn (buffer, i));
      check_block_id (page_buffer, "FEPP");
      blocks += fep_file_block_length (fs, page_buffer);
      deallocate_disk_buffer (page_buffer);
    }
  return (blocks);
}

property_list get_fep_file_properties (fs, address)
file_system fs;
disk_address address;
{
  disk_buffer buffer;
  property_list properties;

  buffer = allocate_disk_buffer ();
  read_disk_block (fs, buffer, address);
  check_block_id (buffer, "FEPF");
  properties = (property_list) malloc (sizeof (struct property_list_representation));
  properties->name = fepf_name (buffer);
  properties->type = fepf_type (buffer);
  properties->version = fepf_version (buffer);
  properties->creation_date = fepf_creation_date (buffer);
  properties->author = fepf_author (buffer);
  properties->comment = fepf_comment (buffer);
  properties->length_in_bytes = fepf_byte_length (buffer);
  properties->length_in_blocks = fep_file_block_length (fs, buffer);
  properties->directory = fepf_directory (buffer);
  properties->deleted = fepf_deleted (buffer);
  properties->dont_delete = fepf_read_only (buffer);
  deallocate_disk_buffer (buffer);

  return (properties);
}

void set_fep_file_properties (fs, address, properties)
file_system fs;
disk_address address;
property_list properties;
{
  disk_buffer buffer;
  integer blocks;
  datamap map;

  buffer = allocate_disk_buffer ();
  read_disk_block (fs, buffer, address);
  check_block_id (buffer, "FEPF");
  set_fepf_name (buffer, properties->name);
  set_fepf_type (buffer, properties->type);
  set_fepf_version (buffer, properties->version);
  set_fepf_creation_date (buffer, properties->creation_date);
  set_fepf_author (buffer, properties->author);
  set_fepf_comment (buffer, properties->comment);
  set_fepf_byte_length (buffer, properties->length_in_bytes);
  set_fepf_directory (buffer, properties->directory);
  set_fepf_deleted (buffer, properties->deleted);
  set_fepf_read_only (buffer, properties->dont_delete);
  write_disk_block (fs, buffer, address);
  /* Now determine if we need to grow or shrink the file and do it */
  blocks = fep_file_block_length (fs, buffer);
  if (properties->length_in_blocks != blocks) {
    map = read_fep_file_datamap (fs, address);
    if (properties->length_in_blocks > blocks)
      grow_fep_file_datamap (fs, address, map, properties->length_in_blocks - blocks);
    else
      shrink_fep_file_datamap (fs, address, map, blocks - properties->length_in_blocks);
  }
  deallocate_disk_buffer (buffer);
}

void show_fep_directory (fs, pathname)
file_system fs;
string pathname;
{
  disk_address directory_address;
  datamap directory_map;
  integer directory_blocks;
  integer block;
  disk_address fepd_address;
  disk_buffer fepd_buffer = allocate_disk_buffer ();
  integer entry;
  property_list properties;
  char name [32+1];
  char type [4+1];
  integer version;
  char filename [48];

  directory_address = lookup_fep_directory (fs, pathname, name, type, &version);
  directory_map = read_fep_file_datamap (fs, directory_address);
  directory_blocks = datamap_blocks (directory_map);
  for (block = 0; block < directory_blocks; block++) {
    fepd_address = datamap_translate (directory_map, block);
    read_disk_block (fs, fepd_buffer, fepd_address);
    check_block_id (fepd_buffer, "FEPD");
    for (entry = 0; entry < fepd_nentries (fepd_buffer); entry++) {
      properties = get_fep_file_properties (fs, fepd_entry_address (fepd_buffer, entry));
      sprintf (filename, "%s.%s.%d", properties->name, properties->type, properties->version);
      if (!properties->directory)
	printf ("  %-42s %4d %8d(8)  %c  %s  %s\n",
		filename,
		properties->length_in_blocks,
		properties->length_in_bytes,
		(properties->dont_delete) ? '@' : ' ',
		properties->comment,
		properties->author);
      else
	printf ("  %-42s Directory    %s  %s\n",
		filename, properties->comment, properties->author);
      free (properties);
    }
  }
  deallocate_disk_buffer (fepd_buffer);
}


/**** `Streams' ****/

fep_stream open_fep_stream (fs, pathname, create_p)
file_system fs;
string pathname;
boolean create_p;
{
  disk_address address;
  fep_stream stream;

  address = lookup_fep_file (fs, pathname);
  if ((address == NULL) && create_p)
    address = create_fep_file (fs, pathname, 0);
  if (address == NULL) return (NULL);
  stream = (fep_stream) malloc (sizeof (struct fep_stream_representation));
  stream->fs = fs;
  stream->pathname = pathname;
  stream->header = address;
  stream->properties = get_fep_file_properties (stream->fs, address);
  stream->map = read_fep_file_datamap (stream->fs, address);
  stream->pointer = 0;

  return (stream);
}

void read_fep_stream (stream, buffer, bytes)
fep_stream stream;
char *buffer;
integer bytes;
{
  integer block_bytes = block_size * 4;
  datamap map = stream->map;
  integer pointer = stream->pointer;
  integer block, offset, count;
  integer extent_address, extent_size;

  while (bytes > 0) {
    /* Compute the logical block and offset of the start of the transfer */
    block = pointer / block_bytes;
    offset = pointer % block_bytes;
    /* Translate that to disk coordinates, and compute how much lies in this extent */
    datamap_translate_block (map, block, &extent_address, &extent_size);
    if ((count = extent_size * block_bytes - offset) > bytes) count = bytes;
    read_disk_block_portion (stream->fs, buffer, extent_address, offset, count);
    pointer += count;
    buffer += count;
    bytes -= count;
  }
  stream->pointer = pointer;
}

void write_fep_stream (stream, buffer, bytes)
fep_stream stream;
char *buffer;
integer bytes;
{
  integer block_bytes = block_size * 4;
  datamap map = stream->map;
  integer pointer = stream->pointer;
  integer block, offset, count, blocks;
  integer extent_address, extent_size;

  /* Extend the file data map if necessary */
  blocks = ceiling (pointer + bytes, block_bytes);
  if (blocks > stream->properties->length_in_blocks) {
    blocks = blocks - stream->properties->length_in_blocks;
    map = stream->map = grow_fep_file_datamap (stream->fs, stream->header, map, blocks);
    stream->properties->length_in_blocks += blocks;
  }
  while (bytes > 0) {
    /* Compute the logical block and offset of the start of the transfer */
    block = pointer / block_bytes;
    offset = pointer % block_bytes;
    /* Translate that to disk coordinates, and compute how much lies in this extent */
    datamap_translate_block (map, block, &extent_address, &extent_size);
    if ((count = extent_size * block_bytes - offset) > bytes) count = bytes;
    write_disk_block_portion (stream->fs, buffer, extent_address, offset, count);
    pointer += count;
    buffer += count;
    bytes -= count;
  }
  stream->pointer = pointer;
}

void seek_fep_stream (stream, offset)
fep_stream stream;
integer offset;
{
  stream->pointer = offset;
}

property_list get_fep_stream_properties (stream)
fep_stream stream;
{
  stream->properties = get_fep_file_properties (stream->fs, stream->header);
  return (stream->properties);
}

void set_fep_stream_properties (stream, properties)
fep_stream stream;
property_list properties;
{
  set_fep_file_properties (stream->fs, stream->header, stream->properties = properties);
  stream->map = read_fep_file_datamap (stream->fs, stream->header);
}


/**** Initial FEP file system (IFS) ****/

disk_address ifs_create_file (fs, name, type, comment, sequence, map, freemap)
file_system fs;
string name;
string type;
string comment;
integer sequence;
datamap map;
bitmap freemap;
{
  disk_buffer buffer = allocate_disk_buffer ();
  disk_address header = datamap_entry_address (find_free_bits (freemap, 1), 0);

  set_feph_id (buffer, "FEPF");
  set_feph_header_version (buffer, current_feph_header_version);
  set_feph_npages (buffer, 1);
  set_feph_timestamp (buffer, get_universal_time());
  set_feph_sequence_number (buffer, sequence);

  set_fepf_owning_directory (buffer, 0);
  set_fepf_name (buffer, name);
  set_fepf_type (buffer, type);
  set_fepf_version (buffer, 1);
  set_fepf_author (buffer, "Benson");
  set_fepf_comment (buffer, comment);
  set_fepf_byte_length (buffer, (!map) ? 0 : datamap_entry_size(map,0)*block_size*4);
  set_fepf_directory (buffer, string_equal (type, "DIR"));
  set_fepf_read_only (buffer, 1);
  set_fepf_fixed (buffer, 0);
  set_fepf_deleted (buffer, 0);
  set_fepf_creation_date (buffer, get_universal_time());

  if (map != NULL) {
    if (datamap_entries (map) != 1)
      report_error ("Improper datamap for ifs_create_file");
    set_fepp_nentries (buffer, 1);
    set_fepp_entry (buffer, 0,
		    datamap_entry_address (map, 0),
		    datamap_entry_size (map, 0),
		    false);
  }

  write_disk_block (fs, buffer, header);
  deallocate_disk_buffer (buffer);

  return (header);
}

void create_initial_file_system (fs)
file_system fs;
{
  bitmap freemap = allocate_bitmap (fs->disk_size);
  datamap labl_datamap, root_datamap, diag_datamap, free_datamap, seqn_datamap;
  disk_address lbel, root, badb, diag, free, seqn;
  disk_buffer buffer = allocate_disk_buffer ();
  register integer i;

  /* Trash the label block so noone can use this file system if we get aborted */
  write_disk_block (fs, buffer, 0);
  force_disk_output (fs);

  labl_datamap = find_free_bits (freemap, 64);
  root_datamap = find_free_bits (freemap, 1);
  diag_datamap = find_free_bits (freemap, 64);
  free_datamap = find_free_bits (freemap, ceiling (fs->disk_size, block_size*32));
  seqn_datamap = find_free_bits (freemap, 1);

  lbel = ifs_create_file (fs, "DISK-LABEL", "FEP", "Disk Label", 1, labl_datamap, freemap);
  root = ifs_create_file (fs, "ROOT-DIRECTORY", "DIR", "His highness", 0, root_datamap, freemap);
  badb = ifs_create_file (fs, "BAD-BLOCKS", "FEP", "Disk defects", 2, NULL, freemap);
  diag = ifs_create_file (fs, "DIAGNOSTIC", "FEP", "Diagnostic area", 3, diag_datamap, freemap);
  free = ifs_create_file (fs, "FREE-PAGES", "FEP", "FEPFS free map", 4, free_datamap, freemap);
  seqn = ifs_create_file (fs, "SEQUENCE-NUMBER", "FEP", "FEPFS sequence number", 5, seqn_datamap, freemap);

  initialize_disk_buffer (buffer);
  set_feph_id (buffer, "FEPD");
  set_feph_header_version (buffer, current_fepd_header_version);
  set_feph_npages (buffer, 1);
  set_feph_timestamp (buffer, get_universal_time());
  set_feph_sequence_number (buffer, 0);
  set_fepd_page_number (buffer, 0);
  set_fepd_nentries (buffer, 6);
  set_fepd_entry (buffer, 0, "ROOT-DIRECTORY", "DIR", 1, root);
  set_fepd_entry (buffer, 1, "DISK-LABEL", "FEP", 1, lbel);
  set_fepd_entry (buffer, 2, "BAD-BLOCKS", "FEP", 1, badb);
  set_fepd_entry (buffer, 3, "FREE-PAGES", "FEP", 1, free);
  set_fepd_entry (buffer, 4, "SEQUENCE-NUMBER", "FEP", 1, seqn);
  set_fepd_entry (buffer, 5, "DIAGNOSTIC", "FEP", 1, diag);
  write_disk_block (fs, buffer, datamap_entry_address (root_datamap, 0));

  /* Write out the free bitmap */
  write_bitmap (fs, freemap, free_datamap);

  /* Initialize and write out the sequence number file */
  initialize_disk_buffer (buffer);
  set_sequence_number (buffer, 8);
  write_disk_block (fs, buffer, datamap_entry_address (seqn_datamap, 0));

  /* Lastly, initialize and write out 64 copies of the disk label */
  initialize_disk_buffer (buffer);
  set_feph_id (buffer, "LABL");
  set_feph_header_version (buffer, current_labl_header_version);
  set_feph_npages (buffer, 1);
  set_feph_timestamp (buffer, get_universal_time ());
  set_feph_sequence_number (buffer, 0);
  set_labl_ncylinders (buffer, 1);
  set_labl_nheads (buffer, 1);
  set_labl_npages_per_track (buffer, fs->disk_size);
  set_labl_creation_date (buffer, get_universal_time ());
  set_labl_author (buffer, "Benson");
  set_labl_pack_name (buffer, "Unknown");
  set_labl_comment (buffer, "Don't tread on me");
  set_labl_root_directory_dpn (buffer, root);
  set_labl_diagnostic_dpn (buffer, diag);
  set_labl_fep_dpn (buffer, 0);
  for (i=0; i<64; i++) write_disk_block (fs, buffer, i);

  force_disk_output (fs);
}
