/* -*-Mode: C; Package: C-user -*- */
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1990 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * Genera X font compilation and installation
 *
 * $Log:	fonts.c,v $
 * Revision 2.5.1.2  93/02/01  12:55:56  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:51:47  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:40:43  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/fonts.c,v 2.5.1.2 93/02/01 12:55:56 kaufman Rel $";
#endif

#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <strings.h>
#include "install.h"

#define FONTPATH_NAME "FONTPATH"
     /* Technically, this should be derived from OPENWINHOME */
#define FONTPATH_DEFAULT "/usr/openwin/lib/fonts:/home/openwin/lib/fonts"

void determine_font_parameters (params)
     struct params *params;
{
  char *cp, buf[BUFSIZ];
  arg arg0, arg1;
  int tmp;

  if (!SW_AVAILABLE_P(params, SW_FONTS)) {
    params->open_windows_fonts_p = FALSE;
    params->mit_fonts_p = FALSE;
    return;
  }
  terpri();

  while (1) {
    /* Determine whether to do Open Windows font compilation.
     * Technically, what we're searching for here is OPENWINHOME. */
    cp = file_search(S_IFDIR, "/usr/local/openwin", "/usr/openwin", "/home/openwin", NULL);
    if ((params->open_windows_fonts_p =
	 y_or_n_p("Do you wish to install the Genera fonts for the Sun Open Windows X server",
		  cp != NULL))) {

      /* Find Open Windows binaries */
      arg0.integer = S_IFDIR;
      arg1.integer = F_OK;
      params->open_windows_bin_directory =
	accept_string("Where do the Open Windows programs live", cp,
		      check_file, &arg0, &arg1);

      /* Find Open Windows Genera font directory.
       * It's possible that there is none, since this might be a new installation,
       * so if we can't find it try to set the default from the location of the
       * Open Windows directory.  Technically, the default here is OPENWINHOME/lib/fonts. */
      cp = file_search(S_IFDIR, "/usr/openwin-genera-fonts",
		       "/usr/local/openwin-genera-fonts",
		       "/usr/openwin/lib/genera-fonts",
		       "/home/openwin/lib/genera-fonts",
		       NULL);
      if (cp == NULL) {
	/* Construct default */
	Sprintf(buf, "%s/lib/genera-fonts", params->open_windows_bin_directory);
	cp = buf;
      }
      arg0.integer = S_IFDIR;
      arg1.integer = NONE_OK|W_OK;
      params->open_windows_font_directory =
	accept_string("Where should the Open Windows Genera fonts go", cp,
		      check_file, &arg0, &arg1);

      /* Check that FONTPATH will find our fonts */
      if (verify_directory_in_path(FONTPATH_NAME, FONTPATH_DEFAULT,
				   params->open_windows_font_directory) < 0) {
	Fprintf(stderr, "\nWARNING: directory %s is not in the search\n",
		params->open_windows_font_directory);
	Fprintf(stderr, "path defined by environment variable %s.  Make sure to update\n",
		FONTPATH_NAME);
	Fprintf(stderr, "%s to to include %s.\n\n",
		FONTPATH_NAME, params->open_windows_font_directory);
      }

      /* To do Open Windows font compilation, we need to be able to run the
       * convertfont and bldfamily programs and to write into the
       * Open Windows Genera fonts directory (checked at accept_string() time).
       */
      Sprintf(buf, "%s/bin/convertfont", params->open_windows_bin_directory);
      if (access(buf, X_OK)) {
	tmp = errno;
	Fprintf(stderr, "\n>> Can't compile the Genera fonts for Open Windows, because you\n");
	Fprintf(stderr, "   don't have permission to run the convertfont program:\n\t");
	errno = tmp;
	perror(buf);
	continue;
      }
      Sprintf(buf, "%s/bin/bldfamily", params->open_windows_bin_directory);
      if (access(buf, X_OK)) {
	tmp = errno;
	Fprintf(stderr, "\n>> Can't compile the Genera fonts for Open Windows, because you\n");
	Fprintf(stderr, "   don't have permission to run the bldfamily program:\n\t");
	errno = tmp;
	perror(buf);
	continue;
      }
    }
    /* This is odd, I know */
    break;
  }

  terpri();

  /* Determine whether to do MIT font compilcation */
  while (1) {
    cp = file_search(S_IFDIR, "/usr/local/bin/X11", "/usr/bin/X11", NULL);
    if ((params->mit_fonts_p =
	 y_or_n_p("Do you wish to install the Genera fonts for the MIT X server",
		  cp != NULL))) {
      /* Find MIT binaries */
      arg0.integer = S_IFDIR;
      arg1.integer = F_OK;
      params->mit_bin_directory =
	accept_string("Where do the MIT X programs live", cp, check_file, &arg0, &arg1);

      /* Find MIT Genera font directory.
       * It's possible that there is none, since this might be a new installation,
       * so if we can't find it try to set the default as a subdirectory of the
       * MIT X fonts directory. */
      cp = file_search(S_IFDIR, "/usr/local/lib/X11/fonts/genera",
		       "/usr/lib/X11/fonts/genera", NULL);
      if (cp == NULL &&
	  (cp = file_search(S_IFDIR, "/usr/local/lib/X11/fonts",
			    "/usr/lib/X11/fonts", NULL)) != NULL) {
	Sprintf(buf, "%s/genera", cp);
	cp = buf;
      }
      arg0.integer = S_IFDIR;
      arg1.integer = NONE_OK|W_OK;
      params->mit_font_directory =
	accept_string("Where should the MIT Genera fonts go", cp, check_file, &arg0, &arg1);

      /* To do MIT font compilation, we need to be able to run the bdftosnf and
       * mkfontdir programs and to write into the MIT Genera fonts directory.
       */
      Sprintf(buf, "%s/bdftosnf", params->mit_bin_directory);
      if (access(buf, X_OK)) {
	tmp = errno;
	Fprintf(stderr, "\n>> Can't compile the Genera fonts for the MIT server, because you\n");
	Fprintf(stderr, "   don't have permission to run the bdftosnf program:\n\t");
	errno = tmp;
	perror(buf);
	continue;
      }
      Sprintf(buf, "%s/mkfontdir", params->mit_bin_directory);
      if (access(buf, X_OK)) {
	tmp = errno;
	Fprintf(stderr, "\n>> Can't compile the Genera fonts for the MIT server, because you\n");
	Fprintf(stderr, "   don't have permission to run the mkfontdir program:\n\t");
	errno = tmp;
	perror(buf);
	continue;
      }
    }
    /* This is odd, I know */
    break;
  }
}



static mit_x_font_select_fn (file, source_directory, dest_directory,
			     source_file, dest_file)
     struct dirent *file;
     char *source_directory;
     char *dest_directory;
     char *source_file;
     char *dest_file;
{
  int i, len = strlen(file->d_name);

  /* Skip directory and parent entries, and anything that isn't a .bdf file */
  if (!strcmp(file->d_name, ".") ||
      !strcmp(file->d_name, "..") ||
      len < 4 ||
      strcmp(file->d_name+len-4, ".bdf")) {
    return(0);
  }
  
  /* Set up source and destination file names */
  Sprintf(source_file, "%s/%s", source_directory, file->d_name);
  file->d_name[len-4] = '\0';
  Sprintf(dest_file, "%s/%s.snf", dest_directory, file->d_name);

  /* Skip if source doesn't exist, or isn't newer */
  if ((i = source_newer_p((char *)NULL, source_file, (char *)NULL, dest_file, 1)) == -1 ||
      i == 0)
    return(0);

  return(1);
}

static mit_x_font_update_fn (source_file, dest_file)
     char *source_file;
     char *dest_file;
{
  char buf[BUFSIZ], *cp;
  int i;

  Sprintf(buf, "%s/bdftosnf %s > %s",
	  ux_installation_params.mit_bin_directory, source_file, dest_file);
  if ((i = shell_command(buf, 0)) == -1) {
    if ((cp = rindex(source_file, '/')) == NULL)
      cp = source_file;
    else
      cp++;
    Fprintf(stderr, ">> Error return code %d from bdftosnf for font %s -- ignored\n",
	    i, cp);
    return(0);
  }

  return(0);
}

compile_x_fonts_for_mit (params)
     struct params *params;
{
  char buf[BUFSIZ];
  int i;

  /* If fonts directory doesn't exist, create it */
  Printf("Installing Genera fonts for MIT X server...");
  (void) fflush(stdout);
  if (require_directory(params->mit_font_directory, 0)) {
    Fprintf(stderr, "  Couldn't create directory %s.\n", params->mit_font_directory);
    return(-1);
  }

  /* Run bdftosnf over new fonts */
  Sprintf(buf, "%s/fonts", params->distribution_directory);
  if ((i = process_directory(buf, params->mit_font_directory,
			     mit_x_font_select_fn, mit_x_font_update_fn, 1)) == -1)
    return(-1);
  else if (i > 0) {
    /* Run mkfontdir */
    Sprintf(buf, "cd %s; %s/mkfontdir",
	    params->mit_font_directory, params->mit_bin_directory);
    if ((i = shell_command(buf, 1)) != 0) {
      Fprintf(stderr, ">> Error return code %d from mkfontdir\n", i);
      return(-1);
    }
  }

  return(0);
}


static open_windows_x_font_select_fn (file, source_directory, dest_directory,
				      source_file, dest_file)
     struct dirent *file;
     char *source_directory;
     char *dest_directory;
     char *source_file;
     char *dest_file;
{
  int i, len = strlen(file->d_name);

  /* Skip directory and parent entries, and anything that isn't a .bdf file */
  if (!strcmp(file->d_name, ".") ||
      !strcmp(file->d_name, "..") ||
      len < 4 ||
      strcmp(file->d_name+len-4, ".bdf")) {
    return(0);
  }
  
  /* Set up source and destination file names */
  Sprintf(source_file, "%s/%s", source_directory, file->d_name);
  file->d_name[len-4] = '\0';
  Sprintf(dest_file, "genera-%s", file->d_name);

  /* Skip if source doesn't exist, or isn't newer.
   * convertfont frobs the filename by some mysterious algorithm,
   * so just check against one of the outputs of bldfamily. */
  i = source_newer_p((char *)NULL, source_file, dest_directory, "Families.list", 1);
  if (i == -1 || i == 0)
    return(0);

  return(1);
}

static open_windows_x_font_update_fn (source_file, dest_file)
     char *source_file;
     char *dest_file;
{
  char buf[BUFSIZ], *cp;
  int i;

  Sprintf(buf, "%s/bin/convertfont -f 100 -n %s -d %s %s > /dev/null",
	  ux_installation_params.open_windows_bin_directory,
	  dest_file, ux_installation_params.open_windows_font_directory,
	  source_file);
  if ((i = shell_command(buf, 0)) == -1) {
    if ((cp = rindex(source_file, '/')) == NULL)
      cp = source_file;
    else
      cp++;
    Fprintf(stderr, ">> Error return code %d from convertfont for font %s -- ignored\n",
	    i, cp);
    return(0);
  }

  return(0);
}

compile_x_fonts_for_open_windows (params)
     struct params *params;
{
  char buf[BUFSIZ];
  int i;

  /* If fonts directory doesn't exist, create it */
  Printf("Installing Genera fonts for Open Windows...");
  (void) fflush(stdout);
  if (require_directory(params->open_windows_font_directory, 0)) {
    Fprintf(stderr, "  Couldn't create directory %s.\n",
	    params->open_windows_font_directory);
    return(-1);
  }

  /* Run convertfont over new fonts */
  Sprintf(buf, "%s/fonts", params->distribution_directory);
  if ((i = process_directory(buf, params->open_windows_font_directory,
			     open_windows_x_font_select_fn,
			     open_windows_x_font_update_fn, 1)) == -1)
    return(-1);
  else if (i > 0) {
    /* Run bldfamily */
    Sprintf(buf, "cd %s; %s/bin/bldfamily -f 100",
		   params->open_windows_font_directory, params->open_windows_bin_directory);
    if ((i = shell_command(buf, 1)) != 0) {
      Fprintf(stderr, ">> Error return code %d from bldfamily\n", i);
      return(-1);
    }
  }

  /* --- Should we edit Families.list here? */

  return(0);
}


compile_x_fonts (params)
     struct params *params;
{

  if (params->open_windows_fonts_p) {
    terpri();
    if (compile_x_fonts_for_open_windows(params) == -1)
      return(-1);
  }

  if (params->mit_fonts_p) {
    terpri();
    if (compile_x_fonts_for_mit(params) == -1)
      return(-1);
  }

  return(0);
}
