/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/* UX400S genera application
 *
 * Flags related to booting are:
 *    -autoboot file
 *    -disable-timer
 *    -force
 *
 * Flags relating to Genera are:
 *    -activity name
 *    -remote-program-name name
 *    -status-line [yes|no]
 *
 * Flags related to the Genera X display are:
 *    -display display
 *    -geometry geometry
 *    -iconic [yes|no]
 *    -foreground-color color
 *    -background-color color
 *    -border-width width
 *    -border-color color
 *
 * Flags related to the cold-load X display are:
 *    -cl-display display
 *    -cl-geometry geometry
 *    -cl-iconic [yes|no]
 *    -cl-foreground-color color
 *    -cl-background-color color
 *    -cl-border-width width
 *    -cl-border-color color
 *
 * If no other arguments are given, then device is /dev/ivory0 (overridden
 * by getenv("IVORY")) and host is whatever host gets dug out of the network
 * channel.  If a pathname is given, it is used as the device, with the host
 * coming from the network channel.  If a host name is given, then no device
 * manipulation will be attempted.
 *
 * $Log:	genera.c,v $
 * Revision 2.5.1.2  93/02/01  12:56:02  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:51:52  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:40:58  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/genera.c,v 2.5.1.2 93/02/01 12:56:02 kaufman Rel $";
#endif
char *genera_version = "@(#)$Revision: 2.5.1.2 $"+15;

#include <sys/types.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <arpa/inet.h>
#include <dirent.h>
#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>

#define INCLUDE_FILE_HEADERS
#include "genera.h"
#include "embed.h"
#include "emblib.h"
#include "RPC.h"
#include "unix-support-client.h"
#include "life.h"

/*#define DEBUG			/**/
#ifdef DEBUG
#define DBG(msg) puts(msg)
#define DBG1(fmt,arg) printf(fmt, arg);
#define DBG2(fmt,arg0,arg1) printf(fmt, arg0, arg1);
#else
#define DBG(msg)
#define DBG1(fmt,arg)
#define DBG2(fmt,arg0,arg1)
#endif

char *note_name = NULL;

char *getenv();			/* Belongs in stdio.h */
extern char *sys_errlist[];
extern int sys_nerr;

#define MATCH(str, min) ((len <= sizeof(str)-1) && !strncmp(str, arg, len > min ? len : min))
#define MAX_CL_OUTPUT_ITEMS 100

void ignore ()
{}

static int nfds, ivory_fd = -1, rpc_fd = -1, x_fd = -1;
static fd_set read_master, write_master;
static struct x_params genera_params, cl_params;
static FILE *autoboot_fp = NULL;
static int who_line_p = UNSPEC, flags = 0;
static char *activity = NULL, *remote_program = NULL, *guest = NULL;
static int started_console_handler = 0;
static long guest_address = 0;
static long window_id = -1;
static int ok_to_jump = 0;

/* Run lights stuff */
int manage_run_lights = 0;
long run_lights_state;

char *get_cold_load_name();
#define update_cold_load_name() set_cold_load_name(get_cold_load_name())

main (argc, argv)
     int argc;
     char *argv[];
{
  int i, len, junk;
  char *arg, local_host[BUFSIZ];
  char *device = NULL;
  EmbColdLoadChannel *cold_channel;
  void close_cold_load();

  /* Setup default display, if any */
  bzero((char *) &genera_params, sizeof(struct x_params));
  bzero((char *) &cl_params, sizeof(struct x_params));
  genera_params.xp_border_width = cl_params.xp_border_width = -1;
  cl_params.xp_name_fn = get_cold_load_name;
  gethostname(local_host, BUFSIZ-1);
  set_param_host_name(&genera_params, local_host);
  set_param_host_name(&cl_params, local_host);
  if ((arg = getenv("DISPLAY")) != NULL) {
    parse_x_param(&genera_params, "display", strdup(arg));
    parse_x_param(&cl_params, "display", strdup(arg));
  }

  /* Parse switches */
  for (i = 1; i < argc; i++)
    if (argv[i][0] == '/' || argv[i][0] == '.') {
      /* If it's got a leading / or ., it's a device */
      if (device != NULL)
	usage(argv[i], "Only one device may be specified");
      device = argv[i];
    } else if (argv[i][0] != '-') {
      /* If it doesn't have a leading -, it's a host */
      if (guest != NULL)
	usage(argv[i], "Only one host may be specified");
      guest = argv[i];
    } else {
      arg = argv[i]+1;
      len = strlen(arg);

      if (MATCH("autoboot", 2)) { 		/* Booting flags */
	if (argc == i+1)
	  usage(NULL, NULL);
	else {
	  if (strcmp(argv[++i], "-")) {
	    if ((autoboot_fp = fopen(argv[i], "r")) == NULL) {
	      perror(argv[i]);
	      exit(1);
	    }
	  } else
	    autoboot_fp = stdin;
	}
      } else if (MATCH("any", 2))
	flags |= FL_ANY_IVORY;
      else if (MATCH("disable-timer", 4))
	flags |= FL_DISABLE_FEP_TIMER;
      else if (MATCH("force", 4))
	flags |= FL_FORCE;
      else if (MATCH("help", 1))
	usage(NULL, NULL);
      else if (MATCH("silent", 1))
	flags |= FL_SILENT;
      else if (MATCH("activity", 2)) 		/* Genera flags */
	activity = argv[++i];
      else if (MATCH("no-console", 4)) {
	flags |= FL_NO_CONSOLE;
	started_console_handler = -1;
      } else if (MATCH("status-line", 4)) {
	if (argc > i+1 && !strcmp(argv[i+1], "yes")) {
	  who_line_p = T;
	  i++;
	} else if (argc > i+1 && !strcmp(argv[i+1], "no")) {
	  who_line_p = NIL;
	  i++;
	} else
	  who_line_p = T;
      } else if (MATCH("remote-program-name", 1))
	remote_program = argv[++i];
      else if (junk = parse_x_param(&genera_params, arg, argc > i+1 ? argv[i+1] : "")) {
	/* Genera console flag */
	if (junk == 2)
	  i++;
      } else if (!strncmp(arg, "cl-", 3) &&
		 (junk = parse_x_param(&cl_params, arg+3, argc > i+1 ? argv[i+1] : ""))) {
	/* Cold-load console flag */
	if (junk == 2)
	  i++;
      } else
	usage(argv[i], "Unknown flag argument");
    }

  /* Must have name (not address) for cold-load display */
  if (cl_params.xp_host_name == NULL && cl_params.xp_host_address != 0)
    usage("-cl-display", "Host argument must be a name (not an address)");

  /* Final defaults */
  if (device == NULL && guest == NULL) {
    if ((device = getenv("IVORY")) == NULL)
      device = DEFAULT_GUEST();
  }
  if (flags & FL_ANY_IVORY)
    if ((ivory_fd = find_unlocked_ivory(&device, flags)) == -1)
      exit(1);
    else
      flags |= FL_IVORY_PRELOCKED;
  if (device != NULL)
    flags |= FL_LOOKUP_GUEST_ADDRESS;
  FD_ZERO(&read_master);
  FD_ZERO(&write_master);

  /* Set up device name and cold load icon name */
  if (device == NULL) {
    struct hostent *hp;

    cl_params.xp_device_name = NULL;
    if ((hp = gethostbyname(guest)) != NULL)
      note_name = strdup(hp->h_name);
    else 
      note_name = guest;
  } else {
    char *cp, buf[BUFSIZ];

    if ((cp = rindex(device, '/')) == NULL) {
      cl_params.xp_device_name = strdup(device);
      cl_params.xp_icon_name = "Cold Load";
    } else {
      struct hostent *hp;

      ++cp;
      cl_params.xp_device_name = strdup(cp);
      if ((hp = gethostbyname(local_host)) != NULL &&
	  cl_params.xp_host_address != *(unsigned long *) hp->h_addr) {
	cl_params.xp_icon_name = calloc(strlen(cp) + strlen(local_host) + 2, 1);
	strcpy(cl_params.xp_icon_name, cp);
	strcat(cl_params.xp_icon_name, "@");
	strcat(cl_params.xp_icon_name, local_host);
	if ((cp = index(cl_params.xp_icon_name, '.')) != NULL)
	  *cp = '\0';
      } else
	cl_params.xp_icon_name = cl_params.xp_device_name;
    }
    note_name = device;
  }

  /* If we're meant to play with the device, go for it */
  if (device != NULL) {
    if ((ivory_fd = boot_ivory(device, ivory_fd, flags)) >= 0) {
      nfds = ivory_fd + 1;
      FD_SET(ivory_fd, &read_master);
      if (autoboot_fp != NULL)
	FD_SET(ivory_fd, &write_master);

      /* We've frobbed the device, maybe we should create a cold load */
      cold_channel = EmbHostPtr(EmbCommAreaPtr->cold_load_channel);
      if (cold_channel) {
	if (cold_channel->selproc) {
	  NOTE("There is already a cold-load display");
	} else {
	  if (cl_params.xp_iconic == UNSPEC)
	    cl_params.xp_iconic = (EmbCommAreaPtr->fep_status == IdleFEPStatus) ? T : NIL;
	  if ((x_fd = open_cold_load_display(&cl_params, get_cold_load_name())) != -1) {
	    FD_SET(x_fd, &read_master);
	    if (x_fd >= nfds)
	      nfds = x_fd+1;
	    signal(SIGINT, close_cold_load);
	    signal(SIGHUP, close_cold_load);
	    signal(SIGTERM, close_cold_load);
	    set_version(&EmbCommAreaPtr->boot_version, genera_version);
	    update_cold_load_name();
	  }
	}
      }
    } else if (ivory_fd == -1)
      /* Fatal error booting Ivory */
      exit(1);

    run_cold_load();
    stop_cold_x();
    if (started_console_handler > 0)
      kill(started_console_handler, SIGINT);
  } else
    run_genera_console(0);
  exit(0);
}

static find_unlocked_ivory (devicep, flags)
     char **devicep;
     int flags;
{
  int fd;
  DIR *dev;
  struct dirent *dir;
  static char *ivory = "/dev/ivoryXXX";

  flags |= FL_LOCK_REQUIRED;

  /* Check preferred device first */
  if (*devicep != NULL) {
    if ((fd = MapEmbCommArea(*devicep)) != -1 &&
	lock_ivory(fd, *devicep, flags) == 0)
      return(fd);
    UnMapEmbCommArea();
  }

  if ((dev = opendir("/dev")) == NULL) {
    perror("/dev");
    return(-1);
  }

  while ((dir = readdir(dev)) != NULL) {
    if (strncmp("ivory", dir->d_name, 5))
      continue;

    strcpy(ivory+5, dir->d_name);
    if (strcmp(ivory, *devicep)) {
      if ((fd = MapEmbCommArea(ivory)) != -1 &&
	  lock_ivory(fd, ivory, flags) == 0) {
	*devicep = ivory;
	closedir(dev);
	return(fd);
      }
      UnMapEmbCommArea();
    }
  }

  closedir(dev);
  puts("All Ivory devices are locked.");
  return(-1);
}

/* Wait for exited children */
static void sigchld_handler ()
{
  union wait status;
  
  DBG("sigchld_handler()");
  while (wait3(&status, WNOHANG, NULL) > 0) {
    if (WIFEXITED(status) && WEXITSTATUS(status) == -1)
      /* If child exited with status -1, we also should exit */
      exit(-1);

    if (WIFEXITED(status) || WIFSIGNALED(status)) {
      /* Child exited, so there's no longer a console handler */
      DBG("child exited");
      started_console_handler = 0;
      guest_address = 0;
      guest = NULL;
      flags |= FL_LOOKUP_GUEST_ADDRESS;
    }
  }

  signal(SIGCHLD, sigchld_handler);
  return;
}

char *get_cold_load_name ()
{
  char *base = NULL;
  static char buf[128];

  switch(EmbCommAreaPtr->fep_status) {

  case HaltedFEPStatus:
    base = "Device PROM";
    break;

  case RunningFEPStatus:
    base = "FEP";
    break;

  case IdleFEPStatus:
    base = "Cold Load Stream";
    break;

  default:
    return(NULL);
  }

  if (cl_params.xp_icon_name == NULL)
    sprintf(buf, "Ivory %s", base);
  else
    sprintf(buf, "%s %s", cl_params.xp_icon_name, base);
  return(buf);
}

void close_cold_load ()
{
  DBG("close_cold_load()");
  stop_cold_x();

  /* Give child process some time to close the X display before we (possibly)
   * turn off life-support */
  if (started_console_handler > 0)
    sleep(2);
  DBG("closed");
  exit(0);
}


run_cold_load ()
{
  int junk;
  struct timeval timeout;
  fd_set read, write;
  int fep_stat = -2, fep_has_run = FALSE;

  FD_ZERO(&read);
  FD_ZERO(&write);

  for (;;) {

    /* Handle pending X input and errors */
    if (x_fd > 0) {
      int old_x_fd = x_fd;

      if (flags & FL_FIRST_PASS) {
	flags ^= FL_FIRST_PASS;
	junk = setup_x_io_error_handler();
      } else
	junk = 0;

      if (junk || FD_ISSET(x_fd, &read)) {
	/* If can read from x_fd, there're events pending */
	x_fd = manage_x_input(&cl_params);
	if (x_fd != old_x_fd) {
	  FD_CLR(old_x_fd, &read_master);
	  FD_SET(x_fd, &read_master);
	}
      }
    }

    if (fep_stat != EmbCommAreaPtr->fep_status) {
      fep_stat = EmbCommAreaPtr->fep_status;
      if (fep_stat == RunningFEPStatus) {
	NOTE("The FEP is running");
	fep_has_run = TRUE;
	update_cold_load_name();
      } else if (fep_stat == IdleFEPStatus) {
	NOTE("Lisp is running");
	update_cold_load_name();
      } else if (fep_has_run && fep_stat == HaltedFEPStatus) {
	NOTE("The FEP has been shut down");
	return;
      }
    }

    if (EmbCommAreaPtr->fep_status == IdleFEPStatus) {
      /* If we're meant to look for the host name, do it now */
      if (flags & FL_LOOKUP_GUEST_ADDRESS)
	switch (find_guest_address(&guest, &guest_address)) {
	case 1:			/* Got an address */
	  flags ^= FL_LOOKUP_GUEST_ADDRESS;
	  break;

	case 0:			/* No address yet */
	  break;

	case -1:		/* Protocol not IP */
	  fprintf(stderr, "Guest primary protocol is not IP, can't talk to it.\n");
	  flags ^= FL_LOOKUP_GUEST_ADDRESS;
	  break;

	case -2:
	  fprintf(stderr, "Can't parse guest address.\n");
	  break;
	}

      /* If we've got a host, start up a display */
      if (guest && !started_console_handler) {
	started_console_handler = run_genera_console(1);
	signal(SIGCHLD, sigchld_handler);
      }
    }

    /* What to do, what to do */
    bcopy((char *) &read_master, (char *) &read, sizeof(fd_set));
    bcopy((char *) &write_master, (char *) &write, sizeof(fd_set));
    timeout.tv_sec = 0;
    timeout.tv_usec = 250000;
    if (select(nfds, &read, &write, NULL, &timeout) == -1 && errno != EINTR) {
      int i;/*---*/

      perror("run_genera select");
#ifdef notdef
      printf("read: ");
      for (i=0;i<FD_SETSIZE;i++)
	if (FD_ISSET(i, &read_master))
	  printf("%d ", i);
      printf("write: ");
      for (i=0;i<FD_SETSIZE;i++)
	if (FD_ISSET(i, &write_master))
	  printf("%d ", i);
      printf("\nnfds %d ivory %d x %d\n", nfds, ivory_fd, x_fd);
#endif
      break;
    }

    /* If can read from ivory_fd, it's cold-load output */
    if (ivory_fd > 0 && x_fd > 0 && FD_ISSET(ivory_fd, &read)) {
      manage_cold_load_output(MAX_CL_OUTPUT_ITEMS);
    }

    /* If can write to ivory_fd, it's ready for autoboot input */
    if (ivory_fd > 0 && FD_ISSET(ivory_fd, &write) && !EmbCommAreaPtr->fep_busy) {
      if (junk = manage_autoboot_input(autoboot_fp)) {
	/* Autoboot finished */
	FD_CLR(ivory_fd, &write_master);
	fclose(autoboot_fp);
	/* -1 returned if finished with error */
	if (junk < 0)
	  fprintf(stderr, "Error during autoboot\n");
	else if (EmbCommAreaPtr->fep_status != IdleFEPStatus)
	  fprintf(stderr, "Autoboot file has finished, but Lisp isn't running\n");
      }
    }

    if (x_fd > 0)
      update_cold_load_blinkers();
  }
  exit(1);
}

set_param_host_name (params, name)
     struct x_params *params;
     char *name;
{
  struct hostent *hp;

  if (*name == '\0' || !strcmp(name, "unix"))
    /* Do nothing, since we've already got the local host wired in */
    ;
  else if ((hp = gethostbyname(name)) != NULL) {
    params->xp_host_address = *(unsigned long *) hp->h_addr;
    params->xp_host_name = strdup(hp->h_name);
  } else if ((params->xp_host_address = inet_addr(name)) == -1) {
    usage(name, "Unknown host");
    /*NOTREACHED*/
  } else
    params->xp_host_name = NULL;
}

/* Return 0 if no match, 1 if match, 2 if match & consumed next */
parse_x_param (params, arg, next)
     struct x_params *params;
     char *arg, *next;
{
  int len = strlen(arg);
  int junk1 = -1, junk2 = -1;
  char *cp;

  if (MATCH("display", 4)) {
    if (cp = index(next, ':')) {
      /* -display host:screen.display */
      *cp++ = '\0';
      set_param_host_name(params, next);
    } else if (*next) {
      /* -display host */
      set_param_host_name(params, next);
      return(2);
    } else {
      /* -display */
      usage(arg-1, "Incorrect argument syntax");
      /*NOTREACHED*/
    }

    sscanf(cp, "%d.%d", &junk1, &junk2);
    params->xp_display = junk1;
    params->xp_screen = junk2;
  } else if (MATCH("geometry", 1)) {
    params->xp_geometry = next;
  } else if (MATCH("iconic", 2)) {
    if (!strcmp(next, "yes")) {
      params->xp_iconic = T;
    } else if (!strcmp(next, "no")) {
      params->xp_iconic = NIL;
    } else {
      params->xp_iconic = T;
      return(1);
    }
  } else if (MATCH("foreground-color", 4))
    params->xp_foreground_color = next;
  else if (MATCH("background-color", 2))
    params->xp_background_color = next;
  else if (MATCH("border-color", 8))
    params->xp_border_color = next;
  else if (MATCH("border-width", 8))
    params->xp_border_width = atoi(next);
  else
    return(0);

  return(2);
}

static char *DFLT = "";
#define DEFAULT(str) (str == NULL ? DFLT : str)
static int embedded_appropriate_p = FALSE;

/* Return -1 for fatal error, 0 for no error, 1 for transient error */
open_genera_display (params, activity, remote_program_name, who_line_p, ret_window_id)
     struct x_params *params;
     char *activity, *remote_program_name;
     int who_line_p;
     long *ret_window_id;
{
  unsigned char *description;
  int reuse_old_screen_p = UNSPEC; /* No way for user to specify this */
  int val, fatalp, junk;

  /* Call out to Genera */
  val = start_x_screen(inet_ntoa(*(struct in_addr *) &params->xp_host_address),
		       "INTERNET",
		       params->xp_display,
		       params->xp_screen,
		       reuse_old_screen_p,
		       DEFAULT(params->xp_geometry),
		       params->xp_iconic,
		       DEFAULT(remote_program_name),
		       DEFAULT(activity),
		       DEFAULT(params->xp_foreground_color),
		       DEFAULT(params->xp_background_color),
		       DEFAULT(params->xp_border_color),
		       params->xp_border_width,
		       who_line_p,
		       embedded_appropriate_p,
		       ret_window_id);
  DBG2("Opened %x, val %d\n", *ret_window_id, val);

  /* Check for error */
  switch (val) {
  case RPC_DENIED:
    fprintf(stderr, "RPC access denied\n");
    return(-1);

  case RPC_PROG_UNAVAIL:
    fprintf(stderr, "Booted version of Genera doesn't support required RPC module\n");
    return(-1);

  case RPC_PROG_MISMATCH:
    fprintf(stderr, "Booted version of Genera doesn't support required version of RPC module\n");
    return(-1);

  case RPC_PROC_UNAVAIL:
    fprintf(stderr, "Booted version of Genera doesn't support required RPC entry\n");
    return(-1);

  case RPC_GARBAGE_ARGS:
    fprintf(stderr, "Genera couldn't parse RPC call\n");
    return(-1);

  case RPC_REMOTE_ERROR:
    if (console_error(&description, &fatalp) == noErr) {
      fprintf(stderr, ">> %s\n", description);
      if (fatalp == TRUE)
	return(-1);
      else
	return(1);
    } else {
      RPCRemoteError(&junk);
      printf("Genera signalled unexpected remote error %d\n", junk);
      return(-1);
    }

  default:		/* Internal implementation error */
    if (val != 0 && val != EINTR) {
      if (errno < sys_nerr)
	fprintf(stderr, "console on %s: %s\n", guest, sys_errlist[val]);
      else
	fprintf(stderr, "console on %s: Unknown error %d\n", guest, val);
      return(-1);
    } else
      return(val ? 1 : 0);
  }
}

close_genera_display (params, window_id, destroy_p)
     struct x_params *params;
     long window_id;
{
  unsigned char *description;
  int val, fatalp;

  manage_run_lights = 0;
  if (destroy_p) {
    DBG1("Destroying %x\n", window_id);
    DestroyWindow(params, window_id);
    val = 0;
  } else {
    DBG1("Closing %x\n", window_id);

    val = halt_x_screen(inet_ntoa(*(struct in_addr *) &params->xp_host_address),
			"INTERNET",
			params->xp_display,
			params->xp_screen,
			window_id);

    /* Check for error.  We only care about a certain class of Remote Error */
    if (val == RPC_REMOTE_ERROR)
      if (console_error(&description, &fatalp) == noErr) {
	val = noErr;
	fprintf(stderr, ">> %s\n", description);
	if (fatalp == TRUE) {
	  DestroyWindow(params, window_id);
	}
      }
  }

  return(val);
}

position_run_lights (address, network, display_number, screen_number, rl_window_id,
		     nlights, width, height, x, y, dx, dy,
		     foreground, background, plane_mask)
     char *address, *network;
     short display_number;
     short screen_number;
     long rl_window_id;
     short nlights;
     unsigned long width;
     unsigned long height;
     unsigned long x;
     unsigned long y;
     unsigned long dx;
     unsigned long dy;
     unsigned long foreground;
     unsigned long background;
     unsigned long plane_mask;
{
  if (ivory_fd != -1) {
#ifdef DEBUG
    printf("window %x n %d wid %d hei %d x %d y %d dx %d dy %d fore %d back %d p_m %d\n",
	   rl_window_id, nlights, width, height, x, y, dx, dy,
	   foreground, background, plane_mask);
#endif
    open_run_lights_display(&genera_params, rl_window_id,
			    nlights, width, height, x, y, dx, dy,
			    foreground, background, plane_mask);
    manage_run_lights = 1;
    return(0);
  } else
    return(1);
}

remove_run_lights(address, network, display_number, screen_number, rl_window_id)
     char *address, *network;
     short display_number;
     short screen_number;
     long rl_window_id;
{
  if (ivory_fd != -1) {
    DBG("run lights disabled");
    manage_run_lights = 0;
    close_run_lights_display();
    return(0);
  } else
    return(1);
}

jmp_buf rpc;

void jump ()
{
  DBG("jump");
  longjmp(rpc, 2);
}

void pipe ()
{
  DBG("pipe");
  longjmp(rpc, 1);
}


run_genera_console (in_child)
     int in_child;
{
  int i, fd;
  int exiting = 0, error_count = 0;
  char buf[BUFSIZ], *cp;
  struct hostent *hp = NULL;

  if (in_child)
    if ((i = fork()) == -1) {
      perror("fork");
      exit(1);
    } else if (i != 0) {
      /* Parent */
      return(i);
    } else {
      /* Child */
      select_disable_child_hook();
      close_display_child_hook();
    }

  /* Set up RPC */
  InitializeRPC();
  initialize_predefined_remote_entries_server();
  initialize_ux400_run_lights_server();
  rpc_server_name = "Genera";
  rpc_server_version = genera_version;
  if (cp = index(rpc_server_version, ' '))
    *cp = '\0';

  /* Allow X access from guest to server */
  if (gethostname(buf, BUFSIZ-1) == 0 &&
      (hp = gethostbyname(buf)) != NULL &&
      genera_params.xp_host_address == *(unsigned long *) hp->h_addr) {
    /* X server is local host, so we can add guest to its access list */
    int address = -1;

    if (guest_address != 0)
      /* Save some computrons */
      address = guest_address;
    else if ((hp = gethostbyname(guest)) != NULL)
      address = *(unsigned long *) hp->h_addr;
    else
      address = inet_addr(guest);

    if (address != -1)
      AllowXAccess(&genera_params, address);
  }

  /* Manage console */
  for (;;) {

    if (in_child) {
      long current_guest_address;
      char *current_guest_name;
      int i;
      
      /* Check that guest IP address hasn't changed */
      if ((i = find_guest_address(&current_guest_name, &current_guest_address)) != 1 ||
	  current_guest_address != guest_address) {
	exit(0);
      }
    }


    while (tcp_agent_open(guest, IPPORT_RPC) != 0) {

      if (errno != EINTR) {
	if (!in_child || (errno != ECONNREFUSED && errno != ETIMEDOUT)) {
	  if (errno < sys_nerr)
	    fprintf(stderr, "connect to %s: %s\n", guest, sys_errlist[errno]);
	  else
	    fprintf(stderr, "connect to %s: Unknown error %d\n", guest, errno);
	  exit(-1);
	} else {
	  if (errno < sys_nerr) {
	    /* Ignore first three refusals/timeouts, to relieve burden on user */
	    if (++error_count > 3)
	      fprintf(stderr, "%s: %s -- retrying...\n", guest, sys_errlist[errno]);
	  } else
	    fprintf(stderr, "%s: Unknown error %d -- retrying...\n", guest, errno);
	  sleep(6);
	}
      }
    }

    /* Enable keep alives (sigh) */
    fd = RPCAgentFileDescriptor();
    i = 1;
    if (setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, (char *) &i, sizeof(i)) < 0)
      perror("enable keep-alives");

    if (window_id == -1 ||
	(in_child && EmbCommAreaPtr->restart_applications)) {

      NOTE("Starting console");
      while (i = open_genera_display(&genera_params, activity, remote_program, who_line_p,
				     &window_id))
	switch (i) {
	case 1:
	  /* Transient error */
	  fprintf(stderr, "\nRetrying...\n");
	  break;

	case -1:
	  fprintf(stderr, "Unable to start console.\n");
	  exit(-1);
	}

      /* Application restarted, so reset this */
      if (in_child && EmbCommAreaPtr->restart_applications != 0) {
	EmbCommAreaPtr->restart_applications = 0;
	EmbCommAreaPtr->restart_applications_count++;
      }
    }
	  

    if ((i = setjmp(rpc)) == 0) {
      /* Normal case */
      ok_to_jump = in_child ? 1 : -1;
      signal(SIGHUP, jump);
      signal(SIGINT, jump);
      signal(SIGTERM, jump);
      signal(SIGPIPE, pipe);
      for (;;)
	BusyWait(true);
    } else if (i == 1) {
      /* Agent closed behind our backs */
      DBG("No agent");
      if (exiting) {
	exiting = 2;
	break;
      } else {
	if (in_child && EmbCommAreaPtr->restart_applications)
	  close_genera_display(&genera_params, window_id, 1);
	ok_to_jump = 0;
	continue;
      }
    } else if (i == 2) {
      /* Signal received */
      DBG("exiting");
      exiting = 1;
      break;
    } else {
      fprintf(stderr, "longjmp %d: internal error\n", i);
      abort();
    }
  }

  if (exiting == 1 &&
      close_genera_display(&genera_params, window_id,
			   in_child && EmbCommAreaPtr->fep_status == HaltedFEPStatus))
    perror("closing display");
  CloseRPC();
  exit(0);
}

void BusyWait(allow_rpc)
     int allow_rpc;
{
  int cc = 0, rpc_fd, junk;
  struct timeval timeout;
  fd_set read;

  /* Set up */
  rpc_fd = RPCAgentFileDescriptor();

  for (; cc == 0 && manage_run_lights != -1;) {

    /* Wait for something interesting to happen */
    timeout.tv_sec = 0;
    timeout.tv_usec = 100000;
    FD_ZERO(&read);
    FD_SET(rpc_fd, &read);
    if ((cc = select(rpc_fd+1, &read, NULL, NULL, &timeout)) < 0 &&
	errno != EINTR && errno != EWOULDBLOCK) {
      perror("BusyWait select");
      exit(-1);
    }

    /* Update run lights */
    if (manage_run_lights > 0 && run_lights_state != EmbCommAreaPtr->run_lights)
      if (update_run_lights(&run_lights_state) == -1)
	/* Window has been destroyed: stop frobbing run lights */
	manage_run_lights = 0;

    /* Check to see if Lisp has been doing disastrous things */
    if (ok_to_jump &&
	((ok_to_jump > 0 && EmbCommAreaPtr->restart_applications) ||
	 (cc > 0 && FD_ISSET(rpc_fd, &read) && !ioctl(rpc_fd, FIONREAD, &junk) && !junk))) {
      /* If the restart_applications word has gone on, or we're ready to read
       * from the RPC fd but there's nothing to read (Unix's way of saying the
       * connection is closed), then close down the agent and open a new one. */
      CloseRPC();
      longjmp(rpc, 1);
    }

    /* RPC recursion */
    if (allow_rpc)
      PollRPC();
  }
}

static char *usage_message[] = {
  "usage: genera [options] [device] [hostname]\n",
  "  program options are: -any, -autoboot file, -disable-timer, -force,\n",
  "\t-help, -no-console\n",
  "  Genera options are: -activity name, -remote-program-name name, \n",
  "\tand -status-line [yes|no]\n",
  "  Genera display options are: -display host:[display.[screen]],\n",
  "\t-geometry geometry, -foreground-color color, -background-color color\n",
  "\t-border-width width, -border-color color, and -iconic [yes|no]\n",
  "  Cold load display options are: -cl-display host:[display.[screen]],\n",
  "\t-cl-geometry geometry, -cl-foreground-color color,\n",
  "\t-cl-background-color color, -cl-border-width width,\n",
  "\t-cl-border-color color, and -cl-iconic [yes|no]\n",
  NULL,
};
  
usage (arg, message)
     char *arg;
     char *message;
{
  int i;

  if (message) {
     if (arg)
       fprintf(stderr, "%s: ", arg);
     fputs(message, stderr);
     fputs("\n\n", stderr);
   }

  for (i = 0; usage_message[i] != NULL; i++)
    fputs(usage_message[i], stderr);

  exit(1);
}



/* Look up primary network address of guest
 * Return values are:
 *    -2 for address not registered in UNIX hosts database
 *    -1 for primary network type not IP
 *     0 for no primary network address set yet
 *     1 for success
 */

find_guest_address (name, address)
     char **name;
     long *address;
{
  EmbNetChannel *net, *find_emb_network_channel();
  struct hostent *hp;

  *name = NULL;
  if ((net = find_emb_network_channel()) != NULL &&
      net->guest_primary_protocol != -1 &&
      net->guest_primary_address != -1)
    if (net->guest_primary_protocol != ETHERTYPE_IP)
      return(-1);
    else {
      *address = net->guest_primary_address;
      if ((hp = gethostbyaddr((char *) &net->guest_primary_address,
			      sizeof(net->guest_primary_address), AF_INET)) != NULL)
	*name = hp->h_name;
      else
	*name = inet_ntoa(*(struct in_addr *) &net->guest_primary_address);
      if (*name != NULL) {
	*name = strdup(*name);
	return(1);
      } else
	return(-2);
    }
  
  return(0);
}
