/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * $Log:	halt.c,v $
 * Revision 2.5.1.2  93/02/01  12:56:06  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:51:57  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:41:07  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/halt.c,v 2.5.1.2 93/02/01 12:56:06 kaufman Rel $";
#endif

#include <sys/types.h>
#include <sys/file.h>
#include <sys/mman.h>
#include <stdio.h>

#define INCLUDE_FILE_HEADERS
#include "embed.h"
#include "life.h"
#undef INCLUDE_FILE_HEADERS

main (argc, argv)
     int argc;
     char *argv[];
{
  int fd;
  int i, force = 0;
  char *ivory, buf[BUFSIZ];
  char *getenv();
  caddr_t ptr;
  EmbCommArea *EmbCommAreaPtr;
  IvoryConfig config;

  if (argc > 2) {
    fprintf(stderr, "usage: %s [device]\n", argv[0]);
    exit(1);
  }

  /* Figure out path to device */
  if (argc > 1) {
    if (index(argv[1], '/') != NULL)
      ivory = argv[1];
    else {
      sprintf(buf, "/dev/%s", argv[1]);
      ivory = buf;
    }
  } else if ((ivory = getenv("IVORY")) == NULL)
    ivory = DEFAULT_GUEST();

  if ((fd = open(ivory, O_RDWR)) < 0) {
    perror(ivory);
    exit(1);
  }
  check_kernel_version();

  if (ioctl(fd, IVORYGCONF, &config)<0) {
    perror("IVORYGCONF");
    exit(1);
  }
  if ((ptr = mmap(0, config.mmap_size, PROT_READ|PROT_WRITE,
		  MAP_SHARED, fd, 0)) == (caddr_t) -1) {
    perror("mmap");
    exit(1);
  }

  EmbCommAreaPtr = (EmbCommArea *) (ptr + config.comm_area_base);
  if (EmbCommAreaPtr->identifier != 'DBME') {
    fprintf(stderr, "Life support is already disabled\n");
    exit(0);
  } else if (EmbCommAreaPtr->fep_status == RunningFEPStatus) {
    printf("Lisp is not running\n");
    exit(0);
  } else if (EmbCommAreaPtr->fep_status == HaltedFEPStatus) {
    printf("Neither Lisp nor the FEP is running\n");
    exit(0);
  }

  printf("Asking Lisp to stop...");
  fflush(stdout);
  EmbCommAreaPtr->stop_request = -1;

  /* Wait a bit before checking the first time */
  usleep(250000);

  /* Give Lisp 5 seconds to stop itself gracefully */
  for (i = 0; i < 10; i++) {
    if (EmbCommAreaPtr->fep_status == RunningFEPStatus) {
      printf("  Lisp stopped.\n");
      exit(0);
    }
    usleep(500000);
  }

  printf("\nLisp not stopped, asking FEP to take control...");
  fflush(stdout);
  if (ioctl(fd, IVORYRESET, 0) == -1) {
    perror("IVORYRESET");
    exit(1);
  }

  for (i = 0; i < 10; i++) {
    if (EmbCommAreaPtr->fep_status == RunningFEPStatus) {
      printf("  Lisp forcibly stopped.\n");
      exit(0);
    }
    usleep(500000);
  }

  printf("Lisp wouldn't stop itself, I've run out of ideas.\n");
  exit(-1);
}
