/* -*- Mode: C; Tab-width: 4 -*- */
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1990 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/* Support for IFEP access to host file system
 *
 * $Log:	host-file.c,v $
 * Revision 2.5.1.2  93/02/01  12:56:07  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:51:59  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:41:11  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */


#ifndef lint
static char *rcsid = "$Header: /usr/share/local/symbolics/src/RCS/host-file.c,v 2.5.1.2 93/02/01 12:56:07 kaufman Rel $";
#endif

#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <dirent.h>
#include <errno.h>
#include <pwd.h>
#include <stdio.h>
#include <syslog.h>

#include "embed.h"
#include "emblib.h"
#include "life.h"
#include "config.h"

#define UNIX_EPOCH 2208988800

#ifdef TEST
#define IVORY_DEBUG
#endif

void HostFileLife (fd, EmbCommAreaPtr, conf)
     int fd;
     EmbCommArea *EmbCommAreaPtr;
     Config *conf;
{
  EmbHostFileChannel *channel, *FindHostFileChannel();
  fd_set read;
  EmbPtr ptr;

  RegisterEmbCommArea(fd, EmbCommAreaPtr);

  /* Locate HostFile channel */
  if ((channel = FindHostFileChannel(EmbCommAreaPtr)) == NULL) {
#ifdef IVORY_DEBUG
    fprintf(stderr, "Couldn't find Host File channel\n");
#endif
    exit(-1);
  }

  /* Set user ID */
  if (conf->emb.host_file_uid != getuid())
    if (setuid(conf->emb.host_file_uid) == -1) {
      log_error(LOG_ERR, "setuid");
      log_message(LOG_ERR, "Host file service disabled");
      pause();
    }

  /* Enable select() */
  ptr = GuestPointer(channel);
  if (ioctl(fd, IVORYSELCHANNEL, &ptr) == -1) {
    log_error(LOG_ERR, "IVORYSELCHANNEL (host file)");
    exit(-1);
  }

  /* Loop waiting for and executing Host File commands */
  while (1) {
    FD_ZERO(&read);
    FD_SET(fd, &read);

    if (select(fd+1, &read, NULL, NULL, NULL) == -1 && errno != EINTR) {
      log_error(LOG_ERR, "select");
      exit(-1);
    }

    if (FD_ISSET(fd, &read))
      CheckHostFile(channel);
  }
}

CheckHostFile(channel)
     EmbHostFileChannel *channel;
{
  EmbPtr command_p;
  EmbHostFileCommand *command;

  if ((command_p = (EmbPtr)EmbQueueTakeWord(HostPointer(channel->command_queue))) == 0)
    return;
  command = (EmbHostFileCommand *)HostPointer(command_p);

  /* Routines can just return if they're happy */
  command->error_code = EmbHostFileNoError;
  command->host_error_code = 0;

#ifdef IVORY_DEBUG
  printf("-> CheckHostFile op 0x%x ref 0x%x\n",
	 command->opcode, command->host_reference);
#endif

  switch (command->opcode) {
  case EmbHostFileCommandNoop:		/* Synchronize */
  case EmbHostFileCommandStartup:	/* Setup for host file operations */
  case EmbHostFileCommandShutdown:	/* Cleanup after host file operations */
    break;
    
  case EmbHostFileCommandOpen:		/* Open a file/directory */
    HostFileOpen(command);
    break;
    
  case EmbHostFileCommandClose:		/* Close the file/directory */
    HostFileClose(command);
    break;

  case EmbHostFileCommandReadData:	/* Read data */
    HostFileReadData(command);
    break;

  case EmbHostFileCommandWriteData: 	/* Write data */
    HostFileWrite(command);
    break;

  case EmbHostFileCommandReadDirectory:	/* Read Directory */
    HostFileReadDir(command);
    break;

  default:
    command->error_code = EmbHostFileUnknownOpcode;
    break;
  }

#ifdef IVORY_DEBUG
  printf("<- CheckHostFile err %d/%d ref 0x%x\n",
	 command->error_code, command->host_error_code, command->host_reference);
#endif

  EmbQueuePutWord(HostPointer(channel->reply_queue), (long)command_p);
}


HostFileOpen(command_arg)
EmbHostFileCommand *command_arg;
{
  EmbHostFileOpenCommand *command = (EmbHostFileOpenCommand*)command_arg;
  char *pathname;

  /* Copy pathname into a usable form */
  if ((pathname = malloc(command->pathname_length+1)) == NULL) {
    command->error_code = EmbHostFileIOError;
    command->host_error_code = ENAMETOOLONG;
    return;
  }
  strncpy(pathname, command->pathname, command->pathname_length);
  pathname[command->pathname_length] = '\0';
  
  switch (command->direction) {
  case EmbHostFileOpenInput:
    HostFileOpenFile(command, pathname, "r");
    break;

  case EmbHostFileOpenOutput:
    HostFileOpenFile(command, pathname, "w");
    break;

  case EmbHostFileOpenDirectory:
    HostFileOpenDirectory(command, pathname);
    break;

  default:
    command->error_code = EmbHostFileUnknownOpenMode;
    break;
  }
}

HostFileOpenFile(command, pathname, mode)
     EmbHostFileOpenCommand *command;
     char *pathname, *mode;
{
  FILE *fp;

  if ((fp = fopen(pathname, mode)) == NULL) {
    switch(errno) {
    case ENOENT:
    case ESRCH:
    case ENXIO:
    case ENODEV:
      command->error_code = EmbHostFileFileNotFound;
      break;
    default:
      command->error_code = EmbHostFileIOError;
      break;
    }
    command->host_error_code = errno;
#ifdef IVORY_DEBUG
    perror(pathname);
#endif
  } else
    command->host_reference = (long) fp;
}

HostFileClose(command_arg)
     EmbHostFileCommand *command_arg;
{
  EmbHostFileCloseCommand *command = (EmbHostFileCloseCommand *)command_arg;

  switch (command->open_direction) {
  case EmbHostFileOpenInput:
  case EmbHostFileOpenOutput:
    if (fclose((FILE *) command->host_reference) == EOF) {
      command->error_code = EmbHostFileIOError;
      command->host_error_code = errno;
    }
    break;

  case EmbHostFileOpenDirectory:
    if (closedir((DIR *) command->host_reference) == EOF) {
      command->error_code = EmbHostFileIOError;
      command->host_error_code = errno;
    }
    break;

  default:
    command->error_code = EmbHostFileUnknownOpenMode;
    break;
  }
}

HostFileReadData(command_arg)
     EmbHostFileCommand *command_arg;
{
  EmbHostFileTransferCommand *command = (EmbHostFileTransferCommand *)command_arg;
  FILE *fp = (FILE *) command->host_reference;
  int cc;

  if (command->transfer_position >= 0) {
    if (fseek(fp, command->transfer_position, L_SET) == -1) {
      command->error_code = EmbHostFileIOError;
      command->host_error_code = errno;
#ifdef IVORY_DEBUG
      perror("fseek");
#endif
      return;
    }
  }

  if ((cc = fread(command->transfer_data, sizeof(char),
		  command->transfer_length, fp)) == 0) {
    if (!feof(fp))
      command->error_code = EmbHostFileIOError;
    command->host_error_code = errno;
#ifdef IVORY_DEBUG
    perror("fread");
#endif
  }
  command->transfer_length = cc;
}

HostFileOpenDirectory(command, pathname)
     EmbHostFileOpenCommand *command;
     char *pathname;
{
  DIR *dp;

  /* See HostFileReadDir() */
  if (chdir(pathname) == -1) {
    command->error_code = EmbHostFileNotADirectory;
    command->host_error_code = errno;
    return;
  }

  if ((dp = opendir(pathname)) == NULL) {
    command->error_code = EmbHostFileIOError; /* --- Not A Directory? */
    command->host_error_code = errno;
  } else
    command->host_reference = (long) dp;
}

HostFileReadDir(command_arg)
     EmbHostFileCommand *command_arg;
{
  EmbHostFileDirectoryEntry *command = (EmbHostFileDirectoryEntry *)command_arg;
  DIR *dp = (DIR *) command->host_reference;
  struct dirent *de;
  struct passwd *pw;
  struct stat st;
  int text_space, cc;
  char linkbuf[BUFSIZ];

  if ((de = readdir(dp)) == NULL) {
    command->error_code = EmbHostFileOutOfBounds;
    return;
  }

  if (lstat(de->d_name, &st) == -1) {
    command->error_code = EmbHostFileIOError;
    command->host_error_code = errno;
#ifdef IVORY_DEBUG
    perror(de->d_name);
#endif
    return;
  }

  command->flags = 0;
  if ((st.st_mode & S_IFMT) == S_IFDIR)
    command->flags |= EmbHostFileDirectoryEntryFlagDirectory;
  if ((st.st_mode & S_IFMT) == S_IFLNK)
    command->flags |= EmbHostFileDirectoryEntryFlagLink;
  if ((st.st_mode & 0444) == 0)
    command->flags |= EmbHostFileDirectoryEntryFlagRead_only;

  command->size_in_blocks = st.st_blocks;
  command->length_in_bytes = st.st_size;

  command->creation_date = st.st_ctime + UNIX_EPOCH;

  command->name_length = strlen(de->d_name);
  if (command->flags & EmbHostFileDirectoryEntryFlagLink) {
    if ((cc = readlink(de->d_name, linkbuf, BUFSIZ)) > 0)
      command->comment_length = cc;
  } else
    command->comment_length = 0;
  if ((pw = getpwuid(st.st_uid)) == NULL)
    command->author_length = 0;
  else
    command->author_length = strlen(pw->pw_name);

  text_space = command->length - sizeof(EmbHostFileDirectoryEntry) + 1;
  if (text_space > 0) {
    strncpy(command->text, de->d_name, text_space);
    text_space -= command->name_length;
  }
  if (text_space > 0 && command->comment_length) {
    strncpy(command->text+command->name_length, linkbuf, text_space);
    text_space -= command->comment_length;
  }
  if (text_space > 0 && command->author_length)
    strncpy(command->text+command->name_length+command->comment_length,
	    pw->pw_name, text_space);
  command->length = sizeof(EmbHostFileDirectoryEntry) - 1 +
    command->name_length + command->comment_length + command->author_length;
}

HostFileWrite(command_arg)
     EmbHostFileCommand *command_arg;
{
  EmbHostFileTransferCommand *command = (EmbHostFileTransferCommand*)command_arg;
  FILE *fp = (FILE *) command->host_reference;
  int cc;

  if (command->transfer_position >= 0) {
    if (fseek(fp, command->transfer_position, L_SET) == -1) {
      command->error_code = EmbHostFileIOError;
      command->host_error_code = errno;
      return;
    }
  }

  if ((cc = fwrite(command->transfer_data, sizeof(char),
		   command->transfer_length, fp)) == 0) {
    command->error_code = EmbHostFileIOError;
    command->host_error_code = errno;
  }
  command->transfer_length = cc;
}



EmbHostFileChannel *FindHostFileChannel(EmbCommAreaPtr)
     EmbCommArea *EmbCommAreaPtr;
{
  EmbPtr cp;
  EmbHostFileChannel *p;

  /* Search for an existing HostFile channel */
  cp = EmbCommAreaPtr->channel_table;
  while (cp != NullEmbPtr) {
    p = HostPointer(cp);
    if (p->type == EmbHostFileChannelType) {
      return(p);
    }
    cp = p->next;
  }

  /* Didn't find one, lose (Solstice doesn't do dynamic channel creation) */
  return(NULL);
}


#ifdef TEST
/* cc -g -DTEST -o host host-file.c embed.o-sun3 */

#include <sys/types.h>
#include <sys/mman.h>

main ()
{
  int fd;
  char *guest = "/dev/ivory0";
  caddr_t ptr;
  EmbHostFileChannel *chan;
  EmbQueue *q;
  IvoryConfig config;

  /* Open device */
  if ((fd = open(guest, O_RDWR|O_EXCL)) < 0) {
    perror(guest);
    exit(1);
  }

  /* Prepare communication area */
  if (ioctl(fd, IVORYGCONF, &config)<0) {
    perror("IVORYGCONF");
    exit(1);
  }
  if ((ptr = mmap((caddr_t) 0, config.mmap_size, PROT_READ|PROT_WRITE,
		  MAP_SHARED, fd, (off_t) 0)) == (caddr_t) -1) {
    perror("mmap");
    exit(1);
  }
  ptr += config.comm_area_base;
  EmbCommAreaPtr = (EmbCommArea *) ptr;

  /* Quickie test */
  if (!(chan = FindHostFileChannel(ptr))) {
    fprintf(stderr, "No Host File channel!\n");
    exit(1);
  }
  if (chan->selproc) {
    puts("slamming selproc & command_queue");
    chan->selproc = 0;
    q = (EmbQueue *) HostPointer(chan->command_queue);
    q->put_index = 0;
    q->take_index = 0;
    q->signal = -1;
  }

  HostFileLife(fd, EmbCommAreaPtr, &config);
  /*NOTREACHED*/

}
#endif

