/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * Mock-interface kludge for networking between an embedded Merlin and
 * its host.  This can be removed if Sun ever fixes their NIT device to
 * work for local packets.
 *
 * This device masquerades as a network interface providing a point-to-point
 * link between the host (Unix) and embedded (Genera) systems.  It accepts
 * packets from the host, build link-layer (Ethernet) headers, and passes
 * them on to the Ivory device driver for eventual delivery to Genera.
 * It also receives packets from Genera destined for the host system, and
 * injects them into the incoming packet processing.  Finally, it maintains
 * the obvious statistics.
 *
 * $Log:	if_emb.c,v $
 * Revision 2.5.1.2  93/02/01  12:56:09  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:52:02  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:41:16  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifdef IVORY_EMBIF

#include "ivory.h"
#if NIVORY > 0

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/if_emb.c,v 2.5.1.2 93/02/01 12:56:09 kaufman Rel $";
#endif

#include "sys/param.h"
#include "sys/types.h"
#include "sys/systm.h"
#include "sys/mbuf.h"
#include "sys/socket.h"
#include "errno.h"
#include "sys/ioctl.h"

#include "net/if.h"
#include "net/route.h"
#include "net/netisr.h"

#include "sys/buf.h"
#include "sundev/mbvar.h"

#ifdef INET
#include "netinet/in.h"
#include "netinet/in_systm.h"
#include "netinet/in_var.h"
#include "netinet/ip.h"
#include "netinet/if_ether.h"
#endif

#include "sys/socket.h"
#include "sys/socketvar.h"
#include "string.h"

#include "embed.h"
#include "life.h"

/* Interface declarations */
extern struct softc if_softc[];

/* This should never be called, since there is no hardware realization of this interface */
emb_attach(md)
     struct mb_device *md;	/*ARGSUSED*/
{
#ifdef IVORY_DEBUG
  printf("emb interface attach - IGNORED");
#endif
}

/* Called by ivoryattach() */
emb_internal_initialize ()
{
  register int i;

  for (i = 0; i < n_emb_if; i++) {
    if_softc[i].EmbCommAreaPtr = (EmbCommArea *) NULL;
    if_softc[i].EmbNetChannel = (EmbNetChannel *) NULL;
    if_softc[i].count = 0;
  }
}

/* Set up kernel structures for the internal interface.
 * hw_addr is currently being passed as NULL; see comment in emb_output() near
 * use of is->is_addr. */
static emb_internal_attach (EmbCommAreaPtr, channel, unit, hw_addr)
     EmbCommArea *EmbCommAreaPtr;
     EmbNetChannel *channel;
     int unit;
     struct ether_addr *hw_addr;
{
  register struct softc *is = &if_softc[unit];
  register struct ifnet *ifp = &is->is_if;
  int already_attached = is->EmbCommAreaPtr == NULL ? 0 : 1;
  int emb_output(), emb_ioctl();

  is->EmbCommAreaPtr = EmbCommAreaPtr;
  is->EmbNetChannel = channel;

  ifp->if_name = "emb";
  ifp->if_unit = unit;
  ifp->if_mtu = EPACKETDATALEN - sizeof(struct ether_header);
  ifp->if_flags = IFF_POINTOPOINT|IFF_RUNNING;
  if (hw_addr != NULL)
    bcopy((caddr_t) hw_addr, (caddr_t) &is->is_addr, sizeof(struct ether_addr));

  ifp->if_ioctl = emb_ioctl;
  ifp->if_output = emb_output;

#ifdef IVORY_DEBUG
  DBG(DBG_IF) printf("emb_internal_attach, unit %d, %s\n",
		     unit, already_attached ? "already attached" : "attaching");
#endif

  if (!already_attached)
    if_attach(ifp);

  return(unit);
}

/* Set up the protocol addresses for both ends of the interface */
static emb_internal_sifaddr(EmbCommAreaPtr, channel, unit)
     EmbCommArea *EmbCommAreaPtr;
     EmbNetChannel *channel;
     int unit;
{
  struct softc *is = &if_softc[unit];
  struct ifnet *ifp = &is->is_if;
  struct ifreq ifr;
  struct socket so;
  struct sockaddr_in *sin;
  struct protosw *pffindtype();
  int val;

  /* Dig up Internet protocol structure */
  so.so_type = SOCK_DGRAM;
  so.so_options = 0;
  so.so_linger = 0;
  so.so_state = SS_NOFDREF|SS_PRIV;
  so.so_proto = pffindtype(AF_INET, SOCK_DGRAM);
  if (so.so_proto == NULL) {
    printf("pffindtype() lossage!\n");
    return(ENETDOWN);
  }

  /* Set remote address */
  bzero(ifr.ifr_name, sizeof(ifr.ifr_name));
  strncpy(ifr.ifr_name, ifp->if_name, sizeof(ifr.ifr_name)-1);
  ifr.ifr_name[strlen(ifr.ifr_name)] = '0' + ifp->if_unit;
  sin = (struct sockaddr_in *) &ifr.ifr_dstaddr;
  bzero((char *) sin, sizeof(struct sockaddr_in));
  sin->sin_family = AF_INET;
  sin->sin_addr.s_addr = channel->guest_primary_address;
  val = ifioctl(&so, SIOCSIFDSTADDR, (caddr_t) &ifr);
  if (val != 0) {
#ifdef IVORY_DEBUG
    DBG(DBG_IF) printf("emb%d SIOCSIFDSTADDR: err %d\n", unit, val);
#endif
    return(val);
  }

  /* Set local address */
  bzero(ifr.ifr_name, sizeof(ifr.ifr_name));
  strncpy(ifr.ifr_name, ifp->if_name, sizeof(ifr.ifr_name)-1);
  ifr.ifr_name[strlen(ifr.ifr_name)] = '0' + ifp->if_unit;
  bzero((char *) sin, sizeof(struct sockaddr_in));
  sin = (struct sockaddr_in *) &ifr.ifr_addr;
  sin->sin_family = AF_INET;
  sin->sin_addr.s_addr = channel->host_primary_address;
  val = ifioctl(&so, SIOCSIFADDR, (caddr_t) &ifr);
  if (val != 0) {
#ifdef IVORY_DEBUG
    DBG(DBG_IF) printf("emb%d SIOCSIFADDR: err %d\n", unit, val);
#endif
    return(val);
  }

#ifdef IVORY_DEBUG
  DBG(DBG_INIT) {
    char buf[20], *inet_ntoa();

    strncpy(buf, inet_ntoa(*(struct in_addr *) &channel->host_primary_address), sizeof(buf));
    printf("%s point-to-point %s to %s\n",
	   ifr.ifr_name,
	   buf,
	   inet_ntoa(*(struct in_addr *) &channel->guest_primary_address));
  }
#endif

  return(0);
}

static same_protocol_addresses (EmbCommAreaPtr, channel, is)
     EmbCommArea *EmbCommAreaPtr; /*ARGSUSED*/
     EmbNetChannel *channel;
     struct softc *is;
{
  struct ifaddr *ifa;

#ifdef INET
  for (ifa = is->is_if.if_addrlist; ifa; ifa = ifa->ifa_next)
    if (ifa->ifa_addr.sa_family == AF_INET &&
	!bcmp((caddr_t) &((struct sockaddr_in *) &ifa->ifa_addr)->sin_addr,
	      (caddr_t) &channel->host_primary_address,
	      sizeof(struct in_addr)) &&
	!bcmp((caddr_t) &((struct sockaddr_in *) &ifa->ifa_dstaddr)->sin_addr,
	      (caddr_t) &channel->guest_primary_address,
	      sizeof(struct in_addr)))
      return(1);
#endif
       return(0);
}

/* Find or create an emb interface appropriate for this channel.
 * "Appropriate" means that it has the same local & remote protocol addresses. */
InitializeEmbInterface (EmbCommAreaPtr, channel)
     EmbCommArea *EmbCommAreaPtr;
     EmbNetChannel *channel;
{
  int s, tmp, unit;

  /* --- Fix this some day.
   * This is hiding INET-dependencies in emb_internal_sifaddr()
   * and same_protocol_addresses(). */
  if (channel->host_primary_protocol != ETHERTYPE_IP ||
      channel->guest_primary_protocol != ETHERTYPE_IP) {
#ifdef IVORY_DEBUG
    DBG(DBG_INIT) printf("EmbNet%d: %s address not IP, init failed\n",
			 channel->unit,
			 channel->host_primary_protocol == ETHERTYPE_IP ? "guest" : "host");
#endif
    return(-1);
  }

  /* No point in not reusing emb interfaces */
  for (unit = 0; unit < n_emb_if; unit++)
    if (if_softc[unit].count > 0 &&
	if_softc[unit].EmbCommAreaPtr == EmbCommAreaPtr &&
	same_protocol_addresses(EmbCommAreaPtr, channel, &if_softc[unit])) {
      if_softc[unit].count++;	/* --- Never gets decremented, sigh */
      return(unit);
    }

  s = splnet();
  for (unit = 0; unit < n_emb_if; unit++)
    if (if_softc[unit].count == 0) {
      if_softc[unit].count = -1;
      break;
    }
  splx(s);
  if (unit == n_emb_if) {
    printf("emb if overflow\n");
    return(-1);
  }

  /* Initialize the internal point-to-point interface */
  emb_internal_attach(EmbCommAreaPtr, channel, unit, (struct ether_addr *) NULL);

  /* Should look up or be passed IP addresses */
  tmp = emb_internal_sifaddr(EmbCommAreaPtr, channel, unit);
#ifdef IVORY_DEBUG
  DBG(DBG_INIT) printf("emb_internal_sifaddr() ==> %d\n", tmp);
#endif
  if (tmp != 0)
    return(-2);
  
  if_softc[unit].count = 1;

  return(unit);
}

/* Receive packet from guest, inject into appropriate host protocol processing */
emb_receive (EmbCommAreaPtr, unit, pkt, pktlen)
     EmbCommArea *EmbCommAreaPtr;
     int unit;
     caddr_t pkt;
     int pktlen;
{
  struct softc *is = &if_softc[unit];
  struct ifnet *ifp = &is->is_if;
  struct ether_header *eh;
  struct mbuf *mhead, *m, *m0;
  struct ifqueue *inq;
  int (*isr)();
  int s, count, pktleft, mlen;

  if (unit >= n_emb_if) {
#ifdef IVORY_DEBUG
    printf("emb%d does not exist!\n", unit);
    return;
#else
    panic("bad unit in emb_receive()");
    /*NOTREACHED*/
#endif
  }

  if ((ifp->if_flags & (IFF_UP|IFF_RUNNING)) != (IFF_UP|IFF_RUNNING)) {
#ifdef IVORY_DEBUG
    DBG(DBG_NET_DROP) printf("emb%d not up&running  flags %d\n", unit, ifp->if_flags);
#endif
    return;
  }
  ifp->if_ipackets++;

  if (pktlen < sizeof(struct ether_header) ||
      pktlen > ifp->if_mtu + sizeof(struct ether_header)) {
#ifdef IVORY_DEBUG
    DBG(DBG_NET_DROP)
      if (pktlen > ifp->if_mtu)
	printf("emb%d pkt size %d greater than if mtu %d\n",
	       unit, pktlen, ifp->if_mtu);
      else
	printf("emb%d pkt size %d less than header size %d\n",
	       unit, pktlen, sizeof(struct ether_header));
#endif
    ifp->if_ierrors++;
    return;
  }

  /* Allocate an mbuf to copy the packet into */
  MGET(m, M_DONTWAIT, MT_DATA);
  if (m == NULL) {
#ifdef IVORY_DEBUG
    DBG(DBG_NET_DROP) printf("emb%d couldn't mget mbuf\n", unit);
#endif
    return;
  }
  m->m_len = 0;
  m->m_next = NULL;
  mlen = MLEN;

  /* Skip past ethernet header */
  eh = (struct ether_header *) pkt;
  pktleft = pktlen - sizeof(struct ether_header);
  pkt += sizeof(struct ether_header);

  /* If packet won't fit in unclustered mbuf, get a cluster */
  if (pktleft > MLEN) {
    mclget(m);
    if (M_HASCL(m)) {
      mlen = m->m_len;
      m->m_len = 0;
    } else
      /* No cluster available, stumble onwards */
      mlen = MLEN;
  }

  /* Prepend back-pointer to embedded interface */
  mhead = m;
  *(mtod(mhead, struct ifnet **)) = ifp;
  mhead->m_off += sizeof(ifp);
  mlen -= sizeof(ifp);

  /* Copy chunks of packet into mbufs */
  while (pktleft > 0) {
    count = MIN(pktleft, mlen);
    bcopy(pkt, mtod(m, char *), count);
    m->m_len += count;
    pkt += count;
    pktleft -= count;
    if (pktleft > 0) {
      MGET(m0, M_DONTWAIT, MT_DATA);
      mlen = MLEN;
      m0->m_len = 0;
      m0->m_next = NULL;
      if (pktleft > MLEN) {
	mclget(m0);
	if (M_HASCL(m0)) {
	  mlen = m0->m_len;
          m0->m_len = 0;
        }
      }
      m->m_next = m0;
      m = m0;
    }
  }

  /* Adjust for backpointer */
  mhead->m_off -= sizeof(ifp);
  mhead->m_len += sizeof(ifp);

#ifdef IVORY_DEBUG
  
  DBG(DBG_IF) {

    printf(":RCV emb%d  bytes %d  type %x  src %x:%x:%x:%x:%x:%x  dst %x:%x:%x:%x:%x:%x\n",
	   unit,
	   pktlen,
	   eh->ether_type,
	   eh->ether_shost.ether_addr_octet[0],
	   eh->ether_shost.ether_addr_octet[1],
	   eh->ether_shost.ether_addr_octet[2],
	   eh->ether_shost.ether_addr_octet[3],
	   eh->ether_shost.ether_addr_octet[4],
	   eh->ether_shost.ether_addr_octet[5],
	   eh->ether_dhost.ether_addr_octet[0],
	   eh->ether_dhost.ether_addr_octet[1],
	   eh->ether_dhost.ether_addr_octet[2],
	   eh->ether_dhost.ether_addr_octet[3],
	   eh->ether_dhost.ether_addr_octet[4],
	   eh->ether_dhost.ether_addr_octet[5]);
  }
#endif

  /* --- Idiom from 4.3BSD interface code, God knows if it will work.
   * We need an entry here for every protocol that Genera and SunOS
   * might try to use to communicate.  Currently this means IP and ARP;
   * in the future it could include DNA, Chaos, or other bizarrities.
   * If you add a protocol here, see also emb_output(). */
  switch (eh->ether_type) {
#ifdef INET
  case ETHERTYPE_IP:
    inq = &ipintrq;
    isr = NETISR_IP;
    break;

  case ETHERTYPE_ARP:
    arpinput(&is->is_ac, mhead);
    return;
#endif INET

  default:
    m_freem(mhead);
    return;
  }

  s = splimp();
  if (IF_QFULL(inq)) {
    IF_DROP(inq);
    m_freem(mhead);
  } else {
    IF_ENQUEUE(inq, mhead);
    schednetisr(isr);
  }
  splx(s);

  return;
}

/* Helper function called from NetReceiver on behalf of emb_output() */
static int fill_in_packet (eh, pkt, m)
     struct ether_header *eh;
     caddr_t pkt;
     struct mbuf *m;
{
 
  /* Copy in the header first */
  bcopy((caddr_t) eh, pkt, sizeof(struct ether_header));
  pkt += sizeof(struct ether_header);

  /* Loop copying in data from mbufs */
  for (; m; m = m->m_next) {
    bcopy(mtod(m, caddr_t), pkt, m->m_len);
    pkt += m->m_len;
  }
}

/* Receive a packet from the host, transmit it to the guest */
emb_output(ifp, m0, dst)
     struct ifnet *ifp;
     struct mbuf *m0;
     struct sockaddr *dst;
{
  struct softc *is = &if_softc[ifp->if_unit];
  struct ether_header ehdr, *eh;
  struct mbuf *n;
  int pktlen;
#ifdef IVORY_DEBUG
  EmbCommArea *EmbCommAreaPtr = is->EmbCommAreaPtr;
#endif

  if ((ifp->if_flags & (IFF_UP|IFF_RUNNING)) != (IFF_UP|IFF_RUNNING)) {
    m_freem(m0);
    return(ENETDOWN);
  }

  /* Calculate packet length (header length added below) */
  for (pktlen = 0, n = m0; n; n = n->m_next)
    pktlen += n->m_len;

  /* Build an ether_header appropriate for the address family.
   * Again, this needs to know about every address family the guest and host
   * might use to communicate with each other. */
  switch (dst->sa_family) {
#ifdef INET
  case AF_INET:
#ifdef notdef
    /* Source and destination hosts have same ether address */
    bcopy((caddr_t) &is->is_addr, (caddr_t) &ehdr.ether_dhost, sizeof(struct ether_addr));
    bcopy((caddr_t) &is->is_addr, (caddr_t) &ehdr.ether_shost, sizeof(struct ether_addr));
#else
    /* Nothing in Genera needs an ethernet address here, really, and the emb if doesn't
     * really have a hardware address, so punt.  If this needs to be changed back,
     * then somebody needs to hand a hardware address to emb_internal_attach(). */
    bzero((caddr_t) &ehdr.ether_dhost, 2*sizeof(struct ether_addr));
#endif
    ehdr.ether_type = ETHERTYPE_IP;
    eh = &ehdr;
    pktlen += sizeof(struct ether_header);
    break;
#endif

  case AF_UNSPEC:
    eh = (struct ether_header *) dst->sa_data;
    pktlen += sizeof(struct ether_header);
    break;

  default:
    printf("emb%d: can't handle af%d\n", ifp->if_unit, dst->sa_family);
    m_freem(m0);
    return(EAFNOSUPPORT);
  }

#ifdef IVORY_DEBUG
  DBG(DBG_IF) {

    printf(":XMT  emb%d  bytes %d  type %x  src %x:%x:%x:%x:%x:%x  dst %x:%x:%x:%x:%x:%x\n",
	   ifp->if_unit,
	   pktlen,
	   eh->ether_type,
	   eh->ether_shost.ether_addr_octet[0],
	   eh->ether_shost.ether_addr_octet[1],
	   eh->ether_shost.ether_addr_octet[2],
	   eh->ether_shost.ether_addr_octet[3],
	   eh->ether_shost.ether_addr_octet[4],
	   eh->ether_shost.ether_addr_octet[5],
	   eh->ether_dhost.ether_addr_octet[0],
	   eh->ether_dhost.ether_addr_octet[1],
	   eh->ether_dhost.ether_addr_octet[2],
	   eh->ether_dhost.ether_addr_octet[3],
	   eh->ether_dhost.ether_addr_octet[4],
	   eh->ether_dhost.ether_addr_octet[5]);
  }
#endif

  if (NetReceiver(is->EmbCommAreaPtr, is->EmbNetChannel, pktlen,
		  fill_in_packet, (caddr_t) eh, (caddr_t) m0))
    ifp->if_oerrors++;
  else
    ifp->if_opackets++;
  m_freem(m0);

  return(0);
}


/* Handle ioctl's.  Most of the work is done by protocol-specific
 * and generic-interface ioctl routines. */
emb_ioctl(ifp, cmd, data)
     struct ifnet *ifp;
     int cmd;
     caddr_t data;		/*ARGSUSED*/
{

  switch (cmd) {
  case SIOCSIFADDR:
    ifp->if_flags |= IFF_UP;
    break;
  }

  return(0);
}

#endif /* NIVORY > 0 */
#endif /* IVORY_EMBIF */
