/* -*-Mode: C; Package: C-user -*- */
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1990 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * Main driver for UX installation
 *
 * $Log:	install.c,v $
 * Revision 2.5.1.2  93/02/01  12:56:13  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:52:07  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:41:30  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/install.c,v 2.5.1.2 93/02/01 12:56:13 kaufman Rel $";
#endif

#include <sys/types.h>
#include <sys/file.h>
#include <stdio.h>
#include "install.h"

#if defined(mc68000)
#define ARCH "sun3"
#elif defined(sparc)
#define ARCH "sun4"
#else
"Unknown architecture"
#endif

  /* Global variable for config_state_transformation() in install-kernel.c */
struct params ux_installation_params;

main (argc, argv)
     int argc;
     char *argv[];
{
  int i;

  /* Initial parameter settings */
  bzero((char *) &ux_installation_params, sizeof(struct params));
  ux_installation_params.architecture = ARCH;
  ux_installation_params.available_software = SW_NONE;
  ux_installation_params.defaulting = FALSE;
  ux_installation_params.distribution_device_type = dd_unknown;
  ux_installation_params.distribution_directory = ".";

  /* Parse command-line arguments */
  for (i = 1; i < argc; i++)
    if (!strncmp(argv[i], "-d", 2)) {
      /* If we were serious about this we'd be checking that the
       * argument to this option is reasonable. */
      if (argv[i][2] == '/') {
	/* Name of filesystem where CD-ROM drive has been mounted */
	ux_installation_params.distribution_device_type = dd_cdrom;
	ux_installation_params.distribution_device = &argv[i][2];
      } else {
	/* Tape drive name ("st0", not "/dev/rst0") */
	ux_installation_params.distribution_device_type = dd_tape;
	ux_installation_params.distribution_device = &argv[i][2];
      }
    } else if (!strncmp(argv[i], "-f", 2))
      /* Called from extract_unbundled */
      ux_installation_params.available_software = SW_NONE;
    else if (!strncmp(argv[i], "-r", 2))
      /* Host that owns tape drive */
      ux_installation_params.distribution_host = &argv[i][2];
    else if (!strncmp(argv[i], "-w", 2))
      /* Work path */
      ux_installation_params.distribution_directory = &argv[i][2];
    else if (!strcmp(argv[i], "-DEFAULT"))
      /* Answer all questions affirmatively */
      ux_installation_params.defaulting = TRUE;
    else if (!strcmp(argv[i], "+fonts"))
      ux_installation_params.available_software |= SW_FONTS;
    else if (!strcmp(argv[i], "+ifep"))
      ux_installation_params.available_software |= SW_IFEP;
    else if (!strcmp(argv[i], "+include"))
      ux_installation_params.available_software |= SW_INCLUDE;
    else if (!strcmp(argv[i], "+source"))
      ux_installation_params.available_software |= SW_SOURCE;
    else if (!strcmp(argv[i], "+sun3"))
      ux_installation_params.available_software |= SW_SUN3;
    else if (!strcmp(argv[i], "+sun4"))
      ux_installation_params.available_software |= SW_SUN4;
    else {
      Fprintf(stderr, "usage: %s [-ddevice] [-rhost] [-wpath] [-DEFAULT] [+sw] [+sw] ...\n",
	      argv[0]);
      Fprintf(stderr, "       sw can be fonts, ifep, include, source, sun3, or sun4.\n");
      exit(1);
    }
  
  /* Pass 1: Determine all the necessary parameters */
  determine_relocation_parameters(&ux_installation_params);
  determine_font_parameters(&ux_installation_params);
  determine_kernel_parameters(&ux_installation_params);
  determine_partition_parameters(&ux_installation_params);
  determine_unix_startup_parameters(&ux_installation_params);

  terpri();
  Printf("About to begin %s software installation, this will probably take a while.\n",
	 UX_NAME);
  if (!y_or_n_p("Okay", 1)) {
    Printf("%s software not installed.\n", UX_NAME);
    exit(0);
  }

  /* Make sure world distribution tape is mounted if we'll need it */
  if (ux_installation_params.create_partitions_p &&
      ux_installation_params.distribution_device_type == dd_tape) {
    terpri();
    do {
      Printf("Is the Genera Distribution World tape mounted");
    } while (!y_or_n_p("", ux_installation_params.defaulting == TRUE ? 0 : 1));
  }

  /* Pass 2: Do the work */
  if (relocate_files(&ux_installation_params) == -1)
    goto lose;
  if (compile_x_fonts(&ux_installation_params) == -1)
    goto lose;
  if (configure_kernel(&ux_installation_params) == -1)
    goto lose;
  if (configure_partitions(&ux_installation_params) == -1)
    goto lose;

  /* This must go last, because it can cause a reboot */
  if (configure_unix_startup(&ux_installation_params) == -1)
    goto lose;

  /* Pass 3: Exit */
  terpri();
  Printf("%s installation completed.\n", UX_NAME);
  exit(0);

 lose:
  Fprintf(stderr, "%s installation aborted.\n", UX_NAME);
  exit(1);
}



static enum distribution_device dispatch_cdrom_or_tape (ch)
     char ch;
{

  switch(ch) {
  case 'c':
  case 'C':
    return(dd_cdrom);

  case 't':
  case 'T':
    return(dd_tape);

  default:
    return(dd_unknown);
  }
}

static int check_cdrom_or_tape (buf, arg0, arg1)
     char *buf;
     arg *arg0, *arg1;		/*ARGSUSED*/
{
  if (dispatch_cdrom_or_tape(*buf) == dd_unknown) {
    Printf("\n>> Please answer CD-ROM or Tape.\n");
    return(-1);
  } else
    return(0);
}

static accept_distribution_device (prompt)
     char *prompt;
{
  return(dispatch_cdrom_or_tape(*accept_string(prompt, "Tape",
					       check_cdrom_or_tape,
					       (arg *) NULL, (arg *) NULL)));
}

check_tape_device (buf, ignore0, ignore1)
     char *buf;
     arg *ignore0, *ignore1;
{
  arg type, access_bits;
  char buf0[BUFSIZ], *cp;

  type.integer = S_IFCHR;
  access_bits.integer = F_OK|R_OK;
  if (buf[0] == '/')
    cp = buf;
  else {
    sprintf(buf0, "/dev/nr%s", buf);
    cp = buf0;
  }
  return(check_file(cp, &type, &access_bits));
}

check_host (buf, ignore0, ignore1)
     char *buf;
     arg *ignore0, *ignore1;
{
  if (gethostbyname(buf) == NULL) {
    Fprintf(stderr, "\n>> %s is not a recognized host name.\n", buf);
    return(1);
  }

  return(0);
}

void require_distribution_device (params)
     struct params *params;
{
  arg arg0, arg1;

  if (params->distribution_device_type != dd_unknown)
    return;

  params->distribution_device_type =
    accept_distribution_device("Is your Symbolics distribution on CD-ROM or Tape");

  if (params->distribution_device_type == dd_cdrom) {
    arg0.integer = S_IFDIR;
    arg1.integer = R_OK;
    params->distribution_device =
      accept_string("Where is the CD-ROM mounted", "/cdrom", check_file, &arg0, &arg1);
  } else {
    if (!y_or_n_p("Is the tape device on the local host", 1)) {
      params->distribution_host =
	accept_string("What host is the tape device on", NULL, check_host,
		      (arg *) NULL, (arg *) NULL);
    }

    params->distribution_device =
      accept_string("What is the tape device",
		    "st0", check_tape_device, (arg *) NULL, (arg *) NULL);
  }
}
