/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * Device driver for embedded Ivory-based processors.
 *
 * $Log:	ivory.c,v $
 * Revision 2.5.1.2  93/02/01  12:56:18  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:52:13  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:41:44  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#include "ivory.h"
#if NIVORY > 0

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/ivory.c,v 2.5.1.2 93/02/01 12:56:18 kaufman Rel $";
#endif

#if defined(NKIVORY) && NKIVORY > 0
#define ivorydriver kivorydriver
#elif defined(NPIVORY) && NPIVORY > 0
#define ivorydriver pivorydriver
#elif defined(NHIVORY) && NHIVORY > 0
#define ivorydriver hivorydriver
#endif

#include "sys/types.h"
#include "sys/time.h"

#include "machine/param.h"
#include "sys/buf.h"
#include "sys/errno.h"
#include "sys/file.h"
#include "sys/filio.h"
#include "sys/map.h"
#include "sys/param.h"
#include "sys/proc.h"
#include "sys/systm.h"
#include "sys/uio.h"
#include "sys/user.h"
#include "sys/vmmac.h"
#include "sundev/mbvar.h"
#include "setjmp.h"
#include "machine/pte.h"

#define INCLUDE_FILE_HEADERS
#include "embed.h"
#include "life.h"
#undef INCLUDE_FILE_HEADERS
#define IVORY_UNIT(dev) minor(dev)

int ivoryprobe(), ivoryattach(), ivoryintr();

extern struct mb_device *ivorydinfo[];

struct mb_driver ivorydriver = {
  ivoryprobe, 0, ivoryattach, 0, 0, ivoryintr,
  MERLIN_II_VME_MMAP_SIZE, "ivory", ivorydinfo, 0, 0, 0, 0,
};

extern struct ivory_state ivories[];

ivoryprobe (reg, unit)
     /*ARGSUSED*/
     caddr_t reg;
     int unit;
{
  long junk;
  
  /* Merlin is there if slave buffer is there.  Check for existence of first and
   * last (longword) locations.  We used to check for non-existence of last
   * location, but SunOS 4.1 kernel doesn't seem to map enough VME address space
   * for us to do that with Merlin-IIs. */
  if (unit<16)
    /* Check for Merlin-I */
    if (peek((long *) reg, &junk) == -1 ||
	peek((long *) (reg + MERLIN_VME_MMAP_SIZE - 4), &junk) == -1)
      return(0);
    else
      return(MERLIN_VME_MMAP_SIZE);
  else
    /* Check for Merlin-II */
    if (peek((long *) reg, &junk) == -1 ||
	peek((long *) (reg + MERLIN_II_VME_MMAP_SIZE - 4), &junk) == -1)
      return(0);
    else
      return(MERLIN_II_VME_MMAP_SIZE);
}

ivoryattach (md)
     struct mb_device *md;
{
  register EmbCommArea *EmbCommAreaPtr;
  int unit = md->md_unit;
  long junk;

  if (unit < 16) {
    /* Merlin-I.  We use the top half of the merlin slave buffer for the comm area.
     * The top 64 longs are reserved for slave triggering and Genera. */
    ivories[unit].config.system_type = SystemTypeUX400S;
    ivories[unit].config.slave_buffer_size = MERLIN_VME_SLAVE_BUFFER_SIZE;
    ivories[unit].config.mmap_size = MERLIN_VME_MMAP_SIZE;
    ivories[unit].config.comm_area_base = 0;
    ivories[unit].config.comm_area_size = MERLIN_VME_COMM_AREA_SIZE;
  } else {
    /* Merlin-II.  We use the top quarter of the merlin slave buffer for the comm area.
     * The top 64 longs are reserved for slave triggering and Genera. */
    ivories[unit].config.system_type = SystemTypeUX1200S;
    ivories[unit].config.slave_buffer_size = MERLIN_II_VME_SLAVE_BUFFER_SIZE;
    ivories[unit].config.mmap_size = MERLIN_II_VME_MMAP_SIZE;
    ivories[unit].config.comm_area_base = 0;
    ivories[unit].config.comm_area_size = MERLIN_II_VME_COMM_AREA_SIZE;
  }
  EmbCommAreaPtr = (EmbCommArea *) (md->md_addr + ivories[unit].config.comm_area_base);
  ivories[unit].EmbCommArea = EmbCommAreaPtr;

  /* Communication area default settings */
  EmbCommAreaPtr->guest_status = UninitializedGuestStatus;	/* Presume no hardware */
  EmbCommAreaPtr->board_major_version = 0;
  EmbCommAreaPtr->board_minor_version = 0;
  EmbCommAreaPtr->spy_status = 2; /* Tell SPY it doesn't know Ivory's state */
  EmbCommAreaPtr->fepstat.fepstat_word = 0;
  EmbCommAreaPtr->fep_status = HaltedFEPStatus;	/* FEP's not running yet */

  /* Configuration defaults */
  ivories[unit].config.enable_Ivory = FALSE;
  ivories[unit].config.slave_trigger_offset = -1;
  ivories[unit].config.device_prom_fd = -1;

#ifdef IVORY_EMBIF
  (void) emb_internal_initialize();
#endif
  ivories[unit].master_proc = NULL;
  ivories[unit].network_proc = NULL;
  ivories[unit].disk_proc = NULL;
  ivories[unit].net_channel = 0;
  ivories[unit].net_channel_ptr = -1;

  /* Map the reset register into kernel virtual memory */
  if (md->md_flags == 0)
    printf("WARNING: No csr specified for ivory%d!\n", unit);
  else if (peek((long *) &md->md_flags, &junk) == -1)
    printf("WARNING: ivory%d csr at 0x%lx does not exist!\n",
	   unit, md->md_flags);
  else {
    ivories[unit].control = map_mailbox(md->md_flags);
  }
}

ivoryopen(dev, flags)
     dev_t dev;
     int flags;			/*ARGSUSED*/
{
  int unit = IVORY_UNIT(dev);

  if (unit > n_ivories || ivories[unit].EmbCommArea == 0)
    return(ENXIO);

  return(0);
}

ivoryclose(dev, flags)
     dev_t dev;
     int flags;			/*ARGSUSED*/
{
  int unit = IVORY_UNIT(dev);

  if (unit > n_ivories)
    return(ENXIO);

  /* --- This is all misplaced, and has nothing to do with opening and closing /dev/ivory */
  /* Scribble on comm area for next time around */
  ivories[unit].EmbCommArea->guest_status = UninitializedGuestStatus;

  /* Scribble on configuration for benefit of next startup, if any */
  ivories[unit].config.enable_Ivory = 0;

  return(0);
}

ivoryintr (unit)
     long unit;
{
  EmbCommArea *EmbCommAreaPtr = ivories[unit].EmbCommArea;

  if (ivories[unit].config.enable_Ivory == 0)
    return;

  if (EmbCommAreaPtr->reset_request != NoResetRequest)
    wakeup((caddr_t) ivories[unit].master_proc);
  else
    DispatchPendingSignals(EmbCommAreaPtr, unit);
}

static void enter_life(unit)
     int unit;
{
  /* Prepare for ivory life */
  ivories[unit].master_proc = u_procp();
  ivories[unit].EmbCommArea->identifier = 'DBME';
  start_ivory_clock(unit);
}

static void exit_life (unit)
     int unit;
{
  stop_ivory_clock(unit);
  ivories[unit].EmbCommArea->identifier = 'DAED';

  /* Make sure helper processes don't out last master process */
  if (u_procp() == ivories[unit].master_proc)
    KillSubProcesses(ivories[unit].EmbCommArea, unit);
}

static ivoryselchannel (EmbCommAreaPtr, chanptr, enable_p)
     EmbCommArea *EmbCommAreaPtr;
     EmbPtr chanptr;
     int enable_p;
{
  EmbChannel *chan;
  EmbColdLoadChannel *cold;
  EmbMessageChannel *message;
  EmbHostFileChannel *hostfile;
  int val, (*fn)(), InstallSelectSignal(), RemoveSelectSignal();
  struct proc *selproc;

  if (chanptr < 0 || chanptr > EmbCommAreaPtr->comm_memory_size)
    return(EINVAL);
  chan = (EmbChannel *) HostPointer(chanptr);

  if (enable_p) {
    fn = InstallSelectSignal;
    selproc = u_procp();
  } else {
    fn = RemoveSelectSignal;
    selproc = NULL;
  }

  switch(chan->type) {
  case EmbColdLoadChannelType:
    cold = (EmbColdLoadChannel *) chan;
    val = (*fn)(EmbCommAreaPtr, cold->display_output_queue);
    if (!val)
      cold->selproc = selproc;
    return (val);
    /*NOTREACHED*/

  case EmbMessageChannelType:
    message = (EmbMessageChannel *) chan;
    val = (*fn)(EmbCommAreaPtr, message->guest_to_host_queue) ||
      (*fn)(EmbCommAreaPtr, message->host_to_guest_supply_queue);
    if (!val)
      message->selproc = selproc;
    return(val);
    /*NOTREACHED*/

  case EmbHostFileChannelType:
    hostfile = (EmbHostFileChannel *) chan;
    val = (*fn)(EmbCommAreaPtr, hostfile->command_queue);
    if (!val)
      hostfile->selproc = selproc;
    return(val);
    /*NOTREACHED*/

  default:
    return(EBADF);
  }
}

ivoryioctl (dev, cmd, data, flag)
     dev_t dev;
     int cmd;
     caddr_t data;
     int flag;
{
  int unit = IVORY_UNIT(dev);
  int retval = 0;

  switch (cmd) {

  case IVORYGCONF:
    bcopy((char *)&(ivories[unit].config), (char *)data, sizeof(IvoryConfig));
    break;

  case IVORYINIT:
    if (!(flag&FWRITE))
      return(EACCES);
    ivories[unit].EmbCommArea->fepstat.fepstat_word = 0;
    ivories[unit].EmbCommArea->fep_status = HaltedFEPStatus;
    ivory_control(ivories[unit].EmbCommArea, ivories[unit].control, cmd);
    break;

  case IVORYRESET:
    if (!(flag&FWRITE))
      return(EACCES);
    ivory_control(ivories[unit].EmbCommArea, ivories[unit].control, cmd);
    break;

  case IVORYLIFE:
    if (!(flag&FWRITE)) {
      retval = EACCES;
      break;
    }
    switch (*(int *)data) {
    case PROC_MASTER:
      FixupSignalHandlers(ivories[unit].EmbCommArea);
      enter_life(unit);
      retval = IvoryLife(ivories[unit].EmbCommArea, unit);
      exit_life(unit);
      break;

    case PROC_DISK:
      FixupSignalHandlers(ivories[unit].EmbCommArea);
      InitializeDiskLifeII(ivories[unit].EmbCommArea, unit);
      ivories[unit].disk_proc = u_procp();
      retval = IvorySubProcessLife(ivories[unit].EmbCommArea, unit, (void (*)()) NULL);
      ivories[unit].disk_proc = NULL;
      break;

    case PROC_NETWORK:
      FixupSignalHandlers(ivories[unit].EmbCommArea);
      InitializeNetworkLifeII(ivories[unit].EmbCommArea, unit);
      ivories[unit].network_proc = u_procp();
      retval = IvorySubProcessLife(ivories[unit].EmbCommArea, unit, (void (*)()) NULL);
      ivories[unit].network_proc = NULL;
      break;

    default:
      retval = EINVAL;
      break;
    }
    break;

  case IVORYSCONF:
    if (!(flag&FWRITE)) {
      retval = EACCES;
    } else {
      EmbCommArea *EmbCommAreaPtr = ivories[unit].EmbCommArea;
      IvoryConfig *conf = (IvoryConfig *) data;
      struct ivory_state *ivory = &ivories[unit];

      if (conf->enable_Ivory != ivory->config.enable_Ivory) {
	ivory->config.enable_Ivory = conf->enable_Ivory;
	if (conf->enable_Ivory)
	  WakeupProcesses(EmbCommAreaPtr, unit);
      }
      if (conf->device_prom_fd != ivory->config.device_prom_fd &&
	  conf->device_prom_fd)
	ivory->config.device_prom_fd = conf->device_prom_fd;
      if (conf->slave_trigger_offset != ivory->config.slave_trigger_offset &&
	  conf->slave_trigger_offset) {
	ivory->config.slave_trigger_offset = conf->slave_trigger_offset;
      }
      EmbCommAreaPtr->slave_trigger =(long *)HostPointer(ivory->config.slave_trigger_offset);
      if (conf->host_ether_addr_offset != ivory->config.host_ether_addr_offset &&
	  conf->host_ether_addr_offset) {
	ivory->config.host_ether_addr_offset = conf->host_ether_addr_offset;
      }
      if (ivory->config.host_ether_addr_offset)
	EmbCommAreaPtr->host_ether_addrs = HostPointer(ivory->config.host_ether_addr_offset);
    }
    break;

  case IVORYSELCHANNEL:
    retval = ivoryselchannel(ivories[unit].EmbCommArea, *(EmbPtr *)data, 1);
    break;

  case IVORYSIGNAL:
    EmbSendSignal(ivories[unit].EmbCommArea, *(SignalNumber *)data);
    break;

  case IVORYUNSELCHANNEL:
    retval = ivoryselchannel(ivories[unit].EmbCommArea, *(EmbPtr *)data, 0);
    break;

  default:
    retval = ENOTTY;
    break;
  }

  return(retval);
}

static ivory_control (EmbCommAreaPtr, reg, action)
     EmbCommArea *EmbCommAreaPtr; /*ARGSUSED*/
     caddr_t reg;
     int action;
{

  if (reg == 0)
    return;

  switch(action) {
  case IVORYINIT:
    *(long *) reg = MERLIN_VME_MAILBOX_INIT;
    break;

  case IVORYRESET:
    *(long *) reg = MERLIN_VME_MAILBOX_RESET;
    break;
  }

  DELAY(1000);
  *(long *) reg = 0;
}

ivorymmap (dev, off, prot)
    dev_t dev;
    off_t off;
    int prot;			/*ARGSUSED*/
{
  if (off > ivories[IVORY_UNIT(dev)].config.mmap_size) {
    off -= ivories[IVORY_UNIT(dev)].config.mmap_size;
    if (off >= MERLIN_VME_MAILBOX_SIZE)
      /* This guy's off in hyper-space */
      return(-1);
    else
      /* Pretend that the mailbox is immediately after the comm area */
      return(hat_getkpfnum(ivories[IVORY_UNIT(dev)].control + off));
  } else
    /* Normal case: map in slave buffer */
    return(hat_getkpfnum(ivorydinfo[IVORY_UNIT(dev)]->md_addr + off));
}

ivoryread (dev, uio)
     dev_t dev;			/*ARGSUSED*/
     struct uio *uio;
{
  return(ENXIO);
}

ivorywrite (dev, uio)
     dev_t dev;			/*ARGSUSED*/
     struct uio *uio;
{
  return(ENXIO);
}

ivoryselect (dev, which)
     dev_t dev;
     int which;
{
  int unit = IVORY_UNIT(dev);
  int selectable_channels_p = 0;
  EmbCommArea *EmbCommAreaPtr;
  EmbPtr ptr;
  EmbChannel *chan;
  int s;

  /* Paranoia? */
  if (unit > n_ivories) {
    set_u_error(ENXIO);
    return(0);
  }
  EmbCommAreaPtr = ivories[unit].EmbCommArea;

#ifdef IVORY_DEBUG
  DBG(DBG_SELECT) 
    printf("ivory%d %s select\n",
       unit, which == FREAD ? "read" : which == FWRITE ? "write" : "exceptional");
#endif

  /* Check each channel this process has enabled */
  s = splnet();
  for (ptr = EmbCommAreaPtr->channel_table; ptr != NullEmbPtr; ptr = chan->next) {
    chan = (EmbChannel *) HostPointer(ptr);
    switch (ivoryselready(EmbCommAreaPtr, chan, which)) {
    case 1:			/* Selecting on channel that's ready */
      splx(s);
#ifdef IVORY_DEBUG
      DBG(DBG_SELECT) printf("ivory%d: channel ready\n", unit);
#endif
      return(1);

    case 0:			/* Selecting on channel that's not ready */
      selectable_channels_p = 1;
      break;

    case -1:			/* Not selecting on this channel */
      break;
    }
  }

  /* If there were no channels on which we are selecting, then they were
   * probably guest-allocated channels that disappeared in an embedded reset.
   * Return from select so that the application can decide what to do. */
  splx(s);
  return (selectable_channels_p ? 0 : 1);
}

ivoryselready(EmbCommAreaPtr, chan, which)
     EmbCommArea *EmbCommAreaPtr;
     EmbChannel *chan;
     int which;
{
  EmbColdLoadChannel *cold;
  EmbMessageChannel *message;
  EmbHostFileChannel *hostfile;

  switch(chan->type) {
  case EmbColdLoadChannelType:
    cold = (EmbColdLoadChannel *) chan;
    if (cold->selproc != u_procp())
      return(0);
    switch(which) {
    case FREAD:
      if (EmbQueueFilled(EmbCommAreaPtr,
			 (EmbQueue *) HostPointer(cold->display_output_queue)))
	return(1);
      return(0);

    case FWRITE:
      if (EmbQueueSpace(EmbCommAreaPtr,
			(EmbQueue *) HostPointer(cold->keyboard_input_queue)))
	return(1);
      return(0);

    default:
      return(0);
    }
    /*NOTREACHED*/

  case EmbMessageChannelType:
    message = (EmbMessageChannel *) chan;
    if (message->selproc != u_procp())
      return(0);
    switch(which) {
    case FREAD:
      if (message->guest_to_host_impulse ||
	  (EmbQueueFilled(EmbCommAreaPtr,
			 (EmbQueue *) HostPointer(message->guest_to_host_queue)) &&
	   EmbQueueSpace(EmbCommAreaPtr,
			 (EmbQueue *) HostPointer(message->guest_to_host_return_queue))))
	return(1);
      return(0);

    case FWRITE:
      if (message->host_to_guest_impulse == 0 &&
	  EmbQueueFilled(EmbCommAreaPtr,
			 (EmbQueue *) HostPointer(message->host_to_guest_supply_queue)) &&
	  EmbQueueSpace(EmbCommAreaPtr,
			(EmbQueue *) HostPointer(message->host_to_guest_queue)))
	return(1);
      return(0);

    default:
      return(0);
    }
    /*NOTREACHED*/

  case EmbHostFileChannelType:
    hostfile = (EmbHostFileChannel *) chan;
    if (hostfile->selproc != u_procp())
      return(0);
    switch(which) {
    case FREAD:
      if (EmbQueueFilled(EmbCommAreaPtr, (EmbQueue *) HostPointer(hostfile->command_queue)))
	return(1);
      return(0);

    case FWRITE:
      if (EmbQueueSpace(EmbCommAreaPtr, (EmbQueue *) HostPointer(hostfile->reply_queue)))
	return(1);
      return(0);

    default:
      return(0);
    }
    /*NOTREACHED*/
    
  default:
    return(0);
  }
  /*NOTREACHED*/
}
#endif				/* NIVORY > 0 */
