/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * Host embedded communication area buffer allocation
 *
 * $Log:	ivory_buf.c,v $
 * Revision 2.5.1.2  93/02/01  12:56:20  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:52:17  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:41:49  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifdef IVORY_EMBBUF

#include "ivory.h"
#if NIVORY > 0

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/ivory_buf.c,v 2.5.1.2 93/02/01 12:56:20 kaufman Rel $";
#endif

#include "sys/errno.h"
#include "sys/param.h"

#include "embed.h"
#include "life.h"

/* Embedded memory allocator, loosely modelled on Genera version */

int buf_sizes_and_counts[] = { 8, 4};
#define NPOOLS (sizeof(buf_sizes_and_counts) / (sizeof(int) * 2))

struct pool {
  int size, count, avail;
  EmbPtr bufs[20];
  bool bitmap[20];
} pools[NIVORY][NPOOLS];


/* Hook for preinitialize_embedding(): return minimum number of bytes for host buffers */
minimum_host_buffer_requirement ()
{
  int i, j;

  for (i = j = 0; i < NPOOLS; i++)
    j += buf_sizes_and_counts[i*2] * buf_sizes_and_counts[(i*2)+1];

  return(j);
}

static void add_buf_to_pool (buf, pool)
     EmbPtr buf;
     struct pool *pool;
{
  int i = ++(pool->count);

  if (i == pool->avail) {
#ifdef IVORY_DEBUG
    printf("Pools too small, some memory will be inaccessible\n");
#endif
  }
  pool->bufs[i] = buf;
  pool->bitmap[i] = FALSE;
}

int InitializeEmbMemory (EmbCommAreaPtr, unit)
     EmbCommArea *EmbCommAreaPtr;
     int unit;
{
  int i, j;
  EmbPtr next, top;
  int size, count;
  int first = 1, found_space;

  /* Initialize pools */
  for (i=0; i < NPOOLS; i++) {
    pools[unit][i].size = buf_sizes_and_counts[i*2];
    pools[unit][i].count = -1;
    pools[unit][i].avail = 20;
  }

  /* Build buffers */
  next = EmbCommAreaPtr->host_buffer_start;
  top = EmbCommAreaPtr->host_buffer_start + EmbCommAreaPtr->host_buffer_size;
  do {
    found_space = FALSE;
    for (i=0; i < NPOOLS; i++) {
      size = buf_sizes_and_counts[i*2];
      count = buf_sizes_and_counts[(i*2)+1];
      for (j = 0; j < count; j++)
	if (next+size > top)
	  break;
	else {
	  add_buf_to_pool(next, &pools[unit][i]);
	  next += size;
	  found_space = TRUE;
	}
    }
    first = 0;
  } while (found_space == TRUE);

  if (first)
    panic("Embedding host buffer space too small");

  return(0);
}

int AllocateEmbMemory (EmbCommAreaPtr, unit, size, ptr)
     EmbCommArea *EmbCommAreaPtr;
     int unit, size;
     void **ptr;
{
  int i, j;

  /* Convert bytes to words */
  size = (size+3)/4;

  for (i=0; i < NPOOLS; i++)
    /* Are buffers in pool big enough? */
    if (pools[unit][i].size >= size)
      /* Is there a free buffer? */
      for (j=0; j < pools[unit][i].count; j++)
	if (pools[unit][i].bitmap[j] == FALSE) {
	  /* Allocate the buffer */
	  pools[unit][i].bitmap[j] = TRUE;
	  *ptr = HostPointer(pools[unit][i].bufs[j] * 4);
#ifdef IVORY_DEBUG
	  DBG(DBG_BUF)
	    printf("Allocated %d-word buf %d gp 0x%x hp 0x%x\n",
		   pools[unit][i].size, j, GuestPointer(*ptr), *ptr);
#endif
	  return(0);
	}

  *ptr = NULL;
  return(ENOMEM);
}

int FreeEmbMemory (EmbCommAreaPtr, unit, ptr)
     EmbCommArea *EmbCommAreaPtr;
     int unit;
     void **ptr;
{
  int i, j;
  EmbPtr buf;

  buf = GuestPointer(*ptr);

  for (i = 0; i < NPOOLS; i++)
    for (j = 0; j < pools[unit][i].avail; j++)
      if (pools[unit][i].bufs[j] == buf)
	if (pools[unit][i].bitmap[j] == FALSE)
	  printf("Ignored attempt to free already-freed embedded buffer 0x%x\n", buf);
	else {
	  pools[unit][i].bitmap[j] = FALSE;
#ifdef IVORY_DEBUG
	  DBG(DBG_BUF)
	    printf("Freed %d-word buf %d gp 0x%x hp 0x%x\n",
		   pools[unit][i].size, j, *ptr, buf);
#endif
	  *ptr = NULL;
	  return(0);
	}

  printf("Ignored attempt to free non-existent embedded buffer 0x%x\n", buf);
  /*NOTREACHED*/
}

#endif				/* NIVORY > 0 */

#endif				/* IVORY_EMBBUF */
