/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * Disk life-support for embedded host
 *
 * $Log:	ivory_disk.c,v $
 * Revision 2.5.1.2  93/02/01  12:56:36  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:52:20  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:41:58  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#include "ivory.h"
#if NIVORY > 0

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/ivory_disk.c,v 2.5.1.2 93/02/01 12:56:36 kaufman Rel $";
#endif

#include "sys/types.h"
#include "sys/param.h"
#include "sys/file.h"
#include "sys/stat.h"
#include "sys/user.h"
#include "sys/uio.h"
#include "sys/vnode.h"
#include "ufs/inode.h"

#include "embed.h"
#include "life.h"

#undef GETF
#define	GETF(fp, fd) { \
	if ((fd) < 0 || (fd) > u_lastfile() || \
	    ((fp) = u_ofile(fd)) == NULL) { \
		set_u_error(EBADF); \
		return; \
	} \
}


/* The control structure is very simple, because the Macintosh disk driver
   is unfortunately not interrupt-driven.  If it were, we could have a
   current operation, an interrupt entry point, and all that hair.
 */

/* ---Disk error handling still needs a great deal of work */

extern void ResetDiskMeters();

/* Process Read or Write request */
/* Later, should figure out how to do this asynchronously so other Life can be overlapped with IO */

IvoryDiskIO (EmbCommAreaPtr, cmd, channel)
     EmbCommArea* EmbCommAreaPtr;
     EmbDiskQueueElement* cmd;
     EmbDiskChannel* channel;
{ 
  struct file* fp = channel->FileStructurePointer;
  struct uio *uio = &channel->UserIOBlock;
  struct vnode *vn = (struct vnode *) fp->f_data;
  enum uio_rw rw;
#ifdef MACIVORY_DISK_HACKS
  long* physctr;
#endif
  
  switch (cmd->op.cmd) {
    case ReadCmd:
      rw = UIO_READ;
#ifdef MACIVORY_DISK_HACKS
      EmbCommAreaPtr->NumberOfLogicalReads += cmd->count;
      physctr = &EmbCommAreaPtr->NumberOfPhysicalReads;
#endif
      break;
    case WriteCmd:
      rw = UIO_WRITE;
#ifdef MACIVORY_DISK_HACKS
      EmbCommAreaPtr->NumberOfLogicalWrites += cmd->count;
      physctr = &EmbCommAreaPtr->NumberOfPhysicalWrites;
#endif
      break;
    }

#ifdef IVORY_DEBUG
  DBG(DBG_DISK) printf("starting disk %s... ", rw == UIO_WRITE ? "write" : "read");
#endif IVORY_DEBUG

  /* Stupid, Simple */
  { int n_daps = cmd->n_addresses;
    struct EmbAddressPair *dap = cmd->addresses;
    off_t off = cmd->page*PageSize;
    struct iovec *iov;
    int iovcnt, count, total = cmd->count*PageSize;
    int err;

    if (cmd->page+cmd->count > channel->number_of_pages)
      return(EINVAL);

    /* fill in iov from daps */
    while(n_daps > 0) {
      for (iov = &channel->IOVectorTable[0], iovcnt = 0, count = 0;
	   n_daps > 0 &&
	   iovcnt < sizeof(channel->IOVectorTable)/sizeof(struct iovec);
	   iov++, iovcnt++) {
	caddr_t base = (caddr_t)HostPointer(dap->address);
	int len = dap->n_words<< 2;

	if (len < 0) return(EINVAL);
	if ((count += len) < 0) return(EINVAL);
      
	iov->iov_base = base;
	iov->iov_len = len;
	dap++;
	n_daps--;
      }

      if ((total -= count) < 0) return(EINVAL);

      uio->uio_iov = channel->IOVectorTable;
      uio->uio_iovcnt = iovcnt;
      uio->uio_offset = off;
      uio->uio_resid = count;
      uio->uio_segflg = UIO_SYSSPACE; /* ??? */
      uio->uio_fmode = 0;	/* ??? */

      /* If the vnode indicates a regular UFS file (that is, it's an inode),
       * hack the i_nextr field in the inode to make the io routines think
       * we're doing sequential reads, so they will do an asynchronous prefetch
       * of the following blocks. */
      if (vn->v_op == &ufs_vnodeops)
	VTOI(vn)->i_nextr = off/8192;

      err = (*fp->f_ops->fo_rw)(fp, rw, uio);
      if (err)
	return(err);
      else if (uio->uio_resid)
	return(EINTR);		/* Didn't get it all */
  
#ifdef MACIVORY_DISK_HACKS
      (*physctr)++;
#endif
      off += count;
    }

#ifdef IVORY_DEBUG
    DBG(DBG_DISK) printf("done\n", rw == UIO_WRITE ? "write" : "read");
#endif IVORY_DEBUG

    return(0);
  }
}


/* Poll a disk channel when a signal is received, also periodically */
void DiskLife (EmbCommAreaPtr, ivory_unit, channel)
     EmbCommArea* EmbCommAreaPtr;
     int ivory_unit;		/*ARGSUSED*/
     EmbDiskChannel* channel;
{ EmbQueue* cq = channel->command_queue_ptr;
  EmbQueue* sq = channel->status_queue_ptr;
  EmbPtr cmd_emb_ptr;
  EmbDiskQueueElement* cmd;
  int err;

  while (EmbQueueFilled(EmbCommAreaPtr, cq)) {
    if ( /* Don't run the disk driver if requested not to */
	EmbCommAreaPtr->inhibit_disk ||
#ifdef MACIVORY_DISK_HACKS
	/* Don't run the disk driver if it has been continuously active for too long */
	EmbCommAreaPtr->remaining_disk_io_throttle_time ||
#endif
	/* Don't run the disk driver if the status queue is full */
	EmbQueueSpace(EmbCommAreaPtr, sq) == 0 ) {
      EmbCommAreaPtr->reawaken |= (long)1 << cq->signal;
      return;
    }

    cmd_emb_ptr = EmbQueueTakeWord(EmbCommAreaPtr, cq);
    if ( cmd_emb_ptr == FALSE ) {
      printf("Disk command queue clobbered");
      return;
    }
    cmd = (EmbDiskQueueElement*)HostPointer(cmd_emb_ptr);
    switch (cmd->op.cmd) {	/* --ignores unimplemented fields */
    case ReadCmd:
    case WriteCmd:
      if ( channel->error_pending)
	cmd->status = AbortStatus;
      else if ( channel->FileStructurePointer == NULL ) {
	cmd->status = LostStatus;
	cmd->error_code = ENXIO;
      } else {
	err = IvoryDiskIO(EmbCommAreaPtr, cmd, channel);
	if (err) {
	  cmd->error_code = err;
	  cmd->status = LostStatus;
	}
	else 
	  cmd->status = WonStatus;
      }
      break;
    case ResetCmd:
      channel->error_pending = 0;
#ifdef MACIVORY_DISK_HACKS
      EmbCommAreaPtr->remaining_disk_io_throttle_time = 0;
#endif
      cmd->status = WonStatus;
      break;
    case InitializeCmd:
#ifdef MACIVORY_DISK_HACKS
      ResetDiskMeters(EmbCommAreaPtr);
#endif
      channel->error_pending = 0;
#ifdef MACIVORY_DISK_HACKS
      EmbCommAreaPtr->remaining_disk_io_throttle_time = 0;
      EmbCommAreaPtr->number_of_disk_throttles = 0;
#endif
      cmd->status = WonStatus;
      break;
    default:
      cmd->status = LostStatus;
      cmd->error_code = ENXIO;
      printf("Disk Error: Unrecognized command %d\n", cmd->op.cmd);
    }
    EmbQueuePutWord(EmbCommAreaPtr, sq, cmd_emb_ptr);
  }
}

/* Initialization entry */

void InitializeDiskLifeII (EmbCommAreaPtr, ivory_unit)
     EmbCommArea* EmbCommAreaPtr;
     int ivory_unit;		/*ARGSUSED*/
{ 
  EmbPtr cptr;
  EmbChannel *cp;
  EmbDiskChannel *disk;

  for (cptr = EmbCommAreaPtr->channel_table; cptr != NullEmbPtr; cptr = cp->next) {
    cp = (EmbChannel *) HostPointer(cptr);
    if (cp->type == EmbDiskChannelType) {
      disk = (EmbDiskChannel *) cp;
      if (!disk->initialized_p) {
	GETF(disk->FileStructurePointer, (int)disk->FileStructurePointer);
	disk->command_queue_ptr = (EmbQueue *) HostPointer(disk->command_queue);
	disk->status_queue_ptr = (EmbQueue *) HostPointer(disk->status_queue);
	disk->initialized_p = 1;
#ifdef IVORY_DEBUG
	DBG(DBG_INIT)
	  printf("initialized disk unit %d\n", disk->unit);
#endif
      }
    }
  }
}

#endif				/* NIVORY > 0 */
