/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * Basic subroutines for embedded host
 *
 * $Log:	ivory_embed.c,v $
 * Revision 2.5.1.2  93/02/01  12:56:37  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:52:22  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:42:02  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#include "ivory.h"
#if NIVORY > 0

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/ivory_embed.c,v 2.5.1.2 93/02/01 12:56:37 kaufman Rel $";
#endif

#include "sys/types.h"
#include "embed.h"
#include "life.h"

#define BlockMove(a,b,c) bcopy((char *)a, (char *)b, (int)c)


/*** Queues ***/

long EmbQueueSpace (EmbCommAreaPtr, qp)	/* Number of free elements */
     EmbCommArea *EmbCommAreaPtr; /*ARGSUSED*/
     EmbQueue* qp;
{
  register long put = qp->put_index;
  register long take = qp->take_index;

  if ( take > put )
    return ( take - put - 1 );
  else return ( take - put - 1 + qp->queue_size );
}

long EmbQueueFilled (EmbCommAreaPtr, qp) /* Number of non-free elements */
     EmbCommArea *EmbCommAreaPtr; /*ARGSUSED*/
     EmbQueue* qp;
{
  register long put = qp->put_index;
  register long take = qp->take_index;

  if ( put >= take )
    return ( put - take );
  else return ( put - take + qp->queue_size );
}

void EmbQueuePut (EmbCommAreaPtr, qp_arg, ep)	/* Put element into queue */
     EmbCommArea *EmbCommAreaPtr;
     EmbQueue* qp_arg;
     PtrV ep;
{
  register EmbQueue* qp = qp_arg;
  register long put = qp->put_index;
  long original_put = put;
  byte* element_array = (byte*) (qp->first_element);

  /* Fill in the element at the put index */
  BlockMove(ep, &element_array[put * qp->element_size], qp->element_size);

  /* Compute the incremented put index */
  put++;
  if (put >= qp->queue_size)
    put = 0;

  /* Wait if queue is full.  Queue sizes will be chosen so that this is rare. */
  while (put == qp->take_index);

  /* Store the incremented put index into the queue header */
  qp->put_index = put;

  /* If the queue had been empty, send a signal to the taker. */
  if (original_put == qp->take_index)
    EmbSendSignal(EmbCommAreaPtr, qp->signal);
}

void EmbQueuePutWord (EmbCommAreaPtr, qp_arg, elt)	/* Put element into word queue */
     EmbCommArea *EmbCommAreaPtr;
     EmbQueue* qp_arg;
     long elt;
{ register EmbQueue* qp = qp_arg;
  register long put = qp->put_index;
  long original_put = put;
  register long* element_array = (long*)(qp->first_element);

  /* Fill in the element at the put index */
  element_array[put] = elt;

  /* Compute the incremented put index */
  put++;
  if ( put >= qp->queue_size )
    put = 0;

  /* Wait if queue is full.  Queue sizes will be chosen so that this is rare. */
  while ( put == qp->take_index );

  /* Store the incremented put index into the queue header */
  qp->put_index = put;

  /* If the queue had been empty, send a signal to the taker. */
  if ( original_put == qp->take_index )
    EmbSendSignal(EmbCommAreaPtr, qp->signal);
}

void EmbQueuePutByte (EmbCommAreaPtr, qp_arg, elt)	/* Put element into byte queue */
     EmbCommArea *EmbCommAreaPtr;
     EmbQueue* qp_arg;
     byte elt;
{ register EmbQueue* qp = qp_arg;
  register long put = qp->put_index;
  long original_put = put;
  register byte* element_array = (byte*)(qp->first_element);

  /* Fill in the element at the put index */
  element_array[put] = elt;

  /* Compute the incremented put index */
  put++;
  if ( put >= qp->queue_size )
    put = 0;

  /* Wait if queue is full.  Queue sizes will be chosen so that this is rare. */
  while ( put == qp->take_index );

  /* Store the incremented put index into the queue header */
  qp->put_index = put;

  /* If the queue had been empty, send a signal to the taker. */
  if ( original_put == qp->take_index )
    EmbSendSignal(EmbCommAreaPtr, qp->signal);
}

bool EmbQueueTake (EmbCommAreaPtr, qp_arg, ep) /* Take element from queue */
     EmbCommArea *EmbCommAreaPtr; /*ARGSUSED*/
     EmbQueue *qp_arg;
     PtrV ep;
{ register EmbQueue* qp = qp_arg;
  register long put = qp->put_index;
  register long take = qp->take_index;
  register byte* element_array = (byte*)(qp->first_element);

  /* Check for empty queue */
  if ( put == take )
    return ( FALSE );

  /* Copy the element at the take index */
  BlockMove(&element_array[take * qp->element_size], ep, qp->element_size);

  /* Increment the take index and store it back */
  take++;
  if ( take >= qp->queue_size )
    take = 0;
  qp->take_index = take;
  return ( TRUE );
}

long EmbQueueTakeWord (EmbCommAreaPtr, qp_arg)	/* Take element from word queue */
     EmbCommArea *EmbCommAreaPtr; /*ARGSUSED*/
     EmbQueue* qp_arg;
{ register EmbQueue* qp = qp_arg;
  register long put = qp->put_index;
  register long take = qp->take_index;
  register long* element_array = (long*)(qp->first_element);
  register long elt;

  /* Check for empty queue--should not happen */
  if ( put == take )
    return ( FALSE );

  /* Copy the element at the take index */
  elt = element_array[take];

  /* Increment the take index and store it back */
  take++;
  if ( take >= qp->queue_size )
    take = 0;
  qp->take_index = take;
  return ( elt );
}

byte EmbQueueTakeByte (EmbCommAreaPtr, qp_arg)	/* Take element from byte queue */
     EmbCommArea *EmbCommAreaPtr; /*ARGSUSED*/
     EmbQueue* qp_arg;
{ register EmbQueue* qp = qp_arg;
  register long put = qp->put_index;
  register long take = qp->take_index;
  register byte* element_array = (byte*)(qp->first_element);
  register byte elt;

  /* Check for empty queue--should not happen */
  if ( put == take )
    return ( FALSE );

  /* Copy the element at the take index */
  elt = element_array[take];

  /* Increment the take index and store it back */
  take++;
  if ( take >= qp->queue_size )
    take = 0;
  qp->take_index = take;
  return ( elt );
}


/*** Transferring multiple queue elements at a time ***/

/* Each of these routines returns the number of elements actually transferred */
/* None of these routines ever waits */

long EmbQueuePutBytes (EmbCommAreaPtr, qp_arg, buffer, length)	/* Put multiple bytes */
     EmbCommArea *EmbCommAreaPtr;
     EmbQueue* qp_arg;
     byte* buffer;
     long length;
{ register EmbQueue* qp = qp_arg;
  register long put = qp->put_index;
  long original_put = put;
  register long take;
  register byte* element_array = (byte*)(qp->first_element);
  long actual_length = 0;
  register long count;

  /* Loop transferring contiguous blocks of elements */
  while ( length > 0 )
    { /* Fill in as many elements as we can */
      take = qp->take_index;
      if ( take > put )
	count = take - put - 1;
      else if ( take == 0 )
	count = qp->queue_size - put - 1;
      else count = qp->queue_size - put;
      if ( count > length ) count = length;
      BlockMove(buffer, &element_array[put], count);
      buffer += count;
      length -= count;
      actual_length += count;
      put += count;
      if ( put == qp->queue_size) put = 0;
      qp->put_index = put;
    };

  /* If the queue had been empty, send a signal to the taker. */
  if ( original_put == qp->take_index )
    EmbSendSignal(EmbCommAreaPtr, qp->signal);
  return ( actual_length );
}

long EmbQueuePutWords (EmbCommAreaPtr, qp_arg, buffer, length)	/* Put multiple words */
     EmbCommArea *EmbCommAreaPtr;
     EmbQueue* qp_arg;
     long* buffer;
     long length;
{ register EmbQueue* qp = qp_arg;
  register long put = qp->put_index;
  long original_put = put;
  register long take;
  register long* element_array = (long*)(qp->first_element);
  long actual_length = 0;
  register long count;

  /* Loop transferring contiguous blocks of elements */
  while ( length > 0 )
    { /* Fill in as many elements as we can */
      take = qp->take_index;
      if ( take > put )
	count = take - put - 1;
      else if ( take == 0 )
	count = qp->queue_size - put - 1;
      else count = qp->queue_size - put;
      if ( count > length ) count = length;
      BlockMove(buffer, &element_array[put], count << 2);
      buffer += count;
      length -= count;
      actual_length += count;
      put += count;
      if ( put == qp->queue_size) put = 0;
      qp->put_index = put;
    };

  /* If the queue had been empty, send a signal to the taker. */
  if ( original_put == qp->take_index )
    EmbSendSignal(EmbCommAreaPtr, qp->signal);
  return ( actual_length );
}

long EmbQueueTakeBytes (EmbCommAreaPtr, qp_arg, buffer, length)	/* Take multiple bytes */
     EmbCommArea *EmbCommAreaPtr; /*ARGSUSED*/
     EmbQueue* qp_arg;
     byte* buffer;
     long length;
{ register EmbQueue* qp = qp_arg;
  register long take = qp->take_index;
  register long put;
  register byte* element_array = (byte*)(qp->first_element);
  long actual_length = 0;
  register long count;

  /* Loop transferring contiguous blocks of elements */
  while ( length > 0 )
    { /* Fill in as many elements as we can */
      put = qp->put_index;
      if ( put >= take )
	count = put - take;
      else count = qp->queue_size - take;
      if ( count > length ) count = length;
      BlockMove(&element_array[take], buffer, count);
      buffer += count;
      length -= count;
      actual_length += count;
      take += count;
      if ( take == qp->queue_size) take = 0;
      qp->take_index = take;
    };
  return ( actual_length );
}

long EmbQueueTakeWords (EmbCommAreaPtr, qp_arg, buffer, length)	/* Take multiple words */
     EmbCommArea *EmbCommAreaPtr; /*ARGSUSED*/
     EmbQueue* qp_arg;
     long* buffer;
     long length;
{ register EmbQueue* qp = qp_arg;
  register long take = qp->take_index;
  register long put;
  register long* element_array = (long*)(qp->first_element);
  long actual_length = 0;
  register long count;

  /* Loop transferring contiguous blocks of elements */
  while ( length > 0 )
    { /* Fill in as many elements as we can */
      put = qp->put_index;
      if ( put >= take )
	count = put - take;
      else count = qp->queue_size - take;
      if ( count > length ) count = length;
      BlockMove(&element_array[take], buffer, count << 2);
      buffer += count;
      length -= count;
      actual_length += count;
      take += count;
      if ( take == qp->queue_size) take = 0;
      qp->take_index = take;
    };
  return ( actual_length );
}


/*** Debugging ***/

#ifdef IVORY_DEBUG
describe_emb_queue (EmbCommAreaPtr, qp, name)
     EmbCommArea *EmbCommAreaPtr;
     EmbQueue *qp;
     char *name;
{
  int n = EmbQueueFilled(EmbCommAreaPtr, qp);
  
  /* Do it all in only two printf() calls, to save csw's */
  printf("Queue %s at emb %d\n  element-size  %d\n  queue-size    %d\n  put-index     %d\n  take-index    %d\n  signal-number %d\n",
	 name, GuestPointer(qp),
	 qp->element_size,
	 qp->queue_size,
	 qp->put_index,
	 qp->take_index,
	 (long) qp->signal);
  if (n)
    printf("  contains %d elements\n", n);
  else
    printf("  is empty\n");
}
#endif

#endif				/* NIVORY > 0 */
