/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * Life-support main loop(s)
 *
 * $Log:	ivory_life.c,v $
 * Revision 2.5.1.2  93/02/01  12:56:39  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:52:24  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:42:07  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#include "ivory.h"
#if NIVORY > 0

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/ivory_life.c,v 2.5.1.2 93/02/01 12:56:39 kaufman Rel $";
#endif

#include "sys/types.h"
#include "sys/time.h"
#include "sys/param.h"
#include "sys/file.h"
#include "sys/proc.h"
#include "sys/stream.h"
#include "sys/uio.h"
#include "sys/user.h"

#include "embed.h"
#include "life.h"

#undef GETF
#ifdef lint
#define GETF(fp, fd)  fp = getf(fd)
#else
#define	GETF(fp, fd) { \
	if ((fd) < 0 || (fd) > u_lastfile() || \
	    ((fp) = u_ofile(fd)) == NULL) { \
		set_u_error(EBADF); \
		return; \
	} \
}
#endif


/* Returns TRUE if the Guest is running either in the IFEP or Lisp */
GuestIsRunning(ep)
  EmbCommArea *ep;
{ 
  return ((ep->spy_status == 0) && (ep->fep_status != HaltedFEPStatus));
}

/* Returns TRUE if the Guest is running Lisp */
GuestIsRunningLisp (ep)
  EmbCommArea *ep;
{
  return (GuestIsRunning(ep) && (ep->fep_status == IdleFEPStatus));
}

/* Updates the Guest status to reflect its current true status */
void UpdateGuestStatus (EmbCommAreaPtr)
  EmbCommArea *EmbCommAreaPtr;
{
  register EmbCommArea *ep = EmbCommAreaPtr;

  switch (ep->guest_status) {
  case NonexistentGuestStatus:
  case BrokenGuestStatus:
    /* Guest stays missing or broken until someone fixes it */
    break;

  case UninitializedGuestStatus:
  case InitializingGuestStatus:
  case InitializedGuestStatus:
    if (GuestIsRunning (ep))
      ep->guest_status = GuestIsRunningLisp(ep) ? RunningGuestStatus : StartedGuestStatus;
    break;

  case StartedGuestStatus:
    /* If Guest's no longer running in the IFEP, act as if it's freshly initialized */
    ep->guest_status = GuestIsRunningLisp(ep) ? RunningGuestStatus :
      GuestIsRunning(ep) ? StartedGuestStatus : InitializedGuestStatus;
    break;

  case CrashedGuestStatus:
  case RunningGuestStatus:
    ep->guest_status = GuestIsRunningLisp(ep) ? RunningGuestStatus :
      GuestIsRunning(ep) ? CrashedGuestStatus : InitializedGuestStatus;
    break;
  }
}


/* Device PROM hackery */

device_prom_rewind (EmbCommAreaPtr, unit)
     EmbCommArea *EmbCommAreaPtr; /*ARGSUSED*/
     int unit;
{
  extern struct ivory_state ivories[];
  struct file *fp;

  GETF(fp, ivories[unit].config.device_prom_fd);
  if (fp) {
    fp->f_offset = 0;
#ifdef IVORY_DEBUG
    DBG(DBG_INIT) printf("ivory%d: device-prom rewound\n", unit);
#endif
    return(0);
  } else {
#ifdef IVORY_DEBUG
    DBG(DBG_INIT) printf("ivory%d: couldn't rewind device-prom (fd %d)\n",
			 unit, ivories[unit].config.device_prom_fd);
#endif
    return(-1);
  }
}

device_prom_read (EmbCommAreaPtr, unit)
     EmbCommArea *EmbCommAreaPtr; /*ARGSUSED*/
     int unit;
{
  extern struct ivory_state ivories[];
  struct file *fp;
  struct uio uio;
  struct iovec iov;
  int err, nbytes;

  GETF(fp, ivories[unit].config.device_prom_fd);
  if (fp) {
    
    nbytes = EmbCommAreaPtr->fep_buffer_size<<2;
    iov.iov_base = (caddr_t) HostPointer(EmbCommAreaPtr->fep_buffer_start);
    iov.iov_len = nbytes;

    uio.uio_iov = &iov;
    uio.uio_iovcnt = 1;
    uio.uio_offset = fp->f_offset;
    uio.uio_resid = nbytes;
    uio.uio_segflg = UIO_SYSSPACE; /* ??? */
    uio.uio_fmode = 0;		/* ??? */
    
    err = (*fp->f_ops->fo_rw)(fp, UIO_READ, &uio);

    if (err) {
      /* --- Not debugging output, should use log() */
      printf("ivory%d: Err%d loading device prom\n", unit, err);
      return(1);
    } else {
      nbytes -= uio.uio_resid;
#ifdef IVORY_DEBUG
      DBG(DBG_INIT) printf("ivory%d: device-prom read %d bytes @%d\n",
			   unit, nbytes, fp->f_offset);
#endif
      fp->f_offset += nbytes;
      return(0);
    }
  } else {
#ifdef IVORY_DEBUG
    DBG(DBG_INIT) printf("ivory%d: couldn't read device-prom (fd %d)\n",
			 unit, ivories[unit].config.device_prom_fd);
#endif
    return(-1);
  }
}



/* Check for requests from the guest */
IvoryLife (EmbCommAreaPtr, unit)
     EmbCommArea *EmbCommAreaPtr;
     int unit;
{
  int wakeup();
  extern int hz;
  extern struct ivory_state ivories[];

  for (;;) {

    if (ivories[unit].config.enable_Ivory) {

      /* Check for guest-requested reset.
       * MacIvory doesn't do this here, but I couldn't figure out where it does do it */
      if (EmbCommAreaPtr->reset_request != NoResetRequest) {
	switch(EmbCommAreaPtr->reset_request) {
	case AreYouThereResetRequest: /* NOOP that proves life support is there */
	  EmbCommAreaPtr->reset_request = NoResetRequest;
	  break;

	case BootResetRequest:	/* Initialize for Device-PROM loading */
	  if (device_prom_rewind(EmbCommAreaPtr, unit) == 0) {
	    EmbCommAreaPtr->reset_request = NoResetRequest;
	    EmbCommAreaPtr->device_prom_reset_request_count++;
	  }
	  break;

	case DevicePROMResetRequest: /* Read in next FEP-buffer-size words of Device-PROM */
	  if (device_prom_read(EmbCommAreaPtr, unit) == 0) {
	    EmbCommAreaPtr->reset_request = NoResetRequest;
	    EmbCommAreaPtr->device_prom_reset_request_count++;
	  }
	  break;

	case FEPResetRequest:
	case LispResetRequest:
	  ResetCommArea(EmbCommAreaPtr, unit);
	  ResetSubProcesses(EmbCommAreaPtr, unit);
	  if (EmbCommAreaPtr->reset_request == LispResetRequest)
	    EmbCommAreaPtr->reset_request_count++;
	  EmbCommAreaPtr->reset_request = NoResetRequest;
	  break;

	case NoResetRequest:
	  break;

	default:
#ifdef IVORY_DEBUG
	  printf("ivory %d: bogus reset request %d\n", unit, EmbCommAreaPtr->reset_request);
#endif
	  EmbCommAreaPtr->reset_request = NoResetRequest;
	  break;
	}
	/* Wait for a tick or an interrupt after a reset */
      } else {

	/* Check for pending emb signals */
	ProcessCheckPendingSignals(EmbCommAreaPtr, unit);

      }

      /* Update private status based on FEP's status */
      UpdateGuestStatus(EmbCommAreaPtr);
    }

    /* Wait for next signal or reset */
    sleep((caddr_t) u_procp(), PSLEP|PCATCH);

    /* Break on signals */
    if (p_cursig(u_procp()) != 0)
      return(EINTR);
  }
}


/* IvoryLife() for sub processes */
IvorySubProcessLife (EmbCommAreaPtr, ivory_unit, reset_fn)
     EmbCommArea *EmbCommAreaPtr;
     int ivory_unit;
     void (*reset_fn)();
{
  int reset_count = EmbCommAreaPtr->reset_request_count;
  int wakeup();
  extern int hz;

  for (;;) {

    if (ivories[ivory_unit].config.enable_Ivory) {

      if (reset_count != EmbCommAreaPtr->reset_request_count) {
	/* Handle a reset */
	reset_count = EmbCommAreaPtr->reset_request_count;
	if (reset_fn != NULL)
	  (*reset_fn)(EmbCommAreaPtr);
      } else {

	/* Check for pending emb signals */
	ProcessCheckPendingSignals(EmbCommAreaPtr, ivory_unit);

	/* We used to check reset_request and guest_to_host_signals here,
	 * and loop again if one of them were non-zero.  This is no good because
	 * some other process might be trying to get scheduled to handle one of
	 * the signals that is still set. */

      }
    }

    /* Wait for next signal or reset */
    sleep((caddr_t) u_procp(), PSLEP|PCATCH);
    
    /* Break on signals */
    if (p_cursig(u_procp()) != 0)
      return(EINTR);
  }
}


static int clock_timeout = 0;

/* Clock interrupt handler */
void IvoryClock (unit)
     long unit;
{
  extern struct ivory_state ivories[];
  register EmbCommArea *ep = ivories[unit].EmbCommArea;
  int nticks;
#ifndef POLL_EMB_COMM_AREA
  int intr = 0;
#endif
#ifdef IVORY_DEBUG
  static int clobbered_p = 0;
#endif

#ifdef POLL_EMB_COMM_AREA
  nticks = 1;
#else
  nticks = clock_timeout;
#endif

#ifdef IVORY_DEBUG
  if (ep->identifier != 'DBME') {
    if (!clobbered_p) {
      clobbered_p = 1;
      printf("Embedding communication area clobbered\n");
    }
  } else if (clobbered_p) {
    clobbered_p = 0;
    printf("Embedding communication area unclobbered\n");
  }
#endif

#ifdef MACIVORY_DISK_HACKS
  if (ep->remaining_disk_io_throttle_time > 0)
    ep->remaining_disk_io_throttle_time--;
#endif

#ifndef POLL_EMB_COMM_AREA
  /* Check for a pending reset */
  if (ep->reset_request != NoResetRequest) {
    /* Service the reset immediately, and check quickly for another. */
    intr = 1;
    nticks = 1;
  }
#endif

  /* Poll all signals four times a second */
  if (ep->poll_tick >= hz/4) {
    ep->poll_tick = 0;
    ep->guest_to_host_signals |= ep->live_guest_to_host_signals;
    ep->reawaken = 0;
#ifndef POLL_EMB_COMM_AREA
    intr = 1;
#endif
  } else if (ep->reawaken) {
    ep->guest_to_host_signals |= ep->reawaken;
    ep->reawaken = 0;
#ifndef POLL_EMB_COMM_AREA
    intr = 1;
#endif
  }
  ep->poll_tick += nticks;
    
#ifdef notdef
  /* Maintain Ivory clock
   * 228 (actually 227.78333) is a magic MacIvory constant */
  ep->signal_tick++;
  if ((ep->signal_tick * 228 * hz) >= ep->clock_interval) {
    ep->signal_tick = 0;
    EmbSendSignal(ep, ep->clock_signal);
  }
#endif

#ifdef POLL_EMB_COMM_AREA
  /* Poll embedding communication area every clock tick */
  ivoryintr(unit);
#else
  if (intr)
    ivoryintr(unit);
#endif

  /* Get called next timeout */
  timeout(IvoryClock, (caddr_t) unit, nticks);
}

start_ivory_clock (unit)
     int unit;
{
  int nticks;

  if (clock_timeout == 0)
    clock_timeout = hz/4;

#ifdef POLL_EMB_COMM_AREA
  nticks = 1;
#else
  nticks = clock_timeout;
#endif

  timeout(IvoryClock, (caddr_t) unit, nticks);
}

stop_ivory_clock (unit)
     int unit;
{
  untimeout(IvoryClock, (caddr_t) unit);
}

#endif				/* NIVORY > 0 */
