/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * Routines for resetting the communication area at the guest's request.
 * Call ResetCommArea if EmbCommAreaPtr->reset_request is nonzero.
 *
 * $Log:	ivory_reset.c,v $
 * Revision 2.5.1.2  93/02/01  12:56:43  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:52:31  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:42:21  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#include "ivory.h"
#if NIVORY > 0

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/ivory_reset.c,v 2.5.1.2 93/02/01 12:56:43 kaufman Rel $";
#endif

#include "sys/types.h"
#include "sys/stream.h"

#include "embed.h"
#include "life.h"

#define ResetQueue(EP, Q) ResetQueueInternal(EP, Q, TRUE)

void ResetQueueInternal (EmbCommAreaPtr, q, host_signal_p)
     EmbCommArea *EmbCommAreaPtr;
     EmbQueue *q;
     int host_signal_p;
{ 
  void MaybeSignal();
  int s;

  q->put_index = 0;
  q->take_index = 0;

  /* Remove signal handlers from queues created by the FEP or guest */
  if (GuestPointer(q) > EmbCommAreaPtr->host_buffer_start+EmbCommAreaPtr->host_buffer_size) {
    s = splhigh();
    if (--EmbCommAreaPtr->SignalHandler[q->signal].count == 0) {
      RemoveSignalHandler(EmbCommAreaPtr, q->signal);
#ifdef IVORY_DEBUG
      DBG(DBG_SIG) printf("Removed handler for signal %d\n", q->signal);
#endif
    }
#ifdef IVORY_DEBUG
    else DBG(DBG_SIG) printf("%d users of sig %d remain\n",
			       EmbCommAreaPtr->SignalHandler[q->signal].count,
			       q->signal);
#endif
    q->signal = -1;
    splx(s);
  }

  if (host_signal_p)
    MaybeSignal(EmbCommAreaPtr, q->signal);
}

/* Call ResetOutgoingQueue when the signal belongs to the guest */
/* Call ResetQueue when the signal belongs to the host */
void ResetOutgoingQueue (EmbCommAreaPtr, q)
     EmbCommArea *EmbCommAreaPtr;
     EmbQueue *q;
{ 
  ResetQueueInternal(EmbCommAreaPtr, q, FALSE);
  q->signal = -1;
}

#ifdef MACIVORY_DISK_HACKS
void PurgeDiskCache (EmbCommAreaPtr)
     EmbCommArea *EmbCommAreaPtr;
{ register long line;
  register long* CacheTag = EmbCommAreaPtr->CacheTag;

  for (line = EmbCommAreaPtr->CacheSize - 1; line >= 0; line--)
    CacheTag[line] = -1;
}

void ResetDiskMeters (EmbCommAreaPtr)
     EmbCommArea *EmbCommAreaPtr;
{ register EmbCommArea* cp = EmbCommAreaPtr;

  cp->NumberOfHits = 0;
  cp->NumberOfMisses = 0;
  cp->NumberOfLogicalReads = 0;
  cp->NumberOfLogicalWrites = 0;
  cp->NumberOfPhysicalReads = 0;
  cp->NumberOfPhysicalWrites = 0;
  cp->NumberOfPhysicalMerges = 0;
  cp->NumberOfPhysicalReadOps = 0;
  cp->NumberOfPhysicalWriteOps = 0;
}
#endif

void ResetDiskChannel (EmbCommAreaPtr, cp)
     EmbCommArea *EmbCommAreaPtr;
     EmbChannel* cp;
{
  register EmbDiskChannel* ch = (EmbDiskChannel*)cp;

  ResetQueue(EmbCommAreaPtr, (EmbQueue *) HostPointer(ch->command_queue));
  ResetOutgoingQueue(EmbCommAreaPtr, (EmbQueue *) HostPointer(ch->status_queue));
  ch->error_pending = 0;
#ifdef MACIVORY_DISK_HACKS
  EmbCommAreaPtr->remaining_disk_io_throttle_time = 0;
  EmbCommAreaPtr->number_of_disk_throttles = 0;
  PurgeDiskCache(EmbCommAreaPtr);
  ResetDiskMeters(EmbCommAreaPtr);
#endif
}

void ResetNetworkChannel (EmbCommAreaPtr, cp)
     EmbCommArea *EmbCommAreaPtr; /*ARGSUSED*/
     EmbChannel* cp;
{
  register EmbNetChannel* ch = (EmbNetChannel*)cp;

  ResetQueue(EmbCommAreaPtr, (EmbQueue *) HostPointer(ch->xmit_guest_to_host_queue));
  ResetOutgoingQueue(EmbCommAreaPtr, (EmbQueue *) HostPointer(ch->xmit_host_to_guest_queue));
  ResetQueue(EmbCommAreaPtr, (EmbQueue *) HostPointer(ch->recv_guest_to_host_queue));
  ResetOutgoingQueue(EmbCommAreaPtr, (EmbQueue *) HostPointer(ch->recv_host_to_guest_queue));

  /* Flush STREAMS queue */
  if (ch->network_queue)
    putctl(ch->network_queue, M_FLUSH);

  /* Turn off guest-ready bit */
  ch->status &= ~EmbNet_status_guest_ready;
}

void ResetColdLoadChannel (EmbCommAreaPtr, cp)
     EmbCommArea *EmbCommAreaPtr;
     EmbChannel* cp;
{ register EmbColdLoadChannel* ch = (EmbColdLoadChannel*)cp;

  /* The kernel doesn't run the cold load, so don't trust the EmbQueue*'s */
  ResetQueue(EmbCommAreaPtr, (EmbQueue *) HostPointer(ch->display_output_queue));
  ResetOutgoingQueue(EmbCommAreaPtr, (EmbQueue *) HostPointer(ch->keyboard_input_queue));
}

void ResetMessageChannel (EmbCommAreaPtr, cp)
     EmbCommArea *EmbCommAreaPtr;
     EmbChannel* cp;
{
  register EmbMessageChannel* ch = (EmbMessageChannel*) cp;

  ch->guest_to_host_impulse = 0;
  ResetQueue(EmbCommAreaPtr, (EmbQueue *) HostPointer(ch->guest_to_host_queue));
  ResetOutgoingQueue(EmbCommAreaPtr, (EmbQueue *) HostPointer(ch->guest_to_host_return_queue));
  ch->host_to_guest_impulse = 0;
  ResetOutgoingQueue(EmbCommAreaPtr, (EmbQueue *) HostPointer(ch->host_to_guest_queue));
  ResetQueue(EmbCommAreaPtr, (EmbQueue *) HostPointer(ch->host_to_guest_supply_queue));

  if (GuestPointer(cp) > EmbCommAreaPtr->host_buffer_start+
      EmbCommAreaPtr->host_buffer_size)
    ch->selproc = 0;
}

void ResetHostFileChannel (EmbCommAreaPtr, cp)
     EmbCommArea *EmbCommAreaPtr;
     EmbChannel* cp;
{
  register EmbHostFileChannel* ch = (EmbHostFileChannel*) cp;

  ResetQueue(EmbCommAreaPtr, (EmbQueue *) HostPointer(ch->command_queue));
  ResetOutgoingQueue(EmbCommAreaPtr, (EmbQueue *) HostPointer(ch->reply_queue));
}

#ifdef unused
void ResetSCSIChannel (EmbCommAreaPtr, cp)
     EmbCommArea *EmbCommAreaPtr;
     EmbChannel* cp;
{ register EmbSCSIChannel* ch = (EmbSCSIChannel*)cp;

  /* The kernel doesn't run these channels, so don't trust the EmbQueue*'s */
  ResetQueue(EmbCommAreaPtr, (EmbQueue *) HostPointer(ch->command_queue));
  ResetOutgoingQueue(EmbCommAreaPtr, (EmbQueue *) HostPointer(ch->reply_queue));
  ch->unit = -1;
}

void ResetCommandChannel (EmbCommAreaPtr, cp)
     EmbCommArea *EmbCommAreaPtr; /*ARGSUSED*/
     EmbChannel* cp;
{ register EmbCommandChannel* ch = (EmbCommandChannel*)cp;

  ResetQueue(EmbCommAreaPtr, ch->guest_to_host_queue_ptr);
  ResetOutgoingQueue(EmbCommAreaPtr, ch->host_to_guest_free_queue_ptr);
  ResetOutgoingQueue(EmbCommAreaPtr, ch->host_to_guest_queue_ptr);
  ResetQueue(EmbCommAreaPtr, ch->guest_to_host_free_queue_ptr);
}
#endif

void ResetCommArea (EmbCommAreaPtr, unit)
     EmbCommArea *EmbCommAreaPtr;
     int unit;			/*ARGSUSED*/
{
  register EmbChannel* cp;
  register EmbPtr ch;

  /* Reset all the channels of types that we know about */
  for (ch = EmbCommAreaPtr->channel_table; ch != NullEmbPtr; ch = cp->next) {
    cp = (EmbChannel *) HostPointer (ch);
    switch (cp->type) {

    case EmbDiskChannelType:
      ResetDiskChannel (EmbCommAreaPtr, cp);
      break;

    case EmbNetworkChannelType:
      if (LispResetRequest == EmbCommAreaPtr->reset_request)
	ResetNetworkChannel(EmbCommAreaPtr, cp);
      break;

    case EmbColdLoadChannelType:
      ResetColdLoadChannel (EmbCommAreaPtr, cp);
      break;

    case EmbMessageChannelType:
      ResetMessageChannel (EmbCommAreaPtr, cp);
      break;

    case EmbHostFileChannelType:
      ResetHostFileChannel(EmbCommAreaPtr, cp);
      break;

#ifdef unused
    case EmbSCSIChannelType:
      ResetSCSIChannel (EmbCommAreaPtr, cp);
      break;
    case EmbCommandChannelType:
      if (LispResetRequest == EmbCommAreaPtr->reset_request)
	ResetCommandChannel (EmbCommAreaPtr, cp);
#endif

    default:
#ifdef IVORY_DEBUG
      printf("Unknown channel type %d at %d, not reset\n", cp->type, ch);
#endif 
      break;
    }
  }

#ifdef IVORY_EMBBUF
  /* Reset the host buffer memory when reset requested by Lisp */
  if (LispResetRequest == EmbCommAreaPtr->reset_request) {
    InitializeEmbMemory(EmbCommAreaPtr, unit);
  }
#endif

  /* Reset complete */
  if (LispResetRequest == EmbCommAreaPtr->reset_request)
    /* Must restart applications whenever reset */
    EmbCommAreaPtr->restart_applications = 1;


#ifdef IVORY_DEBUG
  DBG(DBG_INIT) printf("ivory%d comm area reset\n", unit);
#endif 
}

#endif				/* NIVORY > 0 */
