/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * Life-support signal allocation, reclamation, and other related fun
 *
 * $Log:	ivory_signal.c,v $
 * Revision 2.5.1.2  93/02/01  12:56:45  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:52:33  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:42:26  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#include "ivory.h"
#if NIVORY > 0

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/ivory_signal.c,v 2.5.1.2 93/02/01 12:56:45 kaufman Rel $";
#endif

#include "sys/errno.h"
#include "sys/time.h"
#include "sys/types.h"
#include "sys/param.h"
#include "sys/proc.h"
#include "sys/user.h"

#include "embed.h"
#include "life.h"

/*** Signals ***/

/* Signal handler table management */

void FixupSignalHandlers (EmbCommAreaPtr)
     EmbCommArea *EmbCommAreaPtr;
{
  register EmbCommArea* ep = EmbCommAreaPtr;
  register SignalNumber signal;
  int s = splimp();

  for (signal=0; signal<32; signal++) {
    if (ep->SignalHandler[signal].flag & SIGF_UNINITIALIZED &&
	ep->SignalHandler[signal].proc == (caddr_t) p_pid(u_procp())) {
      ep->SignalHandler[signal].proc = (caddr_t) u_procp();
      ep->SignalHandler[signal].flag &= ~SIGF_UNINITIALIZED;
    }
    if (ep->SignalHandler[signal].flag & SIGF_EMBPTRARG) {
      ep->SignalHandler[signal].arg = HostPointer((int) ep->SignalHandler[signal].arg);
      ep->SignalHandler[signal].flag &= ~SIGF_EMBPTRARG;
    }
  }
  splx(s);
}

/* Allocate a guest-to-host signal, install a handler for it, and return the signal number */
SignalNumber InstallSignalHandlerInternal (EmbCommAreaPtr, handler, argument, flag)
     EmbCommArea *EmbCommAreaPtr;
     void (*handler)();
     PtrV argument;
     short flag;
{
  register EmbCommArea* ep = EmbCommAreaPtr;
  register SignalNumber signal;
  int s = splimp();

  for (signal=0; signal<32; signal++) {
    if ((ep->live_guest_to_host_signals & (1<<signal)) == 0) {
      ep->SignalHandler[signal].fn = handler;
      ep->SignalHandler[signal].arg = argument;
      ep->SignalHandler[signal].proc = (caddr_t) u_procp();
      ep->SignalHandler[signal].flag = flag;
      ep->SignalHandler[signal].count = 1;
      ep->live_guest_to_host_signals |= 1<<signal;
      splx(s);
      return (signal);
    }
  }
  /* ---Become more graceful about this later.  When do, remember splx(s). */
  panic("Signal table overflow");
  /*NOTREACHED*/
}

void RemoveSignalHandler (EmbCommAreaPtr, signal)
     EmbCommArea *EmbCommAreaPtr;
     SignalNumber signal;
{
  int s = splimp();

  EmbCommAreaPtr->SignalHandler[signal].fn = (void (*)()) 0;
  EmbCommAreaPtr->SignalHandler[signal].arg = (PtrV) signal;
  EmbCommAreaPtr->SignalHandler[signal].proc = (caddr_t) 0;
  EmbCommAreaPtr->SignalHandler[signal].flag = 0;
  EmbCommAreaPtr->SignalHandler[signal].count = 0;
  EmbCommAreaPtr->live_guest_to_host_signals &= ~(1<<signal);
  splx(s);
}


/* Signal handling */

void DispatchPendingSignals (EmbCommAreaPtr, unit)
     EmbCommArea *EmbCommAreaPtr;
     long unit;
{
  SignalMask signals;
  int j, sig, s;

  signals = EmbCommAreaPtr->guest_to_host_signals;

#ifdef IVORY_DEBUG
  DBG(DBG_SIG)
    if (signals)
      printf(":in sig mask 0%o\n", signals);
#endif IVORY_DEBUG

  for (sig=1, j = 0; j < 32; sig = sig << 1, j++)
    if (signals&sig && EmbCommAreaPtr->SignalHandler[j].fn != 0 &&
	!(EmbCommAreaPtr->SignalHandler[j].flag & SIGF_UNINITIALIZED))
      if (EmbCommAreaPtr->SignalHandler[j].flag & SIGF_INTERRUPT) {
	s = splimp();
	EmbCommAreaPtr->guest_to_host_signals &= ~sig;
	splx(s);
	(*(EmbCommAreaPtr->SignalHandler[j].fn))
	  (EmbCommAreaPtr, unit, EmbCommAreaPtr->SignalHandler[j].arg);
      } else
	wakeup(EmbCommAreaPtr->SignalHandler[j].proc);
}

void ProcessCheckPendingSignals (EmbCommAreaPtr, ivory_unit)
     register EmbCommArea *EmbCommAreaPtr;
     int ivory_unit;
{
  register SignalMask signals;
  int sig, j, s;

  s = splimp();  
  signals = EmbCommAreaPtr->guest_to_host_signals;
  for (sig=1, j=0; j < 32; sig = sig << 1, j++)
    if (signals&sig && EmbCommAreaPtr->SignalHandler[j].proc != (caddr_t) u_procp())
      signals &= ~sig;
  EmbCommAreaPtr->guest_to_host_signals &= ~signals;
  splx(s);

  for (j=0; j < 32 && signals!=0; j++) {
    if (signals&1 && EmbCommAreaPtr->SignalHandler[j].fn != 0 &&
	!(EmbCommAreaPtr->SignalHandler[j].flag & SIGF_UNINITIALIZED))
	(*(EmbCommAreaPtr->SignalHandler[j].fn))
	  (EmbCommAreaPtr, ivory_unit, EmbCommAreaPtr->SignalHandler[j].arg);

    signals = signals>>1;
  }
}

/* Signal generation */

void EmbSendSignal (EmbCommAreaPtr, num)
     EmbCommArea *EmbCommAreaPtr;
     SignalNumber num;
{
  register EmbCommArea* ep = EmbCommAreaPtr;
  int s;

  if (num >= 0) {			/* -1 is used as the null signal */
    s = splimp();
    ep->host_to_guest_signals |= 1<<num;
    splx(s);
#ifdef IVORY_DEBUG
    DBG(DBG_SIG)
      if (num != ep->clock_signal)
	printf(":out sig %d pending %d live %d\n",
	       num, ep->host_to_guest_signals, ep->live_host_to_guest_signals);
    if (ep->slave_trigger < (long *) EmbCommAreaPtr) {
      printf("Bogus slave-trigger %x\n", ep->slave_trigger);
      return;
    }
#endif
    /* Cause a slave-trigger interrupt */
    *ep->slave_trigger = 0;
  }
}


void MaybeSignal (EmbCommAreaPtr, i)
     EmbCommArea *EmbCommAreaPtr;
     SignalNumber i;
     /*ARGSUSED*/
{
  int unit;
  extern struct ivory_state ivories[];

  /* Ugly, but rare */
  for (unit = 0; unit < n_ivories; unit++)
    if (ivories[unit].EmbCommArea == EmbCommAreaPtr)
      break;
  
  /* Only signal if SIGF_RESET is set */
  if (i > 0 && (EmbCommAreaPtr->SignalHandler[i].flag & SIGF_RESET))
      (*(EmbCommAreaPtr->SignalHandler[i].fn))
	(EmbCommAreaPtr, unit, EmbCommAreaPtr->SignalHandler[i].arg);
}


/* Signals to help implement select() */

void SelectSignalHandler (EmbCommAreaPtr, unit, procp)
     /*ARGSUSED*/
     EmbCommArea *EmbCommAreaPtr;
     int unit;
     struct proc *procp;
{

#ifdef IVORY_DEBUG
  DBG(DBG_SELECT) printf("ivory%d selsig for pid %d\n", unit, p_pid(procp));
#endif

  /* Defined not to allow collisions.  Let ivoryselect() do all the heavy lifting. */
  selwakeup(proc, 1);
}

InstallSelectSignal (EmbCommAreaPtr, qptr)
     register EmbCommArea *EmbCommAreaPtr;
     EmbPtr qptr;
{
  int i;
  EmbQueue *q;
  int s;

  /* Check if queue is already busy */
  if (qptr < 0 || qptr > EmbCommAreaPtr->comm_memory_size)
    return(EINVAL);
  q = (EmbQueue *) HostPointer(qptr);
  if (q->signal != -1)
    return(EBUSY);

  /* See if there's already a select signal for this process */
  s = splhigh();
  for (i = 0; i < 32; i++)
    if (EmbCommAreaPtr->SignalHandler[i].fn == SelectSignalHandler &&
	EmbCommAreaPtr->SignalHandler[i].arg == (PtrV) u_procp()) {
      EmbCommAreaPtr->SignalHandler[i].count++;
      q->signal = i;
#ifdef IVORY_DEBUG
      DBG(DBG_SELECT) printf("reusing sig %d (#%d) for pid %d q %d\n",
			     i, EmbCommAreaPtr->SignalHandler[i].count++,
			     p_pid(u_procp()), qptr);
#endif
      break;
    }

  /* Allocate a signal, if necessary */
  if (i == 32) {
    q->signal = InstallSignalHandlerInternal(EmbCommAreaPtr, SelectSignalHandler,
					     (PtrV) u_procp(), SIGF_INTERRUPT|SIGF_RESET);
#ifdef IVORY_DEBUG
    DBG(DBG_SELECT) printf("allocated sig %d for pid %d q %d\n",
			   q->signal, p_pid(u_procp()), qptr);
#endif
  }

  splx(s);
  return(0);
}

RemoveSelectSignal (EmbCommAreaPtr, qptr)
     register EmbCommArea *EmbCommAreaPtr;
     EmbPtr qptr;
{
  EmbQueue *q;
  int s;

  /* Check that queue is actually in use */
  if (qptr < 0 || qptr > EmbCommAreaPtr->comm_memory_size)
    return(EINVAL);
  q = (EmbQueue *) HostPointer(qptr);
  if (q->signal == -1 ||
      EmbCommAreaPtr->SignalHandler[q->signal].fn != SelectSignalHandler)
    return(ENXIO);
  if (EmbCommAreaPtr->SignalHandler[q->signal].arg != (PtrV) u_procp())
    return(EPERM);

  /* Look for processes watching this queue.  If it's us, remove queue. */
  s = splhigh();
  if (--EmbCommAreaPtr->SignalHandler[q->signal].count == 0) {
    RemoveSignalHandler(EmbCommAreaPtr, q->signal);
#ifdef IVORY_DEBUG
    DBG(DBG_SELECT) printf("Removed handler for signal %d\n", q->signal);
#endif
  }
#ifdef IVORY_DEBUG
  else DBG(DBG_SELECT) printf("%d users of sig %d remain\n",
			      EmbCommAreaPtr->SignalHandler[q->signal].count,
			      q->signal);
#endif
  q->signal = -1;
  splx(s);

  return(0);
}

#endif				/* NIVORY > 0 */
