/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * STREAMS module to provide extra-emb-host network support
 *
 * $Log:	ivory_stream.c,v $
 * Revision 2.5.1.2  93/02/01  12:56:46  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:52:35  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:42:31  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#include "ivory.h"
#if NIVORY > 0

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/ivory_stream.c,v 2.5.1.2 93/02/01 12:56:46 kaufman Rel $";
#endif

/* Basic STREAMS stuff */
#include "sys/types.h"
#include "sys/param.h"
#include "sys/stream.h"
#include "sys/stropts.h"

/* Enough network stuff for sizeof(struct ether_header) */
#include "sys/socket.h"
#include "net/if.h"
#include "net/if_arp.h"
#include "netinet/in.h"
#include "netinet/if_ether.h"

/* Pieces we need for NIOCBIND */
#include "sys/time.h"
#include "net/nit.h"
#include "net/nit_if.h"

#include "embed.h"
#include "life.h"

/* STREAMS structures */

static int ivory_netopen(), ivory_netrput(), ivory_netrsrv(),
  ivory_netwput(), ivory_netclose();

struct module_info ivory_net_module_info = {
  /* Clever folks at SMI implemented an 8-char module name limit in fmodsw */
  0, "IvoryNet", 0, INFPSZ, 150, 50,
};

static struct qinit rinit = {
  ivory_netrput, ivory_netrsrv, ivory_netopen, ivory_netclose,
  NULL, &ivory_net_module_info, NULL,
};

static struct qinit winit = {
  ivory_netwput, NULL, NULL, ivory_netclose, NULL, &ivory_net_module_info, NULL,
};

struct streamtab ivorynetinfo = { &rinit, &winit, NULL, NULL, NULL };

static ivory_netopen(q, dev, flag, sflag)
     queue_t *q;
     dev_t dev;
     int flag, sflag;			/*ARGSUSED*/
{
  return(0);
}

static ivory_netclose(q)
     queue_t *q;			/*ARGSUSED*/
{
  q->q_ptr = NULL;
  return(0);
}

/* Code common to read-side put and service procedures.
 * Pass a packet on to the guest.  Returns 1 for transient failures. */
static receive_pkt (q, mp)
     queue_t *q;
     mblk_t *mp;
{
  struct ivory_state *msp;
  int pktlen, bcopy();

  msp = (struct ivory_state *) q->q_ptr;
  
  if (msp != NULL) {
    if (msp->config.enable_Ivory && msp->EmbCommArea) {
      if (pullupmsg(mp, -1)) {
	pktlen = mp->b_wptr - mp->b_rptr;
	if (NetReceiver(msp->EmbCommArea, msp->net_channel, pktlen,
			bcopy, (caddr_t) mp->b_rptr, (caddr_t) pktlen)) {
	  /* NetReceiver() couldn't handle packet; delay processing */
	  return(1);
	}
      }
    }
  }

  /* Free buffers */
  freemsg(mp);

  return(0);
}

/* If the message contains a packet (message type M_DATA), pass the packet
 * to the guest.  If the forwarding couldn't be accomplished immediately,
 * queue the message for the service procedure.  (Packets can get out of order
 * here, but that shouldn't matter to modern networking protocols.) */
static ivory_netrput (q, mp)
     queue_t *q;
     mblk_t *mp;
{

  switch(mp->b_datap->db_type) {
  case M_DATA:
    if (receive_pkt(q, mp))
      putq(q, mp);
    break;

  case M_IOCTL:
  case M_IOCACK:
  case M_IOCNAK:
    if (q->q_ptr == (caddr_t) &q->q_ptr)
      /* Only true if in network initializations */
      putq(q, mp);
    else
      putnext(q, mp);
    break;
    
  case M_FLUSH:
    /* Canonical flush handling */
    if (*mp->b_rptr & FLUSHR)
      flushq(q, FLUSHDATA);
    /* Fall through */

  default:
    putnext(q, mp);
    break;
  }
}

/* Pass packets on to the guest */
static ivory_netrsrv (q)
     queue_t *q;
{
  mblk_t *mp;

  while ((mp = getq(q)) != NULL) {
  
    switch(mp->b_datap->db_type) {
    case M_DATA:
      if (receive_pkt(q, mp)) {
	putbq(q, mp);
	return;
      }
      break;

    case M_IOCACK:
    case M_IOCNAK:
      if (q->q_ptr == (caddr_t) &q->q_ptr) {
	/* Hold messages for the initialization routines */
	putbq(q, mp);
	return;
      }
      /* By default, fall through */
      
    default:
      /* Drop anything we don't handle */
      freemsg(mp);
      break;
    }
  }
}

/* Pass packets (message type M_PROTO followed by M_DATA block) to NIT
 * for transmission on the Ethernet. */
static ivory_netwput (q, mp)
     queue_t *q;
     mblk_t *mp;
{
  struct iocblk *ic;

  switch(mp->b_datap->db_type) {
  case M_PROTO:
    putnext(q, mp);
    break;

  case M_IOCTL:
    ic = (struct iocblk *) mp->b_rptr;
    if (ic->ioc_cmd == NIOCBIND && ic->ioc_count == sizeof(struct ifreq)) {
      struct ifreq *ifr = (struct ifreq *) mp->b_cont->b_rptr;

      q->q_ptr = (caddr_t) ifunit(ifr->ifr_name);
    }
    /* Fall through */
    
  case M_IOCACK:
  case M_IOCNAK:
    putnext(q, mp);
    break;

  case M_FLUSH:
    /* Canonical flush handling */
    if (*mp->b_rptr & FLUSHR)
      flushq(q, FLUSHDATA);
    /* Fall through */

  default:
    putnext(q, mp);
    break;
  }
}

#endif				/* NIVORY > 0 */
