/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * Kernel-grubbing
 *
 * $Log:	kernel.c,v $
 * Revision 2.5.1.3  93/02/01  12:56:50  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.2  93/02/01  12:01:21  kaufman
 * Update life support version number correspondences
 * for Genera 8.2 NXP and Genera 8.3.
 * 
 * Revision 2.5.1.1  93/02/01  11:52:40  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:42:42  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/kernel.c,v 2.5.1.3 93/02/01 12:56:50 kaufman Rel $";
#endif
static char *release_state = "$State: Rel $"+8;

#include <sys/types.h>
#include <sys/buf.h>
#include <sys/file.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sundev/mbvar.h>
#include <ctype.h>
#include <nlist.h>
#include <stdio.h>
#include <strings.h>

#if defined(TEST)
#define INCLUDE_FILE_HEADERS
#endif
#include "embed.h"

static struct nlist nl[] = {
#define NL_MBDINIT 0
  {"_mbdinit"},
#define NL_IVORYDRIVER 1
  {"_ivorydriver"},
#define NL_HIVORYDRIVER 2
  {"_hivorydriver"},
#define NL_KIVORYDRIVER 3
  {"_kivorydriver"},
#define NL_PIVORYDRIVER 4
  {"_pivorydriver"},
#define NL_IVORY_VERSION 5
  {"_embed_version"},
#define NL_OS_VERSION 6
  {"_version"},
  NULL,
};
static int done_nlist = 0;

#define KMEM "/dev/kmem"
#define MBDSIZE 20
#define VMUNIX "/vmunix"

/* Find the ivory's interrupt level & vector in the kernel config tables */
lookup_interrupt_vector (device_fd, levelp, vectorp)
     int device_fd, *levelp, *vectorp;
{
  struct mb_device mbdinit[MBDSIZE];
  struct mb_driver *ivorydriver;
  struct vec intr;
  int kmem_fd, i, unit;
  struct stat statb;

  /* Figure out device unit number */
  if (fstat(device_fd, &statb) == -1) {
    perror("stat");
    return(-1);
  }
  unit = minor(statb.st_rdev);

  /* Dig useful values out of kernel name list */
  if (!done_nlist) {
    done_nlist = 1;
    if (nlist("/vmunix", nl) == -1) {
      perror("nlist");
      return(-1);
    }
  }
  ivorydriver =
    (struct mb_driver *) (nl[NL_IVORYDRIVER].n_value ? nl[NL_IVORYDRIVER].n_value :
 			  nl[NL_HIVORYDRIVER].n_value ? nl[NL_HIVORYDRIVER].n_value :
			  nl[NL_KIVORYDRIVER].n_value ? nl[NL_KIVORYDRIVER].n_value :
			  nl[NL_PIVORYDRIVER].n_value ? nl[NL_PIVORYDRIVER].n_value : 0);


  /* Seek to the beginning of the configuration tables */
  if ((kmem_fd = open("/dev/kmem", O_RDONLY)) == -1) {
    perror("/dev/kmem");
    return(-1);
  }
  if (lseek(kmem_fd, (off_t) nl[NL_MBDINIT].n_value, L_SET) == (off_t) -1) {
    perror("lseek");
    return(-1);
  }

  for (;;) {
    if ((i = read(kmem_fd, (char *) mbdinit, sizeof(mbdinit))) == -1) {
      perror("read");
      return(-1);
    }

    for (i = 0; i < MBDSIZE; i++) {
      if (mbdinit[i].md_driver == NULL)
	return(-1);
      
      if (mbdinit[i].md_driver == ivorydriver &&
	  mbdinit[i].md_unit == unit) {

	/* Seek to and read interrupt handler */
	if (lseek(kmem_fd, (off_t) mbdinit[i].md_intr, L_SET) == (off_t) -1) {
	  perror("lseek");
	  return(-1);
	}
	if (read(kmem_fd, (char *) &intr, sizeof(intr)) == -1) {
	  perror("read");
	  return(-1);
	}

	*levelp = mbdinit[i].md_intpri;
	*vectorp = intr.v_vec;
	return(0);
      }
    }
  }
}



/* Look up SunOS version number */

get_os_version (maj, min, patch)
     short *maj, *min, *patch;
{
  long majtemp = -1, mintemp = -1, patchtemp = -1;
  int fd;
  char buf[BUFSIZ], *cp;

  /* Dig useful values out of kernel name list */
  if (!done_nlist) {
    done_nlist = 1;
    if (nlist("/vmunix", nl) == -1) {
      perror("nlist");
      return(-1);
    }
  }

  if ((fd = open(KMEM, O_RDONLY)) < 0) {
    perror(KMEM);
    return(-1);
  }

  if (nl[NL_OS_VERSION].n_type == 0) {
    fprintf(stderr, "%s: Symbol not found\n", nl[NL_OS_VERSION].n_name);
    return(-1);
  } else if (lseek(fd, nl[NL_OS_VERSION].n_value, L_SET) == -1) {
    perror("lseek");
    return(-1);
  }

  if (read(fd, buf, BUFSIZ-1) < 0) {
    perror("read");
    return(-1);
  }

  close(fd);

  for (cp = buf; *cp != '\0'; cp++)
    if (isdigit(*cp)) {
      sscanf(cp, "%d.%d.%d", &majtemp, &mintemp, &patchtemp);
      *maj = (short) majtemp;
      *min = (short) mintemp;
      *patch = (short) patchtemp;
      return(1);
    }

  return(-1);
}



char *lookup_kernel_version ()
{
  int fd, off;
  static char buf[BUFSIZ];
  char *cp;

  if (!done_nlist) {
    done_nlist = 1;
    if (nlist("/vmunix", nl) == -1) {
      perror("nlist");
      return(NULL);
    }
  }
  
  if ((fd = open("/dev/kmem", O_RDONLY)) == -1) {
    perror("/dev/kmem");
    return(NULL);
  }

  if (lseek(fd, (off_t) nl[NL_IVORY_VERSION].n_value, L_SET) == (off_t) -1) {
    perror("lseek");
    close(fd);
    return(NULL);
  }

  if (read(fd, (char *) &off, sizeof(long)) == -1) {
    perror("read");
    close(fd);
    return(NULL);
  }

  if (lseek(fd, (off_t) off, L_SET) == (off_t) -1) {
    perror("lseek");
    close(fd);
    return(NULL);
  }

  if (read(fd, buf, BUFSIZ) == -1) {
    perror("read");
    close(fd);
    return(NULL);
  }
  close(fd);

  if ((cp = index(buf, ':')) == NULL) {
    fprintf(stderr, "Version identifier malconstructed: no colon in \"%s\"\n", buf);
    return(NULL);
  }

  return(cp+2);
}

static char *genera_life_support_version[] = {
  "1.3.1", "7.4 Ivory ECO#3",
  "2.1.1", "8.0",
  "2.2.1", "8.0 ECO#1",
  "2.3.1", "8.1 Interim",
  "2.4.1", "8.1",
  "2.4.3", "8.1 ECO#1",
  "2.4.4", "8.2 NXP",
  "2.5.1", "8.3",
  NULL, NULL,
};

static char *life_support_genera_version (life_support_version)
     char *life_support_version;
{
  int i;

  for (i = 0; genera_life_support_version[i] != NULL; i+= 2)
    if (!strcmp(genera_life_support_version[i], life_support_version))
      return(genera_life_support_version[i+1]);

  return(NULL);
}

static check_kernel_version_internal (applic_version, kernel_version)
     char *applic_version;
     char *kernel_version;
{
  int kernel_trunk_major, kernel_trunk_minor;
  int kernel_branch_major = -1, kernel_branch_minor = -1;
  int applic_trunk_major, applic_trunk_minor;
  int applic_branch_major = -1, applic_branch_minor = -1;
  char kbuf[128], abuf[128];
  char *kgenera_version, *agenera_version;

  if (kernel_version == NULL) {
    fprintf(stderr, "WARNING: Can't verify kernel life-support version numbers\n");
    return(-1);
  }

  sscanf(kernel_version, "%d.%d.%d.%d",
	 &kernel_trunk_major, &kernel_trunk_minor,
	 &kernel_branch_major, &kernel_branch_minor);
  sscanf(applic_version, "%d.%d.%d.%d",
	 &applic_trunk_major, &applic_trunk_minor,
	 &applic_branch_major, &applic_branch_minor);

  if (kernel_trunk_major != applic_trunk_major ||
      kernel_trunk_minor != applic_trunk_minor ||
      kernel_branch_major != applic_branch_major) {
    if (kernel_branch_major != -1)
      sprintf(kbuf, "%d.%d.%d", kernel_trunk_major, kernel_trunk_minor, kernel_branch_major);
    else
      sprintf(kbuf, "%d.%d", kernel_trunk_major, kernel_trunk_minor);
    if (applic_branch_major != -1)
      sprintf(abuf, "%d.%d.%d", applic_trunk_major, applic_trunk_minor, applic_branch_major);
    else
      sprintf(abuf, "%d.%d", applic_trunk_major, applic_trunk_minor);

    /* Format warning message */
    agenera_version = life_support_genera_version(abuf);
    kgenera_version = life_support_genera_version(kbuf);
    fprintf(stderr, "WARNING: Program life-support version %s", abuf);
    if (agenera_version != NULL) fprintf(stderr, " (for Genera %s)", agenera_version);
    fprintf(stderr, "\n    does not match /vmunix life-support version %s", kbuf);
    if (kgenera_version != NULL) fprintf(stderr, " (for Genera %s)", kgenera_version);
    fprintf(stderr, "!\n");
    return(1);
  }

  return(0);
}

check_kernel_version ()
{
  extern char *embed_version;

  /* Don't require consistent versions for experimental software */
  if (!strncmp(release_state, "Rel", 3))
    return(check_kernel_version_internal(index(embed_version, ':') + 2,
					 lookup_kernel_version()));
  else
    return(0);
}

set_version (vers, string)
     struct version *vers;
     char *string;
{
  int trunk_major, trunk_minor, branch_major = -1, branch_minor = -1;

  sscanf(string, "%d.%d.%d.%d", &trunk_major, &trunk_minor, &branch_major, &branch_minor);
  vers->trunk_major = trunk_major;
  vers->trunk_minor = trunk_minor;
  vers->branch_major = branch_major;
  vers->branch_minor = branch_minor;
}



#ifdef TEST
main (argc, argv)
     int argc;
     char *argv[];
{
  int fd, level, vector;
  short maj, min, patch;
  char *cp;
  char *ivory = "/dev/ivory0";
  char buf0[BUFSIZ], buf1[BUFSIZ];

  switch (argc) {
/*  case 3:
    vers = argv[2];*/

  case 2:
    ivory = argv[1];
    break;

  default:
    fprintf(stderr, "usage: %s ivory\n", argv[0]);
    exit(1);
  }

  if ((fd = open(ivory, O_RDONLY)) == -1) {
    perror(ivory);
    exit(1);
  }

  if (get_os_version(&maj, &min, &patch) >= 0) {
    printf("SunOS version %d.%d", maj, min);
    if (patch > 0)
      printf(".%d", patch);
    putchar('\n');
  }

  if (lookup_interrupt_vector(fd, &level, &vector) == -1)
    printf("Couldn't find interrupt handler for %s\n", ivory);
  else
    printf("%s interrupt priority %d vector 0x%x\n", ivory, level, vector);

  if (cp = lookup_kernel_version())
    printf("Kernel life-support revision %s\n", cp);

  for (;;) {
    printf("\nKernel version: ");
    if (gets(buf0) == NULL)
      break;
    printf("Application version: ");
    if (gets(buf1) == NULL)
      break;
    check_kernel_version_internal(buf1, buf0);
  }
  printf("\n");
}
#endif
