/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * $Log:	life.c,v $
 * Revision 2.5.1.2  93/02/01  12:56:52  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:52:42  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:42:47  kaufman
 * Baseline for Genera 8.3
 * 
 * 
 */

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/life.c,v 2.5.1.2 93/02/01 12:56:52 kaufman Rel $";
#endif
char *life_version = "@(#)$Revision: 2.5.1.2 $"+15;

#include <sys/types.h>
#include <sys/file.h>
#include <sys/filio.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <errno.h>
#include <setjmp.h>
#include <signal.h>
#include <stdio.h>
#include <strings.h>
#include <syslog.h>

#define INCLUDE_FILE_HEADERS
#include "embed.h"
#include "life.h"
#include "config.h"
#undef INCLUDE_FILE_HEADERS

#define CONFIG_DIR    "/etc"

void InitializeCommunicationArea();
static void restart_life_support();
static void stop_life_support();
static Config config;

static jmp_buf life_support;

/* Useful for debugging */
#define IOCTL(cmd,data,str) { extern int verbose;                 \
                              if (verbose) printf(">%s...", str); \
			      fflush(stdout);                     \
			      if (ioctl(fd, cmd, data) < 0) {     \
                                log_error(LOG_ERR, str);          \
				exit(1);                          \
                              } else                              \
                                if (verbose) puts("  OK");        \
			     }

int verbose = 0;
static char *argv0;
static log_facility = LOG_LOCAL0;

main (argc, argv, envp)
     int argc;
     char *argv[];
     char **envp;		/*ARGSUSED*/
{
  char *guest, *getenv();
  char *conf = NULL;
  char *cp, buf0[128], buf1[128];
  caddr_t ptr;
  int i, fd;
  static int first_time = 1;
  static void sigchld_handler(), register_pid();

  argv0 = argv[0];
  for (i = 1; i < argc && argv[i][0] == '-'; i++)
    switch (argv[i][1]) {
    case 'f':
      conf = argv[++i];
      break;

    case 'F':
      log_facility = atoi(argv[++i]);
      break;

    case 'v':
      verbose = 1;
      break;

    default:
      fprintf(stderr, "usage: ivory-life [-f conf] [-v] [device]\n");
      exit(0);
    }

  /* Figure out path to device */
  if (i < argc) {
    if (index(argv[i], '/') != NULL)
      guest = argv[i];
    else {
      sprintf(buf0, "/dev/%s", argv[i]);
      guest = buf0;
    }
  } else if ((guest = getenv("IVORY")) == NULL)
    guest = DEFAULT_GUEST();

  /* Figure out path to config file */
  if ((cp = rindex(guest, '/')) == NULL) {
    fprintf(stderr, "Can't find configuration file for %s\n", guest);
    exit(1);
  }
  if (conf == NULL) {
    (void) sprintf(buf1, "%s/%s.conf", CONFIG_DIR, cp+1);
    conf = buf1;
  }

  /* Set up syslog() handling */
  if (log_facility > 0)
    openlog(cp+1, LOG_PID, log_facility);

  /* Set up to restart on HUP signals */
  if (setpgrp(0, getpid()) == -1) {
    log_error(LOG_ERR, "setpgrp");
    exit(1);
  }
  signal(SIGCHLD, sigchld_handler);
  setjmp(life_support);
  signal(SIGHUP, restart_life_support);

  /* Open device */
  if ((fd = open(guest, O_RDWR|O_EXCL)) < 0) {
    log_error(LOG_ERR, guest);
    exit(1);
  }
  if (ioctl(fd, FIOCLEX, 0) < 0) {
    log_error(LOG_ERR, "FIOCLEX");
    exit(1);
  }
  if (first_time) {
    first_time = 0;
    check_kernel_version();
  }

  IOCTL(IVORYGCONF, &config.kern, "IVORYGCONF");

  /* Prepare communication area */
  if ((ptr = mmap((caddr_t) 0, config.kern.mmap_size, PROT_READ|PROT_WRITE,
		  MAP_SHARED, fd, (off_t) 0)) == (caddr_t) -1) {
    log_error(LOG_ERR, "mmap");
    exit(1);
  }
  ptr += config.kern.comm_area_base;
  bzero(ptr, config.kern.comm_area_size);

  /* Configure and initialize life-support */
  init_config(&config, fd, config.kern.comm_area_size);
  config.device = guest;
  config.config_filename = conf;
  read_config(conf, &config);
  if (open_device_prom(&config) == -1)
    exit(1);
  InitializeCommunicationArea(ptr, fd, &config);
  IOCTL(IVORYSCONF, &config.kern, "IVORYSCONF");

  /* Try to register our PID, but not too hard */
  register_pid(fd);

  if ((i = fork()) == -1) {
    log_error(LOG_ERR, "fork");
    exit(-1);
  } else if (i == 0) {
    /* Child runs ivory-misc */

    set_process_name("ivory-misc");
    HostFileLife(fd, ptr, &config);

    /*NOTREACHED*/
  } else {
    /* Parent runs master kernel support */

    /* Start Ivory running */
    if (ioctl(fd, IVORYINIT, 0) == -1) {
      log_error(LOG_ERR, "IVORYINIT");
      exit(1);
    }
    IvoryLife(fd, PROC_MASTER);
    
    /*NOTREACHED*/
  }
}

/* Hack for ps */
static set_process_name (name)
     char *name;
{
  char *cp;

  if ((cp = rindex(argv0, '/')) == NULL)
    cp = argv0;
  else
    cp++;
  if (strlen(cp) >= strlen(name))
    strcpy(cp, name);
}

IvoryLife (fd, procno)
     int fd, procno;
{
  int i;

  if (procno != PROC_MASTER)
    set_process_name(procno == PROC_DISK ? "ivory-disk" :
		     procno == PROC_NETWORK ? "ivory-net " : NULL);
		   
  /* Run life support */
  while ((i = ioctl(fd, IVORYLIFE, &procno)) == -1 && errno == EINTR);
  if (i == -1) {
    log_error(LOG_ERR, "IVORYLIFE");
    exit(1);
  }

  exit(0);
}

static void sigchld_handler ()
{
  while (wait3(NULL, WNOHANG, NULL) > 0);
  signal(SIGCHLD, sigchld_handler);
  return;
}

static void stop_life_support ()
{
  int nfds, i;

  if (getpid() != getpgrp(0))
    /* Not the lead process; just exit */
    exit(0);
  
  /* Slam any of our processes that didn't see the signal */
  signal(SIGINT, SIG_IGN);
  if (killpg(getpid(), SIGINT) == -1)
    log_error(LOG_ERR, "killpg");
  else
    /* Reset SIGINT behavior */
    signal(SIGINT, SIG_DFL);

  /* Close all open files */
  nfds = getdtablesize();
  for (i=0; i < nfds; i++)
    if (i != fileno(stdin) &&
	i != fileno(stdout) &&
	i != fileno(stderr))
      close(i);
}  

static void restart_life_support ()
{
  extern char **environ;
  int mask;

  /* Stop life support */
  stop_life_support();

  /* Reset signal mask */
  mask = sigblock(sigmask(SIGKILL)); /* SIGKILL can't be masked... */
  sigsetmask(mask & ~sigmask(SIGHUP));

  /* Restart life-support */
  longjmp(life_support, 1);
  log_error(LOG_ERR, "longjmp");
  sleep(5);
  abort();
}

static void register_pid (device_fd)
     int device_fd;
{
  char buf[128];
  FILE *fp;
  struct stat st;

  if (fstat(device_fd, &st) == -1) {
    log_error(LOG_ERR, "fstat");
    return;
  }
  sprintf(buf, "/etc/ivory%d.pid", minor(st.st_rdev));

  if ((fp = fopen(buf, "w")) != NULL) {
    /* Copy permissions and ownership */
    if (fchown(fileno(fp), st.st_uid, st.st_gid) == -1)
      log_error(LOG_ERR, "fchown");
    if (fchmod(fileno(fp), st.st_mode) == -1)
      log_error(LOG_ERR, "fchmod");
    /* Write PID */
    fprintf(fp, "%d\n", getpid());
    fclose(fp);
  }
}

log_error (level, msg)
     int level;
     char *msg;
{
  int tmp = errno;

  if (log_facility > 0) {
    syslog(level, "%s: %m", msg);
    errno = tmp;
  }
  perror(msg);
}

log_message (level, msg)
     int level;
     char *msg;
{
  if (log_facility > 0) {
    syslog(level, "%s", msg);
  }
  fprintf(stderr, "%s\n", msg);
}
