/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * This file defines the various common data structures, including the
 * communication area, the channels, and the configuration file.
 * For convenience, the private parts of these structures are in this
 * file too, but they shouldn't be accessed outside of Ivory life support
 * and they especially should not be depended upon by the guest.
 *
 * $Log:	life.h,v $
 * Revision 2.5.1.2  93/02/01  12:56:54  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:52:44  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:42:52  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifndef _LIFE_
#define _LIFE_

#if !defined(lint) && defined(INCLUDE_FILE_HEADERS)
static char *life_rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/life.h,v 2.5.1.2 93/02/01 12:56:54 kaufman Rel $";
#endif

/*** Merlin VME slave buffer ***/

#define MERLIN_VME_SLAVE_BUFFER_SIZE (32 * 1024 * sizeof(long)) /* Slave buffer is 32Kx32bits */
#define MERLIN_VME_SLAVE_TRIGGER_SIZE (64*sizeof(long))
#define MERLIN_VME_COMM_AREA_SIZE ((16 * 1024 * sizeof(long)) - MERLIN_VME_SLAVE_TRIGGER_SIZE)
#define MERLIN_VME_COMM_AREA_BASE (MERLIN_VME_SLAVE_BUFFER_SIZE - MERLIN_VME_COMM_AREA_SIZE - MERLIN_VME_SLAVE_TRIGGER_SIZE)
#define MERLIN_VME_MMAP_SIZE (MERLIN_VME_COMM_AREA_SIZE + MERLIN_VME_SLAVE_TRIGGER_SIZE)

#define MERLIN_II_VME_SLAVE_BUFFER_SIZE (256 * 1024 * sizeof(long)) /* Slave buffer is 256Kx32bits */
#define MERLIN_II_VME_SLAVE_TRIGGER_SIZE (64*sizeof(long))
#define MERLIN_II_VME_COMM_AREA_SIZE ((64 * 1024 * sizeof(long)) - MERLIN_II_VME_SLAVE_TRIGGER_SIZE)
#define MERLIN_II_VME_COMM_AREA_BASE (MERLIN_II_VME_SLAVE_BUFFER_SIZE - MERLIN_II_VME_COMM_AREA_SIZE - MERLIN_II_VME_SLAVE_TRIGGER_SIZE)
#define MERLIN_II_VME_MMAP_SIZE (MERLIN_II_VME_COMM_AREA_SIZE + MERLIN_VME_SLAVE_TRIGGER_SIZE)

#define MERLIN_VME_MAILBOX_SIZE  sizeof(long) /* Mailbox is a long word */
#define MERLIN_VME_MAILBOX_RESET 1
#define MERLIN_VME_MAILBOX_INIT  2

#ifndef KERNEL
/* The kernel gets the addresses from the config file by way of auto config */
#define MERLIN_VME_SLAVE_BUFFER_BASE 0xfadc0000
#define MERLIN_VME_MAILBOX_ADDRESS   0xfade0000
#define MERLIN_II_VME_SLAVE_BUFFER_BASE 0xfac00000
#define MERLIN_II_VME_MAILBOX_ADDRESS   0xfad00000
#endif /*KERNEL*/


/*** Configuration constants ***/

#define PageSize 1280		/* Bytes per logical Ivory page */

#ifndef FALSE
#define FALSE 0
#define TRUE 1
#endif


/*** Solstice configuration structure ***/

typedef struct {
  long enable_Ivory;		/* 0 to prevent Ivory life-support from running */
  int device_prom_fd;		/* fd from which device prom is loaded */
  EmbPtr slave_trigger_offset;	/* Offset in comm area of slave-trigger location */
  EmbPtr host_ether_addr_offset; /* Offset in comm area of host ethernet address table */
  int system_type;		/* what kind of board */
  long slave_buffer_size;	/* number of octets in slave buffer */
  long comm_area_base;		/* offset to beginning of the embedding comm area */
  long comm_area_size;		/* number of octets in the comm area */
  long mmap_size;		/* number of slave buffer octets mapped by ivory device */
} IvoryConfig;

struct ivory_state {
  EmbCommArea *EmbCommArea;
  IvoryConfig config;
  struct proc *master_proc, *disk_proc, *network_proc;
  caddr_t control;
  EmbNetChannel *net_channel;	/* ---- Kludge, sigh */
  EmbPtr net_channel_ptr;
};

#ifdef KERNEL
void ResetCommArea(), IvoryClock();
void FixupSignalHandlers(), DispatchPendingSignals(), ProcessCheckPendingSignals();
void KillSubProcesses(), ForkSignalHandler(), RemoveSignalHandler();
void InitializeDiskLifeII(), InitializeNetworkLifeII();
SignalNumber InstallSignalProcess();
int NetReceiver(), IvoryLife();
caddr_t map_mailbox();

extern int n_ivories, n_emb_if;

#ifdef _sys_user_h
struct proc *u_procp();
char p_cursig();
short p_pid();
struct file *u_ofile();
void set_u_error();
#endif
#endif


/*** How to choose the default Ivory ***/

#ifndef KERNEL
#define DEFAULT_UX400S "/dev/ivory0"
#define DEFAULT_UX1200S "/dev/ivory16"

#define DEFAULT_GUEST() access(DEFAULT_UX400S, F_OK) ? DEFAULT_UX1200S : DEFAULT_UX400S
#endif

/*** Ivory ioctl's ***/

#include "sys/ioccom.h"

#define IVORYGCONF	_IOR(I, 1, IvoryConfig)
#define IVORYINIT	_IO(I, 2)
#define IVORYRESET	_IO(I, 3)
#define IVORYSCONF	_IOW(I, 4, IvoryConfig)
#define IVORYLIFE	_IOW(I, 5, int)
#define IVORYSELCHANNEL	_IOW(I, 9, EmbPtr)
#define IVORYUNSELCHANNEL _IOW(I, 10, EmbPtr)
#define IVORYSIGNAL     _IOW(I, 11, int)

/* Arguments to ioctl IVORYLIFE */
#define PROC_MASTER  0
#define PROC_DISK    1
#define PROC_NETWORK 2


/*** Network stuff ***/

#if defined(KERNEL) && defined(_net_if_h)

struct softc {
  struct arpcom is_ac;
#define is_if is_ac.ac_if
#define is_addr is_ac.ac_enaddr
  EmbCommArea *EmbCommAreaPtr;
  EmbNetChannel *EmbNetChannel;
  int count;
};

#endif


/*** Debugging stuff ***/

#ifdef IVORY_DEBUG
#define DBG(level) if ((EmbCommAreaPtr->debug_level & (level)) == (level))

#define DBG_NET_XMIT      1
#define DBG_NET_DROP      2
#define DBG_NET_RECV      4
#define DBG_INIT          8
#define DBG_LIFE         16
#define DBG_BUF          32
#define DBG_IF           64
#define DBG_SIG         128
#define DBG_DBG1        256
#define DBG_DBG2        512
#define DBG_DBG3       1024
#define DBG_DBG4       2048
#define DBG_NET_QUEUE  4096
#define DBG_DISK       8192
#define DBG_SELECT    16384
#define DBG_PROCESS   32768

/* Can't see register values in kernel core dumps */
#define register
#endif

#endif /*_LIFE_*/
