/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * $Log:	network.c,v $
 * Revision 2.5.1.2  93/02/01  12:56:57  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:52:49  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:43:02  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/network.c,v 2.5.1.2 93/02/01 12:56:57 kaufman Rel $";
#endif

#include <sys/types.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <net/if.h>
#include <net/nit_if.h>
#include <net/nit_pf.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <net/packetfilt.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <strings.h>
#include <stropts.h>
#include <stdio.h>
#include <syslog.h>

#include "config.h"

#define NIT "/dev/nit"		/* Provisional as of SunOS 4.0 */
#define NIT_PF "pf"		/* Provisional as of SunOS 4.0 */
#define NIT_BUF "buf"		/* Provisional as of SunOS 4.0 */
#define NETWORK_MODULE "IvoryNet"

#define NINTERFACES 32

static struct fd_interface {
  int fd;
  char name[IFNAMSIZ];
} interfaces[NINTERFACES];
static int interface_n = 0;

char *get_if_name (fd)
     int fd;
{
  int i;

  for (i = 0; i < NINTERFACES; i++)
    if (fd == interfaces[i].fd)
      return(interfaces[i].name);

  return(NULL);
}

nit_init (interface, filter)
     char *interface;
     enum filter_mode filter;
{
  int fd, i;
  struct ifreq ifr;
  struct strioctl si;

  /* Open network tap */
  if ((fd = open(NIT, O_RDWR)) == -1) {
    log_error(LOG_ERR, NIT);
    return(-1);
  }

  /* Set up fd -> interface table */
  interfaces[interface_n].fd = fd;
  strncpy(interfaces[interface_n].name, interface, IFNAMSIZ);

  /* Push on the packet-filterer */
  if (ioctl(fd, I_PUSH, NIT_PF) == -1) {
    log_error(LOG_ERR, NIT_PF);
    return(-1);
  }
  setup_filter(fd, filter);

  /* Push on the embedded network module
   * This must happen before we do the NIOCBIND, so that 
   * the kernel can watch the interface name float by. */
  if (ioctl(fd, I_PUSH, NETWORK_MODULE) == -1) {
    log_error(LOG_ERR, NETWORK_MODULE);
    return(-1);
  }

  /* Bind the NIT device to an appropriate network interface */
  (void) strncpy(ifr.ifr_name, interface, sizeof(ifr.ifr_name));
  ifr.ifr_name[sizeof(ifr.ifr_name) - 1] = '\0';
  si.ic_cmd = NIOCBIND;
  si.ic_timout = 0;
  si.ic_len = sizeof(ifr);
  si.ic_dp = (char *) &ifr; 
  if (ioctl(fd, I_STR, (char *) &si) == -1) {
    log_error(LOG_ERR, "NIOCBIND");
    return(-1);
  }

  /* Receive discrete messages */
  if (ioctl(fd, I_SRDOPT, (char *) RMSGD) == -1) {
    log_error(LOG_ERR, "ioctl RMSGD");
    return(-1);
  }

  /* Set NIT flags s.t. nothing special gets done */
  i = 0;
  if (ioctl(fd, NIOCSFLAGS, (char *) &i) == -1) {
    log_error(LOG_ERR, "NIOCSFLAGS");
    return(-1);
  }

  /* Set snapshot length (0 means infinty) */
  i = 0;
  if (ioctl(fd, NIOCSSNAP, (char *) &i) == -1) {
    log_error(LOG_ERR, "NIOCSSNAP");
    return(-1);
  }

  return(fd);
}

/* Configure packet-filterer.  Allow through all packets not addressed
 * to the host's primary IP address.  Some day, this should be filter
 * out all packets to any of the host's protocol addresses.  Please
 * don't ask about DECNet.
 */
setup_filter (fd, filter)
     int fd;
     enum filter_mode filter;
{
  struct packetfilt pf;
  register u_short *fp;
  struct ether_header eh;
  struct ip ip;
  int etype_offset, ipaddr_offset;
  u_long ipaddr;
  char buf[BUFSIZ];
  struct hostent *hp;
  extern int verbose;

  etype_offset = ((caddr_t)&eh.ether_type - (caddr_t)&eh) / sizeof(u_short);
  fp = pf.Pf_Filter;

  switch (filter) {
  case host_ip:
    /* Look up my IP address */
    if (gethostname(buf, BUFSIZ)) {
      log_error(LOG_ERR, "gethostname");
      exit(1);
    }
    if ((hp = gethostbyname(buf)) == NULL) {
      log_error(LOG_ERR, buf);
      exit(1);
    }
    bcopy(hp->h_addr, (char *) &ipaddr, sizeof(u_long));  
    ipaddr = htonl(ipaddr);

    ipaddr_offset = (((caddr_t)&ip.ip_dst - (caddr_t)&ip) +
		     sizeof(struct ether_header)) / sizeof(u_short);

    *fp++ = ENF_PUSHWORD + etype_offset;
    *fp++ = ENF_PUSHLIT;
    *fp++ = htons(ETHERTYPE_IP);
    *fp++ = ENF_NEQ;
    *fp++ = ENF_PUSHWORD + ipaddr_offset;
    *fp++ = ENF_PUSHLIT;
    *fp++ = (u_short) (ipaddr >> 16) & 0177777;
    *fp++ = ENF_NEQ;
    *fp++ = ENF_OR;
    *fp++ = ENF_PUSHWORD + ipaddr_offset + 1;
    *fp++ = ENF_PUSHLIT;
    *fp++ = (u_short) ipaddr;
    *fp++ = ENF_NEQ;
    *fp++ = ENF_OR;
    break;

  case all_ip:
    *fp++ = ENF_PUSHWORD + etype_offset;
    *fp++ = ENF_PUSHLIT;
    *fp++ = htons(ETHERTYPE_IP);
    *fp++ = ENF_NEQ;
    break;

  case none:
    break;

  default:
    log_message(LOG_ERR, "ivory-life internal error: Unknown packet-filtering mode!");
    exit(1);
  }
  pf.Pf_FilterLen = fp - &pf.Pf_Filter[0];

  if (ioctl(fd, NIOCSETF, (caddr_t) &pf) == -1) {
    log_error(LOG_ERR, "NIOCSETF");
    log_message(LOG_ERR,
		"Embedded network bandwidth will be wasted on host-destined epackets");
  } else if (verbose)
    switch (filter) {
    case host_ip:
      printf(">filtering packets based on host IP address %s\n",
	     inet_ntoa(*(struct in_addr *) &ipaddr));
      break;

    case all_ip:
      printf(">filtering packets based on IP protocol\n");
      break;

    case none:
      printf(">packet filtering disabled\n");
      break;
    }
}
