/* -*-Mode: C; Package: C-user -*- */
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1990 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * UX FEP partition file creation
 *
 * $Log:	partition.c,v $
 * Revision 2.5.1.2  93/02/01  12:56:59  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:52:51  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:43:07  kaufman
 * Baseline for Genera 8.3
 * 
 * 
 */

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/partition.c,v 2.5.1.2 93/02/01 12:56:59 kaufman Rel $";
#endif

#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/vfs.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include "install.h"

  /* Nice size for a FEP partition file, in kbytes */
#define NICE_FEPFS 150000

int check_chaos_address (buf, arg0, arg1)
     char *buf;
     arg *arg0, *arg1;		/*ARGSUSED*/
{
  long address;
  char *end;

  address = strtol(buf, &end, 8);
  if (!*end || isspace(*end))
    if ((address > 0) && (address < 65536))
      return(0);

  Fprintf(stderr, ">> \"%s\" is not a valid Chaosnet address.\n", buf);
  Fprintf(stderr, "   A Chaosnet address is an octal integer that is the concatenation\n");
  Fprintf(stderr, "   of an 8-bit subnet number and an 8-bit host number.\n");
  return(-1);
}

int check_internet_address (buf, arg0, arg1)
     char *buf;
     arg *arg0, *arg1;		/*ARGSUSED*/
{
  int a, b, c, d;

  a = b = c = d = -1;
  (void) sscanf(buf, "%d.%d.%d.%d", &a, &b, &c, &d);
  if ((a >= 0) && (b >= 0) && (c >= 0) && (d >= 0) &&
      (a < 256) && (b < 256) && (c < 256) && (d < 256))
    return(0);

  Fprintf(stderr, ">> \"%s\" is not a valid Internet address.\n", buf);
  Fprintf(stderr, "   An Internet address is a sequence of four decimal integers, each\n");
  Fprintf(stderr, "   less than 256, separated by periods.\n");
  return(-1);
}

void determine_partition_file_parameters (params, part, ux_n)
     struct params *params;
     struct partition_params *part;
     int ux_n;
{
  char *cp, *dp, *path, buf[BUFSIZ];
  int h, i, j;
  int avail;
  arg arg0, arg1;
  struct statfs st;

 top:

  /* For now, always build the damn thing from the ground up */
  part->actions = CREATE_UNIX_FILE|INITIALIZE_PARTITION|RESTORE_DISTIBUTION;

  /* What is the name of the partition called? */
  cp = NULL;
  for (i = 0; i < 'z'; i++) {
    Sprintf(buf, "/ivory/fep%d%c", ux_n, i);
    if (access(buf, F_OK) == -1 && errno == ENOENT) {
      cp = buf;
      break;
    }
    if (i == 0)
      i = 'a' - 1;
  }
  arg0.integer = S_IFREG;
  arg1.integer = NONE_OK;
  path = accept_string("Where do you wish to create the FEP partition file", cp,
		       check_file, &arg0, &arg1);
  part->filename = path;

  /* How much free space is there on this filesystem? */
  dp = cp = strdup(path);
  while (statfs(cp, &st) == -1)
    if (errno != ENOENT) {
      perror("statfs");
      goto top;
    } else if ((dp = rindex(dp, '/')) != NULL)
      *dp-- = '\0';
    else {
      fprintf(stderr, "Unable to figure out how much space is available for this file\n");
      avail = 1000000;
      goto size_query;
    }
  if (getuid() == 0)
    avail = st.f_bfree * st.f_bsize;
  else
    avail = st.f_bavail * st.f_bsize;
  /* Apply magic number from create-fep-partition */
  avail = (avail - 262144) / 1024;
  bcopy((char *) &st.f_fsid, (char *) part->fsid, sizeof(st.f_fsid));
  if (cp != NULL)
    *cp = '/';

  /* How much of the free space has already been allocated? */
  for (i=0; i < params->n_ux400_boards; i++)
    for (j=0; params->ivory_partitions[i][j].filename != NULL && j < MAX_PARTITION_N; j++) {
      if (part == &params->ivory_partitions[i][j]) {
	i = MAX_UX400_N;
	break;
      }
      if (!bcmp((char *) part->fsid, (char *) params->ivory_partitions[i][j].fsid,
		sizeof(st.f_fsid)))
	avail -= params->ivory_partitions[i][j].size_in_kbytes;
    }
  for (h=0,i=MIN_UX1200_DEVNO; h < params->n_ux1200_boards; h++, i++)
    for (j=0; params->ivory_partitions[i][j].filename != NULL && j < MAX_PARTITION_N; j++) {
      if (part == &params->ivory_partitions[i][j]) {
	h = MAX_UX1200_N;
	break;
      }
      if (!bcmp((char *) part->fsid, (char *) params->ivory_partitions[i][j].fsid,
		sizeof(st.f_fsid)))
	avail -= params->ivory_partitions[i][j].size_in_kbytes;
    }
  
  /* How big is it? */
  Printf("There are %d kbytes available for this file.\n", avail);
 size_query:
  arg0.integer = 40000;
  arg1.integer = avail;
  Sprintf(buf, "%ld", avail > NICE_FEPFS ? NICE_FEPFS : avail);
  part->size_in_kbytes = atoi(accept_string("How big do you wish the partition to be", buf,
					    check_integer, &arg0, &arg1));
  
  if (params->distribution_device != NULL &&
      y_or_n_p("Do you wish to create a Hello.boot file", 1)) {
    part->actions |= CREATE_STARTUP_FILES;

    Sprintf(buf, "What is the %s's Internet address", UX_NAME);
    part->internet_address =
      accept_string(buf, (char *) NULL, check_internet_address, (arg *) NULL, (arg *) NULL);
    Sprintf(buf, "What is the %s's Chaosnet address", UX_NAME);
    part->chaos_address =
      accept_string(buf, (char *) NULL, check_chaos_address, (arg *) NULL, (arg *) NULL);
  }
}

void determine_partition_parameters (params)
     struct params *params;
{
  int h, i;
  char buf[BUFSIZ];

  terpri();
  if (!y_or_n_p("Do you wish to create FEP partition file(s)", 1)) {
    params->create_partitions_p = FALSE;
    return;
  }

  require_distribution_device(params);
  require_n_boards(params, FALSE);
  if (!SW_AVAILABLE_P(params, SW_SUN))
    require_etc_directory(params, FALSE);

  /* UX400 partition files */
  for (i = 0; i < params->n_ux400_boards; i++) {
    Sprintf(buf, "Do you wish to create a FEP partition file for %s device #%d",
	    UX400_NAME, i);
    if (!y_or_n_p(buf, 1))
      continue;
    params->create_partitions_p = TRUE;

    /* Only one partition per UX for now */
    determine_partition_file_parameters(params, &params->ivory_partitions[i][0], i);
  }

  /* UX1200 partition files */
  for (h = 0,i = MIN_UX1200_DEVNO; h < params->n_ux1200_boards; h++, i++) {
    Sprintf(buf, "Do you wish to create a FEP partition file for %s device #%d",
	    UX1200_NAME, i);
    if (!y_or_n_p(buf, 1))
      continue;
    params->create_partitions_p = TRUE;

    /* Only one partition per UX for now */
    determine_partition_file_parameters(params, &params->ivory_partitions[i][0], i);
  }
}



configure_partition_file (params, dest, source)
     struct params *params;
     struct partition_params *dest, *source;
{
  int i;
  char *cp, buf[BUFSIZ], etc[BUFSIZ];

  if (dest->filename == NULL)
    return(0);

  if ((cp = rindex(dest->filename, '/')) != NULL) {
    *cp = '\0';
    i = require_directory(dest->filename, 1);
    *cp = '/';
    if (i == -1)
      return(-1);
  }
  if (SW_AVAILABLE_P(params, SW_SUN))
    Sprintf(etc, "%s/%s/etc",
	    params->distribution_directory,
	    params->architecture);
  else
    (void) strcpy(etc, params->relocate_etc_directory);
  Sprintf(buf, "%s/create-fep-partition -automatic %ld %s %s %s %s",
	  etc, dest->size_in_kbytes * 1024,
	  (dest->actions & CREATE_STARTUP_FILES) ? "yes" : "no",
	  (dest->actions & CREATE_STARTUP_FILES) ? dest->chaos_address : "",
	  (dest->actions & CREATE_STARTUP_FILES) ? dest->internet_address : "",
	  dest->filename);
  cp = &buf[strlen(buf)];
  if (source == NULL) {
    /* Restore from tape or CD ROM */
    if (params->distribution_device_type == dd_cdrom) {
      Sprintf(cp, " %s", params->distribution_device);
    } else {
      Sprintf(cp, " %s%s/dev/r%s",
	      params->distribution_host == NULL ? "" : params->distribution_host,
	      params->distribution_host == NULL ? "" : ":",
	      params->distribution_device);
    }
  } else
    /* Restore from source partition */
    Sprintf(cp, " %s", source->filename);
  if ((i = shell_command(buf, TRUE)) != 0) {
    Fprintf(stderr, ">> Error return code %d from create-fep-partition\n", i);
    return(-1);
  }

  return(0);
}

configure_partitions (params)
     struct params *params;
{
  int h, i;
  struct partition_params *source = NULL;

  if (params->create_partitions_p == FALSE)
    return(0);

  /* Create UX400 partitions */
  for (i = 0; i < params->n_ux400_boards; i++) {
    if (params->ivory_partitions[i][0].actions != 0) {
      if (configure_partition_file(params, &params->ivory_partitions[i][0], source) == -1)
	return(-1);
      if (source == NULL)
	source = &params->ivory_partitions[i][0];
    }

#ifdef notdef			/* We only read the tapes once, now */
    if (i+1 < params->n_ux400_boards || params->n_ux1200_boards) {
      /* Assume there are multiple World tapes */
      terpri();
      do {
	Printf("Is the first Genera Distribution World tape mounted");
      } while (!y_or_n_p("", params->defaulting == TRUE ? 0 : 1));
    }
#endif
  }

  /* Create UX1200 partitions */
  for (h = 0, i = MIN_UX1200_DEVNO; h < params->n_ux1200_boards; h++,i++) {
    if (params->ivory_partitions[i][0].actions != 0) {
      if (configure_partition_file(params, &params->ivory_partitions[i][0], source) == -1)
	return(-1);
      if (source == NULL)
	source = &params->ivory_partitions[i][0];
    }

#ifdef notdef			/* We only read the tapes once, now */
    if (i+1 < params->n_ux1200_boards) {
      /* Assume there are multiple World tapes */
      terpri();
      do {
	Printf("Is the first Genera Distribution World tape mounted");
      } while (!y_or_n_p("", params->defaulting == TRUE ? 0 : 1));
    }
#endif
  }

  return(0);
}
