/* -*-Mode: C; Package: C-user -*- */
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1990 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * UX life support file installation
 *
 * $Log:	relocate.c,v $
 * Revision 2.5.1.2  93/02/01  12:57:02  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:52:54  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:43:15  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/relocate.c,v 2.5.1.2 93/02/01 12:57:02 kaufman Rel $";
#endif

#include <sys/file.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <grp.h>
#include <stdio.h>
#include "install.h"

#define EXECPATH_NAME "PATH"
#define EXECPATH_DEFAULT "/bin:/usr/bin:/usr/local/bin"

/* Relocation of files into the UNIX filesystem */

void require_etc_directory (params, relocating_p)
     struct params *params;
{
  char *cp, *prompt, buf[BUFSIZ];
  arg arg0, arg1;

  if (params->relocate_etc_directory != NULL)
    return;

  /* Determine etc directory */
  cp = file_search(S_IFDIR, "/usr/local/etc", "/usr/etc", NULL);
  if (relocating_p) {
    arg1.integer = NONE_OK|W_OK|X_OK;
    prompt = "Where should maintenance and installation programs go";
  } else {
    arg1.integer = X_OK;
    Sprintf(buf, "Where do %s maintenance and installation programs live", UX_NAME);
    prompt = buf;
  }
  arg0.integer = S_IFDIR;
  params->relocate_etc_directory = accept_string(prompt, cp, check_file, &arg0, &arg1);
}

void require_ifep_directory (params, relocating_p)
     struct params *params;
{
  char *cp, *prompt;
  arg arg0, arg1;

  if (params->relocate_ifep_directory != NULL)
    return;

  /* Determine IFEP directory */
  cp = file_search(S_IFDIR, "/usr/share/symbolics/ifep", "/usr/share/local/ivory/ifep",
		   "/usr/share/ivory/ifep", NULL);
  if (cp == NULL)
    cp = "/usr/share/symbolics/ifep";
  if (relocating_p) {
    arg1.integer = NONE_OK|W_OK;
    prompt = "Where should IFEP proms go";
  } else {
    arg1.integer = R_OK;
    prompt = "Where do IFEP proms live";
  }
  arg0.integer = S_IFDIR;
  params->relocate_ifep_directory = accept_string(prompt, cp, check_file, &arg0, &arg1);
}

void determine_relocation_parameters (params)
     struct params *params;
{
  char buf[BUFSIZ], *cp;
  arg arg0, arg1;
  int i, ngroups, groups[NGROUPS];
  int kmem_possible_p = FALSE;
  struct group *kmem;

  if (!SW_AVAILABLE_P(params, SW_IFEP) &&
      !SW_AVAILABLE_P(params, SW_INCLUDE) &&
      !SW_AVAILABLE_P(params, SW_SOURCE) &&
      !SW_AVAILABLE_P(params, SW_SUN)) {
    params->relocate_p = FALSE;
    return;
  }
  terpri();

  Sprintf(buf, "Do you wish to install the %s software", UX_NAME);
  if ((params->relocate_p = y_or_n_p(buf, 1)) == 0)
    return;

  /* Determine whether to make ivory-life setuid */
  while (SW_AVAILABLE_P(params, SW_SUN)) {
    puts("The ivory-life program has to be run by root, the super-user.");
    if (!(params->ivory_life_setuid_p =
	  y_or_n_p("Do you wish ivory-life to be installed setuid to root", !getuid())))
      break;

    if (getuid())
      fputs("\n>> Cannot make ivory-life setuid to root, because you aren't root.\n",
	    stderr);
    else
      break;
  }

  /* Determine whether to make ivory-servicesd setuid */
  while(SW_AVAILABLE_P(params, SW_SUN)) {
    puts("Certain services provided by the ivory-servicesd program require that it run as root.");
    if (!(params->ivory_servicesd_setuid_p =
	  y_or_n_p("Do you wish ivory-servicesd to be installed setuid to root", !getuid())))
      break;

    if (getuid())
      fputs("\n>> Cannot make ivory-servicesd setuid to root, because you aren't root.\n",
	    stderr);
    else
      break;
  }

  /* Determine whether to make programs setgid */
  if ((kmem = getgrnam("kmem")) == NULL)
    params->setgid_p = FALSE;
  else {
    if (!getuid())
      kmem_possible_p = TRUE;
    else if ((ngroups = getgroups(NGROUPS, groups)) == -1) {
      perror("getgroups");
    } else  {
      for (i = 0; i < ngroups; i++)
	if (groups[i] == kmem->gr_gid)
	  kmem_possible_p = TRUE;
    }

    while (SW_AVAILABLE_P(params, SW_SUN)) {
      Printf("Several of the %s life-support programs want to be able to read /dev/kmem.\n",
	     UX_NAME);
      if (!(params->setgid_p =
	    y_or_n_p("Do you wish these programs to be installed setgid to kmem",
		     kmem_possible_p)))
	break;

      for (i = 0; i < ngroups; i++)
	if (groups[i] == kmem->gr_gid)
	  break;

      if (kmem_possible_p == FALSE) {
	fputs("\n>> Cannot install programs setgid to kmem, because you aren't root\n",
	      stderr);
	fputs("   or a member of the kmem group.\n", stderr);
      } else
	break;
    }
  }

  /* Determine bin directory */
  if (SW_AVAILABLE_P(params, SW_SUN)) {
    cp = file_search(S_IFDIR, "/usr/local/bin", "/usr/bin", NULL);
    arg0.integer = S_IFDIR;
    arg1.integer = NONE_OK|W_OK;
    params->relocate_bin_directory =
      accept_string("Where should application programs go", cp, check_file, &arg0, &arg1);
  } else
    params->relocate_bin_directory = NULL;

  /* Check that PATH will find our programs */
  if (verify_directory_in_path(EXECPATH_NAME, EXECPATH_DEFAULT,
			       params->relocate_bin_directory) < 0) {
    Fprintf(stderr, "\nWARNING: directory %s is not in the search path defined\n",
	    params->relocate_bin_directory);
    Fprintf(stderr, "by environment variable %s.  Make sure to update %s to\n",
	    EXECPATH_NAME, EXECPATH_NAME);
    Fprintf(stderr, "to include %s.\n\n",
	    params->relocate_bin_directory);
  }

  /* Determine etc directory */
  if (SW_AVAILABLE_P(params, SW_SUN)) {
    require_etc_directory(params, TRUE);
  } else
    params->relocate_etc_directory = NULL;


  /* Determine include directory */
  if (SW_AVAILABLE_P(params, SW_INCLUDE)) {
    cp = file_search(S_IFDIR, "/usr/local/include/ivory", "/usr/include/ivory", NULL);
    if (cp == NULL) {
      /* This is probably a new installation, that doesn't have an
       * ivory include directory yet.  Find their include heirarchy,
       * and build the include directory later. */
      cp = file_search(S_IFDIR, "/usr/local/include", "/usr/include", NULL);
      if (cp != NULL) {
	Sprintf(buf, "%s/ivory", cp);
	cp = buf;
      }
    }
    arg0.integer = S_IFDIR;
    arg1.integer = NONE_OK|W_OK;
    params->relocate_include_directory =
      accept_string("Where should include files go", cp, check_file, &arg0, &arg1);
  } else
    params->relocate_include_directory = NULL;

  /* Determine lib directory */
  if (SW_AVAILABLE_P(params, SW_SUN)) {
    cp = file_search(S_IFDIR, "/usr/local/lib", "/usr/lib", NULL);
    arg0.integer = S_IFDIR;
    arg1.integer = NONE_OK|W_OK;
    params->relocate_lib_directory =
      accept_string("Where should libraries go", cp, check_file, &arg0, &arg1);
  } else
    params->relocate_lib_directory = NULL;

  /* Determine source directory */
  if (SW_AVAILABLE_P(params, SW_SOURCE)) {
    Sprintf(buf, "Do you wish to install sources to the %s software", UX_NAME);
    if (y_or_n_p(buf, 1)) {
      cp = file_search(S_IFDIR, "/usr/share/symbolics/src", "/usr/share/local/ivory/src",
		       "/usr/share/ivory/src", NULL);
      if (cp == NULL)
	cp = "/usr/share/symbolics/src";
      arg0.integer = S_IFDIR;
      arg1.integer = NONE_OK|W_OK;
      params->relocate_src_directory =
	accept_string("Where should sources go", cp, check_file, &arg0, &arg1);
    } else
      params->relocate_src_directory = NULL;
  } else
    params->relocate_src_directory = NULL;

  /* Determine IFEP directory */
  if (SW_AVAILABLE_P(params, SW_IFEP))
    require_ifep_directory (params, TRUE);
}



relocate_files (params)
     struct params *params;
{
  char source[BUFSIZ];

  if (!params->relocate_p)
    return(0);
  terpri();

  /* Update bin files */
  if (params->relocate_bin_directory != NULL) {
    if (require_directory(params->relocate_bin_directory, 1) == -1)
      return(-1);
    Sprintf(source, "%s/%s/bin", params->distribution_directory, params->architecture);
    Printf("Updating application programs...");
    (void) fflush(stdout);
    if (install_directory(source, params->relocate_bin_directory, 1) == -1)
      return(-1);
  }

  /* Update etc files */
  if (params->relocate_etc_directory != NULL) {
    if (require_directory(params->relocate_etc_directory, 1) == -1)
      return(-1);
    Sprintf(source, "%s/%s/etc", params->distribution_directory, params->architecture);
    Printf("Updating maintenance and installation programs...");
    (void) fflush(stdout);
    if (install_directory(source, params->relocate_etc_directory, 1) == -1)
      return(-1);
  }

  /* Update include files.  */
  if (params->relocate_include_directory != NULL) {
    if (require_directory(params->relocate_include_directory, 1) == -1)
      return(-1);
    Sprintf(source, "%s/h", params->distribution_directory);
    Printf("Updating include files...");
    (void) fflush(stdout);
    if (update_directory(source, params->relocate_include_directory, 1) == -1)
      return(-1);
  }

  /* Update library files */
  if (params->relocate_lib_directory != NULL) {
    if (require_directory(params->relocate_lib_directory, 1) == -1)
      return(-1);
    Sprintf(source, "%s/%s/lib", params->distribution_directory, params->architecture);
    Printf("Updating libraries...");
    (void) fflush(stdout);
    if (update_directory(source, params->relocate_lib_directory, 1) == -1)
      return(-1);
  }

  /* Update source files */
  if (params->relocate_src_directory != NULL) {
    if (require_directory(params->relocate_src_directory, 1) == -1)
      return(-1);
    Sprintf(source, "%s/src", params->distribution_directory);
    Printf("Updating sources...");
    (void) fflush(stdout);
    if (update_directory(source, params->relocate_src_directory, 1) == -1)
      return(-1);
  }

  /* Update IFEP files */
  if (params->relocate_ifep_directory != NULL) {
    if (require_directory(params->relocate_ifep_directory, 1) == -1)
      return(-1);
    Sprintf(source, "%s/ifep", params->distribution_directory);
    Printf("Updating IFEP proms...");
    (void) fflush(stdout);
    if (update_directory(source, params->relocate_ifep_directory, 1) == -1)
      return(-1);
  }

  return(0);
}
