/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1990 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */
/* UNIX side of a Serial Message channel
 *
 * $Log:	serial.c,v $
 * Revision 2.5.1.2  93/02/01  12:57:04  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:52:58  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:43:24  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/serial.c,v 2.5.1.2 93/02/01 12:57:04 kaufman Rel $";
#endif

#include <sys/types.h>
#include <sys/file.h>
#include <sys/mman.h>
#include <sys/termio.h>
#include <sys/termios.h>
#include <sys/time.h>
#include <errno.h>
#include <fcntl.h>
#include <setjmp.h>
#include <signal.h>
#include <stdio.h>
#include "embed.h"
#include "life.h"

/*#define DEBUG /**/
#ifdef DEBUG
#define DBG(a) fputs(dbg_prefix, stdout), printf a
char *dbg_prefix = "open_serial_unit ";
#else
#define DBG(a)
#endif

static EmbSerialDataBufferHeader *pending_input = NULL;
static EmbSerialDataBufferHeader *pending_output = NULL;
static int pending_output_index = 0;
extern int (*rpc_return_hook)();

open_serial_unit_internal (serial_unit, protocol_version, guest_unit, channel,
			   uid, gid, gids,
			   baud_rates, readable_control_lines, writable_control_lines)
     char *serial_unit;
     short protocol_version;
     long guest_unit;
     EmbPtr channel;
     long uid;
     long gid;
     long *gids;
     unsigned long *baud_rates;
     unsigned short *readable_control_lines;
     unsigned short *writable_control_lines;
{
  int serial_fd, ivory_fd, i;
  extern EmbCommArea *EmbCommAreaPtr;
  char device[128];
  caddr_t ptr;
  IvoryConfig config;

  DBG(("Open %s for ivory%d\n", serial_unit, guest_unit));
  DBG(("uid %d gid %d  %d gids: %d %d %d ...\n",
       uid, gid, gids[0], gids[1], gids[2], gids[3]));

  /* Check protocol version */
  if (protocol_version != EMB_SERIAL_PROTOCOL_VERSION) {
    DBG(("Unsupported protocol version %d\n", protocol_version));
    return(ENOPROTOOPT);
  }
  DBG(("proto %d\n", protocol_version));

  /* Fork */
  if ((i = fork()) == -1)
    return(errno);
  else if (i != 0) {
    /* Parent */
    extern jmp_buf skip_rpc;
    int sigchld_handler();

    DBG(("parent pid = %d\n", getpid()));
    signal(SIGCHLD, sigchld_handler);
    rpc_return_hook = NULL;
    longjmp(skip_rpc, 1);
  } else {
    /* Child */
    int exit();
    extern int rpc_inhibit_module_deregistration;

    /* Parent is still handling RPC calls */
    rpc_inhibit_module_deregistration = 1;

    /* If there's an error, make sure the child process goes away */
    rpc_return_hook = exit;
    DBG(("child pid = %d\n", getpid()));

    /* Check permissions */
    if (getuid()) {
      /* If we're running as a particular user, then this had better be that user */
      if (uid != getuid())
	return(EPERM);
    } else if (uid == 0) {
      /* Don't allow LispM to pretend to be root */
      return(EPERM);
    } else {
      /* If we're running as root, then set uid, gid, & groups as specified */
      if (setgroups(gids[0], gids+1) == -1) {
	perror("setgroups");
	return(errno);
      }
      if (setgid(gid) == -1) {
	perror("setgid");
	return(errno);
      }
      if (setuid(uid) == -1) {
	perror("setuid");
	return(errno);
      }
    }

    /* Open the device */
    if ((serial_fd = open(serial_unit, O_RDWR)) == -1) {
      i = errno;
      DBG(("open() ==> %d\n", i));
      return(i);
    }
#ifdef DEBUG
    dbg_prefix = (char *) calloc(strlen(serial_unit), 1) + 2;
    sprintf(dbg_prefix, "%s ", serial_unit);
#endif
    DBG(("fd %d\n", serial_fd));

    /* Grab write lock on device */
    if (flock(serial_fd, LOCK_EX|LOCK_NB) == -1) {
      if (errno == EWOULDBLOCK)
	/* More appropriate error message */
	i = EACCES;
      else
	i = errno;
      DBG(("flock() ==> %d\n", i));
      return(i);
    }
    DBG(("locked\n"));

    /* Map EmbCommArea */
    sprintf(device, "/dev/ivory%d", guest_unit);
    if ((ivory_fd = open(device, O_RDWR)) == -1) {
      i = errno;
#ifdef DEBUG
      perror(device);
#endif
      close(serial_fd);
      return(i);
    }
    if (ioctl(ivory_fd, IVORYGCONF, &config) == -1) {
      i = errno;
#ifdef DEBUG
      perror("IVORYGCONF");
#endif
      close(serial_fd);
      return(i);
    }
    if ((ptr = mmap(0, config.mmap_size, PROT_READ|PROT_WRITE,
		    MAP_SHARED, ivory_fd, 0)) == (caddr_t) -1) {
      i = errno;
#ifdef DEBUG
      perror("mmap");
#endif
      close(serial_fd);
      return(i);
    }
    ptr += config.comm_area_base;
    RegisterEmbCommArea(ivory_fd, (EmbCommArea *) ptr);

    /* Check putative channel EmbPtr */
    if (channel < EmbCommAreaPtr->number_of_slots ||
	channel > EmbCommAreaPtr->comm_memory_size) {
      DBG(("channel %d outside range [%d, %d)\n",
	   channel, EmbCommAreaPtr->number_of_slots, EmbCommAreaPtr->comm_memory_size));
      return(EINVAL);
    }

    /* Enable select() on the emb channel */
    if (ioctl(ivory_fd, IVORYSELCHANNEL, &channel) == -1) {
      i = errno;
      perror("IVORYSELCHANNEL");
      return(i);
    }
    
    IvorySerialLifeSetup(ptr, ivory_fd, channel, serial_fd);

    /* Baud rates and modem control lines as per termios(4) */
    *readable_control_lines = SerialControlLineRTS|SerialControlLineCTS|
      SerialControlLineDSR|SerialControlLineDTR|
	SerialControlLineRI|SerialControlLineCD;
    *writable_control_lines =  SerialControlLineRTS|SerialControlLineCTS|
      SerialControlLineDSR|SerialControlLineDTR|
	SerialControlLineRI|SerialControlLineCD;
    bzero(baud_rates, 30 * sizeof(unsigned long));
    baud_rates[ 0] = 50;
    baud_rates[ 1] = 75;
    baud_rates[ 2] = 110;
    baud_rates[ 3] = 135;	/* 134.5 */
    baud_rates[ 4] = 150;
    baud_rates[ 5] = 200;
    baud_rates[ 6] = 300;
    baud_rates[ 7] = 600;
    baud_rates[ 8] = 1200;
    baud_rates[ 9] = 1800;
    baud_rates[10] = 2400;
    baud_rates[11] = 4800;
    baud_rates[12] = 9600;
    baud_rates[13] = 19200;
    baud_rates[14] = 38400;

    return(0);
  }
}

IvorySerialLifeTrampoline ()
{
  return(IvorySerialLifeSetup(NULL, -1, -1 -1));
}

IvorySerialLifeSetup (EmbCommAreaPtr, ivory_fd, channel, serial_fd)
     EmbCommArea *EmbCommAreaPtr;
     int ivory_fd;
     EmbPtr channel;
     int serial_fd;
{
  static EmbCommArea *trampoline_EmbCommAreaPtr;
  static int trampoline_ivory_fd, trampoline_serial_fd;
  static EmbPtr trampoline_channel;
  int val;

  DBG(("IvorySerialLifeSetup\n"));
  
  if (EmbCommAreaPtr != NULL) {
    trampoline_EmbCommAreaPtr = EmbCommAreaPtr;
    trampoline_ivory_fd = ivory_fd;
    trampoline_channel = channel;
    trampoline_serial_fd = serial_fd;
    rpc_return_hook = IvorySerialLifeTrampoline;
  } else {
    val = IvorySerialLife(trampoline_EmbCommAreaPtr, trampoline_ivory_fd,
			  trampoline_channel, trampoline_serial_fd);

    /* Disable select() on the emb channel */
    if (ioctl(ivory_fd, IVORYUNSELCHANNEL, &channel) == -1) {
      perror("IVORYUNSELCHANNEL");
      return(1);
    }
    if (flock(serial_fd, LOCK_UN) == -1) {
      perror("flock");
      return(1);
    }

    return(val);
  }
  /*NOTREACHED*/
}

IvorySerialLife(EmbCommAreaPtr, ivory_fd, channel, serial_fd)
     EmbCommArea *EmbCommAreaPtr;
     int ivory_fd;
     EmbPtr channel;
     int serial_fd;
{
  int cc, nfds;
  int output_pending_p = FALSE, input_pending_p = FALSE;
  fd_set read, write;
  EmbMessageChannel *serial;
  struct timeval tv;
  int reset_request_count = EmbCommAreaPtr->reset_request_count;

  if (fcntl(serial_fd, F_SETFL, FNDELAY) == -1) {
    /* --- */
    perror("FNDELAY");
    exit(1);
  }

  serial = (EmbMessageChannel *) HostPointer(channel);

  while (1) {
    FD_ZERO(&read);
    FD_ZERO(&write);
    if (output_pending_p)
      FD_SET(serial_fd, &write);
    else
      FD_SET(ivory_fd, &read);
    if (input_pending_p)
      FD_SET(ivory_fd, &write);
    else
      FD_SET(serial_fd, &read);
    tv.tv_sec = 0;
    tv.tv_usec = 250000;
    if ((cc = select(nfds, &read, &write, NULL, &tv)) == -1) {
      /* --- */
      perror("select");
      exit(1);
    }

#ifdef DEBUG_notdef
#define FF(a,b) (FD_ISSET(a,b)?"t":"nil")
    if (cc > 0) {
      printf("read: serial %s, ivory %s\n", FF(serial_fd, &read), FF(ivory_fd, &read));
      printf("write: serial %s, ivory %s\n", FF(serial_fd, &write), FF(ivory_fd, &write));
    }
#endif

    /* If reset_request_count has changed, Lisp must have booted.
     * Bag out before we smash anything. */
    if (reset_request_count != EmbCommAreaPtr->reset_request_count) {
      DBG(("restart_applications shutdown\n"));
      if ((cc = ioctl(ivory_fd, IVORYUNSELCHANNEL, &channel)) == -1)
	perror("IVORYUNSELCHANNEL");
      if (flock(serial_fd, LOCK_UN) == -1)
	perror("flock");
      exit(0);
    }

    /* If the guest to host impulse is set, handle it */
    if (serial->guest_to_host_impulse != SerialImpulseNone) {
      DBG(("SerialImpulse()\n"));
      SerialImpulse(EmbCommAreaPtr, ivory_fd, serial, serial_fd);
    }

    /* Check for control line transitions */
    ReadControlLines(serial_fd, serial, NULL);

    /* Check for data Genera can read */
    if (FD_ISSET(serial_fd, &read) ||
	(input_pending_p == TRUE && FD_ISSET(ivory_fd, &write))) {
      DBG(("SerialReceiver()\n"));
      input_pending_p = SerialReceiver(EmbCommAreaPtr, ivory_fd, serial, serial_fd, FALSE);
    } else if (input_pending_p == FALSE &&
	       pending_input != NULL &&
	       !EmbQueueFilled(HostPointer(serial->host_to_guest_queue))) {
      /* Flush pending buffer */
      EmbQueuePutWord(HostPointer(serial->host_to_guest_queue), GuestPointer(pending_input));
      pending_input = NULL;
    }

    /* Check for data Genera wants to write, or control messages */
    if (FD_ISSET(ivory_fd, &read)) {
      DBG(("SerialTransmitter()\n"));
      output_pending_p = SerialTransmitter(EmbCommAreaPtr, ivory_fd, serial, serial_fd);
    } else if (output_pending_p && FD_ISSET(serial_fd, &write)) {
      EmbPtr temp = GuestPointer(pending_output);
      
      DBG(("TransmitData()\n"));
      if ((output_pending_p = TransmitData(EmbCommAreaPtr, NULL, serial_fd)) == FALSE)
	EmbQueuePutWord(HostPointer(serial->guest_to_host_return_queue), temp);
    }
  }
}



SerialImpulse(EmbCommAreaPtr, ivory_fd, serial, serial_fd)
     EmbCommArea *EmbCommAreaPtr;
     int ivory_fd;
     EmbMessageChannel *serial;
     int serial_fd;
{
  EmbPtr elt;
  EmbQueue *q, *returnq;

  switch (serial->guest_to_host_impulse) {
  case SerialImpulseNone:
    DBG(("SerialImpulseNone\n"));
    break;

  case SerialImpulseWakeup:
    DBG(("SerialImpulseWakeup\n"));
    break;
    
  case SerialImpulseFlush:
    DBG(("SerialImpulseFlush\n"));
    q = HostPointer(serial->guest_to_host_queue);
    returnq = HostPointer(serial->guest_to_host_return_queue);

    /* Flush guest->host queue */
    while (EmbQueueFilled(q)) {
      if ((elt = EmbQueueTakeWord(q)) == NullEmbPtr ||
	  elt == 0)
	break;
      EmbQueuePutWord(returnq, elt);
    }

    /* Drop any pending output */
    if (pending_output != NULL) {
      EmbQueuePutWord(returnq, GuestPointer(pending_output));
      pending_output = NULL;
    }
    break;

  default:
    DBG(("Unknown impulse %d\n", serial->guest_to_host_impulse));
    break;
  }

  /* Restrain impulse */
  serial->guest_to_host_impulse = SerialImpulseNone;
}



SerialReceiver (EmbCommAreaPtr, ivory_fd, serial, serial_fd, force)
     EmbCommArea *EmbCommAreaPtr;
     int ivory_fd;
     EmbMessageChannel *serial;
     int serial_fd;
     int force;
{
  EmbPtr elt;
  EmbSerialDataBufferHeader *data;
  int i, cc;

  if (pending_input == NULL) {
    /* If there's no receive buffer available, or the one we get is bogus,
     * just note that there's input pending and return. */
    if (!EmbQueueFilled(HostPointer(serial->host_to_guest_supply_queue)))
      return(TRUE);
    elt = EmbQueueTakeWord(HostPointer(serial->host_to_guest_supply_queue));
    if (elt == NullEmbPtr || elt == 0)
      return(TRUE);

    /* Set up the receive buffer */
    data = (EmbSerialDataBufferHeader *) HostPointer(elt);
    data->opcode = SerialOpcodeData;
    data->nbytes = 0;
  } else
    data = pending_input;

  /* Try to fill the buffer */
  i = (data->length<<2) - data->nbytes - sizeof(EmbSerialDataBufferHeader);
  DBG(("Trying to read %d bytes into buffer %x at %d...\n",
       i, GuestPointer(data), data->nbytes));
  if ((cc = read(serial_fd, EmbSerialBufferData(data)+data->nbytes, i)) == -1) {
    /* --- */
    perror("read");
    exit(1);
  }
  DBG(("read %d bytes\n", cc));
  data->nbytes += cc;

  if (i == cc || !EmbQueueFilled(HostPointer(serial->host_to_guest_queue))) {
    /* Enqueue the buffer */
    EmbQueuePutWord(HostPointer(serial->host_to_guest_queue), GuestPointer(data));
    pending_input = NULL;
  } else {
    /* Save the buffer for further data */
    pending_input = data;
  }
  return(FALSE);
}

ReceiveBreak(EmbCommAreaPtr, ivory_fd, serial, serial_fd)
     EmbCommArea *EmbCommAreaPtr;
     int ivory_fd;
     EmbMessageChannel *serial;
     int serial_fd;
{
}



SerialTransmitter (EmbCommAreaPtr, ivory_fd, serial, serial_fd)
     EmbCommArea *EmbCommAreaPtr;
     int ivory_fd;
     EmbMessageChannel *serial;
     int serial_fd;
{
  EmbPtr elt;
  EmbSerialBuffer *buf;
  EmbSerialStatusBuffer *status;
  int cc, on, off;

  /* Read EmbPtr from queue */
  elt = EmbQueueTakeWord(HostPointer(serial->guest_to_host_queue));
  if (elt == 0 || elt == NullEmbPtr)
    return;

  /* Parse buffer */
  buf = (EmbSerialBuffer *) HostPointer(elt);
  DBG(("buf %x opcode %d\n", elt, buf->opcode));
  buf->return_code = 0;
  switch (buf->opcode) {
  case SerialOpcodeNOOP:
    DBG(("NOOP\n"));
    break;

  case SerialOpcodeData:
    if (TransmitData(EmbCommAreaPtr, (EmbSerialDataBufferHeader *) buf, serial_fd) == TRUE)
      /* Output pending, don't recycle the emb buffer yet */
      return(TRUE);
    break;

  case SerialOpcodeSetParameters:
    DBG(("SetParameters\n"));
    buf->return_code = SetSerialParameters(EmbCommAreaPtr, serial, serial_fd,
					   (EmbSerialParameterBuffer *) buf);
    break;

  case SerialOpcodeSetControlLines: /* Do lines */
#define HANDLE_CONTROL_LINE(a,b) if (a == 1) on |= b; else if (a == 0) off |= b
    DBG(("SetControlLines\n"));
    status = (EmbSerialStatusBuffer *) buf;
    on = off = 0;
    HANDLE_CONTROL_LINE(status->dtr, TIOCM_DTR);
    HANDLE_CONTROL_LINE(status->dsr, TIOCM_DSR);
    HANDLE_CONTROL_LINE(status->cts, TIOCM_CTS);
    HANDLE_CONTROL_LINE(status->rts, TIOCM_RTS);
    HANDLE_CONTROL_LINE(status->cd, TIOCM_CD);
    HANDLE_CONTROL_LINE(status->ri, TIOCM_RI);
    if (on != 0 && ioctl(serial_fd, TIOCMBIS, &on) == -1)
      perror("TIOCMBIS");
    if (off != 0 && ioctl(serial_fd, TIOCMBIC, &off) == -1)
      perror("TIOCMBIC");
    break;
#undef HANDLE_CONTROL_LINE

  case SerialOpcodeReadControlLines:
    DBG(("ReadControlLines\n"));
    ReadControlLines(serial_fd, NULL, (EmbSerialStatusBuffer *) buf);
    break;

  case SerialOpcodeSendBreak:
    DBG(("SendBreak\n"));
    buf->return_code = SendBreak(serial_fd, (EmbSerialBreakBuffer *) buf);
    break;

  case SerialOpcodeShutdown:
    DBG(("Shutdown\n"));

    /* Disable select() on the emb channel */
    cc = GuestPointer(serial);
    if ((cc = ioctl(ivory_fd, IVORYUNSELCHANNEL, &cc)) == -1)
      perror("IVORYUNSELCHANNEL");
    if (flock(serial_fd, LOCK_UN) == -1)
      perror("flock");

    /* If there's pending input, enqueue the buffer so it doesn't get lost */
    if (pending_input != NULL)
      EmbQueuePutWord(HostPointer(serial->guest_to_host_return_queue),
		      GuestPointer(pending_input));

    /* Finally, ACK the Shutdown message */
    EmbQueuePutWord(HostPointer(serial->guest_to_host_return_queue), elt);
    exit(cc);

  default:
    fprintf(stderr, "Unknown serial opcode %d\n", buf->opcode);
    break;
  }

  /* Recycle the buffer */
  EmbQueuePutWord(HostPointer(serial->guest_to_host_return_queue), elt);
  return(FALSE);
}

TransmitData(EmbCommAreaPtr, data, serial_fd)
     EmbCommArea *EmbCommAreaPtr;
     EmbSerialDataBufferHeader *data;
     int serial_fd;
{
  int cc;
  char *buf;
  int len;

  if (data == NULL) {
    /* Transmit more of pending_output */
    data = pending_output;
    buf = EmbSerialBufferData(pending_output) + pending_output_index;
    len = pending_output->nbytes - pending_output_index;
  } else {
    buf = EmbSerialBufferData(data);
    len = data->nbytes;
  }

  DBG(("%sata (%d bytes)\n", data == pending_output ? "More d" : "D", len));
  if ((cc = write(serial_fd, buf, len)) == -1 && errno != EWOULDBLOCK) {
    data->return_code = errno;
#ifdef DEBUG
    perror("write");
#endif
  } else if ((cc == -1 && errno == EWOULDBLOCK) || cc < len) {
    /* Some data didn't get transmitted */
    if (cc == -1) {
      DBG(("XMIT would block\n"));
      cc = 0;
    } else {
      DBG(("XMIT lost %d bytes\n", data->nbytes - cc));
    }
    pending_output = data;
    pending_output_index += cc;
    return(TRUE);
  } else {
    pending_output = NULL;
    pending_output_index = 0;
  }

  return(FALSE);
}

/* Read modem control lines
 * If serial == NULL, status != NULL, fill in status and return.
 * If serial != NULL, status == NULL, then if the control lines have
 *   changed since last time, grab a  buffer, fill it in, and return it.
 */
ReadControlLines (serial_fd, serial, status)
     int serial_fd;
     EmbMessageChannel *serial;
     EmbSerialStatusBuffer *status;
{
  static long cached_control_lines = 0;
  long control_lines;
  int xmit_buffer = FALSE;

  /* Read control lines */
  if (ioctl(serial_fd, TIOCMGET, &control_lines) == -1) {
    perror("TIOCMGET");
    if (status != NULL)
      status->dtr = status->dsr = status->cts = status->rts = status->cd = status->ri = -1;
    return(-1);
  }

#define HANDLE_CONTROL_LINE(a,b) a = (control_lines & b) ? 1 : 0;
  if (status != NULL || control_lines != cached_control_lines) {
    if (status == NULL) {
      EmbQueue *q;

      /* Grab a buffer */
      q = HostPointer(serial->host_to_guest_supply_queue);
      if (EmbQueueFilled(q)) {
	status = HostPointer(EmbQueueTakeWord(q));
	xmit_buffer = TRUE;
      } else {
	DBG(("control lines changed but no buffer available\n"));
	return(-2);
      }
    }

    /* Set up the status buffer */
    status->opcode = SerialOpcodeSetControlLines;
    HANDLE_CONTROL_LINE(status->dtr, TIOCM_DTR);
    HANDLE_CONTROL_LINE(status->dsr, TIOCM_DSR);
    HANDLE_CONTROL_LINE(status->cts, TIOCM_CTS);
    HANDLE_CONTROL_LINE(status->rts, TIOCM_RTS);
    HANDLE_CONTROL_LINE(status->cd, TIOCM_CD);
    HANDLE_CONTROL_LINE(status->ri, TIOCM_RI);

    /* Xmit the buffer if we grabbed it */
    if (xmit_buffer) {
      EmbQueuePutWord(HostPointer(serial->host_to_guest_queue), GuestPointer(status));
      DBG(("SetControlLines sent\n"));
    }

    cached_control_lines = control_lines;
  }

  return(0);
}

SetSerialParameters(EmbCommAreaPtr, serial, serial_fd, params)
     EmbCommArea *EmbCommAreaPtr;
     EmbMessageChannel *serial;
     int serial_fd;
     EmbSerialParameterBuffer *params;
{
  struct termios tty;

  if (params->mode != SerialModeAsync) {
    DBG(("Unknown mode %d - ignoring params\n", params->mode));
    return;
  }

  bzero(&tty, sizeof(tty));
  tty.c_cflag = CREAD;

  /* Baud Rate */
  DBG(("%d baud\n", params->baud_rate));
  switch (params->baud_rate) {
  case 50:
    tty.c_cflag |= B50;
    break;
  case 75:
    tty.c_cflag |= B75;
    break;
  case 110:
    tty.c_cflag |= B110;
    break;
  case 135:			/* 134.5 */
    tty.c_cflag |= B134;
    break;
  case 150:
    tty.c_cflag |= B150;
    break;
  case 200:
    tty.c_cflag |= B200;
    break;
  case 300:
    tty.c_cflag |= B300;
    break;
  case 600:
    tty.c_cflag |= B600;
    break;
  case 1200:
    tty.c_cflag |= B1200;
    break;
  case 1800:
    tty.c_cflag |= B1800;
    break;
  case 2400:
    tty.c_cflag |= B2400;
    break;
  case 4800:
    tty.c_cflag |= B4800;
    break;
  case 9600:
    tty.c_cflag |= B9600;
    break;
  case 19200:
    tty.c_cflag |= B19200;
    break;
  case 38400:
    tty.c_cflag |= B38400;
    break;
  default:
    DBG(("Unsupported baud_rate = %d\n", params->baud_rate));
    break;

  }

  /* Character Size */
  DBG(("CS%d\n", params->data_bits));
  switch (params->data_bits) {
  case 5:
    tty.c_cflag |= CS5;
    break;
  case 6:
    tty.c_cflag |= CS6;
    break;
  case 7:
    tty.c_cflag |= CS7;
    break;
  case 8:
    tty.c_cflag |= CS8;
    break;
  default:
    DBG(("Unsupported data_bits = %d\n", params->data_bits));
    tty.c_cflag |= CS7;		/* --- */
    break;
  }
  switch (params->stop_bits) {
  case 16:
    /* Default is one stop bit */
    break;
  case 32:
    tty.c_cflag |= CSTOPB;
  default:
    DBG(("Unsupported stop_bits = %d\n", params->stop_bits));
    break;
  }

  /* Flow control
   * Yes, the input_ & output_ names are exactly wrong. */
  if (params->rts_cts_flow_control_p) {
    DBG(("RTS/CTS\n"));
    tty.c_cflag |= CRTSCTS;
  }
  if (params->respect_xon_xoff_p) {
    /* Listen to XON/XOFF */
    DBG(("Respect XON/XOFF (on %d off %d)\n",
	 params->output_xon_char, params->output_xoff_char));
    tty.c_iflag |= IXON;
#ifdef no_confused		/* --- */
    tty.c_cc[VSTART] = params->output_xon_char;
    tty.c_cc[VSTOP] = params->output_xoff_char;
#else
    tty.c_cc[VSTART] = 17;
    tty.c_cc[VSTOP] = 19;
#endif
  }
  if (params->generate_xon_xoff_p) {
    /* Generate XON/XOFF */
    DBG(("Generate XON/XOFF (on %d off %d)\n",
	 params->input_xon_char, params->input_xoff_char));
    tty.c_iflag |= IXOFF;
    tty.c_cc[VSTART] = params->input_xon_char;
    tty.c_cc[VSTOP] = params->input_xoff_char;
  }

  /* Parity */
  switch (params->parity) {
  case SerialParityNone:
    DBG(("No Parity\n"));
    break;
  case SerialParityOdd:
    DBG(("Odd Parity\n"));
    tty.c_cflag |= PARENB|PARODD;
    break;
  case SerialParityEven:
    DBG(("Even Parity\n"));
    tty.c_cflag |= PARENB;
    break;
  case SerialParityMark:	/* --- Unsupported */
  case SerialParitySpace:	/* --- Unsupported */
  default:
    DBG(("Unsupported parity %d\n", params->parity));
    tty.c_cflag |= PARENB;
    break;
  }

  /* --- hack job MIN, TIME */
  tty.c_cc[VMIN] = 0;
  tty.c_cc[VTIME] = 0;

  if (ioctl(serial_fd, TCSETS, &tty) == -1) {
#ifdef DEBUG
    int tmp;
    
    tmp = errno;
    perror("TCSETS");
    errno = tmp;
#endif
    return(errno);
  }
  return(0);
}

SendBreak (serial_fd, brk)
     int serial_fd;
     EmbSerialBreakBuffer *brk;
{
  /* We can only do a 1/4 second break */
  brk->duration = 15;
  
  if (ioctl(serial_fd, TCSBRK, 0) == -1) {
#ifdef DEBUG
    int tmp;
    
    tmp = errno;
    perror("TCSBRK");
    errno = tmp;
#endif
    return(errno);
  }
  return(0);
}
