/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * $Log:	symtab.c,v $
 * Revision 2.5.1.2  93/02/01  12:57:11  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:53:08  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:43:45  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/symtab.c,v 2.5.1.2 93/02/01 12:57:11 kaufman Rel $";
#endif

#include <sys/types.h>
#include <sys/file.h>
#include <fcntl.h>
#include <a.out.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>

static int cached_fd = -1;
struct exec cached_exec;

start_symbol_names (file)
     char *file;
{
  if (cached_fd != -1)
    end_symbol_names();

  if ((cached_fd = open(file, O_RDONLY)) == -1)
    return(-1);

  if (read(cached_fd, (char *) &cached_exec, sizeof(struct exec)) == -1) {
    int tmp = errno;

    close(cached_fd);
    errno = tmp;
    return(-1);
  }

  return(0);
}

end_symbol_names ()
{
  close(cached_fd);
  cached_fd = -1;
}


char *lookup_string(offset)
     long offset;
{
  static char buf[BUFSIZ];
  long stroff = N_STROFF(cached_exec);
  long strsize;

#define NULL1 "1st lseek"
#define NULL2 "1st read"
#define NULL3 "out of range"
#define NULL4 "2nd lseek"
#define NULL5 "2nd read"

  if (lseek(cached_fd, stroff, L_SET) == -1)
    return(NULL1);

  if (read(cached_fd, (char *) &strsize, sizeof(long)) == -1)
    return(NULL2);

  if (offset < 4 || offset > strsize)
/*    return(NULL3);*/
    return((char *) sprintf(buf, "!(4 < %d < %d)", offset, strsize));

  if (lseek(cached_fd, offset-4, L_INCR) == -1)
    return(NULL4);

  if (read(cached_fd, buf, BUFSIZ) == -1)
    return(NULL5);

  return(buf);
}
  

#define NSYMS 128
#define MIN(a,b) ((a) < (b) ? (a) : (b))

get_symbol_names (file, nl, n)
     char *file;
     struct nlist nl[];
     int n;
{
  struct nlist syms[NSYMS];
  long off, stroff;
  int found = 0;
  int i, j, nbytes;
  char *name;

  if (start_symbol_names(file) == -1)
    perror(file);
  off = N_SYMOFF(cached_exec);
  stroff = N_STROFF(cached_exec);

  do {
    if (lseek(cached_fd, off, L_SET) == -1)
      return(-1);

    nbytes = MIN(sizeof(syms), stroff - off);
    if (read(cached_fd, (char *) syms, nbytes) == -1)
      return(-1);
    off += nbytes;

    for (j = 0; j < nbytes/sizeof(struct nlist); j++)
      for (i = 0; i < n; i++)
	if (syms[j].n_value == nl[i].n_value &&
	    (syms[j].n_type & N_TYPE) == N_TEXT &&
	    nl[i].n_un.n_name == NULL) {
	  if ((name = lookup_string(syms[j].n_un.n_strx)) != NULL) {
	    nl[i].n_un.n_name = strdup(name);
	    found++;
	  }
	}
  } while (off < stroff && found < n);

  end_symbol_names();
  return(0);
}

char *symbol_name(value)
     long value;
{
  struct nlist nl;
  long off = N_SYMOFF(cached_exec);
  long stroff = N_STROFF(cached_exec);

  if (cached_fd == -1)
    return(NULL);
  
  do {
    if (lseek(cached_fd, off, L_SET) == -1)
      return(NULL);
    off += sizeof(nl);

    if (read(cached_fd, (char *) &nl, sizeof(nl)) == -1)
      return(NULL);

    if (nl.n_value == value) {
      return(lookup_string(nl.n_un.n_strx));
    }
  } while (off < stroff);

  return(NULL);
}
