/*-*- Mode: C; Package: C-user -*-
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * Ivory diagnostic.  Runs in two modes: a one-pass sanity check, or
 * continuous testing; default is one pass.  First determines whether
 * solstice kernel is running or not.  If it is, disables kernel and
 * saves comm-area (warning luser).
 *
 * First tests comm-area from host, then loads diagnostic prom which
 * will test comm-area memory contention, main-memory, and (perhaps?)
 * interrupts.
 *
 * $Log:	test.c,v $
 * Revision 2.5.1.2  93/02/01  12:57:12  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:53:10  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:43:51  kaufman
 * Baseline for Genera 8.3
 * 
 * 
 */



#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/test.c,v 2.5.1.2 93/02/01 12:57:12 kaufman Rel $";
#endif

#include <sys/types.h>
#include <sys/file.h>
#include <sys/filio.h>
#include <sys/mman.h>
#include <errno.h>
#include <setjmp.h>
#include <signal.h>
#include <strings.h>
#include <stdio.h>
#include <errno.h>
#include <termio.h>

#define INCLUDE_FILE_HEADERS
#include "embed.h"
#include "life.h"
#include "emblib.h"
#undef INCLUDE_FILE_HEADERS

int verbose = 0;
static int ivory_fd;
static int running_our_kernel = 0;
/* 0 to flag not set, defaulted below */
static int system_type = 0;
static char* model = "UX400S";
static long slave_buffer_address = 0;
static long slave_buffer_size;
static long comm_area_base;
static long comm_area_size;
static long mailbox_address = 0;
static long mailbox_size = MERLIN_VME_MAILBOX_SIZE;
static long* mailbox = NULL;
	



#define catch(sig name) if(signal(sig,name/**/_handler)==SIG_ERR) {perror("catch");exit(1);} else
#define uncatch(sig) if(signal(sig, SIG_DFL)==SIG_ERR) {perror("uncatch");exit(1);} else
#define define_handler(name) static jmp_buf name/**/_env; void name/**/_handler(sig){longjmp(name/**/_env,sig);}
#define handle(name) (setjmp(name/**/_env))
#define throw(name, value) name/**/_handler(value)

/* #define DEBUG(level what) if(verbose>=level)(void)fprintf(stderr,what) */
 
#define PEEK_SIGNAL SIGBUS

define_handler(peek)

peek(addr, cont)
     long* addr;
     long* cont;
{
  int error;

  catch(PEEK_SIGNAL, peek);
  if (!(error=handle(peek))) {
    *cont = *addr;
  }
  uncatch(PEEK_SIGNAL);

  return(error?FALSE:TRUE);
}

y_or_n_p (string)
	char *string;
{
  int ans;
  struct termio termio;
  
  do {
    (void)fprintf(stderr,"\n");
    (void)fprintf(stderr, string);
    (void)fprintf(stderr, "(y or n) ", string);
    (void)fflush(stderr);
    /* flush type-ahead */
    if (ioctl(fileno(stdin), TCGETA, &termio) == -1) {
      perror("TCGETA");
      exit(1);
    } 
    if (ioctl(fileno(stdin), TCSETAF, &termio) == -1) {
      perror("TCSETA");
      exit(1);
    }
    
    ans = fgetc(stdin);
    /* flush rest */
    for(;(ans!=EOF) && (fgetc(stdin)!='\n'););
    
    if (isupper(ans)) ans = tolower(ans);
  } while ((ans!=EOF) && (ans!='y') && (ans!= 'n'));

  if (ans==EOF)
    exit(3);

  return(ans=='y');
}

AskDirectQuestion ()
{
  (void) fprintf(stderr, "\n\
The currently running kernel purports to support the %s, but will not\n\
accept commands to configure it for diagnostics.",
		 model);
  return(y_or_n_p("Do you want to continue the tests by accessing the board directly?  "));
}



/*
 * Override the real thing in embed.c so we can work with/without our
 * kernel.
 */

caddr_t SlaveBufferPtr = NULL;
EmbCommArea *EmbCommAreaPtr = NULL;
static long *saved_comm_area = NULL;
static int comm_area_fd;
static int saved_enable_Ivory = 0;

MapEmbCommArea(device)
     char *device;
{
#if unix
  {
    IvoryConfig config;
    caddr_t ptr;
    
    /* Open merlin device, map in slave buffer */
    if ((comm_area_fd = open(device, O_RDWR)) >= 0) {
      /* broken kernel, device not there? */
      if (ioctl(comm_area_fd, IVORYGCONF, &config) == -1) {
	if (AskDirectQuestion())
	  goto brute_force;
	else
	  exit(1);
      } 
    
      if ((saved_enable_Ivory=config.enable_Ivory) == 1) {
	config.enable_Ivory = 0;
	if (ioctl(comm_area_fd, IVORYSCONF, &config) == -1) {
	  perror("IVORYSCONF");
	  exit(1);
	}
	if (verbose) (void)fprintf(stderr, "%s life support disabled\n", model);
      }

      running_our_kernel++;
      system_type = config.system_type;
      switch (system_type) {
      case SystemTypeUX400S:
	model = "UX400S";
	break;
      case SystemTypeUX1200S:
	model = "UX1200S";
	break;
      }
      slave_buffer_address = 0;
      slave_buffer_size = config.mmap_size;
      comm_area_base = config.comm_area_base;
      comm_area_size = config.comm_area_size;
      mailbox_address = config.mmap_size;
    } else {
    brute_force:
      if (!system_type && slave_buffer_address && mailbox_address)
 	switch (mailbox_address - slave_buffer_address) {
	case 000400000:		/* Merlin-I slave buffer size */
	case 000200000:		/* Merlin-I comm area size */
	  system_type = SystemTypeUX400S;
	  break;
	case 010000000:		/* Merlin-II slave buffer size */
	case 002000000:		/* Merlin-II comm area size */
	  system_type = SystemTypeUX1200S;
	  break;
	}
      if (!system_type) system_type = SystemTypeUX400S;
      switch (system_type) {
      case SystemTypeUX400S:
	model = "UX400S";
	if (!slave_buffer_address)
	  slave_buffer_address = MERLIN_VME_SLAVE_BUFFER_BASE;
	slave_buffer_size = MERLIN_VME_SLAVE_BUFFER_SIZE;
	comm_area_base = MERLIN_VME_COMM_AREA_BASE;
	comm_area_size = MERLIN_VME_COMM_AREA_SIZE;
	if (!mailbox_address)
	  mailbox_address = slave_buffer_address + (MERLIN_VME_MAILBOX_ADDRESS - MERLIN_VME_SLAVE_BUFFER_BASE);
	break;
      case SystemTypeUX1200S:
	model = "UX1200S";
	if (!slave_buffer_address)
	  slave_buffer_address = MERLIN_II_VME_SLAVE_BUFFER_BASE;
	slave_buffer_size = MERLIN_II_VME_SLAVE_BUFFER_SIZE;
	comm_area_base = MERLIN_II_VME_COMM_AREA_BASE;
	comm_area_size = MERLIN_II_VME_COMM_AREA_SIZE;
	if (!mailbox_address)
	  mailbox_address = slave_buffer_address + (MERLIN_II_VME_MAILBOX_ADDRESS - MERLIN_II_VME_SLAVE_BUFFER_BASE);
	break;
      }
    
     
      if ((comm_area_fd = open((device = "/dev/vme32d32"), O_RDWR)) < 0) {
	perror("Opening device");
	exit(1);
      }
    }

    if (verbose>2) (void)fprintf(stderr, "  Mapping %s using %s@0x%x\n", model, device, slave_buffer_address);
    /* we map one extra word, so we can peek below to make sure its really us */
    if ((ptr = mmap((char *)0, slave_buffer_size + 4,
  		    PROT_READ|PROT_WRITE, MAP_SHARED,
 		    comm_area_fd, slave_buffer_address)) == (caddr_t) -1) {
      perror("Mapping device");
      exit(1);
    }
    if (verbose>2) (void)fprintf(stderr, "  Mapping mailbox using %s@0x%x\n", device, mailbox_address);
    if ((mailbox = (long *)mmap((char *)0, mailbox_size,
  				PROT_READ|PROT_WRITE, MAP_SHARED,
				comm_area_fd, mailbox_address))
  	== (long *) -1) {
      if (!AskDirectQuestion()) {
  	exit(1);
      } else if ((mailbox = (long *)mmap((char *)0, mailbox_size,
					 PROT_READ|PROT_WRITE, MAP_SHARED,
					 open("/dev/vme32d32", O_RDWR), mailbox_address))
		 == (long *) -1) {
	perror("Opening mailbox");
	exit(1);
      }
    }

    /* sanity check that it's really our board */
    if (!running_our_kernel) {
      long junk;
      /* Merlin is there if slave buffer is there.  Check for existence of */
      /* first and last (longword) locations, and non-existence of last+1 */
      /* location. */
      if (!peek((long *) ptr, &junk) ||
 	  !peek((long *) (ptr + slave_buffer_size - 4), &junk) ||
	  /* Merlin-II responds to last+1 (sigh) */
	  ((system_type == SystemTypeUX400S || system_type == SystemTypeUX400G) &&
	   peek((long *) (ptr + slave_buffer_size), &junk))) {
 	(void) fprintf(stderr, "%s does not appear to be installed\n", model);
	exit(1);
      }
    }

    SlaveBufferPtr = ptr;
    ptr += comm_area_base;
    EmbCommAreaPtr = (EmbCommArea *) ptr;

    /* If the comm area looks live, save it*/
    if (EmbCommAreaPtr->version == 1 &&
	bcmp((char *) &EmbCommAreaPtr->identifier, "DBME", 4) == 0 &&
	EmbCommAreaPtr->system_type == system_type) {
      if (verbose>2) (void)fprintf(stderr, "  Saving embedded communication area\n");
      if((saved_comm_area = (long *)malloc(comm_area_size)) == NULL) {
	perror("Allocating memory to save comm area");
	exit(1);
      }
      bcopy(ptr, (char *)saved_comm_area, comm_area_size);
    }
  }
#endif
  
  return(comm_area_fd);
}



void UnMapEmbCommArea ()
{
  IvoryConfig config;

#ifdef unix
  /* deinit comm area, in case you inited it */
  EmbCommAreaPtr->identifier = 'GAID';	/* Set to 'DBME' while Life Support is running */

  if (saved_comm_area != NULL) {
    if (verbose>2) (void)fprintf(stderr, "  Restoring embedded communication area\n");
    bcopy((char *)saved_comm_area, (char *)EmbCommAreaPtr, comm_area_size);
  }

  if (running_our_kernel) {
    if (ioctl(comm_area_fd, IVORYGCONF, &config) == -1) {
      perror("Trying to reenable life support");
      exit(1);
    }

    if (saved_enable_Ivory) {
      config.enable_Ivory = saved_enable_Ivory;
      if (ioctl(comm_area_fd, IVORYSCONF, &config) == -1) {
	perror("IVORYSCONF");
	exit(1);
      }
      if (verbose) (void)fprintf(stderr, "%s life support re-enabled\n", model);
    }
  }


  /* Probably superfluous, but be neat */
  (void)munmap((caddr_t) EmbCommAreaPtr, slave_buffer_size + 4);
  EmbCommAreaPtr = NULL;
  (void)munmap((caddr_t) mailbox, mailbox_size);
  mailbox = NULL;
  (void)close(comm_area_fd);
  comm_area_fd = -1;
#endif
}



/*
 * Based on embinit.c, but does only enough to let us con the Boot ROM
 * into loading a Device PROM
 */


void InitializeCommunicationArea (EmbCommAreaPtr)
  register EmbCommArea* EmbCommAreaPtr;
{
  if (verbose>2) (void)fprintf(stderr, "  Initializing comm area...");
  /* Initialize miscellaneous communication area fields */
  EmbCommAreaPtr->identifier = 'ENON';	/* Set to 'DBME' while Life Support is running */
  EmbCommAreaPtr->version = 1;
  EmbCommAreaPtr->number_of_slots = ((Ptr)&EmbCommAreaPtr->pad0 - (Ptr)EmbCommAreaPtr) >>2; /* Ivory-visible portion only */
  EmbCommAreaPtr->comm_memory_size = MERLIN_VME_COMM_AREA_SIZE>>2;	/* words */

  EmbCommAreaPtr->system_type = system_type;
/*  EmbCommAreaPtr->host_major_version = config->host_major_version;*/
/*  EmbCommAreaPtr->host_minor_version = config->host_minor_version;*/

  EmbCommAreaPtr->siv.siv_word = -1;	/* Don't interrupt */

  EmbCommAreaPtr->channel_table = NullEmbPtr;
  EmbCommAreaPtr->boot_disk_channel = NullEmbPtr;
  EmbCommAreaPtr->command_channel = NullEmbPtr;
  EmbCommAreaPtr->clock_signal = -1;	/* No signal allocated until guest needs it */
/*  EmbCommAreaPtr->debug_level = config->debug_level; */

  /* Set up for dynamic allocation within the communication area */
  {
  int EmbCommAreaAllocPtr = sizeof(*EmbCommAreaPtr)>>2;
  int EmbCommAreaAllocSize = (MERLIN_VME_COMM_AREA_SIZE>>2) - EmbCommAreaAllocPtr;

  /* Allocate remaining portion of communication area to buffers */
  EmbCommAreaPtr->host_buffer_start = EmbCommAreaAllocPtr;
  EmbCommAreaPtr->host_buffer_size = 0;
  EmbCommAreaPtr->fep_buffer_start = EmbCommAreaAllocPtr + EmbCommAreaPtr->host_buffer_size;
  /* Give the whole thing to the FEP */
  EmbCommAreaPtr->fep_buffer_size = EmbCommAreaAllocSize;
  EmbCommAreaPtr->guest_buffer_start = EmbCommAreaAllocPtr;
  EmbCommAreaPtr->guest_buffer_size = 0;
  }
  EmbCommAreaPtr->identifier = 'DBME';
  if (verbose>2) (void)fprintf(stderr, " done\n");
}




/*
 * Beat on the unused part of slave buffer from Unix.  Ivory is halted
 * for this test.
 */

define_handler(slave)

static int SlaveBufferPasses = 0;
static int SlaveBufferFailures = 0;

/*
 * Currently the test ends on the first failure, if you
 * take out the throw, it will continue with other tests
 */
void slave_failure(address, written, read, reread)
     caddr_t address;
     u_long written, read, reread;
{
  SlaveBufferFailures++;
  if (verbose>1)
    (void)fprintf(stderr, "Failed transfer, slave address #x%08lX, wrote #x%08lX, read #x%08lX, reread #x%08lX\n",
		  address-SlaveBufferPtr, written, read, reread);
  throw(slave, -1);
}


void slave_write_32(address, words, value, increment)
     register long* address;
     long words;
     register long value;
     register long increment;
{
  register long* limit = address + words;

  if (increment)
    for (; address < limit; value += increment)
      *address++ = value;
  else for (; address < limit;)
    *address++ = value;
}

slave_compare_32(address, words, value, increment)
     register long* address;
     long words;
     register long value;
     long increment;
{
  register long* limit = address + words;
  register long read;

  if (increment) {
    for (; address < limit; value += increment)
      if ((read = *address++) != value)
	goto fail;
  }
  else for (; address < limit;) {
    if ((read = *address++) != value)
      goto fail;
  }
  return(0);
 fail:
  --address;
  slave_failure((caddr_t)address, (u_long)value, (u_long)read, (u_long)*address);
  return(1);
}

void slave_write_16(address, words, value, increment)
     register short* address;
     long words;
     long value;
     register long increment;
{
  register short* limit = address + words * (sizeof(long)/sizeof(short));
  register short* values;
  register int n;
  
  for (; address < limit; value += increment) 
    for(values = (short*)&value, n = 0; n < (sizeof(long)/sizeof(short)); n++)
      *address++ = *values++;
}

slave_compare_16(address, words, value, increment)
     register short* address;
     long words;
     long value;
     register long increment;
{
  register short* limit = address + words * (sizeof(long)/sizeof(short));
  register short* values;
  register short read;
  register int n;

  for (; address < limit; value += increment)
    for(values = (short*)&value, n = 0; n < (sizeof(long)/sizeof(short)); n++)
      if ((read = *address++) != *values++)
	goto fail;
  return(0);
 fail:
  --address;
  slave_failure((caddr_t)address, (u_long)*--values, (u_long)read, (u_long)*address);
  return(1);
}

void slave_write_8(address, words, value, increment)
     register char* address;
     long words;
     long value;
     register long increment;
{
  register char* limit = address + words * (sizeof(long)/sizeof(char));
  register char* values;
  register int n;
  
  for (; address < limit; value += increment) 
    for(values = (char*)&value, n = 0; n < (sizeof(long)/sizeof(char)); n++)
      *address++ = *values++;
}

slave_compare_8(address, words, value, increment)
     register char* address;
     long words;
     long value;
     register long increment;
{
  register char* limit = address + words * (sizeof(long)/sizeof(char));
  register char* values;
  register char read;
  register int n;

  for (; address < limit; value += increment)
    for(values = (char*)&value, n = 0; n < (sizeof(long)/sizeof(char)); n++)
      if ((read = *address++) != *values++)
	goto fail;
  return(0);
 fail:
  --address;
  slave_failure((caddr_t)address, (u_long)*--values, (u_long)read, (u_long)*address);
  return(1);
}


/* Test Patterns */
static long checkerboard[] = {0x55555555, 
				-0x55555556, 
				0x33333333, 
				-0x33333334, 
				0x0F0F0F0F, 
				-0xF0F0F10, 
				0x00FF00FF, 
				-0xFF0100, 
				0x0000FFFF, 
				-0x10000}; 
static long inverse_checkerboard[] = {-0x55555556, 
				0x55555555, 
				-0x33333334, 
				0x33333333, 
				-0xF0F0F10, 
				0x0F0F0F0F, 
				-0xFF0100, 
				0x00FF00FF, 
				-0x10000,
				0x0000FFFF}; 
static long walking_bit[] = {-0x80000000, 
			       0x40000000, 
			       0x20000000, 
			       0x10000000, 
			       0x08000000, 
			       0x04000000, 
			       0x02000000, 
			       0x01000000, 
			       0x00800000, 
			       0x00400000, 
			       0x00200000, 
			       0x00100000, 
			       0x00080000, 
			       0x00040000, 
			       0x00020000, 
			       0x00010000, 
			       0x00008000, 
			       0x00004000, 
			       0x00002000, 
			       0x00001000, 
			       0x00000800, 
			       0x00000400, 
			       0x00000200, 
			       0x00000100, 
			       0x00000080, 
			       0x00000040, 
			       0x00000020, 
			       0x00000010, 
			       0x00000008, 
			       0x00000004, 
			       0x00000002, 
			       0x00000001};

long zero[] = {0};
long minus_one[] = {-1};

struct pattern_list {
  char* name;
  long* patterns;
  long increment;
  int npatterns;
} pattern_list[] = {{"checkerboard", checkerboard, 0L, sizeof(checkerboard)/sizeof(checkerboard[0])},
		      {"inverse checkerboard", inverse_checkerboard, 0L, sizeof(inverse_checkerboard)/sizeof(inverse_checkerboard[0])},
		      {"walking bit", walking_bit, 0L, sizeof(walking_bit)/sizeof(walking_bit[0])},
		      {"incrementing value", zero, 1L, 1},
		      {"decrementing value", minus_one, -1L, 1}};

#define NPATTERNS (sizeof(pattern_list)/sizeof(pattern_list[0]))

struct test_list {
  char* name;
  void (*write)();
  int (*compare)();
} test_list[] = {{"32", slave_write_32, slave_compare_32},
		   {"16", slave_write_16, slave_compare_16},
		   {"8", slave_write_8, slave_compare_8}};

#define NTESTS (sizeof(test_list)/sizeof(test_list[0]))

static caddr_t buffer_start()
{
  register EmbCommArea* ep = EmbCommAreaPtr;
  long min = ep->host_buffer_start;

  if (min > ep->guest_buffer_start)
    min = ep->guest_buffer_start;
  if (min > ep->fep_buffer_start)
    min = ep->fep_buffer_start;

  return(HostPointer(min));
}

static long buffer_size()
{
  register EmbCommArea* ep = EmbCommAreaPtr;

  return(ep->host_buffer_size +
	 ep->guest_buffer_size +
	 ep->fep_buffer_size);
}

/*
 * Currently only test the unused portion of the slave buffer, for
 * fear that Ivory would start doing wierd things if we randomly write
 * into the comm area.  Later: also test slave buffer above
 * host_buffer_start, which being just buffers should be safe to
 * clobber if Ivory is stopped (waiting for a device-prom).  At least
 * one board failed in only the top half of its memory.
 */

void SlaveBufferTest ()
{
  caddr_t slave_low = SlaveBufferPtr;
  long low_size = comm_area_base/sizeof(long);
  caddr_t slave_high = buffer_start();
  long high_size = buffer_size();
  struct pattern_list* pl;
  struct test_list* wt;
  struct test_list* ct;
  int failures = 0;

  SlaveBufferPasses++;
  if (!handle(slave)) {
    if(verbose>1) (void)fprintf(stderr, "Starting Slave-buffer tests...\n");
    for(pl = pattern_list; pl < &pattern_list[NPATTERNS]; pl++)
      for(wt = test_list; wt < &test_list[NTESTS]; wt++)
	for(ct = test_list; ct < &test_list[NTESTS]; ct++) {
	  register long* pattern;

	  if (verbose>2)
	    (void)fprintf(stderr, "  %sx%s %s... ", wt->name, ct->name, pl->name);

	  for(pattern = pl->patterns; pattern < &pl->patterns[pl->npatterns]; pattern++) {
	    if (low_size > 0) {
	      (*wt->write)((long*)slave_low, low_size, *pattern, pl->increment);
	      (*ct->compare)((long*)slave_low, low_size, *pattern, pl->increment);
	    }
	    (*wt->write)((long*)slave_high, high_size, *pattern, pl->increment);
	    (*ct->compare)((long*)slave_high, high_size, *pattern, pl->increment);
	  }

	  if (verbose>2) (void)fprintf(stderr, "done.\n");
	}
  }
  else
    failures++;

  if (failures)
    (void)fprintf(stderr, "Slave-buffer test failed.\n");
  else if (verbose)
    (void)fprintf(stderr, "Slave-buffer test passed.\n");
}



/*
 * Set up enough of the comm area to send an init and see that the
 * BootROM responds with a reset request.
 */

define_handler(init)

static char* reset_request_name[] = {"Are You There", "Boot ROM", "Device PROM", "FEP", "none", "Lisp"};
#define decode_reset(state) (((int)AreYouThereResetRequest<=(int)(state)&&(int)(state)<=(int)LispResetRequest)?reset_request_name[(int)(state)-(int)AreYouThereResetRequest]:"?")
static char* boot_status_name[] = {"Initialize registers", "Initialize Stack Backing", "Build Initial Frame", "Load Device Prom", "Start Device Prom", "Initialize Interleave", "Initialize Memory", "Fatal Error", "xGYY", "xGYx", "xGxY", "xGxx", "xxYY", "xxYx", "xxxY", "xxxx"};
#define decode_status(state) ((0<=state&&state<=0xf)?boot_status_name[0xf-(state)]:"?")

static int InitTestPasses = 0;
static int InitTestFailures = 0;

void send_init ()
{
  register EmbCommArea *ptr = EmbCommAreaPtr;
  register long* mb = mailbox;
    

  *mb = MERLIN_VME_MAILBOX_INIT;
  ptr->reset_request = NoResetRequest;
  if((verbose>2)) {
    (void)fprintf(stderr, "  Spy: %d; Status: %s; Reset: %s\n",
		  ptr->spy_status,
		  decode_status(ptr->run_lights),
		  decode_reset(ptr->reset_request));
    (void)fprintf(stderr, "  INIT\n");
  } 
  usleep(1000);
  *mb = 0;
}

void InitTest ()
{
  enum ResetRequest last = NoResetRequest, reset;
  long last_spy = 0, spy;

  catch(SIGALRM, init);
  (void)alarm(5);				/* give time for memory test */	
  
  InitTestPasses++;
  if(!handle(init)) {
    register EmbCommArea *ptr = EmbCommAreaPtr;

    if (verbose>1) (void)fprintf(stderr, "Starting Init test...\n");
    send_init();
    do {
      reset = ptr->reset_request;
      ptr->reset_request = NoResetRequest;
      spy = ptr->spy_status;
      if(last != reset || last_spy != spy) {
	if (verbose>2) 
          (void)fprintf(stderr, "  Spy: %d; Status: %s; Reset: %s\n",
		        spy,
		        decode_status(ptr->run_lights),
		        decode_reset(reset));

        alarm(5);				/* give more time, as long as progressing */
	last = reset;
        last_spy = spy;
      }
    } while (reset != BootResetRequest);

    if (verbose) (void)fprintf(stderr, "Init test passed\n");
  } else {
    (void)fprintf(stderr, "Init test timed out\n");
    InitTestFailures++;
  }
  
  (void)alarm(0);
  uncatch(SIGALRM);
}



LoadPROM (prom, prom_limit, prom_name) 
  char* prom;
  register char* prom_limit;
  char* prom_name;
{
  register EmbCommArea* ptr = EmbCommAreaPtr;
  register char* pp = prom;
  bool timeout;

  /* Timeout */
  catch(SIGALRM, init);
  (void)alarm(5);
  
  if(!(timeout = handle(init))) {
    enum ResetRequest reset;

    send_init();

    /* Stay in this loop 'till PROM comes up and writes a FEP reset */
    do
      if ((reset = ptr->reset_request) != NoResetRequest) {
	(void)alarm(5);				/* reset timeout for each command */
	switch(reset) {
	case BootResetRequest:
	  if (verbose>2)
	    (void)fprintf(stderr, "  Rewinding %s PROM\n", prom_name);
	  pp = prom;
	  break;

#define min(a b) ((a)<(b)?(a):(b))
	case DevicePROMResetRequest: 
	  { register int nbytes = min(ptr->fep_buffer_size<<2, prom_limit-pp);
	    register char* buffer = (caddr_t) HostPointer(ptr->fep_buffer_start);

	    if (verbose>2)
	      (void)fprintf(stderr, "  Sending %d %s PROM bytes\n", nbytes, prom_name);
	    for(; nbytes--;)
	      *buffer++ = *pp++;
          }
	  break;

	case FEPResetRequest:
          if (verbose>2)
	    (void)fprintf(stderr, "  %s PROM %d.%d running\n",
	                  prom_name,
			  ptr->fep_major_version,
			  ptr->fep_minor_version);
	  break;

	case AreYouThereResetRequest:
          if (verbose>2)
	    (void)fprintf(stderr, "  Boot ROM running\n");
	  break;

        default:
          (void)fprintf(stderr, "Invalid Reset Request %d, loading %s\n", 
			ptr->reset_request,
			prom_name);
	  throw(init, -1);
	}
	ptr->reset_request = NoResetRequest;
      }  while (reset != FEPResetRequest);

  }

  (void)alarm(0);
  uncatch(SIGALRM);

  if(timeout && (verbose>2)) (void)fprintf(stderr, "  %s PROM failed to load\n", prom_name);
  
  return(timeout?FALSE:TRUE);
}



/*
 * Load the diagnostic prom and start it going.
 */

static int DiagnosticPasses = 0;
static int DiagnosticFailures = 0;

/* A transcription of the .flod generated from SYS:IFEP;IVORY-DIAGNOSTIC-PROM */
char diagnostic_prom[] = {
'I', 'F', 'E', 'P', '\040', 'p', 'r', 'o', 'm', '\073', '\001', '\000', '\000', '\000', '\020',
'\004', '\370', '\001', '\000', '\000', '\000', '\073', '\001', '\000', '\000', '\010', 'D',
'\000', '\000', '\000', '\000', '\301', '\000', '\370', 'p', '\000', '\000', '\000', '\033',
'\000', '\000', 'P', 'C', 'B', 'O', 'O', 'T', 'H', '\055', 'E', 'M', 'B', 'H', '\055', 'C',
'O', 'M', 'H', 'M', 'U', 'N', 'I', 'H', 'C', 'A', 'T', 'I', 'H', 'O', 'N', '\055', 'A', 'H',
'R', 'E', 'A', '\000', 'H', '\000', '\301', '\000', '\370', 'X', '\014', '\301', '\000',
'\370', '\134', '\012', '\000', '\004', '\000', '\003', '\014', '\301', '\000', '\370', '\134',
'\002', '\374', '\011', '\000', '\360', '\001', '\000', '\376', '\377', '\304', '\340', '\340',
'\021', '\214', '\073', '\004', '\340', '\000', '\330', '\364', '\000', '\100', '\360', '\217',
'\240', '\000', '\000', '\376', '\377', '\304', '\000', '\066', '\375', '\033', '\376', '\010',
'\301', '\000', '\370', 'U', '\020', '\000', '\000', 'P', 'C', 'I', 'V', 'O', 'R', 'H', 'Y',
'\055', 'D', 'I', 'H', 'A', 'G', 'N', 'O', 'H', 'S', 'T', 'I', 'C', 'H', '\003', '\000',
'\000', 'P', 'C', 'N', 'I', 'L', '\000', 'H', '\024', '\301', '\000', '\370', 'X', '\037',
'\301', '\000', '\370', '\134', 'S', '\000', '\004', '\000', '\003', '\037', '\301', '\000',
'\370', '\134', '\002', '\374', '\011', '\000', '\360', '\011', '\301', '\000', '\370', '\356',
'\001', 'q', '\011', '\000', '\064', '\017', 'c', '\302', '\340', '\064', '\377', '\323',
'\006', '\024', '\062', '\005', '\364', '\004', '\304', '\365', '\011', '\301', '\000', '\370',
'\252', '\003', '\140', '\372', '\253', '\066', '\002', '\000', '\045', '\214', '\071', '\073',
'\001', '\000', '\000', '\310', '\000', 'r', '\012', '\000', '\064', '\012', 'c', '\002',
'\314', '\071', '\002', '\000', '\201', '\214', '\071', '\060', '\070', '\001', '\310', '\061',
'\374', '\343', '\010', '\000', '\064', '\040', 'c', '\376', '\317', '\071', '\002', '\000',
'\201', '\214', '\071', '\060', '\070', '\001', '\310', '\061', '\374', '\343', '\010', '\034',
'\064', '\000', '\003', '\021', '\020', '\076', '\376', '\252', '\015', '\000', '\064', '\001',
'c', '\376', '\337', '\071', '\003', '\000', '\301', '\340', '\064', '\000', 'r', '\364',
'\217', '\063', '\003', '\000', '\011', '\214', '\071', '\060', '\070', '\015', '\000', '\064',
'\003', 'c', '\302', '\340', '\064', '\003', '\000', '\021', '\214', '\071', '\060', '\070',
'\015', '\000', '\064', '\005', 'c', '\302', '\340', '\064', '\003', '\000', '\051', '\214',
'\071', '\060', '\070', '\015', '\000', '\064', '\013', 'c', '\302', '\340', '\064', '\003',
'\000', '\061', '\214', '\071', '\060', '\070', '\015', '\000', '\064', '\015', 'c', '\302',
'\340', '\064', '\003', '\000', '\005', '\214', '\071', '\000', 'w', '\032', '\304', '\066',
'\003', '\000', '\031', '\214', '\071', '\206', '\264', '\015', '\000', '\064', '\006', 'c',
'\032', '\324', '\066', '\006', 'h', '\034', '\240', '\061', '\010', 'h', '\044', '\240',
'\061', '\004', '\000', '\021', '\014', '\074', '\004', '\200', 'C', '\020', '\065', '\020',
'D', 'A', '\020', '\065', '\020', 'D', '\075', '\000', '\064', '\006', '\300', '\026', '\000',
'\063', '\005', '\214', '\075', '\000', '\064', '\012', '\300', 'R', '\200', '\063', '\020',
'\000', '\035', '\000', '\073', '\005', '\300', '\024', '\060', '\066', '\020', '\000', '\055',
'\000', '\073', '\015', '\340', 'D', '\000', '\064', '\010', '\300', '\026', '\000', '\063',
'\005', '\214', 'E', '\000', '\064', '\014', '\300', '\032', '\200', '\063', '\011', '\320',
'F', 'P', '\063', '\005', '\214', '\065', '\100', '\073', '\016', '\324', '\014', '\000',
'\064', '\016', 'c', '\076', '\320', '\071', '\003', '\000', '\075', '\214', '\071', '\020',
't', '\016', '\000', '\064', '\020', 'c', 'F', '\320', '\071', '\003', '\000', 'E', '\214',
'\071', '\022', 't', '\006', '\314', '\061', '\373', '\252', '\015', '\000', '\063', '\001',
's', '\030', '\300', '\067', '\003', '\000', '\301', '\340', '\064', '\000', 'r', '\004',
'\017', '\063', '\000', 's', '\070', '\000', '\064', '\006', '\265', '\371', '\253', '\066',
'\003', '\000', 'I', '\214', '\071', '\004', 't', '\016', '\000', '\064', '\023', 'c', '\026',
'\320', '\071', '\367', '\252', '\075', '\016', '\063', '\003', '\000', '\005', '\214', '\071',
'\377', 'w', '\016', '\000', '\064', '\060', '\070', '\001', '\310', '\061', '\210', '\343',
'\360', '\317', '\067', '\033', '\301', '\000', '\370', 'U', '\336', '\247', '\000', '\000',
'\037', '\301', '\000', '\370', '\377', '\377', '\377', '\377'
};

struct contention_buffer {
  long host_flag;
  long guest_flag;
  long guest_0, guest_1, guest_2, guest_3;
  long share_0, share_1, share_2, share_3;
  long host_0, host_1, host_2, host_3;
  long seen_0, seen_1, seen_2, seen_3;
  long guest_passes, host_values;
} *contention_buffer;

void LoadDiagnosticPROM () 
{

  /* Set up buffer */
  contention_buffer = (struct contention_buffer *) HostPointer(EmbCommAreaPtr->fep_buffer_start);

  /* Since INIT is broken, we have to try alot */
  while(!LoadPROM(diagnostic_prom, &diagnostic_prom[sizeof(diagnostic_prom)], "Diagnostic"));

  /* Halt program initially */
  contention_buffer->host_flag = -1;
  contention_buffer->guest_flag = -1;
}


define_handler(diagnose)

/*
 * The host and guest both beat on a 4-word area of the slave buffer
 * writing different values and looking for either value; halting on a
 * miscompare.  The host tells the guest what values each will write via
 * other slots in the buffer; the guest reports back via yet more slots.
 */
ContentionTest (pattern, name)
  long * pattern;
  char * name;
{
  register struct contention_buffer* cb = contention_buffer;
  register long* rwp;
  register long* gp = &pattern[0];
  register long* hp = &pattern[7];
  register long read;
    register int guest_values = 0;
    int host_passes = 0;
    int guest_error;
    int error_seen = 0;

    /* Initialize values */
    cb->guest_0 = *gp++;
    cb->guest_1 = *gp++;
    cb->guest_2 = *gp++;
    cb->guest_3 = *gp;
   
    cb->host_3 = *hp;
    cb->host_2 = *--hp;
    cb->host_1 = *--hp;
    cb->host_0 = *--hp;

    catch(SIGALRM, diagnose);
    (void)alarm(5);

    /* run each pattern for 5 seconds */
    if (!handle(diagnose)) {
      /* start the test */
      cb->host_flag = 0;
      /* wait for guest to start */
      while(cb->guest_flag);
      if(verbose>2)
        (void)fprintf(stderr, "  Guest started\n");

      do {
	rwp = &cb->share_3;
        hp = &pattern[7];
	gp = &pattern[3];     /* from end!!! */

	*rwp = *hp; *--rwp = *--hp; *--rwp = *--hp; *--rwp = *--hp;

	rwp = &cb->share_3;
	hp = &pattern[7];
        host_passes += 4;
      } while ((((read = *rwp) == *hp) || (guest_values++, read == *gp)) &&
	       ((--gp, (read = *--rwp) == *--hp) || (guest_values++, read == *gp)) &&
	       ((--gp, (read = *--rwp) == *--hp) || (guest_values++, read == *gp)) &&
	       ((--gp, (read = *--rwp) == *--hp) || (guest_values++, read == *gp)) &&
	       (!cb->guest_flag));

      guest_error = cb->guest_flag;
      /* stop the guest */
      cb->host_flag = -1;
      /* wait for guest to report */
      while(!cb->guest_flag);

      DiagnosticFailures++;
      error_seen++;
      (void)fprintf(stderr, "Remote Diagnostic test failed %s\n", name);
      if (verbose>1) {
	if(guest_error) 
	  (void)fprintf(stderr, "Miscompare on guest side\n");
	else
	  (void)fprintf(stderr, "Miscompare on host, slot #x%08lX, wrote #x%08lX (or #x%08lX), read #x%08lX, reread #x%08lX\n",
			rwp-&cb->share_0, *hp, *gp, read, *rwp);
	/* maybe >1? */
	if (verbose>1) {
	  int i = 4;
	  long* sp;

	  for (hp = &cb->host_0,
	       gp = &cb->guest_0,
	       rwp = &cb->share_0,
	       sp = &cb->seen_0; i--; hp++, gp++, rwp++, sp++)
	    (void)fprintf(stderr,
			  "  Host value: #x%08lX; Share value: #x%08lX; Guest value: #x%08lX; Guest seen #x%08lX\n",
			  *hp, *rwp, *gp, *sp);
	}
      }
    }
    else {
      /* stop the guest */
      cb->host_flag = -1;
      /* wait for guest to report */
      while(!cb->guest_flag);
    }
    if(verbose>1 || (verbose && error_seen))
      (void)fprintf(stderr, "For %s, host passes: %d, (saw %d guest values); Guest passes: %d, (saw %d host values)\n", name, host_passes, guest_values, cb->guest_passes, cb->host_values);

    (void)alarm(0);
    uncatch(SIGALRM);
    return(!error_seen);
  }

void DiagnosticPROM ()
{
  struct pattern_list* pl;
  long * pattern;

  if(verbose>1) (void)fprintf(stderr, "Starting Remote Diagnostic test ...\n");
  LoadDiagnosticPROM();

  DiagnosticPasses++;

  for (pl = pattern_list; pl < &pattern_list[NPATTERNS]; pl++)
    for (pattern = pl->patterns; &pattern[8] <&pl->patterns[pl->npatterns]; pattern += 8)
     if (!ContentionTest(pattern, pl->name))
	return;

  if (verbose)
    (void)fprintf(stderr, "Remote Diagnostic test passed\n");
}




static int TestPasses = 0;

void SummarizeResults()
{

  (void)fprintf(stderr,"\n\
Completed %d pass%s of tests.", TestPasses, TestPasses!=1?"es":"");
  if(TestPasses>0)
    (void)fprintf(stderr,"  Summary results:\n");
  if(InitTestPasses>0)
    (void)fprintf(stderr," Init Test:         %8d Failures (%3d%%)\n",
		  InitTestFailures, InitTestFailures*100/InitTestPasses);
  if(SlaveBufferPasses>0)
    (void)fprintf(stderr," Slave Buffer Test: %8d Failures (%3d%%)\n",
		  SlaveBufferFailures, SlaveBufferFailures*100/SlaveBufferPasses);
  if(DiagnosticPasses>0)
    (void)fprintf(stderr," Remote Diagnostic: %8d Failures (%3d%%)\n",
		  DiagnosticFailures, DiagnosticFailures*100/DiagnosticPasses);
  (void)fprintf(stderr, "\n");
}
	  


/*
 * See "usage" below for documentation of options
 */

define_handler(interrupt)

main (argc, argv)
     int argc;
     char *argv[];
{
  char *device = NULL;
  int continuous = 0;
  extern char *optarg;
  extern int optind;
  int c, bogus = 0, selected = 0, halt = 0;
  int init = 0, buffer = 0, remote = 0;
  extern char* getenv();

  while ((c = getopt(argc, argv, "bchim:rs:t:v")) != -1)
    switch (c)
      {
      case 'b':
      case 'i':
      case 'r':
        selected++;
        switch (c)
	{
	case 'b': buffer++; break;
	case 'i': init++; break;
        case 'r': remote++; break;
	}
	break;
      case 'c':
	continuous++;
	break;
      case 'h':
        halt++;
	break;
      case 'm':
	mailbox_address = strtol(optarg, NULL, 0);
	break;
      case 's':
	slave_buffer_address = strtol(optarg, NULL, 0);
	break;
      case 't':
	if (!strcmp(optarg, "UX400S"))
	  system_type = SystemTypeUX400S;
	else if (!strcmp(optarg, "UX1200S"))
	  system_type = SystemTypeUX1200S;
	else bogus++;
	break;
      case 'v':
	verbose++;
	break;
      case '?':
	bogus++;
	break;
      }

  if ((optind < argc) && (access(argv[optind], R_OK|W_OK) == 0)) 
    device = argv[optind++];
  
  if (bogus || (optind < argc)) {
    (void) fprintf(stderr, "usage: %s -bchirv -t <model> -m <addr> -s <addr> [device]\n\
  -c         run test in continuous mode\n\
  -h         halt on error\n\
  -s  <addr> override default address for slave-buffer\n\
  -m  <addr> override default address for mailbox\n\
  -t  {UX400S | UX1200S}\n\
  -v         jabber, -vv for test details, -vvv for debugging info\n\
  The test options all default to on, if you select any though,\n\
  only the ones you mention will be done\n\
  -i         run init test\n\
  -b         run buffer test\n\
  -r         run remote test\n\
", argv[0]);
    exit(2);
  }

  /* Defaults */
  if (device == NULL) {
    if ((device = getenv("IVORY")) == NULL)
      device = DEFAULT_GUEST();
  }

  /* Open device, map in communication area */
  ivory_fd = MapEmbCommArea(device);

  InitializeCommunicationArea(EmbCommAreaPtr);

  {

    catch(SIGINT, interrupt);
    catch(SIGQUIT, interrupt);
    if (verbose) (void)fprintf(stderr, "Starting %stesting\n", continuous?"continuous ":"");
    
    if (!handle(interrupt))
      do {
	TestPasses++;
	if (!selected || init) InitTest();
	if (InitTestFailures && halt) throw(interrupt, -1);
	if (!selected || buffer) SlaveBufferTest();
        if (SlaveBufferFailures && halt) throw(interrupt, -1);
	if (!selected || remote) DiagnosticPROM();
        if (DiagnosticFailures && halt) throw(interrupt, -1);
	if (verbose && continuous) (void)fprintf(stderr, "Finished pass %d\n", TestPasses);
      } while (continuous);
    uncatch(SIGINT);
    uncatch(SIGQUIT);
  }
  if (verbose) (void)fprintf(stderr, "Done testing\n");
  SummarizeResults();
  UnMapEmbCommArea();
  exit(0);
}
