/* -*-Mode: C; Package: C-user -*- */
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1990 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * Utilities for UX installation
 *
 * $Log:	unix-config.c,v $
 * Revision 2.5.1.2  93/02/01  12:57:17  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:53:16  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:44:06  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/unix-config.c,v 2.5.1.2 93/02/01 12:57:17 kaufman Rel $";
#endif

#include <sys/types.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <net/if.h>
#include <netinet/in.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include "install.h"

static void determine_ux_interface (uxif, ifr, nif, ux_name, ux_n)
     char *uxif[];
     struct ifreq *ifr;
     int nif;
     char *ux_name;
     int ux_n;
{
  char buf[BUFSIZ];
  int i, j;

  while (uxif[0] == NULL) {
    for (i = j = 0; j < nif && i < MAX_INTERFACE_N; j++) {
      if (ifr[j].ifr_name[0] == '\0')
	continue;

      Sprintf(buf, "Do you wish %s #%d to use network interface %s",
	      ux_name, ux_n, ifr[j].ifr_name);
      if (y_or_n_p(buf, 1))
	uxif[i++] = strdup(ifr[j].ifr_name);
    }

    if (uxif[0] == NULL) {
      Sprintf(buf, "If %s #%d can't use any network interfaces, it will only be able\nto communicate with this Sun and any other UX400Ses or UX1200Ses\nembedded in this Sun.  Okay",
	      ux_name, ux_n);
      if (y_or_n_p(buf, 0))
	break;

#ifdef notdef			/* We allow no network interfaces now */
      Fprintf(stderr, "\n>> You must select a network interface for %s %s\n",
	      "the", ux_name);
#endif
    }
  }
  uxif[i] = NULL;
}

static void determine_interface_list(params)
     struct params *params;
{
  int h, i, j, fd, nif, available_interfaces = 0;
  struct ifconf ifc;
  struct ifreq ifr[MAX_INTERFACE_N];

  /* Get interface list */
  if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    i = errno;
    Fprintf(stderr, "Unable to get interface list:\n\t");
    errno = i;
    perror("socket");
    return;
  }
  ifc.ifc_len = sizeof(ifr);
  ifc.ifc_req = ifr;
  if (ioctl(fd, SIOCGIFCONF, &ifc) == -1) {
    i = errno;
    Fprintf(stderr, "Unable to get interface list:\n\t");
    errno = i;
    perror("SIOCGIFCONF");
    return;
  }
  (void) close(fd);
  nif = ifc.ifc_len/sizeof(struct ifreq);

  /* Don't offer loopback or emb interfaces */
  for (j = 0; j < nif; j++) {
    if ((!strncmp(ifr[j].ifr_name, "lo", 2) &&
	 isdigit(ifr[j].ifr_name[2])) ||
	(!strncmp(ifr[j].ifr_name, "emb", 3) &&
	 isdigit(ifr[j].ifr_name[3])))
      ifr[j].ifr_name[0] = '\0';
    else
      available_interfaces++;
  }

  if (!available_interfaces) {
    Fprintf(stderr, "No suitable network interfaces have been configured!\n");
    return;
  }

  /* Determine a single interface for each UX board.
   * Don't support multiple interfaces unless/until create-fep-partition does.
   */
  for (i = 0; i < params->n_ux400_boards; i++)
    determine_ux_interface(params->ivory_interfaces[i], ifr, nif,
			   UX400_NAME, i);
  for (h = 0, i = MIN_UX1200_DEVNO; h < params->n_ux1200_boards; h++, i++)
    determine_ux_interface(params->ivory_interfaces[i], ifr, nif,
			   UX1200_NAME, h);
}

static void determine_ux_partition_file_name(part, ux_devno, ux_n, ux_name)
     struct partition_params *part;
     int ux_devno, ux_n;
     char *ux_name;
{
  char *cp, buf[BUFSIZ], buf0[BUFSIZ];
  arg arg0, arg1;

  if (part->filename == NULL) {
    arg0.integer = S_IFREG;
    arg1.integer = F_OK;
    Sprintf(buf0, "/ivory/fep%d", ux_devno);
    Sprintf(buf, "Where is the partition file for %s #%d", ux_name, ux_n);
    cp = buf;
    part->filename = accept_string(cp, buf0, check_file, &arg0, &arg1);
  }
}

void determine_unix_startup_parameters (params)
     struct params *params;
{
  char buf[BUFSIZ], buf0[BUFSIZ];
  char *ux_kernel_name;
  arg arg0, arg1;
  int h, i, tmp;
  int check_kernel_name();

  terpri();

  /* Determine whether to install a kernel.
   * Doing this requires write access to the root directory,
   * and read access to the kernel build directory in question.
   */
  while (1) {
    Sprintf(buf, "Do you wish to install %s %s kernel",
		   params->configure_kernel_p ? "the" : "a", UX_NAME);
    if (!(params->install_kernel_p = y_or_n_p(buf, 1)))
      break;

    /* Need write access to the root directory */
    if (access("/", W_OK)) {
      tmp = errno;
      Fprintf(stderr, ">> Can't install the kernel, because you don't have\n");
      Fprintf(stderr, "   permission to write to the root directory:\n\t");
      errno = tmp;
      perror("/");
      continue;
    }

    /* If we aren't configuring the kernel we're going to install, find
     * out which one to install and where to get it. */
    if (!params->configure_kernel_p) {
      /* Determine kernel build directory */
      determine_kernel_build_directory(params);

      /* Determine kernel name.  Prefer a UX kernel.
       * IWBNI there was a way to arrange these questions so we would know n_boards
       * by now, and use that information to generate the kernel name, but it doesn't
       * flow nicely if you do that. */
      Sprintf(buf, "%s/%s/conf/%s",
		     params->kernel_build_directory, params->kernel_architecture, UX_NAME);
      if (!access(buf, R_OK))
	ux_kernel_name = UX400_NAME;
      else
	ux_kernel_name = NULL;
      arg0.string = params->kernel_build_directory;
      arg1.string = params->kernel_architecture;
      ux_kernel_name = accept_string("What kernel do you wish to install",
				     ux_kernel_name, check_kernel_name,
				     &arg0, &arg1);
      params->ux_kernel_name = ux_kernel_name;
    
      /* Make we sure we can read the kernel in question */
      Sprintf(buf, "%s/%s/%s/vmunix",
		     params->kernel_build_directory, params->kernel_architecture,
		     params->ux_kernel_name);
      Sprintf(buf0, "%s/%s/%s/vmunix_small",
		     params->kernel_build_directory, params->kernel_architecture,
		     params->ux_kernel_name);
      i = 0;
      if (access(buf, R_OK) &&
	  (errno == ENOENT && (i = errno) && access(buf0, R_OK))) {
	tmp = errno;
	fputs("\n>> Can't install that kernel, because you don't have permission to read it:\n\t",
	      stderr);
	if (i) {
	  errno = i;
	  perror(buf);
	  putc('\t', stderr);
	  fflush(stderr);
	  errno = tmp;
	  perror(buf0);
	} else {
	  errno = tmp;
	  perror(buf);
	}
	continue;
      }
    }
    break;
  }

  /* Determine whether to create ivory config files */
  while (1) {
    Sprintf(buf, "Do you wish to create the %s configuration file(s)", UX_NAME);
    if (!(params->configure_ivory_life_p = y_or_n_p(buf, 1)))
      break;

    /* Need write access to /etc */
    if (access("/etc", W_OK)) {
      tmp = errno;
      Fprintf(stderr,
		     "\n>> Can't update the %s configuration files, because you don't\n",
		     UX_NAME);
      Fprintf(stderr, "   have permission to write to /etc:\n\t");
      errno = tmp;
      perror("/etc");
      continue;
    }
    break;
  }

  /* Determine location of ifep directory, interface and partition lists. */
  if (params->configure_ivory_life_p) {
    require_n_boards(params, FALSE);
    require_ifep_directory(params, FALSE);
    determine_interface_list(params);
    for (i = 0; i < params->n_ux400_boards; i++)
      determine_ux_partition_file_name(params->ivory_partitions[i], i, i, UX400_NAME);
    for (h = 0, i = MIN_UX1200_DEVNO; h < params->n_ux1200_boards; h++, i++)
       determine_ux_partition_file_name(params->ivory_partitions[i], i, h, UX1200_NAME);
  }

  /* Determine whether to create /dev/ivoryN */
  while (1) {
    Sprintf(buf, "Do you wish to create/update the %s device files in /dev", UX_NAME);
    if (!(params->device_special_files_p = y_or_n_p(buf, 1)))
      break;

    if (getuid()) {
      Fprintf(stderr,
		     "\n>> Can't update the %s device files, because you're not root.\n",
		     UX_NAME);
      continue;
    }
    break;
  }

  if (params->device_special_files_p && !params->configure_kernel_p) {
    require_n_boards(params, FALSE);
#ifdef FINDDEVNO
    Sprintf(buf, "%s/%s/%s/vmunix",
	    params->kernel_build_directory, params->kernel_architecture,
	    params->ux_kernel_name);
    if ((tmp = find_ivory_device_number(buf)) == -1)
      (void) strcpy(buf, "72");
    else {
      Printf("Found ivory dev no %d\n", tmp);
      Sprintf(buf, "%d", tmp);
    }
#else
    (void) strcpy(buf, "104");
#endif
    arg0.integer = 1;
    arg1.integer = 1000;
    params->ivory_major_device_number =
      atoi(accept_string("What is the Ivory major device number", buf,
			 check_integer, &arg0, &arg1));
  }

  /* Determine whether to hack rc.local */
  while (1) {
    Sprintf(buf, "Do you wish /etc/rc.local to start %s life support", UX_NAME);
    if (!(params->configure_startup_p = y_or_n_p(buf, 1)))
      break;

    /* Need write access to /etc */
    if (access("/etc", W_OK)) {
      tmp = errno;
      (void) fputs("\n>> Can't update /etc/rc.local, because you don't have permission to write to /etc:\n\t", stderr);
      errno = tmp;
      perror("/etc");
      continue;
    }
    break;
  }
  if (params->configure_startup_p) {
    require_etc_directory(params, FALSE);
    require_n_boards(params, FALSE);
  }


  /* Determine whether to install and boot the kernel */
  while (1) {
    if (!(params->boot_p =
	  y_or_n_p("Do you wish to reboot UNIX after installation", 1)))
      break;

    if (getuid()) {
      (void) fputs("\n>> Can't reboot UNIX, because you're not root.\n", stderr);
      continue;
    }
    break;
  }
}



/* State machine for rc.local */

rc_next_state (state, line)
     int state;
     char *line;
{
  static int first_time = TRUE;
  static char life[BUFSIZ];

  if (first_time) {
    Sprintf(life, "%s/ivory-life", ux_installation_params.relocate_etc_directory);
    first_time = FALSE;
  }

  switch (state) {
  case 0:
    if (token_match(line, "if", "[", "-f", life, NULL) ||
	token_match(line, "if", "[", "-f", "/etc/ivory", NULL) ||
	token_match(line, "if", "[", "-c", "/dev/ivory", NULL))
      return(10);

  case 10:
    if (state == 10)
      if (token_match(line, "fi", NULL))
	return(0);
      else if (token_match(line, "if", "[", NULL))
	/* Nested if */
	return(20);

  case 20:
    if (state == 20 && token_match(line, "fi", NULL))
      return(10);

  case 10000:
    return(state);
  }
  /*NOTREACHED*/
}

static void rc_start_ivory_life (devno, out)
     int devno;
     FILE *out;
{
  Fprintf(out, "    if [ -f /etc/ivory%d.conf -a -c /dev/ivory%d ]; then\n",
	  devno, devno);
  Fprintf(out, "\t(echo -n ' ivory%d')\t\t\t\t>/dev/console\n", devno);
  Fprintf(out, "\t%s/ivory-life /dev/ivory%d\t\t>/dev/console 2>&1 &\n",
	  ux_installation_params.relocate_etc_directory, devno);
  Fprintf(out, "    fi\n");
}

rc_state_transition (to, direction, line, out)
     int to, direction;
     char *line;		/*ARGSUSED*/
     FILE *out;
{
  int h, i, copy_lines_p = 1;

#ifdef notdef
  switch(to) {
  case 0:
  case 10:
  case 20:
  case 10000:
    Printf("state %s to %d", direction > 0 ? "up" : "down", to);
    if (line != NULL)
      Printf(" at %s", line);
    else
      terpri();
    break;
  }
#endif

  switch(to) {
  case 0:			/* Normal case */
    break;

  case 10:			/* In ivory-life if clause */
  case 20:			/* In nested if */
    copy_lines_p = 0;
    break;

  case 10000:			/* EOF */
    Fprintf(out, "if [ -f %s/ivory-life ]; then\n",
		   ux_installation_params.relocate_etc_directory);
    Fprintf(out, "    (echo -n \"starting %s life support:\")\t\t>/dev/console\n",
		   UX_NAME);

    /* Start ivory-servicesd */
    Fprintf(out, "    if [ -f %s/ivory-servicesd ]; then\n",
		   ux_installation_params.relocate_etc_directory);
    Fprintf(out, "\t(echo -n ' rpc services ')\t\t\t>/dev/console\n");
    Fprintf(out, "\t%s/ivory-servicesd\t\t\t>/dev/console 2>&1 &\n",
		   ux_installation_params.relocate_etc_directory);
    Fprintf(out, "    fi\n");

    /* Start ivory-life for each /dev/ivoryN */
    for (i = 0; i < ux_installation_params.n_ux400_boards; i++)
      rc_start_ivory_life(i, out);
    for (h = 0, i = MIN_UX1200_DEVNO; h < ux_installation_params.n_ux1200_boards; h++, i++)
      rc_start_ivory_life(i, out);
    Fprintf(out, "    (echo '.')\t\t\t\t\t>/dev/console\n");
    Fprintf(out, "fi\n");
    break;
  }

  return(copy_lines_p);
}



static int update_ux_configuration_file (params, devno)
     struct params *params;
{
  char buf[64], buf1[64];
  FILE *fp;
  int j;

  Sprintf(buf, "/etc/ivory%d.conf", devno);
  Printf("Writing %s...", buf);
  (void) fflush(stdout);

  if (!access(buf, F_OK)) {
    Sprintf(buf1, "%s.orig", buf);
    if (rename(buf, buf1) == -1) {
      Sprintf(buf1, "Renaming %s to %s.orig", buf, buf);
      perror(buf1);
      return(-1);
    }
  }

  if ((fp = fopen(buf, "w")) == NULL) {
    perror(buf);
    return(-1);
  }

  Fprintf(fp, "device /dev/ivory%d\n", devno);
  Fprintf(fp, "prom-directory %s\n", params->relocate_ifep_directory);
  for (j = 0; params->ivory_interfaces[devno][j] != NULL && j < MAX_INTERFACE_N; j++)
    Fprintf(fp, "interface %s\n", params->ivory_interfaces[devno][j]);
  for (j = 0; params->ivory_partitions[devno][j].filename != NULL && j < MAX_PARTITION_N; j++)
    Fprintf(fp, "partition %s\n", params->ivory_partitions[devno][j].filename);

  (void) fclose(fp);
  terpri();
  return(0);
}

static update_ivory_configuration_files (params)
     struct params *params;
{
  int i;

  for (i = 0; i < params->n_ux400_boards; i++)
    if (update_ux_configuration_file(params, i) == -1)
      return(-1);
  for (i = 0; i < params->n_ux1200_boards; i++)
    if (update_ux_configuration_file(params, i+MIN_UX1200_DEVNO) == -1)
      return(-1);

  return(0);
}

static int update_ux_device_special_file (params, devno)
     struct params *params;
     int devno;
{
  char buf[35];
  struct stat st;

  Sprintf(buf, "/dev/ivory%d", devno);

  /* Check to see if there's already a file with this name */
  if (stat(buf, &st) == -1) {
    if (errno != ENOENT) {
      perror(buf);
      return(-1);
    }
  } else
    if (major(st.st_rdev) == params->ivory_major_device_number &&
	minor(st.st_rdev) == devno)
      /* Device-special file is up to date */
      return(0);
    else
      /* Remove old file */
      if (unlink(buf) == -1) {
	Sprintf(buf, "Deleting old /dev/ivory%d", devno);
	perror(buf);
	return(-1);
      }

  /* Create the device-special file */
  if (mknod(buf, S_IFCHR, makedev(params->ivory_major_device_number, devno)) == -1) {
    perror(buf);
    return(-1);
  }
  (void) chmod(buf, 0666);
  return(0);
}

static update_device_special_files (params)
     struct params *params;
{
  int i;

  Printf("Updating Ivory device files...");
  (void) fflush(stdout);

  for (i = 0; i < params->n_ux400_boards; i++)
    if (update_ux_device_special_file(params, i) == -1)
      return(-1);
  for (i = 0; i < params->n_ux1200_boards; i++)
    if (update_ux_device_special_file(params, i+MIN_UX1200_DEVNO) == -1)
      return(-1);

  Printf("  Done.\n");
  return(0);
}

static install_kernel (params)
     struct params *params;
{
  char buf[BUFSIZ];

  Printf("Installing kernel as /vmunix...");
  (void) fflush(stdout);

  /* Rename old kernel */
  Sprintf(buf, "/vmunix.pre-%s", params->ux_kernel_name);
  if (rename("/vmunix", buf) == -1) {
    Sprintf(buf, "Renaming /vmunix to /vmunix.pre-%s", params->ux_kernel_name);
    perror(buf);
    return(-1);
  }

  /* Copy in new kernel */
  Sprintf(buf, "%s/%s/%s/vmunix",
		 params->kernel_build_directory, params->kernel_architecture,
		 params->ux_kernel_name);
  if (access(buf, R_OK))
    strcat(buf, "_small");
  if (copy_file(buf, "/vmunix") != 0)
    return(-1);

  Printf("  Done.\n");

  return(0);
}

static boot_kernel (params)
     struct params *params;	/*ARGSUSED*/
{

  Printf("Rebooting...");
  (void) fflush(stdout);

  /* Give a little while for panicky users to abort */
  sleep(10);

  terpri();
  (void) shell_command("/usr/etc/fastboot", 1);

  return(0);
}



configure_unix_startup (params)
     struct params *params;
{

  if (params->install_kernel_p ||
      params->configure_ivory_life_p ||
      params->device_special_files_p ||
      params->configure_startup_p ||
      params->boot_p)
    terpri();
  else
    return(0);

  /* Copy UX kernel to /vmunix */
  if (params->install_kernel_p &&
      install_kernel(params) == -1)
    return(-1);

  /* Update /etc/ivoryN.conf */
  if (params->configure_ivory_life_p &&
      update_ivory_configuration_files(params) == -1)
    return(-1);

  /* Update /dev/ivoryN */
  if (params->device_special_files_p &&
      update_device_special_files(params) == -1)
    return(-1);

  /* Update rc.local */
  if (params->configure_startup_p) {
    Printf("Updating /etc/rc.local...");
    (void) fflush(stdout);
    if (update_file("/etc", "rc.local", (char *) NULL, 10000,
		    rc_next_state, rc_state_transition) == -1)
      return(-1);
    terpri();
  }

  /* This clause must come last */
  if (params->boot_p && boot_kernel(params) == -1)
    return(-1);

  return(0);
}
