/* -*-Mode: C; Package: C-user -*- */
/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1990 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */

/*
 * Utilities for UX installation
 *
 * $Log:	util.c,v $
 * Revision 2.5.1.2  93/02/01  12:57:19  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:53:18  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:44:11  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/util.c,v 2.5.1.2 93/02/01 12:57:19 kaufman Rel $";
#endif

#include <sys/types.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <varargs.h>
#include "install.h"

char *getenv();			/* Should be in stdio.h */

/* Read a string, with defaults and an acceptance check */

char *accept_string(prompt, defaultt, checker, arg0, arg1)
     char *prompt;
     char *defaultt;
     int (*checker)();
     arg *arg0, *arg1;
{
  static char buf[BUFSIZ];

  while (1) {
    /* Print prompt */
    fputs(prompt, stdout);
    if (defaultt != NULL)
      Printf(" (default %s)", defaultt);
    Printf("? ");

    /* If we're defaulting, return the default */
    if (ux_installation_params.defaulting &&
	defaultt != NULL) {
      printf("%s\n", defaultt);
      return(strdup(defaultt));
    }

    /* Read an answer and check it */
    if (fgets(buf, BUFSIZ, stdin) == NULL) {
      clearerr(stdin);
      fputs("\n", stdout);
    } else {
      buf[strlen(buf)-1] = '\0';
      if (buf[0] == '\0')
	if (defaultt == NULL)
	  continue;
	else
	  return(strdup(defaultt));
      else if (!checker(buf, arg0, arg1))
	return(strdup(buf));
    }
  }
}

/* Acceptance check for integers */

check_integer (buf, min, max)
     char *buf;
     arg *min, *max;
{
  long input;
  char *end;

  input = strtol(buf, &end, 10);
  if (!*end || isspace(*end))
    if (min->integer <= input && input <= max->integer)
      /* Success */
      return(0);

  Fprintf(stderr, "\n>> Please enter an integer between %d and %d.\n",
	  min->integer, max->integer);
  return(-1);
}

/* Acceptance check for files */

#define ACCESS_BITS (F_OK|R_OK|W_OK|X_OK)

check_file (buf, type, access_bits)
     char *buf;
     arg *type, *access_bits;
{
  int tmp;
  char *cp;
  struct stat st;

  if (!stat(buf, &st)) {
    if (type->integer == 0 || (st.st_mode & S_IFMT) == type->integer) {
      /* File exists: Check access */
      if (!(access_bits->integer & ACCESS_BITS) && (access_bits->integer & NONE_OK)) {
	/* NONE_OK is only bit set, but file exists */
	tmp = EEXIST;
	goto lose;
      } else if (access(buf, (int) (access_bits->integer & ACCESS_BITS)) == -1) {
	/* File exists, but has wrong access */
	tmp = errno;
	goto lose;
      }
      return(0);
    } else
      /* Wrong type of file */
      Fprintf(stderr, "\n>> %s is not a %s.\n", buf,
	      type->integer == S_IFDIR ? "directory" :
	      type->integer == S_IFREG ? "regular file" :
	      type->integer == S_IFCHR ? "character special file" :
	      "file of an unexpected type");
  } else if (errno == ENOENT &&
	     (access_bits->integer & NONE_OK) &&
	     (cp = rindex(buf, '/')) != NULL) {
    /* Hair for directories we might have to create */
    tmp = errno;
    *cp = '\0';
    if ((stat(buf, &st) == -1 && errno == ENOENT) ||
	((st.st_mode & S_IFMT) == S_IFDIR &&
	 /* Check access to parent */
	 (!(access_bits->integer & ACCESS_BITS) ||
	  !access(buf, (int) (access_bits->integer & ACCESS_BITS))))) {
      /* Path doesn't exist, but parent directory does */
      *cp = '/';
      return(0);
    }
    goto lose;
  } else {
    tmp = errno;
  lose:
    terpri();
    errno = tmp;
    perror(buf);
  }

  return(-1);
}



/* How to ask yes-or-no questions */

static int dispatch_y_or_n (ch)
     char ch;
{
  switch(ch) {
  case 't':
  case 'T':
  case 'y':
  case 'Y':
    return(1);

  case 'n':
  case 'N':
    return(0);

  default:
    return(-1);
  }
}

static int check_y_or_n (buf, arg0, arg1)
     char *buf;
     arg *arg0, *arg1;		/*ARGSUSED*/
{
  if (dispatch_y_or_n(*buf) == -1) {
    Printf("\n>> Please answer Y or N.\n");
    return(-1);
  } else
    return(0);
}

y_or_n_p (prompt, defaultt)
     char *prompt;
     int defaultt;
{
  return(dispatch_y_or_n(*accept_string(prompt, defaultt ? "Yes" : "No",
					check_y_or_n, (arg *) NULL, (arg *) NULL)));
}


/* Search for a file of a particular type, when it might have one of several names */

/*VARARGS0*/
char *file_search (va_alist)
     va_dcl
{
  va_list pvar;
  int file_type;
  char *file;
  struct stat st;

  va_start(pvar);
  file_type = va_arg(pvar, int);

  while ((file = va_arg(pvar, char *)) != NULL)
    if (!stat(file, &st) &&
	(!file_type || file_type == (st.st_mode & S_IFMT))) {
      va_end(pvar);
      return(file);
    }

  va_end(pvar);
  return(NULL);
}


/* Copy files from source that are newer than their counterparts in dest (if any) */

process_directory (source, dest, select_fn, update_fn, verbose_p)
     char *source;
     char *dest;
     int (*select_fn)();
     int (*update_fn)();
     int verbose_p;
{
  struct dirent *file;
  DIR *dp;
  char source_file[BUFSIZ];
  char dest_file[BUFSIZ];
  int updated = 0, failed = 0;

  if ((dp = opendir(source)) == NULL) {
    perror(source);
    if (verbose_p)
      Printf("  Update failed.\n");
    return(-1);
  }

  while ((file = readdir(dp)) != (struct dirent *) NULL) {

    if ((*select_fn) (file, source, dest, source_file, dest_file) == 0)
      continue;

    if ((*update_fn)(source_file, dest_file) == 0)
      updated++;
    else
      failed++;
  }

  if (verbose_p) {
    if (updated == 0)
      Printf("  No files updated");
    else
      Printf("  Updated %d file%s", updated, updated > 1 ? "s" : "");
    if (failed == 0)
      Printf(".\n");
    else
      Printf(", %d update%s failed.\n", failed, failed > 1 ? "s" : "");
  }

  (void) closedir(dp);
  return(updated);
}

/* Return non-zero if dest doesn't exist, or is older than source */

source_newer_p (source_directory, source_file, dest_directory, dest_file, verbose_p)
     /*ARGSUSED*/
     char *source_directory, *source_file;
     char *dest_directory, *dest_file;
     int verbose_p;
{

  /* Always copy files */
  return(1);

#ifdef notdef
  char source_buf[BUFSIZ], dest_buf[BUFSIZ];
  char *source, *dest;
  struct stat source_st, dest_st;
  int i;

  if (source_directory != NULL) {
    Sprintf(source_buf, "%s/%s", source_directory, source_file);
    source = source_buf;
  } else
    source = source_file;

  if (dest_directory != NULL) {
    Sprintf(dest_buf, "%s/%s", dest_directory, dest_file);
    dest = dest_buf;
  } else
    dest = dest_file;

  if (stat(source, &source_st) == -1) {
    if (verbose_p)
      perror(source);
    return(-1);
  }

  i = stat(dest, &dest_st);
  if ((i == -1 && errno == ENOENT) ||
      (i == 0 && source_st.st_mtime > dest_st.st_mtime))
    return(1);
  else if (i == -1) {
    if (verbose_p)
      perror(dest);
    return(-1);
  }

  return(0);
#endif
}



void set_file_times (source, dest)
     char *source;
     char *dest;
{
  struct stat st;
  struct timeval tvs[2];

  if (stat(source, &st) == -1)
    return;
  tvs[0].tv_sec = st.st_atime;
  tvs[0].tv_usec = 0;
  tvs[1].tv_sec = st.st_mtime;
  tvs[1].tv_usec = 0;
  (void) utimes(dest, tvs);
}


/* Copy a single file */

copy_file (source, dest)
     char *source;
     char *dest;
{
  int val;
  union wait status;

  if ((val = vfork()) == -1) {
    perror("fork");
    return(-1);
  } else if (val != 0) {
    /* Parent */
    wait4(val, &status, 0, NULL);
    if (WIFEXITED(status)) {
      if (!status.w_retcode)
	set_file_times(source, dest);
      return(status.w_retcode);
    } else
      return(1);
  } else {
    /* Child */
    execlp("cp", "cp", source, dest, NULL);
    perror("cp");
    exit(-1);
  }
  /*NOTREACHED*/
}

static update_select_fn  (file, source_directory, dest_directory,
				      source_file, dest_file)
     struct dirent *file;
     char *source_directory;
     char *dest_directory;
     char *source_file;
     char *dest_file;
{
  int i;

  /* Skip directory and parent entries */
  if (!strcmp(file->d_name, ".") ||
      !strcmp(file->d_name, ".."))
    return(0);
  
  /* Set up source and destination file names */
  Sprintf(source_file, "%s/%s", source_directory, file->d_name);
  Sprintf(dest_file, "%s/%s", dest_directory, file->d_name);

  /* Skip if source doesn't exist, or isn't newer */
  if ((i = source_newer_p((char *) NULL, source_file, (char *) NULL, dest_file, 1)) == -1 ||
      i == 0)
    return(0);

  return(1);
}

update_directory (source, dest, verbose_p)
     char *source;
     char *dest;
     int verbose_p;
{
  return(process_directory(source, dest, update_select_fn, copy_file, verbose_p));
}



install_file (source, dest)
     char *source;
     char *dest;
{
  int val;
  char *cp;
  union wait status;

  if ((val = vfork()) == -1) {
    perror("fork");
    return(-1);
  } else if (val != 0) {
    /* Parent */
    wait4(val, &status, 0, NULL);
    if (WIFEXITED(status)) {
      if (!status.w_retcode)
	set_file_times(source, dest);
      return(status.w_retcode);
    } else
      return(1);
  } else {
    /* Child */
    
    if ((cp = rindex(source, '/')) != NULL &&
	((!strcmp(cp+1, "ivory-life") &&
	  ux_installation_params.ivory_life_setuid_p) ||
	 (!strcmp(cp+1, "ivory-servicesd") &&
	  ux_installation_params.ivory_servicesd_setuid_p)))
      execlp("install", "install", "-o", "root", "-m", "4555", source, dest, NULL); 
    else if (ux_installation_params.setgid_p)
      execlp("install", "install", "-g", "kmem", "-m", "2555", source, dest, NULL); 
    else
      execlp("install", "install", "-m", "555", source, dest, NULL);

    perror("install");
    exit(-1);
  }
  /*NOTREACHED*/
}

/* Like update_directory(), but uses install instead of cp */
install_directory (source, dest, verbose_p)
     char *source;
     char *dest;
     int verbose_p;
{
  return(process_directory(source, dest, update_select_fn, install_file, verbose_p));
}



/* Support for file-processing state machines */

static state_transition (from, to, line, out, state_transition_fn)
     int from, to;
     char *line;
     FILE *out;
     int (*state_transition_fn)();
{
  int copy_lines_p;

  if (from > to) {
    /* Downwards state transition */
    do {
      from--;
      copy_lines_p = (*state_transition_fn)(from, -1, from == to ? line : NULL, out);
      if (copy_lines_p == -1) {
	/* Error */
	return(-1);
      }
    } while (from > to); 
  } else {
    /* Upwards state transition */
    do {
      from++;
      copy_lines_p = (*state_transition_fn)(from, 1, from == to ? line : NULL, out);
      if (copy_lines_p == -1) {
	/* Error */
	return(-1);
      }
    } while (to > from); 
  }

  return(copy_lines_p);
}

update_file (directory, file_in, file_out, final_state, next_state_fn, state_transition_fn)
     char *directory, *file_in, *file_out;
     int final_state;
     int (*next_state_fn)();
     int (*state_transition_fn)();
{
  FILE *in, *out;
  char *temp, source[BUFSIZ], buf[BUFSIZ];
  int state = 0, copy_lines_p = 1;
  int new;

  /* Open input file */
  Sprintf(source, "%s/%s", directory, file_in);
  if ((in = fopen(source, "r")) == NULL) {
    perror(source);
    return(-1);
  }

  /* Open temporary output file */
  temp = tempnam(directory, "SMBX");
  if ((out = fopen(temp, "w")) == NULL) {
    perror(temp);
    (void) fclose(in);
    return(-1);
  }

  /* Process contents */
  while (fgets(buf, BUFSIZ, in) != NULL) {
    if ((new = (*next_state_fn)(state, buf)) == -1) {
      /* Error */
      (void) fclose(in);
      (void) fclose(out);
      (void) unlink(temp);
      return(-1);
    } else if (state != new) {
      copy_lines_p = state_transition(state, new, buf, out, state_transition_fn);
      if (copy_lines_p == -1) {
	/* Error */
	(void) fclose(in);
	(void) fclose(out);
	(void) unlink(temp);
	return(-1);
      }
      state = new;
    } else
      /* State unchanged */
      if (copy_lines_p)
	fputs(buf, out);
  }

  /* Final state transitions */
  if (final_state != 0)
    if (state_transition(state, final_state, (char *) NULL,
			 out, state_transition_fn) == -1) {
      /* Error */
      (void) fclose(in);
      (void) fclose(out);
      (void) unlink(temp);
      return(-1);
    }

  /* Close files */
  (void) fclose(in);
  (void) fclose(out);

  /* Rename files as appropriate */
  if (file_out == NULL) {
    /* Update source with temporary file */
    Sprintf(buf, "%s.orig", source);
    if (rename(source, buf) == -1) {
      Sprintf(buf, "rename %s to %s.orig", source, source);
      perror(buf);
      return(-1);
    }
    if (rename(temp, source) == -1) {
      Sprintf(buf, "rename %s to %s", temp, source);
      perror(buf);
      return(-1);
    }
  } else {
    /* Temporary file is new */
    Sprintf(buf, "%s/%s", directory, file_out);
    if (rename(temp, buf) == -1) {
      Sprintf(buf, "rename %s to %s/%s", temp, directory, file_out);
      perror(buf);
      return(-1);
    }
  }

  /* Final cleanup and return */
  free(temp);
  return(0);
}

/* Return non-zero if a line is "equivalent" to the specified series of tokens.
 * This doesn't do anything like C tokenizing, we could eventually lose because of that. */

/*VARARGS0*/
token_match (va_alist)
     va_dcl
{
  va_list args;
  char *line, *token;
  int token_len;

  va_start(args);
  line = va_arg(args, char *);

  while ((token = va_arg(args, char *)) != NULL) {

    /* Find start of next token */
    while (*line && isspace(*line))
      line++;

    /* Have we run out of input?  Does token match? */
    token_len = strlen(token);
    if (*line == '\0' || strncmp(line, token, token_len)) {
      /* Didn't match */
      va_end(args);
      return(0);
    } else
      line += token_len;
  }

  va_end(args);
  return(1);
}



/* Sugar for system() */

shell_command(command, verbose_p)
     char *command;
     int verbose_p;
{
  int ret;

  if (verbose_p)
    Printf("%% %s\n", command);
  ret = system(command);
  return(ret >> 8);
}

require_directory (directory, verbose_p)
     char *directory;
     int verbose_p;
{
  char *cp, buf[BUFSIZ];
  int i;

  if (access(directory, F_OK)) {

    /* Check parent */
    if ((cp = rindex(directory, '/')) != NULL &&
	cp != directory) {
      *cp = '\0';
      if (require_directory(directory, verbose_p) == -1)
	return(-1);
      *cp = '/';
    }

    Sprintf(buf, "mkdir %s", directory);
    if ((i = shell_command(buf, verbose_p)) != 0) {
      if (verbose_p)
	Fprintf(stderr, ">> Error return code %d from mkdir.\n", i);
      return(-1);
    }
  }
  return(0);
}



/* Check that a given directory is in the search path
 * defined by an environment variable.  Return values
 * are:
 *    -3 for NULL default path
 *    -2 for directory not in default path
 *    -1 for directory not in path
 *     1 for directory in path
 *     2 for directory in default path
 *
 * Negative return values are considered failure results,
 * positive values are success results. */

verify_directory_in_path (path_name, path_default, directory)
     char *path_name, *path_default, *directory;
{
  char *cp, *path;
  int using_default_path = FALSE;

  if ((path = getenv(path_name)) == NULL)
    if ((path = path_default) == NULL)
      return(-3);
    else
      using_default_path = TRUE;

  do {

    if ((cp = index(path, ':')) != NULL)
      *cp = '\0';

    if (!strcmp(path, directory))
      return(using_default_path == TRUE ? 2: 1);

    if (cp == NULL)
      path = NULL;
    else {
      *cp = ':';
      path = cp+1;
    }
  } while (path != NULL);

  return(using_default_path == TRUE ? -2 : -1);
}
