/*>
 *> *****************************************************************************************
 *> ** (c) Copyright 1993-1989 Symbolics, Inc.  All rights reserved.
 *> ** Portions of font library Copyright (c) 1984 Bitstream, Inc.  All Rights Reserved.
 *>
 *>    The software, data, and information contained herein are proprietary 
 *> to, and comprise valuable trade secrets of, Symbolics, Inc., which intends 
 *> to keep such software, data, and information confidential and to preserve 
 *> them as trade secrets.  They are given in confidence by Symbolics pursuant 
 *> to a written license agreement, and may be used, copied, transmitted, and 
 *> stored only in accordance with the terms of such license.
 *> 
 *> Symbolics, Symbolics 3600, Symbolics 3670 (R), Symbolics 3675 (R), Symbolics
 *> 3630, Symbolics 3640, Symbolics 3645 (R), Symbolics 3650 (R), Symbolics 3653,
 *> Symbolics 3620 (R), Symbolics 3610 (R), Symbolics Common Lisp (R),
 *> Symbolics-Lisp (R), Zetalisp (R), Genera (R), Wheels (R), Dynamic Windows (R),
 *> SmartStore (R), Semanticue (R), Frame-Up (R), Firewall (R), Document Examiner (R),
 *> Delivery Document Examiner, "Your Next Step in Computing" (R), Ivory, MacIvory,
 *> MacIvory model 1, MacIvory model 2, MacIvory model 3, XL400, XL1200, XL1201,
 *> Symbolics UX400S, Symbolics UX1200S, NXP1000, Symbolics C, Symbolics Pascal (R),
 *> Symbolics Prolog, Symbolics Fortran (R), CLOE (R), CLOE Application Generator,
 *> CLOE Developer, CLOE Runtime, Common Lisp Developer, Symbolics Concordia, Joshua,
 *> Statice (R), and Minima are trademarks of Symbolics, Inc.
 *> 
 *> RESTRICTED RIGHTS LEGEND
 *>    Use, duplication, and disclosure by the Government are subject to restrictions 
 *> as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and Computer 
 *> Software Clause at DFAR 52.227-7013.
 *> 
 *>      Symbolics, Inc.
 *>      6 Concord Farms
 *>      555 Virginia Road
 *>      Concord, Massachusetts 01742-2727
 *>      United States of America
 *>      508-287-1000
 *>
 *> *****************************************************************************************
 *>
 */
/*
 * $Log:	write-fep-file.c,v $
 * Revision 2.5.1.2  93/02/01  12:57:21  kaufman
 * Update copyright notices
 * 
 * Revision 2.5.1.1  93/02/01  11:53:21  kaufman
 * Baseline for distributed software
 * 
 * Revision 2.5  93/02/01  11:44:15  kaufman
 * Baseline for Genera 8.3
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Header: /usr/share/local/symbolics/src/RCS/write-fep-file.c,v 2.5.1.2 93/02/01 12:57:21 kaufman Rel $";
#endif

#include <stdio.h>
#include "fepfs.h"

void write_fep_file (fs, pathname, descriptor, characters)
file_system fs;
string pathname;
integer descriptor;
{
  fep_stream stream;
  char buffer [4096];
  integer count;
  integer total = 0;
  property_list properties;
  integer i;

  if ((stream = open_fep_stream (fs, pathname, true)) == NULL) {
    fprintf (stderr, "Couldn't open fep file %s\n", pathname);
    exit (2);
  }
  while (true) {
    if ((count = read (descriptor, buffer, 4096)) < 0)
      perror ("write_fep_file error reading descriptor");
    if (count == 0) break;
    if (characters)
      translate_unix_to_genera (buffer, count);
    /* For byte swapping purposes, we round the byte count down to the next word boundary */
    bswap (buffer, buffer, count & ~3);
    write_fep_stream (stream, buffer, count & ~3);
    /* This works only for trailing bytes */
    for (i = count & ~3; i < count; i++) {
      seek_fep_stream (stream, (total + i) ^ 3);
      write_fep_stream (stream, buffer + i, 1);
    }
    total += count;
    seek_fep_stream (stream, total);
  }
  properties = get_fep_stream_properties (stream);
  properties->length_in_bytes = total;
  set_fep_stream_properties (stream, properties);
}

void usage (name)
string name;
{
  fprintf (stderr, "Usage: %s [-characters] partition-filename fep-filename\n", name);
  exit (2);
}

void main (argc, argv)
int argc;
char **argv;
{
  file_system fs;
  boolean characters = false;
  string partition = NULL;
  string filename = NULL;
  integer i;

  for (i = 1; i < argc && argv[i][0] == '-'; i++)
    switch (argv[i][1]) {
    case 'c':
      characters = true;
      break;
    default:
      usage (argv[0]);
    }
  if (i == argc) usage (argv[0]);         /* Not enough args */
  partition = argv[i++];
  if (i == argc) usage (argv[0]);         /* Not enough args */
  filename = argv[i++];
  if (i != argc) usage (argv[0]);          /* Too many args */

  if ((fs = bootstrap_disk (partition, false)) == NULL) {
    fprintf (stderr, "Couldn't open FEP partition file %s: ", partition);
    perror (NULL);
    exit (2);
  }
  if (!bootstrap_file_system (fs)) {
    fprintf (stderr, "Sorry, partition %s does not contain a viable FEP file system.\n",
	     partition);
    exit (2);
  }
  write_fep_file (fs, filename, fileno (stdin), characters);
}
